#include <iostream>
#include <torch/script.h>
#include <torch/torch.h>
#include <iomanip>
#include <Utils/Geometry/ElementInfo.h>
#include <Utils/Geometry/AtomCollection.h>
#include <Utils/IO/ChemicalFileFormats/XyzStreamHandler.h>
#include "Utils/Constants.h"

int main(int argc, const char *argv[]) {

  // Get the path of model
  if (argc != 3) {
    std::cerr << "usage: model xyzFile\n";
    return -1;
  }

  torch::jit::script::Module model;
  try {
    model = torch::jit::load(argv[1]);
    //std::cout << "successfully loaded the model\n";
  }

  catch (const c10::Error &e) {
    std::cerr << "error loading the model\n";
    return -1;
  }
  
  std::string filename = argv[2];
  std::ifstream structureFile(filename);
  auto structure = Scine::Utils::XyzStreamHandler::read(structureFile);
  
  auto positions = structure.getPositions();
  auto elements = structure.getElements();
  int natom = structure.size();
  long int Znumber[natom];
  for (int atom = 0; atom < natom; ++atom) {
    const auto e = elements[atom];
    Znumber[atom] = Scine::Utils::ElementInfo::Z(e);
  }
  
  torch::Tensor species = torch::from_blob(
      Znumber, {1, natom}, 
      torch::TensorOptions().requires_grad(false).dtype(torch::kLong));

  positions *= Scine::Utils::Constants::angstrom_per_bohr;
  torch::Tensor coords = torch::from_blob(
      positions.data(), {1, natom, 3}, 
      torch::TensorOptions().requires_grad(false).dtype(torch::kFloat64));
  
  std::vector<torch::jit::IValue> inputs;

  inputs.push_back(species);
  inputs.push_back(coords);
  inputs.push_back(false);
  inputs.push_back(false);

  auto outputs = model.forward(inputs).toTuple();
  auto energies = outputs->elements()[0].toTensor().item().toFloat();
  auto saes = outputs->elements()[1].toTensor().item().toFloat();
  auto stds = outputs->elements()[2].toTensor().item().toFloat();
  
  std::cout << "Standard deviation of NN contribution   :";
  std::cout << std::right << std::setw(15) << std::fixed << std::setprecision(8) << stds;
  std::cout << " Hartree" << std::right << std::setw(15) << std::fixed << std::setprecision(5);
  std::cout << stds * Scine::Utils::Constants::kCalPerMol_per_hartree << " kcal/mol" << std::endl;

  std::cout << "NN contribution                         :";
  std::cout << std::right << std::setw(15) << std::fixed << std::setprecision(8) << energies - saes;
  std::cout << " Hartree" << std::endl;

  std::cout << "Sum of atomic self energies             :";
  std::cout << std::right << std::setw(15) << std::fixed << std::setprecision(8) << saes;
  std::cout << " Hartree" << std::endl;

  std::string energyName = ".Enn";
  std::ofstream energyOut(energyName);
  energyOut << std::defaultfloat << std::fixed << std::setprecision(10);
  energyOut << energies << std::endl;
  energyOut.close();
  
  return 0;
}
