/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#include "AbInitioDensityMatrixGuessGenerator.h"
#include "Libint.h"
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Utils/DataStructures/DensityMatrix.h>
#include <Utils/Geometry/ElementInfo.h>
#include <Utils/Scf/MethodInterfaces/OverlapCalculator.h>
#include <Eigen/Eigenvalues>

namespace Scine {
namespace Sparrow {
namespace Integrals {

AbInitioDensityMatrixGuessGenerator::AbInitioDensityMatrixGuessGenerator(const BasisSetController& basisSet,
                                                                         const int& nElectrons,
                                                                         Utils::OverlapCalculator& overlapCalculator)
  : initialGuessType_(InitialGuessType::ExtendedHueckel),
    nElectrons_(nElectrons),
    basisSet_(basisSet),
    overlapCalculator_(overlapCalculator) {
}

AbInitioDensityMatrixGuessGenerator::~AbInitioDensityMatrixGuessGenerator() = default;

void AbInitioDensityMatrixGuessGenerator::setGuessType(InitialGuessType guessType) {
  initialGuessType_ = guessType;
}

Utils::DensityMatrix AbInitioDensityMatrixGuessGenerator::calculateGuess() const {
  Utils::DensityMatrix dm;
  if (initialGuessType_ == InitialGuessType::ExtendedHueckel) {
    dm = calculateExtendedHueckelGuess();
  }
  else if (initialGuessType_ == InitialGuessType::HCore) {
    dm = calculateHCoreGuess();
  }
  return dm;
}

Utils::DensityMatrix AbInitioDensityMatrixGuessGenerator::calculateExtendedHueckelGuess() const {
  const auto nbf = basisSet_.nBasisFunction();
  Eigen::MatrixXd densityMatrix = Eigen::MatrixXd::Zero(nbf, nbf);
  auto aoindex = basisSet_.toAtomsOrbitalsIndexes();

  if (nbf != 0) {
    overlapCalculator_.calculateOverlap(Utils::DerivativeOrder::Zero);
    densityMatrix =
        (overlapCalculator_.getOverlap().getMatrixXd() * nElectrons_ / (2 * nbf)).selfadjointView<Eigen::Upper>();
  }
  auto elements = basisSet_.getStructure().getElements();
  int basisFunctionIndex = 0;
  for (int el = 0; el < int(elements.size()); ++el) {
    double nElectronsOnAtom = Utils::ElementInfo::valElectrons(elements[el]);
    auto nAOs = aoindex.getNOrbitals(el);
    for (int j = 0; j < nAOs; ++j, ++basisFunctionIndex) {
      densityMatrix.diagonal()(basisFunctionIndex) = nElectronsOnAtom / nAOs;
    }
  }

  Utils::DensityMatrix dm;
  dm.setDensity(std::move(densityMatrix), nElectrons_);
  return dm;
}

Utils::DensityMatrix AbInitioDensityMatrixGuessGenerator::calculateHCoreGuess() const {
  auto& libint = Libint::getInstance();
  Eigen::MatrixXd hcore = libint.computeHCore(basisSet_);
  overlapCalculator_.calculateOverlap(Utils::DerivativeOrder::Zero);
  Eigen::GeneralizedSelfAdjointEigenSolver<Eigen::MatrixXd> es(hcore, overlapCalculator_.getOverlap().getMatrixXd());

  Eigen::MatrixXd occupiedMOs = es.eigenvectors().leftCols(nElectrons_ / 2);

  Utils::DensityMatrix dm;
  dm.setDensity(2 * occupiedMOs * occupiedMOs.transpose(), nElectrons_);
  return dm;
}

/*
Utils::DensityMatrix AbInitioDensityMatrixGuessGenerator::calculateSemiempiricalGuess() const {
  auto& manager = Core::ModuleManager::getInstance();
  try {
    manager.load("sparrow");
  }
  catch (const std::exception& e) {
    throw std::runtime_error("Sparrow module for semiempirical guess was not found.");
  }

  auto guessCalculator = manager.get<Core::Calculator>("PM6");

  guessCalculator->setStructure(basisSet_.getStructure());
  return guessCalculator->results().get<Utils::Property::DensityMatrix>();
}*/
} // namespace Integrals
} // namespace Sparrow
} // namespace Scine
