/**
 * @file Calculator.h
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher
 * Group.\n See LICENSE.txt for details.
 */

#include "BasisSetController.h"
#include "Libint.h"
#include <Utils/DataStructures/AtomicGtos.h>
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Utils/Geometry/ElementInfo.h>
#include <algorithm>

namespace Scine {
namespace Sparrow {
namespace Integrals {

BasisSetController::BasisSetController() = default;

BasisSetController::~BasisSetController() = default;

BasisSetController::BasisSetController(const BasisSetController& rhs) : BasisSetController() {
  *this = rhs;
}

BasisSetController& BasisSetController::operator=(const BasisSetController& rhs) {
  basisSetName_ = rhs.basisSetName_;
  elements_ = rhs.elements_;
  positions_ = rhs.positions_;
  basisSet_ = std::make_unique<libint2::BasisSet>(*rhs.basisSet_);
  return *this;
}

BasisSetController::BasisSetController(std::string basisSetName, const Utils::AtomCollection& structure)
  : basisSetName_(std::move(basisSetName)), elements_(structure.getElements()), positions_(structure.getPositions()) {
  generateBasisSetFromStructure(structure);
}

BasisSetController::BasisSetController(const Utils::AtomCollection& structure,
                                       const std::vector<Utils::AtomicGtos>& gtos, std::string basisSetName)
  : basisSetName_(std::move(basisSetName)), elements_(structure.getElements()), positions_(structure.getPositions()) {
  generateBasisSet(structure, gtos);
}

void BasisSetController::setStructure(const Utils::AtomCollection& structure) {
  elements_ = structure.getElements();
  positions_ = structure.getPositions();
  generateBasisSetFromStructure(structure);
}

void BasisSetController::setStructure(const Utils::AtomCollection& structure, const std::vector<Utils::AtomicGtos>& gtos) {
  elements_ = structure.getElements();
  positions_ = structure.getPositions();
  generateBasisSet(structure, gtos);
}

void BasisSetController::modifyPositions(const Utils::PositionCollection& positions) {
  std::vector<libint2::Atom> atoms(positions_.rows());

  for (size_t i = 0; i < atoms.size(); ++i) {
    const Eigen::RowVectorXd& position = positions_.row(i);
    atoms[i] = {static_cast<int>(Utils::ElementInfo::Z(elements_[i])), position.x(), position.y(), position.z()};
  }

  auto correspondingAtomList = basisSet_->shell2atom(atoms);

  positions_ = positions;
  auto& basis = *basisSet_;

  for (size_t i = 0; i < basis.size(); ++i) {
    const Eigen::RowVectorXd& position = positions_.row(correspondingAtomList[i]);
    basis[i].move({{position.x(), position.y(), position.z()}});
  }
}

void BasisSetController::setBasisSet(std::string basisSetName) {
  basisSetName_ = std::move(basisSetName);
}

void BasisSetController::setBasisSet(const std::vector<Utils::AtomicGtos>& gtos, const Utils::PositionCollection& positions) {
  generateBasisSet({elements_, positions}, gtos);
}

const libint2::BasisSet& BasisSetController::getBasisSet() const {
  if (!basisSet_)
    throw BasisSetNotSetException();
  return *basisSet_;
}

auto BasisSetController::convertToShell(const Utils::Position& position, const Utils::GtoExpansion& gto, double scaling)
    -> libint2::Shell {
  libint2::svector<double> exponents;
  libint2::svector<double> coefficients;
  for (const auto gtf : gto.gtfs) {
    exponents.push_back(gtf.exponent * scaling);
    coefficients.push_back(gtf.coefficient);
  }
  assert(exponents.size() != 0);
  assert(coefficients.size() != 0);
  assert(exponents.size() == coefficients.size());
  libint2::Shell shell{exponents, {{gto.angularMomentum, false, coefficients}}, {{position.x(), position.y(), position.z()}}};
  return shell;
}

void BasisSetController::generateBasisSet(const Utils::AtomCollection& structure, const std::vector<Utils::AtomicGtos>& gtos) {
  if (int(gtos.size()) != structure.size()) {
    throw std::runtime_error("Size mismatch between how many AtomicGtos there are and how many atoms there are. Basis "
                             "set cannot be constructed.");
  }
  if (gtos.size() == 0) {
    throw std::runtime_error("No gtos given in construction of the Basis Set. Aborting.");
  }
  const auto atoms = Libint::toLibintAtoms(structure);

  int maxZ = std::max_element(atoms.begin(), atoms.end(), [](const auto& atom_i, const auto& atom_j) {
               return atom_i.atomic_number < atom_j.atomic_number;
             })->atomic_number;

  std::vector<std::vector<libint2::Shell>> shells;
  shells.resize(maxZ + 1);

  auto createShell = [&](const Utils::GtoExpansion& gto, const libint2::Atom& atom) -> libint2::Shell {
    libint2::svector<double> exponents;
    libint2::svector<double> coefficients;
    for (const auto gtf : gto.gtfs) {
      exponents.push_back(gtf.exponent);
      coefficients.push_back(gtf.coefficient);
    }
    assert(exponents.size() != 0);
    assert(coefficients.size() != 0);
    assert(exponents.size() == coefficients.size());
    libint2::Shell shell{exponents, {{gto.angularMomentum, false, coefficients}}, {{atom.x, atom.y, atom.z}}};
    return shell;
  };

  int index = 0;
  for (const auto& atomicGto : gtos) {
    assert(int(atoms[index].atomic_number) < int(shells.size()));
    std::vector<libint2::Shell> atomicShells;
    if (atomicGto.s) {
      atomicShells.push_back(createShell(*atomicGto.s, atoms[index]));
    }
    if (atomicGto.p) {
      atomicShells.push_back(createShell(*atomicGto.p, atoms[index]));
    }
    if (atomicGto.d) {
      atomicShells.push_back(createShell(*atomicGto.d, atoms[index]));
    }
    shells.at(atoms[index].atomic_number) = std::move(atomicShells);
    ++index;
  }

  basisSet_ = std::make_unique<libint2::BasisSet>(atoms, shells, basisSetName_, true);
}

void BasisSetController::generateBasisSetFromStructure(const Utils::AtomCollection& structure) {
  const auto atoms = Libint::toLibintAtoms(structure);
  basisSet_ = std::make_unique<libint2::BasisSet>(basisSetName_, atoms, true);
}

int BasisSetController::nBasisFunction() const {
  return basisSet_->nbf();
}

int BasisSetController::maximumNumberOfPrimitives() const {
  return basisSet_->max_nprim();
}

int BasisSetController::maxAngularMomentum() const {
  return basisSet_->max_l();
}

int BasisSetController::nShells() const {
  return basisSet_->size();
}

std::vector<long> BasisSetController::shellsToAtoms() const {
  auto atoms = Libint::toLibintAtoms({elements_, positions_});
  return basisSet_->shell2atom(atoms);
}

std::map<int, std::vector<long>> BasisSetController::atomsToShells() const {
  auto atoms = Libint::toLibintAtoms({elements_, positions_});

  auto const& basis = *basisSet_;
  auto correspondingAtomList = basis.atom2shell(atoms);

  std::map<int, std::vector<long>> atomsMap;

  int atomIndex = 0;
  for (const auto& atomShellList : correspondingAtomList) {
    atomsMap.insert({atomIndex++, atomShellList});
  }
  return atomsMap;
}

Utils::AtomsOrbitalsIndexes BasisSetController::toAtomsOrbitalsIndexes() const {
  auto atoms = Libint::toLibintAtoms({elements_, positions_});

  auto const& basis = *basisSet_;
  auto correspondingAtomList = basis.atom2shell(atoms);
  Utils::AtomsOrbitalsIndexes aoInfos(elements_.size());

  for (const auto& atomShellList : correspondingAtomList) {
    int nAOs = 0;
    for (const auto& shell : atomShellList) {
      nAOs += basis[shell].size();
    }
    aoInfos.addAtom(nAOs);
  }
  return aoInfos;
}

Utils::AtomCollection BasisSetController::getStructure() const {
  return {elements_, positions_};
}

} // namespace Integrals
} // namespace Sparrow
} // namespace Scine
