/**
 * @file Calculator.h
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher
 * Group.\n See LICENSE.txt for details.
 */

#ifndef INTEGRALS_BASISSETCONTROLLER_H
#define INTEGRALS_BASISSETCONTROLLER_H

#include <Utils/Geometry/AtomCollection.h>
#include <map>
#include <memory>

namespace libint2 {
class BasisSet;
class Shell;
} // namespace libint2

namespace Scine {
namespace Utils {
class AtomsOrbitalsIndexes;
struct AtomicGtos;
struct GtoExpansion;
} // namespace Utils
namespace Sparrow {
namespace Integrals {

class BasisSetNotSetException : public std::exception {
  const char* what() const noexcept final {
    return "No basis set was set.";
  }
};

/**
 * @brief Class representing an atom-centered basis set. Uses privately Libint2.
 */
class BasisSetController {
 public:
  BasisSetController();
  ~BasisSetController();
  BasisSetController(const BasisSetController& rhs);
  BasisSetController& operator=(const BasisSetController& rhs);
  BasisSetController(BasisSetController&&) = default;
  BasisSetController& operator=(BasisSetController&& rhs) noexcept = default;

  /**
   * @brief Constructor
   * @param basisSetName String with the basis set name. Case sensitive.
   * @param structure Elements and positions composing the molecular structure.
   */
  BasisSetController(std::string basisSetName, const Utils::AtomCollection& structure);
  /**
   * @brief Constructor
   * @param basisSetName String with the basis set name. Case sensitive.
   * @param structure Elements and positions composing the molecular structure.
   * @param gtos Internal representation of the basis.
   */
  BasisSetController(const Utils::AtomCollection& structure, const std::vector<Utils::AtomicGtos>& gtos,
                     std::string basisSetName = "");

  void setStructure(const Utils::AtomCollection& structure);
  void setStructure(const Utils::AtomCollection& structure, const std::vector<Utils::AtomicGtos>& gtos);
  Utils::AtomCollection getStructure() const;
  void modifyPositions(const Utils::PositionCollection& positions);

  /**
   * @brief Creates Libint-friendly basis from basis name.
   */
  void setBasisSet(std::string basisSetName);

  /**
   * @brief Converts the internal AtomicGtos basis format to libint-friendly one.
   */
  void setBasisSet(const std::vector<Utils::AtomicGtos>& gtos, const Utils::PositionCollection& positions);

  const libint2::BasisSet& getBasisSet() const;

  int nBasisFunction() const;
  /**
   * @brief Maximal primitive size in this basis.
   */
  int maximumNumberOfPrimitives() const;
  int maxAngularMomentum() const;
  int nShells() const;
  /**
   * @brief Returns a vector with the indices of the atom each shell belongs to.
   */
  std::vector<long> shellsToAtoms() const;
  /**
   * @brief Returns a map form atom index -> list of shell indices.
   */
  std::map<int, std::vector<long>> atomsToShells() const;

  /**
   * @brief Generate a Scine::Utils::AtomsOrbitalsIndexes needed for some properties.
   */
  Utils::AtomsOrbitalsIndexes toAtomsOrbitalsIndexes() const;

  /**
   * @brief Convert a Scine gto to a libint shell.
   */
  static auto convertToShell(const Utils::Position& position, const Utils::GtoExpansion& gto, double scaling = 1.0)
      -> libint2::Shell;

 private:
  void generateBasisSetFromStructure(const Utils::AtomCollection& structure);
  void generateBasisSet(const Utils::AtomCollection& structure, const std::vector<Utils::AtomicGtos>& gtos);
  std::string basisSetName_;
  Utils::ElementTypeCollection elements_;
  Utils::PositionCollection positions_;

  std::unique_ptr<libint2::BasisSet> basisSet_;
};
} // namespace Integrals
} // namespace Sparrow
} // namespace Scine
#endif // INTEGRALS_BASISSETCONTROLLER_H
