/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef INTEGRALS_INTEGRALTYPESPECIFIER_H
#define INTEGRALS_INTEGRALTYPESPECIFIER_H

#include <Utils/Math/DerivOrderEnum.h>
#include <array>
#include <libint2.hpp>

namespace Scine {
namespace Sparrow {
namespace Integrals {

enum class Operator : unsigned {
  overlap = 0,
  kinetic,
  nuclear,
  emultipole1,
  emultipole2,
  emultipole3,
  delta,
  coulomb,
  cgtg,
  cgtg_x_coulomb,
  // delcgtg2,
  erf_coulomb,
  r12,
  erfc_coulomb
};

constexpr std::array<Operator, 14> availableOperators{
    {Operator::overlap, Operator::kinetic, Operator::nuclear, Operator::emultipole1, Operator::emultipole2,
     Operator::emultipole3, Operator::delta, Operator::coulomb, Operator::cgtg, Operator::cgtg_x_coulomb,
     /*Operator::delcgtg2,*/ Operator::erf_coulomb, Operator::r12, Operator::erfc_coulomb}};

/**
 * @brief Function to map between libint2 Operator and AbInitio ones. Hides Libint2 dependency.
 * The function works using a switch statement. Technically one could cast the underlying unsigned type,
 * but I wanted to avoid that in case the libint2 enum changes in the future.
 */
constexpr inline libint2::Operator toLibintOperator(const Operator op) {
  using lOp = libint2::Operator;
  switch (op) {
    case Operator::overlap:
      return lOp::overlap;
    case Operator::kinetic:
      return lOp::kinetic;
    case Operator::nuclear:
      return lOp::nuclear;
    case Operator::emultipole1:
      return lOp::emultipole1;
    case Operator::emultipole2:
      return lOp::emultipole2;
    case Operator::emultipole3:
      return lOp::emultipole3;
    case Operator::delta:
      return lOp::delta;
    case Operator::coulomb:
      return lOp::coulomb;
    case Operator::cgtg:
      return lOp::cgtg;
    case Operator::cgtg_x_coulomb:
      return lOp::cgtg_x_coulomb;
    // case Operator::delcgtg2:
    //  return lOp::delcgtg2;
    case Operator::erf_coulomb:
      return lOp::erf_coulomb;
    case Operator::r12:
      return lOp::r12;
    case Operator::erfc_coulomb:
      return lOp::erfc_coulomb;
    default:
      return lOp::invalid;
  }
}
/*
inline void addDefaultParametersToEngine(Operator op, libint2::Engine& engine) {
  using lOp = libint2::Operator;
  using
  switch (op) {
    case Operator::nuclear:
    case Operator::emultipole1:
    case Operator::emultipole2:
    case Operator::emultipole3:
    case Operator::delta:
    case Operator::coulomb:
      engine.set_params(libint2::operator_traits<lOp::cgtg>::default_params());
    case Operator::cgtg:
      engine.set_params(libint2::operator_traits<lOp::cgtg>::default_params());
    case Operator::cgtg_x_coulomb:
    case Operator::delcgtg2:
    case Operator::erf_coulomb:
    case Operator::r12:
    case Operator::erfc_coulomb:
    default:
      return;
  }
}*/

struct IntegralTypeSpecifier {
  // Use libint2 macro
  static constexpr Utils::DerivativeOrder maxDerivOrder = static_cast<Utils::DerivativeOrder>(LIBINT2_MAX_DERIV_ORDER);

  bool operator<(IntegralTypeSpecifier rhs) const noexcept {
    if (static_cast<int>(op) < static_cast<int>(rhs.op)) {
      return true;
    }
    if (static_cast<int>(op) == static_cast<int>(rhs.op)) {
      return static_cast<int>(derivOrder) < static_cast<int>(rhs.derivOrder);
    }
    return false;
  }

  Operator op;
  Utils::DerivativeOrder derivOrder;
};

} // namespace Integrals
} // namespace Sparrow
} // namespace Scine

#endif // INTEGRALS_INTEGRALTYPESPECIFIER_H
