/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#include "Libint.h"
#include "BasisSetController.h"
#include "IntegralTypeSpecifier.h"
#include "OneBodyIntegralsEvaluator.h"
#include "TwoBodiesIntegrals/Orbitals.h"
#include <Utils/DataStructures/GtoExpansion.h>
#include <Utils/Geometry/AtomCollection.h>
#include <Utils/Geometry/ElementInfo.h>
#include <libint2.hpp>

namespace Scine {
namespace Sparrow {
namespace Integrals {

int Libint::nThreads_ = 1;

Libint::Libint() {
  libint2::initialize();
  nThreads_ = omp_get_max_threads();
}

Libint::~Libint() {
  libint2::finalize();
}

Libint& Libint::getInstance() {
  static Libint libint;
  return libint;
}

std::vector<libint2::Atom> Libint::toLibintAtoms(const Utils::AtomCollection& structure) {
  std::vector<libint2::Atom> libint2Atoms;
  libint2Atoms.reserve(structure.size());

  for (const auto& atom : structure) {
    const auto& position = atom.getPosition();
    libint2Atoms.push_back(
        {static_cast<int>(Utils::ElementInfo::Z(atom.getElementType())), position.x(), position.y(), position.z()});
  }
  return libint2Atoms;
}

int Libint::getMaxNumberThreads() {
  return nThreads_;
}

template<Utils::DerivativeOrder order>
auto Libint::computeOverlapWithCoreOrbitalOmx(int atom_a, const BasisSetController& basisSet,
                                              const Utils::GtoExpansion& coreOrbital) -> Eigen::MatrixXd {
  static_assert(order != Utils::DerivativeOrder::Two, "No second derivatives implemented");
  int nCols = order == Utils::DerivativeOrder::Zero ? 1 : 6;
  Eigen::MatrixXd result = Eigen::MatrixXd::Zero(basisSet.nBasisFunction(), nCols);

  const auto& localBasis = basisSet.getBasisSet();

  const auto& indicesOfFirstBFinShell = localBasis.shell2bf();
  auto atomsToShells = basisSet.atomsToShells();
  // Record the index of the basis on A to skip them in the
  // calculation of the overlap S_{mu,nu} with the core orbital on A
  auto index_shells_on_a = atomsToShells[atom_a];
  libint2::Shell coreShell = BasisSetController::convertToShell(basisSet.getStructure().getPosition(atom_a), coreOrbital);
  auto engine = libint2::Engine(toLibintOperator(Operator::overlap), static_cast<size_t>(localBasis.max_nprim()),
                                static_cast<int>(localBasis.max_l()), static_cast<int>(order));

  const auto& buffer = engine.results();
  for (size_t s1 = 0; s1 < localBasis.size(); ++s1) {
    int firstIndexOfS1 = indicesOfFirstBFinShell[s1];
    int sizeOfShell1 = localBasis[s1].size();

    if (std::find(index_shells_on_a.begin(), index_shells_on_a.end(), s1) == index_shells_on_a.end()) {
      engine.compute(localBasis[s1], coreShell);

      // If everything is screened out break out
      if (buffer[0] == nullptr) {
        break;
      }
      // Digest
      for (int shellIndex = 0; shellIndex < nCols; ++shellIndex) {
        Eigen::Map<const Eigen::Matrix<double, -1, -1, Eigen::RowMajor>> bufferMatrix(buffer[shellIndex], sizeOfShell1, 1);
        result.block(firstIndexOfS1, shellIndex, sizeOfShell1, 1) = bufferMatrix;
      }
    }
  }
  return result;
}

Eigen::MatrixXd Libint::computeHCore(const BasisSetController& basisSet) const {
  Eigen::MatrixXd hCore =
      compute1BodyIntegrals<Operator::kinetic, Utils::DerivativeOrder::Zero>(basisSet).get<Utils::DerivativeOrder::Zero>();
  hCore +=
      compute1BodyIntegrals<Operator::nuclear, Utils::DerivativeOrder::Zero>(basisSet).get<Utils::DerivativeOrder::Zero>();
  return hCore;
}

auto Libint::createLocalBasisOnAtom(int atom_a, const BasisSetController& basisSet) -> libint2::BasisSet {
  auto elements = Utils::ElementTypeCollection{basisSet.getStructure().getElement(atom_a)};
  Utils::PositionCollection pos(1, 3);
  pos.row(0) = basisSet.getStructure().getPosition(atom_a);

  auto libintAtoms = Libint::toLibintAtoms({elements, pos});
  auto atomsToShells = basisSet.atomsToShells();
  // Create a local basis with the basis functions of A
  // to calculate the one-center core-electron attraction V_{mu, nu}, B
  // mu, nu on A
  auto index_shells_on_a = atomsToShells[atom_a];

  std::vector<libint2::Shell> shellsOnA;
  std::transform(index_shells_on_a.begin(), index_shells_on_a.end(), std::back_inserter(shellsOnA),
                 [&](long idx) { return basisSet.getBasisSet()[idx]; });

  auto z_a = Utils::ElementInfo::Z(elements[0]);
  std::vector<std::vector<libint2::Shell>> shellsInBasis(z_a + 1);
  shellsInBasis[z_a] = shellsOnA;
  return libint2::BasisSet(libintAtoms, shellsInBasis);
}

template<Utils::DerivativeOrder order>
auto Libint::calculateVuvBImpl(const libint2::BasisSet& localBasis,
                               const std::vector<std::pair<double, std::array<double, 3>>>& nuclearCharges)
    -> std::vector<Eigen::MatrixXd> {
  // Indices of first basis funcion in shell.
  const auto& indicesOfFirstBFinShell = localBasis.shell2bf();

  int nResults = libint2::num_geometrical_derivatives(1, static_cast<int>(order));
  std::vector<Eigen::MatrixXd> results(nResults);
  std::fill(results.begin(), results.end(), Eigen::MatrixXd::Zero(localBasis.nbf(), localBasis.nbf()));

  auto localEngine = libint2::Engine(toLibintOperator(Operator::nuclear), static_cast<size_t>(localBasis.max_nprim()),
                                     static_cast<int>(localBasis.max_l()), static_cast<int>(order));
  localEngine.set_params(nuclearCharges);
  const auto& buffer = localEngine.results();

  // number of operators * number of derivatives
  // const auto resultsDimension = localEngine.nshellsets();

  // If oparators are hermitian (here, they are) there is index permutational symmetry:
  // <a|O|b> = <b|O|a>, so loop only on half the matrix
  for (size_t s1 = 0; s1 < localBasis.size(); ++s1) {
    int firstIndexOfS1 = indicesOfFirstBFinShell[s1];
    int sizeOfShell1 = localBasis[s1].size();

    for (size_t s2 = 0; s2 <= s1; ++s2) {
      int firstIndexOfS2 = indicesOfFirstBFinShell[s2];
      int sizeOfShell2 = localBasis[s2].size();

      localEngine.compute(localBasis[s1], localBasis[s2]);

      // If everything is screened out break out
      if (buffer[0] == nullptr) {
        break;
      }
      // Digest
      using MapType = Eigen::Map<const Eigen::Matrix<double, -1, -1, Eigen::RowMajor>>;
      if (nResults == 1) {
        MapType bufferMatrix(buffer[0], sizeOfShell1, sizeOfShell2);
        results[0].block(firstIndexOfS1, firstIndexOfS2, sizeOfShell1, sizeOfShell2) = bufferMatrix;
        results[0].block(firstIndexOfS2, firstIndexOfS1, sizeOfShell2, sizeOfShell1) = bufferMatrix.transpose();
      }
      else if (nResults == 3) {
        for (int xyz = 0; xyz < nResults; ++xyz) {
          // Since one center, derivs on mu and on nu are on the same center: sum them up!
          MapType bufferMatrix(buffer[xyz], sizeOfShell1, sizeOfShell2);
          MapType bufferMatrix2(buffer[xyz + 3], sizeOfShell1, sizeOfShell2);
          Eigen::Matrix<double, -1, -1, Eigen::RowMajor> buf = bufferMatrix + bufferMatrix2;
          results[xyz].block(firstIndexOfS1, firstIndexOfS2, sizeOfShell1, sizeOfShell2) = buf;
          results[xyz].block(firstIndexOfS2, firstIndexOfS1, sizeOfShell2, sizeOfShell1) = buf.transpose();
        }
      }
      else {
        throw std::runtime_error("Second derivatives not implemented");
      }
    }
  }
  return results;
}

template<Utils::DerivativeOrder order>
auto Libint::computeVuvB(int atom_a, int atom_b, double chargeOnB, const BasisSetController& basisSet) const
    -> std::vector<Eigen::MatrixXd> {
  DiatomicKey index{atom_a, atom_b};
  auto atomsToShells = basisSet.atomsToShells();

  auto elements = basisSet.getStructure().getElements();
  auto positions = basisSet.getStructure().getPositions();

  std::vector<std::pair<double, std::array<double, 3>>> nuclearCharges;
  // Define a point charge at the position of atom B with charge Z_B
  std::pair<double, std::array<double, 3>> nuclearCharge{
      chargeOnB, {{positions(index.second, 0), positions(index.second, 1), positions(index.second, 2)}}};
  nuclearCharges.emplace_back(std::move(nuclearCharge));

  // Create a local basis with the basis functions of A
  // to calculate the one-center core-electron attraction V_{mu, nu}, B
  // mu, nu on A
  return calculateVuvBImpl<order>(createLocalBasisOnAtom(atom_a, basisSet), nuclearCharges);
}

libint2::Engine Libint::getEngine(IntegralTypeSpecifier intType, const BasisSetController& basisSet) const {
  return {toLibintOperator(intType.op), static_cast<size_t>(basisSet.maximumNumberOfPrimitives()),
          basisSet.maxAngularMomentum(), static_cast<int>(intType.derivOrder)};
}

libint2::Engine Libint::getEngine(IntegralTypeSpecifier intType, const libint2::Shell& shell1, const libint2::Shell& shell2,
                                  const libint2::Shell& shell3, const libint2::Shell& shell4) const {
  int angularMomentum = 0;
  for (const auto& primitive : shell1.contr) {
    angularMomentum = std::max(primitive.l, angularMomentum);
  }
  for (const auto& primitive : shell2.contr) {
    angularMomentum = std::max(primitive.l, angularMomentum);
  }
  for (const auto& primitive : shell3.contr) {
    angularMomentum = std::max(primitive.l, angularMomentum);
  }
  for (const auto& primitive : shell4.contr) {
    angularMomentum = std::max(primitive.l, angularMomentum);
  }
  auto nPrimitives = shell1.nprim();
  nPrimitives = std::max(shell2.nprim(), nPrimitives);
  nPrimitives = std::max(shell3.nprim(), nPrimitives);
  nPrimitives = std::max(shell4.nprim(), nPrimitives);

  return {toLibintOperator(intType.op), nPrimitives, angularMomentum, static_cast<int>(intType.derivOrder)};
}
template auto Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(int, const BasisSetController&,
                                                                                     const Utils::GtoExpansion&)
    -> Eigen::MatrixXd;
template auto Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::One>(int, const BasisSetController&,
                                                                                    const Utils::GtoExpansion&)
    -> Eigen::MatrixXd;
template auto Libint::computeVuvB<Utils::DerivativeOrder::Zero>(int, int, double, const BasisSetController&) const
    -> std::vector<Eigen::MatrixXd>;
template auto Libint::computeVuvB<Utils::DerivativeOrder::One>(int, int, double, const BasisSetController&) const
    -> std::vector<Eigen::MatrixXd>;
template auto Libint::computeVuvB<Utils::DerivativeOrder::Two>(int, int, double, const BasisSetController&) const
    -> std::vector<Eigen::MatrixXd>;
template auto
Libint::calculateVuvBImpl<Utils::DerivativeOrder::Zero>(const libint2::BasisSet&,
                                                        const std::vector<std::pair<double, std::array<double, 3>>>&)
    -> std::vector<Eigen::MatrixXd>;
template auto
Libint::calculateVuvBImpl<Utils::DerivativeOrder::One>(const libint2::BasisSet&,
                                                       const std::vector<std::pair<double, std::array<double, 3>>>&)
    -> std::vector<Eigen::MatrixXd>;
template auto
Libint::calculateVuvBImpl<Utils::DerivativeOrder::Two>(const libint2::BasisSet&,
                                                       const std::vector<std::pair<double, std::array<double, 3>>>&)
    -> std::vector<Eigen::MatrixXd>;
} // namespace Integrals
} // namespace Sparrow
} // namespace Scine
