/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef HARTREE_FOCK_LIBINT_H
#define HARTREE_FOCK_LIBINT_H

#include "IntegralTypeSpecifier.h"
#include "OneBodyIntegralsEvaluator.h"
#include <Eigen/Core>
#include <libint2.hpp>
#include <string>

namespace libint2 {
class Atom;
}

namespace Scine {
namespace Utils {
class AtomCollection;
struct GtoExpansion;
} // namespace Utils
namespace Sparrow {
namespace Integrals {

/**
 * Typedefs from
 * https://github.com/evaleev/libint
 * hartree-fock++.cc example.
 *
 * Import dense, dynamically
 * sized Matrix type from Eigen;
 * this is a matrix with
 * row-major storage
 * (http://en.wikipedia.org/wiki/Row-major_order)
 * to meet the layout of the integrals returned by the Libint integral library
 */
// using LibintMatrix = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>;
using DiagonalLibintMatrix = Eigen::DiagonalMatrix<double, Eigen::Dynamic, Eigen::Dynamic>;

/**
 * @brief Exception to throw if a non available operator is requested.
 */
class InvalidOperatorException : public std::exception {
 public:
  explicit InvalidOperatorException(unsigned operatorID)
    : errorMessage_("Invalid operator with ID " + std::to_string(operatorID) + "."){};
  const char* what() const noexcept final {
    return errorMessage_.c_str();
  }

 private:
  std::string errorMessage_;
};

class BasisSetController;

/**
 * @class Libint @file Libint.h
 * @brief Singleton to initialize and wrap up Libint.
 * Precalculates a set of libint2::Engines to be used throughout the calculation.
 * The libint2::Engines are pImpled away to make libint dependency private.
 * This class assumes that the maximal number of primitives per contracted Gaussian
 * shell is 20.
 */
class Libint {
 public:
  /**
   * @brief Getter for an instance of the Libint singleton.
   * The private constructor must initialize the Libint environment.
   */
  static Libint& getInstance();
  /**
   * Deleted functions neede for the singleton pattern.
   * @{
   */
  Libint(const Libint&) = delete;
  Libint& operator=(const Libint&) = delete;
  //! @}

  /**
   * @brief Converter from the Utils::AtomCollection to the std::vector<libint2::Atom>
   * @param structure A Scine::Utils::AtomCollection const-ref.
   * @return a const std::vector to libint2::Atom describing the structure.
   */
  static std::vector<libint2::Atom> toLibintAtoms(const Utils::AtomCollection& structure);

  /**
   * @brief Function returning the max number of threads in use.
   * @return the value of omp_get_max_threads() at initialization time of this class.
   */
  static int getMaxNumberThreads();
  /**
   * @brief Destructor. Must finalize Libint.
   */
  ~Libint();

  template<Operator op, Utils::DerivativeOrder order>
  typename OneBodyIntegralsResult<op>::type compute1BodyIntegrals(const BasisSetController& basisSet) const;

  /**
   * @brief returns the S_{\mu, \alpha} matrix with the corresponding derivatives if necessary
   */
  template<Utils::DerivativeOrder order>
  static auto computeOverlapWithCoreOrbitalOmx(int atom_a, const BasisSetController& basisSet,
                                               const Utils::GtoExpansion& coreOrbital) -> Eigen::MatrixXd;

  /*  template<Operator op, Utils::DerivativeOrder order>
    Eigen::MatrixXd compute(const libint2::Shell& shell1, const libint2::Shell& shell2, const libint2::Shell& shell3,
                            const libint2::Shell& shell4) const;*/
  Eigen::MatrixXd computeHCore(const BasisSetController& basisSet) const;

  /**
   * @brief Create a local basis with the basis functions of A
   * to calculate the one-center core-electron attraction V_{mu, nu}, B
   * mu, nu on A
   */
  static auto createLocalBasisOnAtom(int atom_a, const BasisSetController& basisSet) -> libint2::BasisSet;
  /**
   * Calculate the analytical 2-center core-electron attraction.
   */
  template<Utils::DerivativeOrder order>
  auto computeVuvB(int atom_a, int atom_b, double chargeOnB, const BasisSetController& basisSet) const
      -> std::vector<Eigen::MatrixXd>;

  libint2::Engine getEngine(IntegralTypeSpecifier intType, const BasisSetController& basisSet) const;
  libint2::Engine getEngine(IntegralTypeSpecifier intType, const libint2::Shell& shell1, const libint2::Shell& shell2,
                            const libint2::Shell& shell3, const libint2::Shell& shell4) const;

 private:
  template<Utils::DerivativeOrder order>
  static auto calculateVuvBImpl(const libint2::BasisSet& localBasis,
                                const std::vector<std::pair<double, std::array<double, 3>>>& nuclearCharges)
      -> std::vector<Eigen::MatrixXd>;
  /**
   * @brief Private constructor, part of the singleton pattern.
   */
  Libint();
  // Dependent on the basis.
  static constexpr int maxPrimitiveNumber = 20;
  // Dependent on how the libint2 library was generated. Use Libint2 macro
  static constexpr int maxAngularMomentum = LIBINT2_MAX_AM;
  // Saves the number of OMP threads to use
  static int nThreads_;
}; // namespace Integrals

template<Operator op, Utils::DerivativeOrder order>
typename OneBodyIntegralsResult<op>::type Libint::compute1BodyIntegrals(const BasisSetController& basisSet) const {
  auto result = OneBodyIntegralsEvaluator<op, order>::compute1BodyIntegrals(basisSet);
  return result;
}

} // namespace Integrals
} // namespace Sparrow
} // namespace Scine

#endif // HARTREE_FOCK_LIBINT_H
