/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#include "OneBodyIntegralsEvaluator.h"
#include "Libint.h"

namespace Scine {
namespace Sparrow {
namespace Integrals {
constexpr std::array<int, 1> OneBodyIntegralsResult<Operator::overlap>::relevantIndices;
constexpr std::array<int, 1> OneBodyIntegralsResult<Operator::kinetic>::relevantIndices;
constexpr std::array<int, 1> OneBodyIntegralsResult<Operator::nuclear>::relevantIndices;
constexpr std::array<int, 3> OneBodyIntegralsResult<Operator::emultipole1>::relevantIndices;
constexpr std::array<int, 6> OneBodyIntegralsResult<Operator::emultipole2>::relevantIndices;
constexpr std::array<int, 10> OneBodyIntegralsResult<Operator::emultipole3>::relevantIndices;

template<Operator op, Utils::DerivativeOrder order>
typename OneBodyIntegralsEvaluator<op, order>::ResultType
OneBodyIntegralsEvaluator<op, order>::compute1BodyIntegrals(const BasisSetController& basisSet) {
  // Initialize result (Template Parameter dependent!)
  ResultType result;

  if (order >= Utils::DerivativeOrder::Zero)
    computeZerothDerivative(basisSet, result);
  if (order >= Utils::DerivativeOrder::One)
    computeFirstDerivative(basisSet, result);
  if (order == Utils::DerivativeOrder::Two)
    computeSecondDerivative(basisSet, result);

  return result;
}

template<Operator op, Utils::DerivativeOrder order>
void OneBodyIntegralsEvaluator<op, order>::computeZerothDerivative(const BasisSetController& basisSet, ResultType& result) {
  const auto numberBasisFunctions = basisSet.nBasisFunction();
  // Indices of first basis funcion in shell.
  const auto indicesOfFirstBFinShell = basisSet.getBasisSet().shell2bf();
  result.reset(numberBasisFunctions);

  auto& libint = Libint::getInstance();

#pragma omp parallel
  {
    auto localEngine = libint.getEngine({op, order}, basisSet);
    // If nuclear, you need to have nuclear charges
    if (op == Operator::nuclear) {
      auto nuclearCharges = libint2::make_point_charges(Libint::toLibintAtoms(basisSet.getStructure()));
      localEngine.set_params(nuclearCharges);
    }
    const auto& buffer = localEngine.results();
    // number of operators * number of derivatives
    // const auto resultsDimension = localEngine.nshellsets();

    // If oparators are hermitian (here, they are) there is index permutational symmetry:
    // <a|O|b> = <b|O|a>, so loop only on half the matrix
#pragma omp for schedule(dynamic)
    for (int s1 = 0; s1 < basisSet.nShells(); ++s1) {
      int firstIndexOfS1 = indicesOfFirstBFinShell[s1];
      int sizeOfShell1 = basisSet.getBasisSet()[s1].size();

      for (int s2 = 0; s2 <= s1; ++s2) {
        int firstIndexOfS2 = indicesOfFirstBFinShell[s2];
        int sizeOfShell2 = basisSet.getBasisSet()[s2].size();

        localEngine.compute(basisSet.getBasisSet()[s1], basisSet.getBasisSet()[s2]);

        // If everything is screened out break out
        if (buffer[0] == nullptr) {
          break;
        }
        // Digest
        // Create an array with size equal to the number of required integral types
        // (1 for overlap, kinetic and nuclear, 3 for dipole, 6 for quadrupole, 10 for octupole)
        auto const& relevantIndices = OneBodyIntegralsResult<op>::relevantIndices;
        std::array<Eigen::MatrixXd, relevantIndices.size()> opDependentBuffer;
        for (int index : relevantIndices) {
          Eigen::Map<const Eigen::Matrix<double, -1, -1, Eigen::RowMajor>> bufferMatrix(buffer[index], sizeOfShell1,
                                                                                        sizeOfShell2);
          opDependentBuffer[index - relevantIndices[0]] = bufferMatrix;
        }
        fill<Utils::DerivativeOrder::Zero>(result, firstIndexOfS1, firstIndexOfS2, sizeOfShell1, sizeOfShell2, opDependentBuffer);
      }
    }
  }
}

template<Operator op, Utils::DerivativeOrder order>
void OneBodyIntegralsEvaluator<op, order>::computeFirstDerivative(const BasisSetController& /*basisSet*/, ResultType& /*result*/) {
  throw std::runtime_error("First derivatives not yet implemented!");
}
template<Operator op, Utils::DerivativeOrder order>
void OneBodyIntegralsEvaluator<op, order>::computeSecondDerivative(const BasisSetController& /*basisSet*/,
                                                                   ResultType& /*result*/) {
  throw std::runtime_error("Second derivatives not yet implemented!");
}

template class OneBodyIntegralsEvaluator<Operator::overlap, Utils::DerivativeOrder::Zero>;
template class OneBodyIntegralsEvaluator<Operator::overlap, Utils::DerivativeOrder::One>;
template class OneBodyIntegralsEvaluator<Operator::overlap, Utils::DerivativeOrder::Two>;
template class OneBodyIntegralsEvaluator<Operator::kinetic, Utils::DerivativeOrder::Zero>;
template class OneBodyIntegralsEvaluator<Operator::kinetic, Utils::DerivativeOrder::One>;
template class OneBodyIntegralsEvaluator<Operator::kinetic, Utils::DerivativeOrder::Two>;
template class OneBodyIntegralsEvaluator<Operator::nuclear, Utils::DerivativeOrder::Zero>;
template class OneBodyIntegralsEvaluator<Operator::nuclear, Utils::DerivativeOrder::One>;
template class OneBodyIntegralsEvaluator<Operator::nuclear, Utils::DerivativeOrder::Two>;
template class OneBodyIntegralsEvaluator<Operator::emultipole1, Utils::DerivativeOrder::Zero>;
template class OneBodyIntegralsEvaluator<Operator::emultipole1, Utils::DerivativeOrder::One>;
template class OneBodyIntegralsEvaluator<Operator::emultipole1, Utils::DerivativeOrder::Two>;
template class OneBodyIntegralsEvaluator<Operator::emultipole2, Utils::DerivativeOrder::Zero>;
template class OneBodyIntegralsEvaluator<Operator::emultipole2, Utils::DerivativeOrder::One>;
template class OneBodyIntegralsEvaluator<Operator::emultipole2, Utils::DerivativeOrder::Two>;
template class OneBodyIntegralsEvaluator<Operator::emultipole3, Utils::DerivativeOrder::Zero>;
template class OneBodyIntegralsEvaluator<Operator::emultipole3, Utils::DerivativeOrder::One>;
template class OneBodyIntegralsEvaluator<Operator::emultipole3, Utils::DerivativeOrder::Two>;

} // namespace Integrals
} // namespace Sparrow
} // namespace Scine
