/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef INTEGRALS_ONEBODYINTEGRALSEVALUATOR_H
#define INTEGRALS_ONEBODYINTEGRALSEVALUATOR_H

#include "BasisSetController.h"
#include "IntegralTypeSpecifier.h"
#include <Utils/DataStructures/DipoleMatrix.h>
#include <Utils/DataStructures/OctupoleMatrix.h>
#include <Utils/DataStructures/QuadrupoleMatrix.h>
namespace Scine {
namespace Sparrow {
namespace Integrals {

template<Operator op>
struct OneBodyIntegralsResult {
  static constexpr int size = 1;
  static constexpr std::array<int, size> relevantIndices{{0}};
  using type = Utils::MatrixWithDerivatives;
};

template<Operator op, Utils::DerivativeOrder order>
class OneBodyIntegralsEvaluator {
 public:
  /**
   * @brief Type of the result.
   */
  using ResultType = typename OneBodyIntegralsResult<op>::type;

  static ResultType compute1BodyIntegrals(const BasisSetController& basisSet);

 private:
  static void computeZerothDerivative(const BasisSetController& basisSet, ResultType& result);
  static void computeFirstDerivative(const BasisSetController& basisSet, ResultType& result);
  static void computeSecondDerivative(const BasisSetController& basisSet, ResultType& result);
};

//! Overloads for different kind of data structures containing the integrals.
template<Utils::DerivativeOrder order>
inline void fill(Utils::MatrixWithDerivatives& /*result*/, int /*firstIndexOfS1*/, int /*firstIndexOfS2*/,
                 int /*sizeOfShell1*/, int /*sizeOfShell2*/, const std::array<Eigen::MatrixXd, 1>& /*bufferMatrix*/) {
  // sum up all the derivative contributions (find right indices...)
  throw std::runtime_error("Derivatives not yet implemented!");
}
template<Utils::DerivativeOrder order>
inline void fill(Utils::DipoleMatrix& /*result*/, int /*firstIndexOfS1*/, int /*firstIndexOfS2*/, int /*sizeOfShell1*/,
                 int /*sizeOfShell2*/, const std::array<Eigen::MatrixXd, 3>& /*bufferMatrix*/) {
  throw std::runtime_error("Derivatives not yet implemented!");
}
template<Utils::DerivativeOrder order>
inline void fill(Utils::QuadrupoleMatrix& /*result*/, int /*firstIndexOfS1*/, int /*firstIndexOfS2*/,
                 int /*sizeOfShell1*/, int /*sizeOfShell2*/, const std::array<Eigen::MatrixXd, 6>& /*bufferMatrix*/) {
  throw std::runtime_error("Derivatives not yet implemented!");
}
template<Utils::DerivativeOrder order>
inline void fill(Utils::OctupoleMatrix& /*result*/, int /*firstIndexOfS1*/, int /*firstIndexOfS2*/,
                 int /*sizeOfShell1*/, int /*sizeOfShell2*/, const std::array<Eigen::MatrixXd, 10>& /*bufferMatrix*/) {
  throw std::runtime_error("Derivatives not yet implemented!");
}
//! Template specializations for the OneBodyIntegralsResult struct
//! @{
template<>
struct OneBodyIntegralsResult<Operator::overlap> {
  static constexpr int size = 1;
  static constexpr std::array<int, size> relevantIndices{{0}};
  using type = Utils::MatrixWithDerivatives;
};

template<>
struct OneBodyIntegralsResult<Operator::kinetic> {
  static constexpr int size = 1;
  static constexpr std::array<int, size> relevantIndices{{0}};
  using type = Utils::MatrixWithDerivatives;
};

template<>
struct OneBodyIntegralsResult<Operator::nuclear> {
  static constexpr int size = 1;
  static constexpr std::array<int, size> relevantIndices{{0}};
  using type = Utils::MatrixWithDerivatives;
};
template<>
struct OneBodyIntegralsResult<Operator::emultipole1> {
  static constexpr int size = 3;
  static constexpr std::array<int, size> relevantIndices{{1, 2, 3}};
  using type = Utils::DipoleMatrix;
};

template<>
struct OneBodyIntegralsResult<Operator::emultipole2> {
  static constexpr int size = 6;
  static constexpr std::array<int, size> relevantIndices{{4, 5, 6, 7, 8, 9}};
  using type = Utils::QuadrupoleMatrix;
};

template<>
struct OneBodyIntegralsResult<Operator::emultipole3> {
  static constexpr int size = 10;
  static constexpr std::array<int, size> relevantIndices{{10, 11, 12, 13, 14, 15, 16, 17, 18, 19}};
  using type = Utils::OctupoleMatrix;
};
//! @}

//! Template specialization for the fill overloads
//! @{
template<>
inline void fill<Utils::DerivativeOrder::Zero>(Utils::MatrixWithDerivatives& result, int firstIndexOfS1,
                                               int firstIndexOfS2, int sizeOfShell1, int sizeOfShell2,
                                               const std::array<Eigen::MatrixXd, 1>& bufferMatrix) {
  result.get<Utils::DerivativeOrder::Zero>().block(firstIndexOfS1, firstIndexOfS2, sizeOfShell1, sizeOfShell2) =
      bufferMatrix[0];
  result.get<Utils::DerivativeOrder::Zero>().block(firstIndexOfS2, firstIndexOfS1, sizeOfShell2, sizeOfShell1) =
      bufferMatrix[0].transpose();
}
template<>
inline void fill<Utils::DerivativeOrder::Zero>(Utils::DipoleMatrix& result, int firstIndexOfS1, int firstIndexOfS2,
                                               int sizeOfShell1, int sizeOfShell2,
                                               const std::array<Eigen::MatrixXd, 3>& bufferMatrix) {
  for (int dimension = 0; dimension < 3; ++dimension) {
    result.at(dimension).get<Utils::DerivativeOrder::Zero>().block(firstIndexOfS1, firstIndexOfS2, sizeOfShell1,
                                                                   sizeOfShell2) = bufferMatrix[dimension];
    result.at(dimension).get<Utils::DerivativeOrder::Zero>().block(firstIndexOfS2, firstIndexOfS1, sizeOfShell2,
                                                                   sizeOfShell1) = bufferMatrix[dimension].transpose();
  }
}
template<>
inline void fill<Utils::DerivativeOrder::Zero>(Utils::QuadrupoleMatrix& result, int firstIndexOfS1, int firstIndexOfS2,
                                               int sizeOfShell1, int sizeOfShell2,
                                               const std::array<Eigen::MatrixXd, 6>& bufferMatrix) {
  for (int dimension = 0; dimension < 6; ++dimension) {
    result.at(dimension).get<Utils::DerivativeOrder::Zero>().block(firstIndexOfS1, firstIndexOfS2, sizeOfShell1,
                                                                   sizeOfShell2) = bufferMatrix[dimension];
    result.at(dimension).get<Utils::DerivativeOrder::Zero>().block(firstIndexOfS2, firstIndexOfS1, sizeOfShell2,
                                                                   sizeOfShell1) = bufferMatrix[dimension].transpose();
  }
}
template<>
inline void fill<Utils::DerivativeOrder::Zero>(Utils::OctupoleMatrix& result, int firstIndexOfS1, int firstIndexOfS2,
                                               int sizeOfShell1, int sizeOfShell2,
                                               const std::array<Eigen::MatrixXd, 10>& bufferMatrix) {
  for (int dimension = 0; dimension < 10; ++dimension) {
    result.at(dimension).get<Utils::DerivativeOrder::Zero>().block(firstIndexOfS1, firstIndexOfS2, sizeOfShell1,
                                                                   sizeOfShell2) = bufferMatrix[dimension];
    result.at(dimension).get<Utils::DerivativeOrder::Zero>().block(firstIndexOfS2, firstIndexOfS1, sizeOfShell2,
                                                                   sizeOfShell1) = bufferMatrix[dimension].transpose();
  }
}
//! @}

} // namespace Integrals
} // namespace Sparrow
} // namespace Scine

#endif // INTEGRALS_ONEBODYINTEGRALSEVALUATOR_H
