/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "AO2MOTransformation.h"
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Utils/DataStructures/MolecularOrbitals.h>

namespace Scine {
namespace Sparrow {
namespace Integrals {
namespace AO2MOTransformation {

template<>
auto transform1eIntegrals<Utils::Reference::Restricted>(const Eigen::MatrixXd& hCore,
                                                        const Utils::MolecularOrbitals& molecularOrbitals)
    -> Utils::SecondQuantization::MOTypeTrait<Utils::Reference::Restricted>::OneElectronMatrixType {
  if (molecularOrbitals.isUnrestricted()) {
    throw std::runtime_error("Restricted AO2MO requested with unrestricted Molecular Orbitals");
  }
  return molecularOrbitals.restrictedMatrix().transpose() * hCore * molecularOrbitals.restrictedMatrix();
}
template<>
auto transform1eIntegrals<Utils::Reference::Unrestricted>(const Eigen::MatrixXd& hCore,
                                                          const Utils::MolecularOrbitals& molecularOrbitals)
    -> Utils::SecondQuantization::MOTypeTrait<Utils::Reference::Unrestricted>::OneElectronMatrixType {
  Utils::SecondQuantization::MOTypeTrait<Utils::Reference::Unrestricted>::OneElectronMatrixType moMatrix;
  moMatrix[Utils::SecondQuantization::SpinComponent::Alpha] =
      molecularOrbitals.alphaMatrix().transpose() * hCore * molecularOrbitals.alphaMatrix();
  moMatrix[Utils::SecondQuantization::SpinComponent::Beta] =
      molecularOrbitals.betaMatrix().transpose() * hCore * molecularOrbitals.betaMatrix();
  return moMatrix;
}

auto generateBlockTransformation(int atomToTransform, const Eigen::MatrixXd& mos, const Utils::AtomsOrbitalsIndexes& aoInfos)
    -> Eigen::MatrixXd {
  int nOrb = int(mos.cols());
  int nAOs = aoInfos.getNOrbitals(atomToTransform);
  int dimAO = (nAOs + 1) * nAOs / 2;
  int dimMO = nOrb * nOrb;
  int firstAO = aoInfos.getFirstOrbitalIndex(atomToTransform);
  Eigen::MatrixXd transformationMatrix = Eigen::MatrixXd::Zero(dimAO, dimMO);
  for (int p = 0; p < nOrb; ++p) {
    for (int q = 0; q < nOrb; ++q) {
      for (int mu = 0; mu < nAOs; ++mu) {
        for (int nu = 0; nu < nAOs; ++nu) {
          auto bra = Orbitals::pair(Orbitals::Orbital(mu), Orbitals::Orbital(nu));
          transformationMatrix(bra, p * nOrb + q) += mos(mu + firstAO, p) * mos(nu + firstAO, q);
        }
      }
    }
  }
  return transformationMatrix;
}

auto transformBlock(const Eigen::MatrixXd& eriBlock, const Eigen::MatrixXd& transMatA, const Eigen::MatrixXd& transMatB)
    -> Eigen::MatrixXd {
  return transMatA.transpose() * eriBlock * transMatB;
}

template<>
auto generateTransformationMatrices<Utils::Reference::Restricted>(const Utils::MolecularOrbitals& mos,
                                                                  const Utils::AtomsOrbitalsIndexes& aoInfos)
    -> std::vector<Utils::SpinAdaptedContainer<Utils::Reference::Restricted, Eigen::MatrixXd>> {
  if (mos.isUnrestricted()) {
    throw std::runtime_error("Restricted AO2MO requested with unrestricted Molecular Orbitals");
  }
  using ContainerType = Utils::SpinAdaptedContainer<Utils::Reference::Restricted, Eigen::MatrixXd>;
  std::vector<ContainerType> transformationMatrices(aoInfos.getNAtoms());
#pragma omp parallel for
  for (int i = 0; i < aoInfos.getNAtoms(); ++i) {
    transformationMatrices[i] = ContainerType{generateBlockTransformation(i, mos.restrictedMatrix(), aoInfos)};
  }
  return transformationMatrices;
}

template<>
auto generateTransformationMatrices<Utils::Reference::Unrestricted>(const Utils::MolecularOrbitals& mos,
                                                                    const Utils::AtomsOrbitalsIndexes& aoInfos)
    -> std::vector<Utils::SpinAdaptedContainer<Utils::Reference::Unrestricted, Eigen::MatrixXd>> {
  auto mosTmp = mos;
  if (mos.isRestricted()) { // split the matrix in alpha and beta densities if restricted
    mosTmp = mos.toUnrestricted();
  }
  using ContainerType = Utils::SpinAdaptedContainer<Utils::Reference::Unrestricted, Eigen::MatrixXd>;
  std::vector<ContainerType> transformationMatrices(aoInfos.getNAtoms());
#pragma omp parallel for
  for (int i = 0; i < aoInfos.getNAtoms(); ++i) {
    transformationMatrices[i] = ContainerType{generateBlockTransformation(i, mosTmp.alphaMatrix(), aoInfos),
                                              generateBlockTransformation(i, mosTmp.betaMatrix(), aoInfos)};
  }
  return transformationMatrices;
}

template<>
auto ao2mo<Utils::Reference::Restricted>(const EriType& eri, const Utils::MolecularOrbitals& mos,
                                         const Utils::AtomsOrbitalsIndexes& aoInfos)
    -> Utils::SecondQuantization::MOTypeTrait<Utils::Reference::Restricted>::Type {
  int nOrb = mos.numberOrbitals();
  int dimMO = nOrb * nOrb;
  Eigen::MatrixXd moEri = Eigen::MatrixXd::Zero(dimMO, dimMO);

  if (mos.isUnrestricted()) {
    throw std::runtime_error("Restricted AO2MO requested with unrestricted Molecular Orbitals");
  }

  auto transformationMatrices = generateTransformationMatrices<Utils::Reference::Restricted>(mos, aoInfos);

  for (const auto& eriBlock : eri) {
    Eigen::MatrixXd transformedBlock =
        transformBlock(eriBlock.second.get0(), transformationMatrices[eriBlock.first.first].restricted,
                       transformationMatrices[eriBlock.first.second].restricted);
    moEri += transformedBlock;
    if (eriBlock.first.first != eriBlock.first.second) {
      moEri += transformedBlock.transpose();
    }
  }
  return moEri;
}

template<>
auto ao2mo<Utils::Reference::Unrestricted>(const EriType& eri, const Utils::MolecularOrbitals& mos,
                                           const Utils::AtomsOrbitalsIndexes& aoInfos)
    -> Utils::SecondQuantization::MOTypeTrait<Utils::Reference::Unrestricted>::Type {
  int nOrb = mos.numberOrbitals();
  int dimMO = nOrb * nOrb;
  Utils::SecondQuantization::MOTypeTrait<Utils::Reference::Unrestricted>::Type moEri;
  moEri[{Utils::SecondQuantization::SpinComponent::Alpha, Utils::SecondQuantization::SpinComponent::Alpha}] =
      Eigen::MatrixXd::Zero(dimMO, dimMO);
  moEri[{Utils::SecondQuantization::SpinComponent::Alpha, Utils::SecondQuantization::SpinComponent::Beta}] =
      Eigen::MatrixXd::Zero(dimMO, dimMO);
  moEri[{Utils::SecondQuantization::SpinComponent::Beta, Utils::SecondQuantization::SpinComponent::Alpha}] =
      Eigen::MatrixXd::Zero(dimMO, dimMO);
  moEri[{Utils::SecondQuantization::SpinComponent::Beta, Utils::SecondQuantization::SpinComponent::Beta}] =
      Eigen::MatrixXd::Zero(dimMO, dimMO);
  auto transformationMatrices = generateTransformationMatrices<Utils::Reference::Unrestricted>(mos, aoInfos);

  for (const auto& eriBlock : eri) {
    Eigen::MatrixXd transformedBlockAlphaAlpha =
        transformBlock(eriBlock.second.get0(), transformationMatrices[eriBlock.first.first].alpha,
                       transformationMatrices[eriBlock.first.second].alpha);
    Eigen::MatrixXd transformedBlockAlphaBeta =
        transformBlock(eriBlock.second.get0(), transformationMatrices[eriBlock.first.first].alpha,
                       transformationMatrices[eriBlock.first.second].beta);
    Eigen::MatrixXd transformedBlockBetaAlpha =
        transformBlock(eriBlock.second.get0(), transformationMatrices[eriBlock.first.first].beta,
                       transformationMatrices[eriBlock.first.second].alpha);
    Eigen::MatrixXd transformedBlockBetaBeta =
        transformBlock(eriBlock.second.get0(), transformationMatrices[eriBlock.first.first].beta,
                       transformationMatrices[eriBlock.first.second].beta);
    moEri[{Utils::SecondQuantization::SpinComponent::Alpha, Utils::SecondQuantization::SpinComponent::Alpha}] +=
        transformedBlockAlphaAlpha;
    moEri[{Utils::SecondQuantization::SpinComponent::Alpha, Utils::SecondQuantization::SpinComponent::Beta}] +=
        transformedBlockAlphaBeta;
    moEri[{Utils::SecondQuantization::SpinComponent::Beta, Utils::SecondQuantization::SpinComponent::Alpha}] +=
        transformedBlockBetaAlpha;
    moEri[{Utils::SecondQuantization::SpinComponent::Beta, Utils::SecondQuantization::SpinComponent::Beta}] +=
        transformedBlockBetaBeta;
    if (eriBlock.first.first != eriBlock.first.second) {
      moEri[{Utils::SecondQuantization::SpinComponent::Alpha, Utils::SecondQuantization::SpinComponent::Alpha}] +=
          transformedBlockAlphaAlpha.transpose();
      moEri[{Utils::SecondQuantization::SpinComponent::Alpha, Utils::SecondQuantization::SpinComponent::Beta}] +=
          transformedBlockAlphaBeta.transpose();
      moEri[{Utils::SecondQuantization::SpinComponent::Beta, Utils::SecondQuantization::SpinComponent::Alpha}] +=
          transformedBlockBetaAlpha.transpose();
      moEri[{Utils::SecondQuantization::SpinComponent::Beta, Utils::SecondQuantization::SpinComponent::Beta}] +=
          transformedBlockBetaBeta.transpose();
    }
  }
  return moEri;
}
} // namespace AO2MOTransformation
} // namespace Integrals
} // namespace Sparrow
} // namespace Scine
