/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef INTEGRALS_AO2MOTRANSFORMER_H
#define INTEGRALS_AO2MOTRANSFORMER_H

#include "Orbitals.h"
#include <Utils/DataStructures/SecondQuantization/CasSpecifier.h>
#include <Utils/DataStructures/SecondQuantization/EriUtilities.h>

namespace Scine {
namespace Utils {
class MolecularOrbitals;
class AtomsOrbitalsIndexes;
} // namespace Utils
namespace Sparrow {
namespace Integrals {

/**
 * @class AO2MOTransformer @file AO2MOTransformer.h
 * @brief Class performing the 4-index transformation of the two electron integrals.
 * This class can take either eris and molecular orbitals, or eris and a CasSpecifier<Restrictedness::restricted>.
 * With the CasSpecifier, only the integrals where all 4 molecular orbitals are within the cas are transformed.
 * Indices after the transformation are not consistent anymore.
 * @tparam symmetry The symmetry with which the eris are saved.
 */
namespace AO2MOTransformation {

/**
 * @brief Transforms a (self-adjoint) matrix containing the integrals in AO to MO.
 * @param hCore Matrix containing the one-electron integrals.
 * @param molecularOrbitals The expansion coefficients AO to MO.
 * @return A matrix containing the transformed integrals in MO basis.
 */
template<Utils::Reference restrictedness>
auto transform1eIntegrals(const Eigen::MatrixXd& hCore, const Utils::MolecularOrbitals& molecularOrbitals) ->
    typename Utils::SecondQuantization::MOTypeTrait<restrictedness>::OneElectronMatrixType;

/**
 * @brief Transforms a (self-adjoint) matrix containing the integrals in AO to MO within a CAS.
 * @param hCore Matrix containing the one-electron integrals.
 * @param specifier A restricted CasSpecifier containing the infos about a CAS.
 * @return A matrix containing the transformed integrals in MO basis within a cas.
 * The dimension of the returned matrix is the same as the input matrix, but the columns and rows corresponding to
 * molecular orbitals outside the CAS are set to 0.
 */
template<Utils::Reference restrictedness>
auto transform1eIntegrals(const Eigen::MatrixXd& hCore, const Utils::SecondQuantization::CasSpecifier<restrictedness>& specifier)
    -> typename Utils::SecondQuantization::MOTypeTrait<restrictedness>::OneElectronMatrixType {
  return transform1eIntegrals<restrictedness>(hCore, specifier.getActiveOrbitalsCoefficients());
}

/**
 * @brief transforms the two-electrons integrals from AO to MO basis (eventually, within a CAS).
 * The algorithm is better than performing 4 one-index-transformations. Scales better than O(N^5).
 * Theoretically, this takes advantage of the NDDO approximation and eventual CAS.
 * Works like this:
 * - create all transformation matrices (1 or 2 per atom, depending on restrictedness).
 *   This loops over all MO pairs and all ERI on a diagonal {atom_a, atom_a} block.
 *   Since the number of AOs on an atom is at most 4, this has a cost that is linear in the
 *   number of atoms and quadratic in the number of MO -> O(N MO^2)
 * - loop over atom pairs and transform the {atom_a, atom_b} block (atom_a can be also equal
 *   to atom_b) in MO, and sum it up. There are O(N^2) atom pairs and each transformation costs
 *   2 matrix multiply O(AOOnOneAtom^2 * MO^2) + O(AOOnOneAtom^2 * MO^2), and since the AOs^2 dimension on one
 *   atom is a constant number (max 10 if s and p basis functions are present), then the transformation scales
 *   as O(MO^2).
 *   This needs to be done O(N^2) times, so the scaling is O(N^2 * MO^2)
 * - The total scaling is then the sum of the two, which gives an overall scaling of O(N^2 * MO^2).
 */
template<Utils::Reference restrictedness>
auto ao2mo(const EriType& eri, const Utils::MolecularOrbitals& mos, const Utils::AtomsOrbitalsIndexes& aoInfos) ->
    typename Utils::SecondQuantization::MOTypeTrait<restrictedness>::Type;

/**
 * @brief Generates the AOAO->MOMO 1-el transformation supermatrix
 * The super matrix has as rows the AOxAO indices (ss or ss,sx,sy,sz,xx,xy,xz,yy,yz,zz, or ...).
 * of the AOs centered on a single atom (atomToTransform).
 * MOs are transformed without symmetry -> nMOs*nMOs column dimension.
 */
auto generateBlockTransformation(int atomToTransform, const Eigen::MatrixXd& mos, const Utils::AtomsOrbitalsIndexes& aoInfos)
    -> Eigen::MatrixXd;
/**
 * @brief Transforms a diatomic block of the eri tensor.
 * In NDDO only 1 and 2 center integrals exist. Therefore, one can save them
 * in single- and diatomic blocks. These blocks can be independently transformed.
 * This function transforms a diatomic block from AO to MOs.
 */
auto transformBlock(const Eigen::MatrixXd& eriBlock, const Eigen::MatrixXd& transMatA, const Eigen::MatrixXd& transMatB)
    -> Eigen::MatrixXd;

/**
 * @brief Generates the transformation matrices for all atoms
 * The transformation matrices are the matrices that transform the AOxAO block on each
 * atom into MOxMO
 * @tparamm restrictedness Restricted or Unrestricted
 * If Restricted, only a single matrix per atom will be calculated, if Unrestricted Alpha and Beta
 * transformation matrices are calculated.
 */
template<Utils::Reference restrictedness>
auto generateTransformationMatrices(const Utils::MolecularOrbitals& mos, const Utils::AtomsOrbitalsIndexes& aoInfos)
    -> std::vector<Utils::SpinAdaptedContainer<restrictedness, Eigen::MatrixXd>>;

} // namespace AO2MOTransformation
} // namespace Integrals
} // namespace Sparrow
} // namespace Scine

#endif // INTEGRALS_AO2MOTRANSFORMER_H
