/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#include "NddoEriSaverDigester.h"
#include "../Libint.h"
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>

namespace Scine {
namespace Sparrow {
namespace Integrals {

NddoEriSaverDigester::NddoEriSaverDigester(const BasisSetController& basisSet)
  : basisSet_(basisSet),
    indexFirstBFInShell_(basisSet_.getBasisSet().shell2bf()),
    aoInfos_(std::make_unique<Utils::AtomsOrbitalsIndexes>(basisSet_.toAtomsOrbitalsIndexes())) {
}

NddoEriSaverDigester::~NddoEriSaverDigester() = default;

NddoEriSaverDigester::NddoEriSaverDigester(NddoEriSaverDigester&& rhs) noexcept = default;

const EriType& NddoEriSaverDigester::getResult() const {
  return eri_;
}

void NddoEriSaverDigester::initialize(int /*numberThreads*/) {
  auto orbitalPairSize = [&](int nOrbitals) {
    return nOrbitals == 1
               ? 1
               : (nOrbitals == 4 ? 10 : throw std::runtime_error("Only s and p orbitals supported in frozen-core minimal basis!"));
  };
  for (int atom_i = 0; atom_i < aoInfos_->getNAtoms(); ++atom_i) {
    for (int atom_j = 0; atom_j < atom_i; ++atom_j) {
      const int sizeI = orbitalPairSize(aoInfos_->getNOrbitals(atom_i));
      const int sizeJ = orbitalPairSize(aoInfos_->getNOrbitals(atom_j));
      eri_[{atom_i, atom_j}].setOrder(Utils::DerivativeOrder::Zero);
      eri_[{atom_i, atom_j}].setDimension(sizeI, sizeJ);
    }
  }
}

void NddoEriSaverDigester::digest(const Eigen::RowVectorXd& bufferVector, IndexType shellIndices, int atom_i, int atom_j) {
  size_t bufferIndex = 0;
  const auto& basis = basisSet_.getBasisSet();

  Eigen::MatrixXd& pairwiseMatrix = eri_[{atom_i, atom_j}].get0();

  const int firstBfOfAtomI = aoInfos_->getFirstOrbitalIndex(atom_i);
  const int firstBfOfAtomJ = aoInfos_->getFirstOrbitalIndex(atom_j);

  // first atom
  for (int functionInShell1 = 0; functionInShell1 < int(basis[shellIndices[0]].size()); ++functionInShell1) {
    const int basisFunction1 = functionInShell1 + int(indexFirstBFInShell_[shellIndices[0]]);
    auto orb1 = Orbitals::Orbital(basisFunction1 - firstBfOfAtomI);
    for (int functionInShell2 = 0; functionInShell2 < int(basis[shellIndices[1]].size()); ++functionInShell2) {
      const int basisFunction2 = functionInShell2 + int(indexFirstBFInShell_[shellIndices[1]]);
      auto orb2 = Orbitals::Orbital(basisFunction2 - firstBfOfAtomI);
      auto braTerm = Orbitals::pair(orb1, orb2);

      // second atom
      for (int functionInShell3 = 0; functionInShell3 < int(basis[shellIndices[2]].size()); ++functionInShell3) {
        const int basisFunction3 = functionInShell3 + int(indexFirstBFInShell_[shellIndices[2]]);
        auto orb3 = Orbitals::Orbital(basisFunction3 - firstBfOfAtomJ);
        for (int functionInShell4 = 0; functionInShell4 < int(basis[shellIndices[3]].size()); ++functionInShell4) {
          const int basisFunction4 = functionInShell4 + int(indexFirstBFInShell_[shellIndices[3]]);
          auto orb4 = Orbitals::Orbital(basisFunction4 - firstBfOfAtomJ);
          auto ketTerm = Orbitals::pair(orb3, orb4);

          auto const integral = bufferVector(bufferIndex);
          ++bufferIndex;

          assert(braTerm < pairwiseMatrix.rows());
          assert(ketTerm < pairwiseMatrix.cols());

          pairwiseMatrix(braTerm, ketTerm) = integral;
        }
      }
    }
  }
}

void NddoEriSaverDigester::digest(const Eigen::MatrixXd& oneCenterBlock, int atom_i) {
  assert(oneCenterBlock.size() != 0);
  eri_[{atom_i, atom_i}].setBaseMatrix(oneCenterBlock);
}

} // namespace Integrals
} // namespace Sparrow
} // namespace Scine
