/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef INTEGRALS_TWOBODIESINTEGRALSEVALUATOR_H
#define INTEGRALS_TWOBODIESINTEGRALSEVALUATOR_H

#include "NddoEriSaverDigester.h"
#include "NddoOneCenterTwoElectronEvaluator.h"
#include "Utils/Math/DerivOrderEnum.h"
#include <Integrals/IntegralTypeSpecifier.h>
#include <Integrals/Libint.h>
#include <memory>
namespace Scine {
namespace Sparrow {
namespace Integrals {

class NddoTwoBodiesIntegralsEvaluator {
 public:
  NddoTwoBodiesIntegralsEvaluator(const BasisSetController& basisSet) : basisSet_(basisSet), digester_(basisSet_) {
  }

  void set(std::shared_ptr<NddoOneCenterTwoElectronEvaluator> evaluator) {
    oneCenterEvaluator_ = std::move(evaluator);
  }

  const auto& getResult() const {
    return digester_.getResult();
  }

  template<Operator op, Utils::DerivativeOrder order>
  void evaluate() {
    digester_.initialize(Libint::getMaxNumberThreads());
    evaluateOneCenter<op>();
    evaluateTwoCenter<op, order>();
  }

  template<Operator op>
  void evaluateOneCenter() {
    if (oneCenterEvaluator_) {
      const int nAtoms = basisSet_.getStructure().getElements().size();
      for (int atomIndex = 0; atomIndex < nAtoms; ++atomIndex) {
        oneCenterEvaluator_->evaluate(digester_, op, atomIndex);
      }
    }
  }

  template<Operator op, Utils::DerivativeOrder order>
  void evaluateTwoCenter() {
    constexpr const int orderAsInt =
        order == Utils::DerivativeOrder::Zero ? 0 : (order == Utils::DerivativeOrder::One ? 1 : 2);
    const int nDerivsOnShellset = libint2::num_geometrical_derivatives(4, orderAsInt);
    const auto& basis = basisSet_.getBasisSet();
    auto& libint = Libint::getInstance();

    auto localEngine = libint.getEngine({op, order}, basisSet_);
    auto const& buffer = localEngine.results();
    const int nAtoms = basisSet_.getStructure().getElements().size();
    auto atomsToShells = basisSet_.atomsToShells();

    auto evaluateAtomPair = [&](int atomI, int atomJ) {
      auto shells_i = atomsToShells[atomI];
      auto shells_j = atomsToShells[atomJ];
      using IntType = decltype(shells_i)::size_type;
      for (IntType i1 = 0; i1 < shells_i.size(); ++i1) {
        const auto shellIndex1 = int(shells_i[i1]);
        const auto i1Size = basis[shellIndex1].size();
        for (IntType i2 = 0; i2 <= i1; ++i2) {
          const auto shellIndex2 = int(shells_i[i2]);
          const auto i2Size = basis[shellIndex2].size();
          for (IntType j1 = 0; j1 < shells_j.size(); ++j1) {
            const auto shellIndex3 = int(shells_j[j1]);
            const auto j1Size = basis[shellIndex3].size();
            for (IntType j2 = 0; j2 <= j1; ++j2) {
              const auto shellIndex4 = int(shells_j[j2]);
              const auto j2Size = basis[shellIndex4].size();
              localEngine.compute2<toLibintOperator(op), libint2::BraKet::xx_xx, orderAsInt>(
                  basis[shellIndex1], basis[shellIndex2], basis[shellIndex3], basis[shellIndex4]);

              /* Everything is screened out */
              if (buffer[0] == nullptr) {
                continue;
              }

              if (order == Utils::DerivativeOrder::Zero) {
                Eigen::RowVectorXd bufferVector(i1Size * i2Size * j1Size * j2Size);
                bufferVector = Eigen::Map<const Eigen::RowVectorXd>(buffer[0], i1Size * i2Size * j1Size * j2Size);
                digester_.digest(bufferVector, {{shellIndex1, shellIndex2, shellIndex3, shellIndex4}}, atomI, atomJ);
              }
              else if (order == Utils::DerivativeOrder::One) {
                throw std::runtime_error("No first derivatives implemented yet!");
              }
              else {
                throw std::runtime_error("No second derivatives implemented yet!");
              }
            }
          }
        }
      }
    };

    for (int atom_i = 0; atom_i < nAtoms; ++atom_i) {
      for (int atom_j = 0; atom_j < atom_i; ++atom_j) {
        evaluateAtomPair(atom_i, atom_j);
      }
    }
  }

 private:
  const BasisSetController& basisSet_;
  std::shared_ptr<NddoOneCenterTwoElectronEvaluator> oneCenterEvaluator_;
  NddoEriSaverDigester digester_;
};

} // namespace Integrals
} // namespace Sparrow
} // namespace Scine

#endif // INTEGRALS_TWOBODIESINTEGRALSEVALUATOR_H
