/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef INTEGRALS_ERIUTILITIES_H
#define INTEGRALS_ERIUTILITIES_H

#include <Utils/DataStructures/MatrixWithDerivatives.h>
#include <Eigen/Core>
#include <array>
#include <boost/functional/hash.hpp>
#include <map>
#include <unordered_map>
#include <vector>

namespace Scine {
namespace Sparrow {
namespace Integrals {

namespace Orbitals {
/**
 *  @brief enum indicating the possible orbitals
 *
 *  orb_t::s A s orbital
 *  orb_t::x A p_x orbital
 *  orb_t::y A p_y orbital
 *  orb_t::z A p_z orbital
 *  orb_t::x2y2 A d_{x^2-y^2} orbital
 *  orb_t::xz A d_{xz} orbital
 *  orb_t::z2 A d_{z^2} orbital
 *  orb_t::yz A d_{yz} orbital
 *  orb_t::xy A d_{xy} orbital
 */
enum Orbital : int { s = 0, x = 1, y = 2, z = 3, x2y2 = 4, xz = 5, z2 = 6, yz = 7, xy = 8, none_orbital = -1 };

/**
 * @brief enum listing all of the orbital pairs up to orbital p
 * They are ordered in such a way that if a model only has s and p orbitals,
 * they can just work with the first 10 elements.
 */
enum OrbitalPair : int { // 1 ss, 9 sp/pp, 30 sd/pd/dd
  s_s = 0,
  s_x = 1,
  s_y = 2,
  s_z = 3,
  x_x = 4,
  x_y = 5,
  x_z = 6,
  y_y = 7,
  y_z = 8,
  z_z = 9,
  s_x2y2 = 10,
  s_xz = 11,
  s_z2 = 12,
  s_yz = 13,
  s_xy = 14,
  x_x2y2 = 15,
  x_xz = 16,
  x_z2 = 17,
  // x_yz = 18,
  x_xy = 18,
  y_x2y2 = 19,
  // y_xz = 21,
  y_z2 = 20,
  y_yz = 21,
  y_xy = 22,
  // z_x2y2 = 25,
  z_xz = 23,
  z_z2 = 24,
  z_yz = 25,
  // z_xy = 29,
  x2y2_x2y2 = 26,
  x2y2_xz = 27,
  x2y2_z2 = 28,
  x2y2_yz = 29,
  // x2y2_xy = 34,
  xz_xz = 30,
  xz_z2 = 31,
  xz_yz = 32,
  xz_xy = 33,
  z2_z2 = 34,
  z2_yz = 35,
  z2_xy = 36,
  yz_yz = 37,
  yz_xy = 38,
  xy_xy = 39,
  none_pair_orbital = -1
};

// clang-format off
static constexpr std::array<std::array<OrbitalPair, 9>, 9> orbitalPairs{
    {{OrbitalPair::s_s              , OrbitalPair::s_x              , OrbitalPair::s_y              , OrbitalPair::s_z, OrbitalPair::s_x2y2, OrbitalPair::s_xz, OrbitalPair::s_z2, OrbitalPair::s_yz, OrbitalPair::s_xy},
     {OrbitalPair::none_pair_orbital, OrbitalPair::x_x              , OrbitalPair::x_y              , OrbitalPair::x_z, OrbitalPair::x_x2y2, OrbitalPair::x_xz, OrbitalPair::x_z2, OrbitalPair::none_pair_orbital, OrbitalPair::x_xy},
     {OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::y_y              , OrbitalPair::y_z, OrbitalPair::y_x2y2, OrbitalPair::none_pair_orbital, OrbitalPair::y_z2, OrbitalPair::y_yz, OrbitalPair::y_xy},
     {OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::z_z, OrbitalPair::none_pair_orbital, OrbitalPair::z_xz, OrbitalPair::z_z2, OrbitalPair::z_yz, OrbitalPair::none_pair_orbital},
     {OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::x2y2_x2y2, OrbitalPair::x2y2_xz, OrbitalPair::x2y2_z2, OrbitalPair::x2y2_yz, OrbitalPair::none_pair_orbital},
     {OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::xz_xz, OrbitalPair::xz_z2, OrbitalPair::xz_yz, OrbitalPair::xz_xy},
     {OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::z2_z2, OrbitalPair::z2_yz, OrbitalPair::z2_xy},
     {OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::yz_yz, OrbitalPair::yz_xy},
     {OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::none_pair_orbital, OrbitalPair::xy_xy}
    }};
// clang-format on

/**
 * @brief Function to get a bra or ket from two orbitals. Just up to p shells.
 */
inline constexpr OrbitalPair pair(Orbital first, Orbital second) {
  return orbitalPairs[std::min(first, second)][std::max(first, second)];
}

} // namespace Orbitals

using DiatomicKey = std::pair<int, int>;
using EriType = std::unordered_map<DiatomicKey, Utils::MatrixWithDerivatives, boost::hash<DiatomicKey>>;

/**
 * @brief forces structure i >= j.
 */
inline void canonicalize(DiatomicKey& index) {
  if (index.first < index.second) {
    std::swap(index.first, index.second);
  }
}

} // namespace Integrals
} // namespace Sparrow
} // namespace Scine

#endif // INTEGRALS_ERIUTILITIES_H
