/**
 * @file SeconQuantizationUtilsTest.cpp
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include <Integrals/TwoBodiesIntegrals/AO2MOTransformation.h>
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Utils/DataStructures/MolecularOrbitals.h>
#include <Utils/DataStructures/SecondQuantization/CasSpecifier.h>
#include <Utils/DataStructures/SingleParticleEnergies.h>
#include <gmock/gmock.h>
#include <vector>

namespace Scine {
namespace Sparrow {
namespace Integrals {

using namespace testing;
using namespace Orbitals;

/**
 * @test ACasTest @file CasTest.cpp
 * Tests for the generation of CAS spaces
 */

class AnAO2MOTest : public Test {
 public:
  EriType arbitraryEris;
  int arbitraryNumberOfFullMOs;
  int arbitraryNumberOfElectrons = 4;
  int arbitraryNumberOfAOs;
  Utils::AtomsOrbitalsIndexes aoInfos;
  Utils::MolecularOrbitals arbitraryMos;
  void SetUp() final {
    Utils::MatrixWithDerivatives blockAA, blockAB, blockBB;
    blockAA.get0() = Eigen::MatrixXd::Random(10, 10);
    blockAB.get0() = Eigen::MatrixXd::Random(1, 10);
    blockBB.get0() = Eigen::MatrixXd::Random(1, 1);
    blockAA.get0() = blockAA.get0().selfadjointView<Eigen::Lower>();
    blockBB.get0() = blockBB.get0().selfadjointView<Eigen::Lower>();
    DiatomicKey AA{0, 0}, AB{1, 0}, BB{1, 1};
    canonicalize(AA);
    canonicalize(AB);
    canonicalize(BB);
    arbitraryEris[AA] = blockAA;
    arbitraryEris[AB] = blockAB;
    arbitraryEris[BB] = blockBB;
    arbitraryNumberOfAOs = 5;
    arbitraryNumberOfFullMOs = 5;

    arbitraryMos = Utils::MolecularOrbitals::createFromRestrictedCoefficients(
        Eigen::MatrixXd::Random(arbitraryNumberOfFullMOs, arbitraryNumberOfFullMOs));
    aoInfos.addAtom(4);
    aoInfos.addAtom(1);
  }

 private:
};

TEST_F(AnAO2MOTest, AO2MORestricted) {
  int nMOs = arbitraryNumberOfFullMOs;

  Eigen::MatrixXd erisMO = Eigen::MatrixXd::Zero(nMOs * nMOs, nMOs * nMOs);

  std::vector<int> firstAOs{0, 4};
  std::vector<int> nAOs{4, 1};
  for (int p = 0; p < nMOs; ++p) {
    for (int q = 0; q < nMOs; ++q) {
      for (int r = 0; r < nMOs; ++r) {
        for (int s = 0; s < nMOs; ++s) {
          for (const auto& block : arbitraryEris) {
            for (int mu = 0; mu < nAOs[block.first.first]; ++mu) {
              double cmup = arbitraryMos.restrictedMatrix()(firstAOs[block.first.first] + mu, p);
              for (int nu = 0; nu < nAOs[block.first.first]; ++nu) {
                double cnuq = arbitraryMos.restrictedMatrix()(firstAOs[block.first.first] + nu, q);
                auto bra = pair(Orbital(mu), Orbital(nu));

                for (int lambda = 0; lambda < nAOs[block.first.second]; ++lambda) {
                  double clambdar = arbitraryMos.restrictedMatrix()(firstAOs[block.first.second] + lambda, r);
                  for (int sigma = 0; sigma < nAOs[block.first.second]; ++sigma) {
                    double csigmas = arbitraryMos.restrictedMatrix()(firstAOs[block.first.second] + sigma, s);
                    auto ket = pair(Orbital(lambda), Orbital(sigma));
                    erisMO(p * nMOs + q, r * nMOs + s) += block.second.get0()(bra, ket) * cmup * cnuq * clambdar * csigmas;
                    if (block.first.first != block.first.second) {
                      erisMO(r * nMOs + s, p * nMOs + q) += block.second.get0()(bra, ket) * cmup * cnuq * clambdar * csigmas;
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  auto atomMOTransf = [&](DiatomicKey idx) -> Eigen::MatrixXd {
    Eigen::MatrixXd moeris = Eigen::MatrixXd::Zero(nMOs * nMOs, nMOs * nMOs);
    for (int p = 0; p < nMOs; ++p) {
      for (int q = 0; q < nMOs; ++q) {
        for (int r = 0; r < nMOs; ++r) {
          for (int s = 0; s < nMOs; ++s) {
            auto block = arbitraryEris[idx];
            for (int mu = 0; mu < nAOs[idx.first]; ++mu) {
              double cmup = arbitraryMos.restrictedMatrix()(firstAOs[idx.first] + mu, p);
              for (int nu = 0; nu < nAOs[idx.first]; ++nu) {
                double cnuq = arbitraryMos.restrictedMatrix()(firstAOs[idx.first] + nu, q);
                auto bra = pair(Orbital(mu), Orbital(nu));

                for (int lambda = 0; lambda < nAOs[idx.second]; ++lambda) {
                  double clambdar = arbitraryMos.restrictedMatrix()(firstAOs[idx.second] + lambda, r);
                  for (int sigma = 0; sigma < nAOs[idx.second]; ++sigma) {
                    double csigmas = arbitraryMos.restrictedMatrix()(firstAOs[idx.second] + sigma, s);
                    auto ket = pair(Orbital(lambda), Orbital(sigma));
                    moeris(p * nMOs + q, r * nMOs + s) += block.get0()(bra, ket) * cmup * cnuq * clambdar * csigmas;
                  }
                }
              }
            }
          }
        }
      }
    }
    return moeris;
  };
  // TEST TRANSFORMATION MATRIX BY HAND
  EXPECT_NEAR((atomMOTransf({0, 0}) -
               AO2MOTransformation::transformBlock(
                   arbitraryEris[{0, 0}].get0(),
                   AO2MOTransformation::generateBlockTransformation(0, arbitraryMos.restrictedMatrix(), aoInfos),
                   AO2MOTransformation::generateBlockTransformation(0, arbitraryMos.restrictedMatrix(), aoInfos)))
                  .norm(),
              0, 1e-12);
  EXPECT_NEAR((atomMOTransf({1, 1}) -
               AO2MOTransformation::transformBlock(
                   arbitraryEris[{1, 1}].get0(),
                   AO2MOTransformation::generateBlockTransformation(1, arbitraryMos.restrictedMatrix(), aoInfos),
                   AO2MOTransformation::generateBlockTransformation(1, arbitraryMos.restrictedMatrix(), aoInfos)))
                  .norm(),
              0, 1e-12);
  EXPECT_NEAR((atomMOTransf({1, 0}) -
               AO2MOTransformation::transformBlock(
                   arbitraryEris[{1, 0}].get0(),
                   AO2MOTransformation::generateBlockTransformation(1, arbitraryMos.restrictedMatrix(), aoInfos),
                   AO2MOTransformation::generateBlockTransformation(0, arbitraryMos.restrictedMatrix(), aoInfos)))
                  .norm(),
              0, 1e-12);

  Eigen::MatrixXd moEri = AO2MOTransformation::ao2mo<Utils::Reference::Restricted>(arbitraryEris, arbitraryMos, aoInfos);
  for (int i = 0; i < nMOs; ++i) {
    for (int j = 0; j < nMOs; ++j) {
      EXPECT_NEAR(moEri(i, j), erisMO(i, j), 1e-14);
    }
  }
}

TEST_F(AnAO2MOTest, AO2MORestrictedInCAS) {
  auto casIndices = Utils::SecondQuantization::CasGenerator::parseCasString("1,2,3");
  int nMOs = 5;
  int casSize = casIndices.size();
  auto generator = Utils::SecondQuantization::CasGenerator(arbitraryMos);
  auto cas = generator.generateRestricted(arbitraryNumberOfElectrons, casIndices);
  auto casMos = cas.getActiveOrbitalsCoefficients();

  Eigen::MatrixXd erisMO = Eigen::MatrixXd::Zero(nMOs * nMOs, nMOs * nMOs);

  std::vector<int> firstAOs{0, 4};
  std::vector<int> nAOs{4, 1};
  for (int p = 0; p < nMOs; ++p) {
    for (int q = 0; q < nMOs; ++q) {
      for (int r = 0; r < nMOs; ++r) {
        for (int s = 0; s < nMOs; ++s) {
          for (auto block : arbitraryEris) {
            for (int mu = 0; mu < nAOs[block.first.first]; ++mu) {
              double cmup = arbitraryMos.restrictedMatrix()(firstAOs[block.first.first] + mu, p);
              for (int nu = 0; nu < nAOs[block.first.first]; ++nu) {
                double cnuq = arbitraryMos.restrictedMatrix()(firstAOs[block.first.first] + nu, q);
                auto bra = pair(Orbital(mu), Orbital(nu));

                for (int lambda = 0; lambda < nAOs[block.first.second]; ++lambda) {
                  double clambdar = arbitraryMos.restrictedMatrix()(firstAOs[block.first.second] + lambda, r);
                  for (int sigma = 0; sigma < nAOs[block.first.second]; ++sigma) {
                    double csigmas = arbitraryMos.restrictedMatrix()(firstAOs[block.first.second] + sigma, s);
                    auto ket = pair(Orbital(lambda), Orbital(sigma));
                    erisMO(p * nMOs + q, r * nMOs + s) += block.second.get0()(bra, ket) * cmup * cnuq * clambdar * csigmas;
                    if (block.first.first != block.first.second) {
                      erisMO(r * nMOs + s, p * nMOs + q) += block.second.get0()(bra, ket) * cmup * cnuq * clambdar * csigmas;
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  auto casEris =
      AO2MOTransformation::ao2mo<Utils::Reference::Restricted>(arbitraryEris, cas.getActiveOrbitalsCoefficients(), aoInfos);

  for (int row = 0; row < casEris.rows(); ++row) {
    int p = cas.getActiveIndices().restricted[row / casSize];
    int q = cas.getActiveIndices().restricted[row % casSize];
    for (int col = 0; col < casEris.cols(); ++col) {
      int r = cas.getActiveIndices().restricted[col / casSize];
      int s = cas.getActiveIndices().restricted[col % casSize];
      EXPECT_NEAR(casEris(row, col), erisMO(p * nMOs + q, r * nMOs + s), 1e-12);
    }
  }
}

TEST_F(AnAO2MOTest, AO2MO1ElectronRestricted) {
  Eigen::MatrixXd arbitrary1ElMatrix = Eigen::MatrixXd::Random(arbitraryNumberOfAOs, arbitraryNumberOfAOs);
  arbitrary1ElMatrix = arbitrary1ElMatrix.selfadjointView<Eigen::Upper>();
  Eigen::MatrixXd moMatrix =
      AO2MOTransformation::transform1eIntegrals<Utils::Reference::Restricted>(arbitrary1ElMatrix, arbitraryMos);
  Eigen::MatrixXd transformed =
      arbitraryMos.restrictedMatrix().transpose() * arbitrary1ElMatrix * arbitraryMos.restrictedMatrix();
  for (int row = 0; row < transformed.rows(); ++row) {
    for (int col = 0; col < transformed.cols(); ++col) {
      EXPECT_NEAR(moMatrix(row, col), transformed(row, col), 1e-14);
    }
  }
}

TEST_F(AnAO2MOTest, AO2MO1ElectronInCASRestricted) {
  Eigen::MatrixXd arbitrary1ElMatrix = Eigen::MatrixXd::Random(arbitraryNumberOfAOs, arbitraryNumberOfAOs);
  arbitrary1ElMatrix = arbitrary1ElMatrix.selfadjointView<Eigen::Upper>();
  auto casIndices = Utils::SecondQuantization::CasGenerator::parseCasString("1,2,3");
  auto generator = Utils::SecondQuantization::CasGenerator(arbitraryMos);
  auto cas = generator.generateRestricted(arbitraryNumberOfElectrons, casIndices);
  Eigen::MatrixXd moMatrix = AO2MOTransformation::transform1eIntegrals(arbitrary1ElMatrix, cas);
  Eigen::MatrixXd transformed = arbitraryMos.restrictedMatrix().middleCols(1, 3).transpose() * arbitrary1ElMatrix *
                                arbitraryMos.restrictedMatrix().middleCols(1, 3);
  for (int row = 0; row < transformed.rows(); ++row) {
    for (int col = 0; col < transformed.cols(); ++col) {
      EXPECT_NEAR(moMatrix(row, col), transformed(row, col), 1e-14);
    }
  }
}

TEST_F(AnAO2MOTest, AO2MOUnrestricted) {
  auto moEriUnrestricted = AO2MOTransformation::ao2mo<Utils::Reference::Unrestricted>(arbitraryEris, arbitraryMos, aoInfos);
  Eigen::MatrixXd moEriRestricted =
      AO2MOTransformation::ao2mo<Utils::Reference::Restricted>(arbitraryEris, arbitraryMos, aoInfos);
  EXPECT_EQ(moEriUnrestricted.size(), 4);
  EXPECT_TRUE(moEriUnrestricted.find({Utils::SecondQuantization::SpinComponent::Alpha,
                                      Utils::SecondQuantization::SpinComponent::Alpha}) != moEriUnrestricted.end());
  EXPECT_TRUE(moEriUnrestricted.find({Utils::SecondQuantization::SpinComponent::Alpha,
                                      Utils::SecondQuantization::SpinComponent::Beta}) != moEriUnrestricted.end());
  EXPECT_TRUE(moEriUnrestricted.find({Utils::SecondQuantization::SpinComponent::Beta,
                                      Utils::SecondQuantization::SpinComponent::Alpha}) != moEriUnrestricted.end());
  EXPECT_TRUE(moEriUnrestricted.find({Utils::SecondQuantization::SpinComponent::Beta,
                                      Utils::SecondQuantization::SpinComponent::Beta}) != moEriUnrestricted.end());
  EXPECT_EQ(moEriUnrestricted
                .at({Utils::SecondQuantization::SpinComponent::Alpha, Utils::SecondQuantization::SpinComponent::Alpha})
                .rows(),
            std::pow(arbitraryNumberOfFullMOs, 2));
  EXPECT_EQ(moEriUnrestricted
                .at({Utils::SecondQuantization::SpinComponent::Alpha, Utils::SecondQuantization::SpinComponent::Alpha})
                .cols(),
            std::pow(arbitraryNumberOfFullMOs, 2));
  EXPECT_EQ(moEriUnrestricted
                .at({Utils::SecondQuantization::SpinComponent::Alpha, Utils::SecondQuantization::SpinComponent::Beta})
                .rows(),
            std::pow(arbitraryNumberOfFullMOs, 2));
  EXPECT_EQ(moEriUnrestricted
                .at({Utils::SecondQuantization::SpinComponent::Alpha, Utils::SecondQuantization::SpinComponent::Beta})
                .cols(),
            std::pow(arbitraryNumberOfFullMOs, 2));
  EXPECT_EQ(moEriUnrestricted
                .at({Utils::SecondQuantization::SpinComponent::Beta, Utils::SecondQuantization::SpinComponent::Alpha})
                .rows(),
            std::pow(arbitraryNumberOfFullMOs, 2));
  EXPECT_EQ(moEriUnrestricted
                .at({Utils::SecondQuantization::SpinComponent::Beta, Utils::SecondQuantization::SpinComponent::Alpha})
                .cols(),
            std::pow(arbitraryNumberOfFullMOs, 2));
  EXPECT_EQ(moEriUnrestricted
                .at({Utils::SecondQuantization::SpinComponent::Beta, Utils::SecondQuantization::SpinComponent::Beta})
                .rows(),
            std::pow(arbitraryNumberOfFullMOs, 2));
  EXPECT_EQ(moEriUnrestricted
                .at({Utils::SecondQuantization::SpinComponent::Beta, Utils::SecondQuantization::SpinComponent::Beta})
                .cols(),
            std::pow(arbitraryNumberOfFullMOs, 2));
  for (int p = 0; p < arbitraryNumberOfFullMOs; ++p) {
    for (int q = 0; q < arbitraryNumberOfFullMOs; ++q) {
      int row = p * arbitraryNumberOfFullMOs + q;
      for (int r = 0; r < arbitraryNumberOfFullMOs; ++r) {
        for (int s = 0; s < arbitraryNumberOfFullMOs; ++s) {
          int col = r * arbitraryNumberOfFullMOs + s;
          for (auto spinEri : moEriUnrestricted) {
            EXPECT_NEAR(spinEri.second(row, col), moEriRestricted(row, col), 1e-12);
          }
        }
      }
    }
  }
}

TEST_F(AnAO2MOTest, AO2MOUnrestrictedInCAS) {
  auto casIndices = Utils::SecondQuantization::CasGenerator::parseCasString("1,2,3");
  auto generator = Utils::SecondQuantization::CasGenerator(arbitraryMos);
  auto cas = generator.generateRestricted(arbitraryNumberOfElectrons, casIndices);
  auto casUnrestricted = generator.generateUnrestricted(5, 2, casIndices, casIndices);
  auto casMos = cas.getActiveOrbitalsCoefficients();
  auto casMosUnrestricted = casUnrestricted.getActiveOrbitalsCoefficients();
  auto casErisUnrestricted =
      AO2MOTransformation::ao2mo<Utils::Reference::Unrestricted>(arbitraryEris, casMosUnrestricted, aoInfos);
  auto casEris = AO2MOTransformation::ao2mo<Utils::Reference::Restricted>(arbitraryEris, casMos, aoInfos);
}

TEST_F(AnAO2MOTest, AO2MO1ElectronUnrestricted) {
  Eigen::MatrixXd arbitrary1ElMatrix = Eigen::MatrixXd::Random(arbitraryNumberOfAOs, arbitraryNumberOfAOs);
  arbitrary1ElMatrix = arbitrary1ElMatrix.selfadjointView<Eigen::Upper>();
  auto moMatrices = AO2MOTransformation::transform1eIntegrals<Utils::Reference::Unrestricted>(
      arbitrary1ElMatrix, arbitraryMos.toUnrestricted());

  Eigen::MatrixXd alphaMatrix = arbitraryMos.toUnrestricted().alphaMatrix().transpose() * arbitrary1ElMatrix *
                                arbitraryMos.toUnrestricted().alphaMatrix();
  Eigen::MatrixXd betaMatrix = arbitraryMos.toUnrestricted().betaMatrix().transpose() * arbitrary1ElMatrix *
                               arbitraryMos.toUnrestricted().betaMatrix();

  for (int row = 0; row < alphaMatrix.rows(); ++row) {
    for (int col = 0; col < alphaMatrix.cols(); ++col) {
      EXPECT_NEAR(moMatrices.at(Utils::SecondQuantization::SpinComponent::Alpha)(row, col), alphaMatrix(row, col), 1e-14);
      EXPECT_NEAR(moMatrices.at(Utils::SecondQuantization::SpinComponent::Beta)(row, col), betaMatrix(row, col), 1e-14);
    }
  }
}
TEST_F(AnAO2MOTest, AO2MO1ElectronUnrestrictedInCAS) {
  Eigen::MatrixXd arbitrary1ElMatrix = Eigen::MatrixXd::Random(arbitraryNumberOfAOs, arbitraryNumberOfAOs);
  arbitrary1ElMatrix = arbitrary1ElMatrix.selfadjointView<Eigen::Upper>();
  auto casIndices = Utils::SecondQuantization::CasGenerator::parseCasString("1,2,3");
  auto generator = Utils::SecondQuantization::CasGenerator(arbitraryMos.toUnrestricted());
  auto cas = generator.generateUnrestricted(3, 2, casIndices, casIndices);
  auto moMatrices = AO2MOTransformation::transform1eIntegrals(arbitrary1ElMatrix, cas);

  Eigen::MatrixXd alphaMatrix = arbitraryMos.toUnrestricted().alphaMatrix().middleCols(1, 3).transpose() *
                                arbitrary1ElMatrix * arbitraryMos.toUnrestricted().alphaMatrix().middleCols(1, 3);
  Eigen::MatrixXd betaMatrix = arbitraryMos.toUnrestricted().betaMatrix().middleCols(1, 3).transpose() *
                               arbitrary1ElMatrix * arbitraryMos.toUnrestricted().betaMatrix().middleCols(1, 3);

  for (int row = 0; row < alphaMatrix.rows(); ++row) {
    for (int col = 0; col < alphaMatrix.cols(); ++col) {
      EXPECT_NEAR(moMatrices.at(Utils::SecondQuantization::SpinComponent::Alpha)(row, col), alphaMatrix(row, col), 1e-14);
      EXPECT_NEAR(moMatrices.at(Utils::SecondQuantization::SpinComponent::Beta)(row, col), betaMatrix(row, col), 1e-14);
    }
  }
}

} // namespace Integrals
} // namespace Sparrow
} // namespace Scine
