/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include <Integrals/BasisSetController.h>
#include <Integrals/Libint.h>
#include <Integrals/TwoBodiesIntegrals/AO2MOTransformation.h>
#include <Integrals/TwoBodiesIntegrals/NddoEriSaverDigester.h>
#include <Integrals/TwoBodiesIntegrals/NddoOneCenterTwoElectronEvaluator.h>
#include <Integrals/TwoBodiesIntegrals/NddoTwoBodiesIntegralsEvaluator.h>
#include <Utils/Constants.h>
#include <Utils/DataStructures/AtomicGtos.h>
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Utils/DataStructures/DensityMatrix.h>
#include <Utils/DataStructures/MolecularOrbitals.h>
#include <Utils/DataStructures/SecondQuantization/EriUtilities.h>
#include <Utils/DataStructures/SlaterToGaussian.h>
#include <Utils/Geometry/AtomCollection.h>
#include <Utils/Geometry/ElementInfo.h>
#include <Utils/IO/ChemicalFileFormats/XyzStreamHandler.h>
#include <gmock/gmock.h>
#include <libint2.hpp>

namespace Scine {
namespace Sparrow {
namespace Integrals {

using namespace testing;
using namespace Orbitals;

/// OneCenterIntegralsEvaluator hardcoded for HF
struct MockHFOneCenterIntegralsEvaluator final : NddoOneCenterTwoElectronEvaluator {
  void evaluate(NddoEriSaverDigester& digester, Operator /*op*/, int atomIndex) final {
    if (atomIndex == 0) { // H
      Eigen::MatrixXd blockHH = Eigen::MatrixXd::Identity(1, 1);
      digester.digest(blockHH, 0);
    }
    else if (atomIndex == 1) { // F
      Eigen::MatrixXd blockHF = Eigen::MatrixXd::Identity(10, 10);
      digester.digest(blockHF, 1);
    }
  }
};

/**
 * @test AnIntegralsTest @file IntegralsTest.cpp
 * Reference data by pyScf
 * Geometry:
 * 2
 *
 * H 0 0 0
 * H 1.2 0 0
 *
 * basis = 'def2-svp'
 * Hartree--Fock
 *
 */
class AnIntegralsTest : public Test {
 public:
  Utils::AtomCollection structure;
  Eigen::MatrixXd pyScfOverlapH2Def2SVP, pyScfHCoreH2Def2SVP, pyScfDipoleH2Def2SVP, pyScfDensityH2Def2SVP,
      pyScfDensityH2Sto3G, pyScf2eMatH2Def2SVP, pyScf2eMatH2Sto3G, libintDensityH2Def2SVP, libint2eMatH2Def2SVP;
  std::map<std::string, double> pyScfEriMapString;
  std::map<Utils::SecondQuantization::IndexType<4>, double> pyScfMOEriMap, pyScfMOEriMap2, pyScfEriMap;
  BasisSetController basisSet;

 private:
  void SetUp() final {
    std::stringstream h2("2\n\n"
                         "H 0 0 0\n"
                         "H 1.2 0 0");
    structure = Utils::XyzStreamHandler::read(h2);

    pyScfOverlapH2Def2SVP.resize(10, 10);
    pyScfHCoreH2Def2SVP.resize(10, 10);
    pyScfDipoleH2Def2SVP.resize(10, 10);
    pyScfDensityH2Def2SVP.resize(10, 10);
    pyScfDensityH2Sto3G.resize(2, 2);
    pyScf2eMatH2Def2SVP.resize(10, 10);
    pyScf2eMatH2Sto3G.resize(2, 2);
    libint2eMatH2Def2SVP.resize(10, 10);
    libintDensityH2Def2SVP.resize(10, 10);
    pyScfOverlapH2Def2SVP << 1.00000000e+00, 6.84799825e-01, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
        2.53303682e-01, 4.14114249e-01, -2.86782459e-01, 0.00000000e+00, 0.00000000e+00, 6.84799825e-01, 1.00000000e+00,
        0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 4.14114249e-01, 7.30845790e-01, -1.73676647e-01, 0.00000000e+00,
        0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 1.00000000e+00, 0.00000000e+00, 0.00000000e+00, 2.86782459e-01,
        1.73676647e-01, -3.98093618e-01, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
        1.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 1.27845428e-01, 0.00000000e+00,
        0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 1.00000000e+00, 0.00000000e+00, 0.00000000e+00,
        0.00000000e+00, 0.00000000e+00, 1.27845428e-01, 2.53303682e-01, 4.14114249e-01, 2.86782459e-01, 0.00000000e+00,
        0.00000000e+00, 1.00000000e+00, 6.84799825e-01, -8.49505555e-18, 0.00000000e+00, 0.00000000e+00, 4.14114249e-01,
        7.30845790e-01, 1.73676647e-01, 0.00000000e+00, 0.00000000e+00, 6.84799825e-01, 1.00000000e+00, 0.00000000e+00,
        0.00000000e+00, 0.00000000e+00, -2.86782459e-01, -1.73676647e-01, -3.98093618e-01, 0.00000000e+00,
        0.00000000e+00, -8.49505555e-18, 0.00000000e+00, 1.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
        0.00000000e+00, 0.00000000e+00, 1.27845428e-01, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
        1.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 1.27845428e-01,
        0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 1.00000000e+00;
    pyScfHCoreH2Def2SVP << -0.84872143, -0.74618239, -0.11906097, 0., 0., -0.39465425, -0.51067364, 0.34020547, 0., 0.,
        -0.74618239, -0.76538332, -0.10274864, 0., 0., -0.51067364, -0.63111362, 0.21472884, 0., 0., -0.11906097,
        -0.10274864, 0.55434149, 0., 0., -0.34020547, -0.21472884, 0.07485182, 0., 0., 0., 0., 0., 0.63428204, 0., 0.,
        0., 0., -0.12945872, 0., 0., 0., 0., 0., 0.63428204, 0., 0., 0., 0., -0.12945872, -0.39465425, -0.51067364,
        -0.34020547, 0., 0., -0.84872143, -0.74618239, 0.11906097, 0., 0., -0.51067364, -0.63111362, -0.21472884, 0.,
        0., -0.74618239, -0.76538332, 0.10274864, 0., 0., 0.34020547, 0.21472884, 0.07485182, 0., 0., 0.11906097,
        0.10274864, 0.55434149, 0., 0., 0., 0., 0., -0.12945872, 0., 0., 0., 0., 0.63428204, 0., 0., 0., 0., 0.,
        -0.12945872, 0., 0., 0., 0., 0.63428204;
    pyScfDipoleH2Def2SVP << 0., 0., 0.61504448, 0., 0., 0.28720475, 0.18658214, -0.26195336, 0., 0., 0., 0., 0.54109665,
        0., 0., 0.75249287, 0.82865903, -0.02773136, 0., 0., 0.61504448, 0.54109665, 0., 0., 0., 0.38837501, 0.36611019,
        -0.45137275, 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.14495571, 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0.14495571, 0.28720475, 0.75249287, 0.38837501, 0., 0., 2.26767135, 1.55290094, 0.61504448, 0., 0., 0.18658214,
        0.82865903, 0.36611019, 0., 0., 1.55290094, 2.26767135, 0.54109665, 0., 0., -0.26195336, -0.02773136,
        -0.45137275, 0., 0., 0.61504448, 0.54109665, 2.26767135, 0., 0., 0., 0., 0., 0.14495571, 0., 0., 0., 0.,
        2.26767135, 0., 0., 0., 0., 0., 0.14495571, 0., 0., 0., 0., 2.26767135;

    // clang-format off
/*
    pyScfMOEriMap2.insert({{{ 0 , 0 , 0 , 0 }},  5.364249605498416 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 0 , 1 }},  -0.5200835340041695 }); pyScfMOEriMap2.insert({{{ 0 , 0 , 0 , 2 }},  -0.10783712187747321 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 0 , 5 }},  0.11375854689257645 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 1 , 0 }},  -0.5200835340041694 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 1 , 1 }},  1.2293757987783787 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 1 , 2 }},  0.12432091685874537 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 1 , 5 }},  -0.16555143955044374 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 2 , 0 }},  -0.10783712187747321 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 2 , 1 }},  0.12432091685874536 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 2 , 2 }},  0.9010678616246691 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 2 , 5 }},  0.4133422610947352 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 3 , 3 }},  1.2640235169140817 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 4 , 4 }},  1.2640235169140817 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 5 , 0 }},  0.11375854689257647 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 5 , 1 }},  -0.1655514395504437 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 5 , 2 }},  0.4133422610947352 });
    pyScfMOEriMap2.insert({{{ 0 , 0 , 5 , 5 }},  0.8458292894066548 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 0 , 0 }},  -0.5200835340041695 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 0 , 1 }},  0.07950548681826038 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 0 , 2 }},  0.015852949346668536 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 0 , 5 }},  -0.018219197221434117 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 1 , 0 }},  0.07950548681826042 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 1 , 1 }},  -0.02244554864321646 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 1 , 2 }},  -0.005488899793502205 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 1 , 5 }},  0.004281226271006021 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 2 , 0 }},  0.015852949346668543 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 2 , 1 }},  -0.005488899793502205 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 2 , 2 }},  -0.008736027503191789 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 2 , 5 }},  -0.00695887946963582 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 3 , 3 }},  -0.015015483866924709 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 4 , 4 }},  -0.015015483866924709 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 5 , 0 }},  -0.018219197221434127 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 5 , 1 }},  0.0042812262710060205 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 5 , 2 }},  -0.00695887946963582 });
    pyScfMOEriMap2.insert({{{ 0 , 1 , 5 , 5 }},  -0.009008571190298494 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 0 , 0 }},  -0.10783712187747325 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 0 , 1 }},  0.015852949346668557 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 0 , 2 }},  0.017931096953950003 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 0 , 5 }},  0.011906519639453234 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 1 , 0 }},  0.015852949346668557 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 1 , 1 }},  -0.007243776338423317 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 1 , 2 }},  0.020592754123666 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 1 , 5 }},  0.019959753048813016 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 2 , 0 }},  0.01793109695395 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 2 , 1 }},  0.020592754123666 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 2 , 2 }},  0.0024413251013918193 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 2 , 5 }},  0.0015348520934502685 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 3 , 3 }},  -0.0031747829737657205 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 4 , 4 }},  -0.0031747829737657205 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 5 , 0 }},  0.011906519639453234 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 5 , 1 }},  0.01995975304881302 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 5 , 2 }},  0.0015348520934502677 });
    pyScfMOEriMap2.insert({{{ 0 , 2 , 5 , 5 }},  -0.01088488511810909 });
    pyScfMOEriMap2.insert({{{ 0 , 3 , 0 , 3 }},  0.02954291948275415 });
    pyScfMOEriMap2.insert({{{ 0 , 3 , 1 , 3 }},  0.039997864086739325 });
    pyScfMOEriMap2.insert({{{ 0 , 3 , 2 , 3 }},  0.007551732287551679 });
    pyScfMOEriMap2.insert({{{ 0 , 3 , 3 , 0 }},  0.029542919482754153 });
    pyScfMOEriMap2.insert({{{ 0 , 3 , 3 , 1 }},  0.039997864086739325 });
    pyScfMOEriMap2.insert({{{ 0 , 3 , 3 , 2 }},  0.007551732287551679 });
    pyScfMOEriMap2.insert({{{ 0 , 3 , 3 , 5 }},  -0.007432393683356122 });
    pyScfMOEriMap2.insert({{{ 0 , 3 , 5 , 3 }},  -0.007432393683356122 });
    pyScfMOEriMap2.insert({{{ 0 , 4 , 0 , 4 }},  0.02954291948275415 });
    pyScfMOEriMap2.insert({{{ 0 , 4 , 1 , 4 }},  0.039997864086739325 });
    pyScfMOEriMap2.insert({{{ 0 , 4 , 2 , 4 }},  0.007551732287551679 });
    pyScfMOEriMap2.insert({{{ 0 , 4 , 4 , 0 }},  0.029542919482754153 });
    pyScfMOEriMap2.insert({{{ 0 , 4 , 4 , 1 }},  0.039997864086739325 });
    pyScfMOEriMap2.insert({{{ 0 , 4 , 4 , 2 }},  0.007551732287551679 });
    pyScfMOEriMap2.insert({{{ 0 , 4 , 4 , 5 }},  -0.007432393683356122 });
    pyScfMOEriMap2.insert({{{ 0 , 4 , 5 , 4 }},  -0.007432393683356122 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 0 , 0 }},  0.11375854689257656 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 0 , 1 }},  -0.018219197221434152 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 0 , 2 }},  0.01190651963945323 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 0 , 5 }},  0.02053661143861543 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 1 , 0 }},  -0.018219197221434145 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 1 , 1 }},  0.0025529728476517076 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 1 , 2 }},  0.023520278873163904 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 1 , 5 }},  0.018611468969205975 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 2 , 0 }},  0.011906519639453233 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 2 , 1 }},  0.023520278873163904 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 2 , 2 }},  0.006406501796788169 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 2 , 5 }},  0.0050551342830337164 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 3 , 3 }},  0.0032880948899024934 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 4 , 4 }},  0.0032880948899024934 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 5 , 0 }},  0.020536611438615432 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 5 , 1 }},  0.018611468969205975 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 5 , 2 }},  0.0050551342830337164 });
    pyScfMOEriMap2.insert({{{ 0 , 5 , 5 , 5 }},  -0.0076169653286799486 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 0 , 0 }},  -0.5200835340041693 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 0 , 1 }},  0.07950548681826032 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 0 , 2 }},  0.015852949346668532 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 0 , 5 }},  -0.018219197221434114 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 1 , 0 }},  0.07950548681826036 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 1 , 1 }},  -0.02244554864321638 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 1 , 2 }},  -0.005488899793502174 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 1 , 5 }},  0.00428122627100598 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 2 , 0 }},  0.015852949346668543 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 2 , 1 }},  -0.005488899793502174 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 2 , 2 }},  -0.008736027503191782 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 2 , 5 }},  -0.006958879469635846 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 3 , 3 }},  -0.015015483866924735 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 4 , 4 }},  -0.015015483866924735 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 5 , 0 }},  -0.01821919722143412 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 5 , 1 }},  0.004281226271005979 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 5 , 2 }},  -0.006958879469635846 });
    pyScfMOEriMap2.insert({{{ 1 , 0 , 5 , 5 }},  -0.009008571190298503 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 0 , 0 }},  1.2293757987783784 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 0 , 1 }},  -0.022445548643216403 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 0 , 2 }},  -0.007243776338423319 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 0 , 5 }},  0.0025529728476517063 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 1 , 0 }},  -0.02244554864321645 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 1 , 1 }},  0.8698237471107465 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 1 , 2 }},  0.044147246479907205 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 1 , 5 }},  -0.09134768769024355 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 2 , 0 }},  -0.007243776338423316 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 2 , 1 }},  0.04414724647990724 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 2 , 2 }},  0.7008395580679245 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 2 , 5 }},  0.22815287734219125 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 3 , 3 }},  0.8930214699008083 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 4 , 4 }},  0.8930214699008083 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 5 , 0 }},  0.002552972847651703 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 5 , 1 }},  -0.09134768769024354 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 5 , 2 }},  0.22815287734219125 });
    pyScfMOEriMap2.insert({{{ 1 , 1 , 5 , 5 }},  0.6447460447663704 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 0 , 0 }},  0.12432091685874538 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 0 , 1 }},  -0.005488899793502189 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 0 , 2 }},  0.02059275412366601 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 0 , 5 }},  0.023520278873163908 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 1 , 0 }},  -0.005488899793502197 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 1 , 1 }},  0.044147246479907254 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 1 , 2 }},  0.1386454257776287 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 1 , 5 }},  0.07334459830249487 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 2 , 0 }},  0.02059275412366601 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 2 , 1 }},  0.1386454257776287 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 2 , 2 }},  0.01787337888005427 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 2 , 5 }},  0.0834063384426554 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 3 , 3 }},  0.06842315023589612 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 4 , 4 }},  0.06842315023589612 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 5 , 0 }},  0.02352027887316391 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 5 , 1 }},  0.07334459830249487 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 5 , 2 }},  0.08340633844265538 });
    pyScfMOEriMap2.insert({{{ 1 , 2 , 5 , 5 }},  -0.049052465549102225 });
    pyScfMOEriMap2.insert({{{ 1 , 3 , 0 , 3 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 1 , 3 , 1 , 3 }},  0.18845834038484022 });
    pyScfMOEriMap2.insert({{{ 1 , 3 , 2 , 3 }},  0.02993450933726156 });
    pyScfMOEriMap2.insert({{{ 1 , 3 , 3 , 0 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 1 , 3 , 3 , 1 }},  0.1884583403848402 });
    pyScfMOEriMap2.insert({{{ 1 , 3 , 3 , 2 }},  0.02993450933726156 });
    pyScfMOEriMap2.insert({{{ 1 , 3 , 3 , 5 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 1 , 3 , 5 , 3 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 1 , 4 , 0 , 4 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 1 , 4 , 1 , 4 }},  0.18845834038484022 });
    pyScfMOEriMap2.insert({{{ 1 , 4 , 2 , 4 }},  0.02993450933726156 });
    pyScfMOEriMap2.insert({{{ 1 , 4 , 4 , 0 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 1 , 4 , 4 , 1 }},  0.1884583403848402 });
    pyScfMOEriMap2.insert({{{ 1 , 4 , 4 , 2 }},  0.02993450933726156 });
    pyScfMOEriMap2.insert({{{ 1 , 4 , 4 , 5 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 1 , 4 , 5 , 4 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 0 , 0 }},  -0.1655514395504438 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 0 , 1 }},  0.004281226271006029 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 0 , 2 }},  0.019959753048813027 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 0 , 5 }},  0.01861146896920598 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 1 , 0 }},  0.0042812262710060145 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 1 , 1 }},  -0.09134768769024358 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 1 , 2 }},  0.07334459830249489 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 1 , 5 }},  0.0917621045797013 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 2 , 0 }},  0.019959753048813027 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 2 , 1 }},  0.07334459830249487 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 2 , 2 }},  -0.02333878806899988 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 2 , 5 }},  -0.039485391983394805 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 3 , 3 }},  -0.09008816936079046 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 4 , 4 }},  -0.09008816936079046 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 5 , 0 }},  0.018611468969205985 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 5 , 1 }},  0.0917621045797013 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 5 , 2 }},  -0.03948539198339482 });
    pyScfMOEriMap2.insert({{{ 1 , 5 , 5 , 5 }},  -0.05567568117736145 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 0 , 0 }},  -0.10783712187747317 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 0 , 1 }},  0.015852949346668546 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 0 , 2 }},  0.01793109695395 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 0 , 5 }},  0.011906519639453236 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 1 , 0 }},  0.015852949346668543 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 1 , 1 }},  -0.00724377633842329 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 1 , 2 }},  0.020592754123666006 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 1 , 5 }},  0.01995975304881301 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 2 , 0 }},  0.01793109695395 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 2 , 1 }},  0.020592754123666006 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 2 , 2 }},  0.0024413251013918175 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 2 , 5 }},  0.0015348520934502596 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 3 , 3 }},  -0.0031747829737657136 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 4 , 4 }},  -0.0031747829737657136 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 5 , 0 }},  0.011906519639453236 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 5 , 1 }},  0.01995975304881301 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 5 , 2 }},  0.0015348520934502588 });
    pyScfMOEriMap2.insert({{{ 2 , 0 , 5 , 5 }},  -0.010884885118109096 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 0 , 0 }},  0.12432091685874537 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 0 , 1 }},  -0.005488899793502185 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 0 , 2 }},  0.02059275412366601 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 0 , 5 }},  0.0235202788731639 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 1 , 0 }},  -0.005488899793502186 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 1 , 1 }},  0.04414724647990725 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 1 , 2 }},  0.1386454257776287 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 1 , 5 }},  0.07334459830249487 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 2 , 0 }},  0.020592754123666006 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 2 , 1 }},  0.1386454257776287 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 2 , 2 }},  0.0178733788800543 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 2 , 5 }},  0.08340633844265535 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 3 , 3 }},  0.06842315023589612 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 4 , 4 }},  0.06842315023589612 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 5 , 0 }},  0.023520278873163904 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 5 , 1 }},  0.07334459830249486 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 5 , 2 }},  0.08340633844265535 });
    pyScfMOEriMap2.insert({{{ 2 , 1 , 5 , 5 }},  -0.04905246554910223 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 0 , 0 }},  0.9010678616246693 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 0 , 1 }},  -0.008736027503191905 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 0 , 2 }},  0.002441325101391799 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 0 , 5 }},  0.006406501796788193 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 1 , 0 }},  -0.00873602750319188 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 1 , 1 }},  0.7008395580679245 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 1 , 2 }},  0.017873378880054254 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 1 , 5 }},  -0.02333878806899983 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 2 , 0 }},  0.002441325101391792 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 2 , 1 }},  0.017873378880054244 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 2 , 2 }},  0.686987995121938 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 2 , 5 }},  0.09695021862088712 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 3 , 3 }},  0.688008820494766 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 4 , 4 }},  0.688008820494766 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 5 , 0 }},  0.006406501796788199 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 5 , 1 }},  -0.02333878806899986 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 5 , 2 }},  0.09695021862088711 });
    pyScfMOEriMap2.insert({{{ 2 , 2 , 5 , 5 }},  0.652485222395791 });
    pyScfMOEriMap2.insert({{{ 2 , 3 , 0 , 3 }},  0.007551732287551679 });
    pyScfMOEriMap2.insert({{{ 2 , 3 , 1 , 3 }},  0.02993450933726155 });
    pyScfMOEriMap2.insert({{{ 2 , 3 , 2 , 3 }},  0.03669680124491041 });
    pyScfMOEriMap2.insert({{{ 2 , 3 , 3 , 0 }},  0.00755173228755168 });
    pyScfMOEriMap2.insert({{{ 2 , 3 , 3 , 1 }},  0.029934509337261557 });
    pyScfMOEriMap2.insert({{{ 2 , 3 , 3 , 2 }},  0.036696801244910406 });
    pyScfMOEriMap2.insert({{{ 2 , 3 , 3 , 5 }},  0.021474892456275575 });
    pyScfMOEriMap2.insert({{{ 2 , 3 , 5 , 3 }},  0.02147489245627558 });
    pyScfMOEriMap2.insert({{{ 2 , 4 , 0 , 4 }},  0.007551732287551679 });
    pyScfMOEriMap2.insert({{{ 2 , 4 , 1 , 4 }},  0.02993450933726155 });
    pyScfMOEriMap2.insert({{{ 2 , 4 , 2 , 4 }},  0.03669680124491041 });
    pyScfMOEriMap2.insert({{{ 2 , 4 , 4 , 0 }},  0.00755173228755168 });
    pyScfMOEriMap2.insert({{{ 2 , 4 , 4 , 1 }},  0.029934509337261557 });
    pyScfMOEriMap2.insert({{{ 2 , 4 , 4 , 2 }},  0.036696801244910406 });
    pyScfMOEriMap2.insert({{{ 2 , 4 , 4 , 5 }},  0.021474892456275575 });
    pyScfMOEriMap2.insert({{{ 2 , 4 , 5 , 4 }},  0.02147489245627558 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 0 , 0 }},  0.41334226109473543 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 0 , 1 }},  -0.0069588794696359155 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 0 , 2 }},  0.0015348520934502414 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 0 , 5 }},  0.005055134283033745 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 1 , 0 }},  -0.006958879469635914 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 1 , 1 }},  0.228152877342191 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 1 , 2 }},  0.08340633844265533 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 1 , 5 }},  -0.03948539198339476 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 2 , 0 }},  0.001534852093450244 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 2 , 1 }},  0.08340633844265534 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 2 , 2 }},  0.09695021862088714 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 2 , 5 }},  0.22480696893504218 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 3 , 3 }},  0.23977430999163057 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 4 , 4 }},  0.23977430999163057 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 5 , 0 }},  0.005055134283033743 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 5 , 1 }},  -0.03948539198339477 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 5 , 2 }},  0.22480696893504218 });
    pyScfMOEriMap2.insert({{{ 2 , 5 , 5 , 5 }},  0.02454804807381984 });
    pyScfMOEriMap2.insert({{{ 3 , 0 , 0 , 3 }},  0.02954291948275415 });
    pyScfMOEriMap2.insert({{{ 3 , 0 , 1 , 3 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 3 , 0 , 2 , 3 }},  0.007551732287551681 });
    pyScfMOEriMap2.insert({{{ 3 , 0 , 3 , 0 }},  0.029542919482754153 });
    pyScfMOEriMap2.insert({{{ 3 , 0 , 3 , 1 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 3 , 0 , 3 , 2 }},  0.007551732287551681 });
    pyScfMOEriMap2.insert({{{ 3 , 0 , 3 , 5 }},  -0.007432393683356124 });
    pyScfMOEriMap2.insert({{{ 3 , 0 , 5 , 3 }},  -0.007432393683356122 });
    pyScfMOEriMap2.insert({{{ 3 , 1 , 0 , 3 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 3 , 1 , 1 , 3 }},  0.18845834038484022 });
    pyScfMOEriMap2.insert({{{ 3 , 1 , 2 , 3 }},  0.02993450933726156 });
    pyScfMOEriMap2.insert({{{ 3 , 1 , 3 , 0 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 3 , 1 , 3 , 1 }},  0.1884583403848402 });
    pyScfMOEriMap2.insert({{{ 3 , 1 , 3 , 2 }},  0.02993450933726156 });
    pyScfMOEriMap2.insert({{{ 3 , 1 , 3 , 5 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 3 , 1 , 5 , 3 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 3 , 2 , 0 , 3 }},  0.007551732287551679 });
    pyScfMOEriMap2.insert({{{ 3 , 2 , 1 , 3 }},  0.02993450933726155 });
    pyScfMOEriMap2.insert({{{ 3 , 2 , 2 , 3 }},  0.03669680124491041 });
    pyScfMOEriMap2.insert({{{ 3 , 2 , 3 , 0 }},  0.00755173228755168 });
    pyScfMOEriMap2.insert({{{ 3 , 2 , 3 , 1 }},  0.029934509337261557 });
    pyScfMOEriMap2.insert({{{ 3 , 2 , 3 , 2 }},  0.036696801244910406 });
    pyScfMOEriMap2.insert({{{ 3 , 2 , 3 , 5 }},  0.021474892456275575 });
    pyScfMOEriMap2.insert({{{ 3 , 2 , 5 , 3 }},  0.02147489245627558 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 0 , 0 }},  1.2640235169140823 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 0 , 1 }},  -0.015015483866924875 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 0 , 2 }},  -0.003174782973765762 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 0 , 5 }},  0.003288094889902535 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 1 , 0 }},  -0.015015483866924846 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 1 , 1 }},  0.8930214699008083 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 1 , 2 }},  0.06842315023589612 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 1 , 5 }},  -0.09008816936079043 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 2 , 0 }},  -0.0031747829737657544 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 2 , 1 }},  0.06842315023589612 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 2 , 2 }},  0.688008820494766 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 2 , 5 }},  0.23977430999163044 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 3 , 3 }},  0.9975136259296558 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 4 , 4 }},  0.8899728857331787 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 5 , 0 }},  0.0032880948899025202 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 5 , 1 }},  -0.09008816936079049 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 5 , 2 }},  0.2397743099916305 });
    pyScfMOEriMap2.insert({{{ 3 , 3 , 5 , 5 }},  0.627865737872027 });
    pyScfMOEriMap2.insert({{{ 3 , 4 , 3 , 4 }},  0.05377037009823802 });
    pyScfMOEriMap2.insert({{{ 3 , 4 , 4 , 3 }},  0.05377037009823802 });
    pyScfMOEriMap2.insert({{{ 3 , 5 , 0 , 3 }},  -0.007432393683356122 });
    pyScfMOEriMap2.insert({{{ 3 , 5 , 1 , 3 }},  -0.030267551761869946 });
    pyScfMOEriMap2.insert({{{ 3 , 5 , 2 , 3 }},  0.021474892456275575 });
    pyScfMOEriMap2.insert({{{ 3 , 5 , 3 , 0 }},  -0.007432393683356122 });
    pyScfMOEriMap2.insert({{{ 3 , 5 , 3 , 1 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 3 , 5 , 3 , 2 }},  0.021474892456275575 });
    pyScfMOEriMap2.insert({{{ 3 , 5 , 3 , 5 }},  0.029373444901067365 });
    pyScfMOEriMap2.insert({{{ 3 , 5 , 5 , 3 }},  0.02937344490106737 });
    pyScfMOEriMap2.insert({{{ 4 , 0 , 0 , 4 }},  0.02954291948275415 });
    pyScfMOEriMap2.insert({{{ 4 , 0 , 1 , 4 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 4 , 0 , 2 , 4 }},  0.007551732287551681 });
    pyScfMOEriMap2.insert({{{ 4 , 0 , 4 , 0 }},  0.029542919482754153 });
    pyScfMOEriMap2.insert({{{ 4 , 0 , 4 , 1 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 4 , 0 , 4 , 2 }},  0.007551732287551681 });
    pyScfMOEriMap2.insert({{{ 4 , 0 , 4 , 5 }},  -0.007432393683356124 });
    pyScfMOEriMap2.insert({{{ 4 , 0 , 5 , 4 }},  -0.007432393683356122 });
    pyScfMOEriMap2.insert({{{ 4 , 1 , 0 , 4 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 4 , 1 , 1 , 4 }},  0.18845834038484022 });
    pyScfMOEriMap2.insert({{{ 4 , 1 , 2 , 4 }},  0.02993450933726156 });
    pyScfMOEriMap2.insert({{{ 4 , 1 , 4 , 0 }},  0.03999786408673933 });
    pyScfMOEriMap2.insert({{{ 4 , 1 , 4 , 1 }},  0.1884583403848402 });
    pyScfMOEriMap2.insert({{{ 4 , 1 , 4 , 2 }},  0.02993450933726156 });
    pyScfMOEriMap2.insert({{{ 4 , 1 , 4 , 5 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 4 , 1 , 5 , 4 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 4 , 2 , 0 , 4 }},  0.007551732287551679 });
    pyScfMOEriMap2.insert({{{ 4 , 2 , 1 , 4 }},  0.02993450933726155 });
    pyScfMOEriMap2.insert({{{ 4 , 2 , 2 , 4 }},  0.03669680124491041 });
    pyScfMOEriMap2.insert({{{ 4 , 2 , 4 , 0 }},  0.00755173228755168 });
    pyScfMOEriMap2.insert({{{ 4 , 2 , 4 , 1 }},  0.029934509337261557 });
    pyScfMOEriMap2.insert({{{ 4 , 2 , 4 , 2 }},  0.036696801244910406 });
    pyScfMOEriMap2.insert({{{ 4 , 2 , 4 , 5 }},  0.021474892456275575 });
    pyScfMOEriMap2.insert({{{ 4 , 2 , 5 , 4 }},  0.02147489245627558 });
    pyScfMOEriMap2.insert({{{ 4 , 3 , 3 , 4 }},  0.05377037009823802 });
    pyScfMOEriMap2.insert({{{ 4 , 3 , 4 , 3 }},  0.05377037009823802 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 0 , 0 }},  1.2640235169140823 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 0 , 1 }},  -0.015015483866924875 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 0 , 2 }},  -0.003174782973765762 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 0 , 5 }},  0.003288094889902535 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 1 , 0 }},  -0.015015483866924846 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 1 , 1 }},  0.8930214699008083 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 1 , 2 }},  0.06842315023589612 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 1 , 5 }},  -0.09008816936079043 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 2 , 0 }},  -0.0031747829737657544 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 2 , 1 }},  0.06842315023589612 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 2 , 2 }},  0.688008820494766 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 2 , 5 }},  0.23977430999163044 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 3 , 3 }},  0.8899728857331787 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 4 , 4 }},  0.9975136259296558 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 5 , 0 }},  0.0032880948899025202 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 5 , 1 }},  -0.09008816936079049 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 5 , 2 }},  0.2397743099916305 });
    pyScfMOEriMap2.insert({{{ 4 , 4 , 5 , 5 }},  0.627865737872027 });
    pyScfMOEriMap2.insert({{{ 4 , 5 , 0 , 4 }},  -0.007432393683356122 });
    pyScfMOEriMap2.insert({{{ 4 , 5 , 1 , 4 }},  -0.030267551761869946 });
    pyScfMOEriMap2.insert({{{ 4 , 5 , 2 , 4 }},  0.021474892456275575 });
    pyScfMOEriMap2.insert({{{ 4 , 5 , 4 , 0 }},  -0.007432393683356122 });
    pyScfMOEriMap2.insert({{{ 4 , 5 , 4 , 1 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 4 , 5 , 4 , 2 }},  0.021474892456275575 });
    pyScfMOEriMap2.insert({{{ 4 , 5 , 4 , 5 }},  0.029373444901067365 });
    pyScfMOEriMap2.insert({{{ 4 , 5 , 5 , 4 }},  0.02937344490106737 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 0 , 0 }},  0.11375854689257647 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 0 , 1 }},  -0.01821919722143413 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 0 , 2 }},  0.011906519639453236 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 0 , 5 }},  0.020536611438615425 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 1 , 0 }},  -0.018219197221434124 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 1 , 1 }},  0.002552972847651684 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 1 , 2 }},  0.023520278873163904 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 1 , 5 }},  0.018611468969205992 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 2 , 0 }},  0.011906519639453236 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 2 , 1 }},  0.023520278873163904 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 2 , 2 }},  0.0064065017967881764 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 2 , 5 }},  0.005055134283033726 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 3 , 3 }},  0.0032880948899024977 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 4 , 4 }},  0.0032880948899024977 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 5 , 0 }},  0.02053661143861543 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 5 , 1 }},  0.018611468969205992 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 5 , 2 }},  0.005055134283033725 });
    pyScfMOEriMap2.insert({{{ 5 , 0 , 5 , 5 }},  -0.007616965328679943 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 0 , 0 }},  -0.1655514395504438 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 0 , 1 }},  0.004281226271006015 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 0 , 2 }},  0.01995975304881302 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 0 , 5 }},  0.018611468969205982 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 1 , 0 }},  0.0042812262710060075 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 1 , 1 }},  -0.09134768769024358 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 1 , 2 }},  0.07334459830249487 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 1 , 5 }},  0.0917621045797013 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 2 , 0 }},  0.01995975304881302 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 2 , 1 }},  0.07334459830249487 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 2 , 2 }},  -0.023338788068999937 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 2 , 5 }},  -0.039485391983394846 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 3 , 3 }},  -0.09008816936079052 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 4 , 4 }},  -0.09008816936079052 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 5 , 0 }},  0.018611468969205985 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 5 , 1 }},  0.0917621045797013 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 5 , 2 }},  -0.03948539198339484 });
    pyScfMOEriMap2.insert({{{ 5 , 1 , 5 , 5 }},  -0.0556756811773615 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 0 , 0 }},  0.4133422610947353 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 0 , 1 }},  -0.006958879469635846 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 0 , 2 }},  0.0015348520934502633 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 0 , 5 }},  0.005055134283033718 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 1 , 0 }},  -0.006958879469635845 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 1 , 1 }},  0.22815287734219106 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 1 , 2 }},  0.08340633844265535 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 1 , 5 }},  -0.03948539198339481 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 2 , 0 }},  0.0015348520934502596 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 2 , 1 }},  0.08340633844265537 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 2 , 2 }},  0.09695021862088718 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 2 , 5 }},  0.22480696893504218 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 3 , 3 }},  0.23977430999163063 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 4 , 4 }},  0.23977430999163063 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 5 , 0 }},  0.005055134283033715 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 5 , 1 }},  -0.03948539198339482 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 5 , 2 }},  0.2248069689350422 });
    pyScfMOEriMap2.insert({{{ 5 , 2 , 5 , 5 }},  0.02454804807381991 });
    pyScfMOEriMap2.insert({{{ 5 , 3 , 0 , 3 }},  -0.0074323936833561215 });
    pyScfMOEriMap2.insert({{{ 5 , 3 , 1 , 3 }},  -0.030267551761869946 });
    pyScfMOEriMap2.insert({{{ 5 , 3 , 2 , 3 }},  0.021474892456275575 });
    pyScfMOEriMap2.insert({{{ 5 , 3 , 3 , 0 }},  -0.0074323936833561215 });
    pyScfMOEriMap2.insert({{{ 5 , 3 , 3 , 1 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 5 , 3 , 3 , 2 }},  0.02147489245627558 });
    pyScfMOEriMap2.insert({{{ 5 , 3 , 3 , 5 }},  0.029373444901067365 });
    pyScfMOEriMap2.insert({{{ 5 , 3 , 5 , 3 }},  0.029373444901067365 });
    pyScfMOEriMap2.insert({{{ 5 , 4 , 0 , 4 }},  -0.0074323936833561215 });
    pyScfMOEriMap2.insert({{{ 5 , 4 , 1 , 4 }},  -0.030267551761869946 });
    pyScfMOEriMap2.insert({{{ 5 , 4 , 2 , 4 }},  0.021474892456275575 });
    pyScfMOEriMap2.insert({{{ 5 , 4 , 4 , 0 }},  -0.0074323936833561215 });
    pyScfMOEriMap2.insert({{{ 5 , 4 , 4 , 1 }},  -0.03026755176186995 });
    pyScfMOEriMap2.insert({{{ 5 , 4 , 4 , 2 }},  0.02147489245627558 });
    pyScfMOEriMap2.insert({{{ 5 , 4 , 4 , 5 }},  0.029373444901067365 });
    pyScfMOEriMap2.insert({{{ 5 , 4 , 5 , 4 }},  0.029373444901067365 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 0 , 0 }},  0.8458292894066555 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 0 , 1 }},  -0.009008571190298603 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 0 , 2 }},  -0.010884885118109113 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 0 , 5 }},  -0.00761696532867993 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 1 , 0 }},  -0.009008571190298631 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 1 , 1 }},  0.6447460447663703 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 1 , 2 }},  -0.049052465549102295 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 1 , 5 }},  -0.055675681177361415 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 2 , 0 }},  -0.010884885118109117 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 2 , 1 }},  -0.04905246554910225 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 2 , 2 }},  0.6524852223957908 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 2 , 5 }},  0.02454804807381995 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 3 , 3 }},  0.6278657378720268 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 4 , 4 }},  0.6278657378720268 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 5 , 0 }},  -0.0076169653286799295 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 5 , 1 }},  -0.05567568117736146 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 5 , 2 }},  0.024548048073819877 });
    pyScfMOEriMap2.insert({{{ 5 , 5 , 5 , 5 }},  0.6809317903002591 });
*/


  /*
   * Integrals obtained from pyscf with the script:
   * from pyscf import gto, scf
   * HF = 'H 0 0 0; F 1.2 0 0'
   *
   * mol = gto.M(atom=HF, basis='sto-3g')
   * mf = scf.RHF(mol)
   * mf.init_guess='hcore'
   * mf.diis = False
   *
   * mf.kernel()
   *
   * import numpy as np
   * np.set_printoptions(precision=16)
   * print(mol.atom_coords())
   *
   * eri = mol.intor('int2e', aosym='s1')
   *
   * for mu in [0,2,3,4,5]:
   *         for nu in [0,2,3,4,5]:
   *                 for lam in [0,2,3,4,5]:
   *                     for sig in [0,2,3,4,5]:
   *                         if eri[mu,nu,lam,sig] != 0:
   *                             print("pyScfEriMap.insert({{{",mu, ", ", nu, ", ", lam, ",", sig, "}}, ", eri[mu,nu,lam,sig], "});")
   *
   */
    pyScfEriMap.insert({{{ 0 ,  0 ,  0 , 0 }},  0.7746059439198978 });
    pyScfEriMap.insert({{{ 0 ,  0 ,  0 , 2 }},  0.16214629701101607 });
    pyScfEriMap.insert({{{ 0 ,  0 ,  0 , 3 }},  -0.1797280146835231 });
    pyScfEriMap.insert({{{ 0 ,  0 ,  2 , 0 }},  0.1621462970110161 });
    pyScfEriMap.insert({{{ 0 ,  0 ,  2 , 2 }},  0.4271362781981994 });
    pyScfEriMap.insert({{{ 0 ,  0 ,  2 , 3 }},  -0.09281459666767564 });
    pyScfEriMap.insert({{{ 0 ,  0 ,  3 , 0 }},  -0.17972801468352306 });
    pyScfEriMap.insert({{{ 0 ,  0 ,  3 , 2 }},  -0.09281459666767562 });
    pyScfEriMap.insert({{{ 0 ,  0 ,  3 , 3 }},  0.4533821936331979 });
    pyScfEriMap.insert({{{ 0 ,  0 ,  4 , 4 }},  0.4138158434935222 });
    pyScfEriMap.insert({{{ 0 ,  0 ,  5 , 5 }},  0.4138158434935222 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  0 , 0 }},  0.16214629701101604 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  0 , 2 }},  0.05876884334250608 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  0 , 3 }},  -0.0558286200205692 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  2 , 0 }},  0.05876884334250608 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  2 , 2 }},  0.20689357902047986 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  2 , 3 }},  -0.031740491816135474 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  3 , 0 }},  -0.055828620020569195 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  3 , 2 }},  -0.031740491816135474 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  3 , 3 }},  0.2129201282937293 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  4 , 4 }},  0.20377685411996843 });
    pyScfEriMap.insert({{{ 0 ,  2 ,  5 , 5 }},  0.20377685411996843 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  0 , 0 }},  -0.17972801468352317 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  0 , 2 }},  -0.055828620020569215 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  0 , 3 }},  0.06109983072455267 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  2 , 0 }},  -0.055828620020569215 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  2 , 2 }},  -0.1598343201880774 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  2 , 3 }},  0.05842269218108925 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  3 , 0 }},  0.06109983072455271 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  3 , 2 }},  0.058422692181089274 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  3 , 3 }},  -0.17293644755814708 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  4 , 4 }},  -0.15319127169355665 });
    pyScfEriMap.insert({{{ 0 ,  3 ,  5 , 5 }},  -0.15319127169355665 });
    pyScfEriMap.insert({{{ 0 ,  4 ,  0 , 4 }},  0.008391975738708628 });
    pyScfEriMap.insert({{{ 0 ,  4 ,  2 , 4 }},  0.03540434128645107 });
    pyScfEriMap.insert({{{ 0 ,  4 ,  3 , 4 }},  -0.006535353518258114 });
    pyScfEriMap.insert({{{ 0 ,  4 ,  4 , 0 }},  0.008391975738708632 });
    pyScfEriMap.insert({{{ 0 ,  4 ,  4 , 2 }},  0.035404341286451084 });
    pyScfEriMap.insert({{{ 0 ,  4 ,  4 , 3 }},  -0.006535353518258114 });
    pyScfEriMap.insert({{{ 0 ,  5 ,  0 , 5 }},  0.008391975738708628 });
    pyScfEriMap.insert({{{ 0 ,  5 ,  2 , 5 }},  0.03540434128645107 });
    pyScfEriMap.insert({{{ 0 ,  5 ,  3 , 5 }},  -0.006535353518258114 });
    pyScfEriMap.insert({{{ 0 ,  5 ,  5 , 0 }},  0.008391975738708632 });
    pyScfEriMap.insert({{{ 0 ,  5 ,  5 , 2 }},  0.035404341286451084 });
    pyScfEriMap.insert({{{ 0 ,  5 ,  5 , 3 }},  -0.006535353518258114 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  0 , 0 }},  0.16214629701101604 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  0 , 2 }},  0.05876884334250608 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  0 , 3 }},  -0.05582862002056921 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  2 , 0 }},  0.05876884334250608 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  2 , 2 }},  0.20689357902047986 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  2 , 3 }},  -0.031740491816135474 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  3 , 0 }},  -0.055828620020569195 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  3 , 2 }},  -0.03174049181613548 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  3 , 3 }},  0.21292012829372936 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  4 , 4 }},  0.20377685411996846 });
    pyScfEriMap.insert({{{ 2 ,  0 ,  5 , 5 }},  0.20377685411996846 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  0 , 0 }},  0.4271362781981994 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  0 , 2 }},  0.2068935790204799 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  0 , 3 }},  -0.15983432018807736 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  2 , 0 }},  0.2068935790204799 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  2 , 2 }},  0.9261671399811631 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  2 , 3 }},  -1.6190681720401595e-16 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  3 , 0 }},  -0.1598343201880773 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  3 , 2 }},  -1.6190681720401597e-16 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  3 , 3 }},  0.9259589345574372 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  4 , 4 }},  0.9259589345574372 });
    pyScfEriMap.insert({{{ 2 ,  2 ,  5 , 5 }},  0.9259589345574372 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  0 , 0 }},  -0.0928145966676756 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  0 , 2 }},  -0.03174049181613548 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  0 , 3 }},  0.05842269218108927 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  2 , 0 }},  -0.03174049181613549 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  2 , 2 }},  -1.6190681720401595e-16 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  2 , 3 }},  0.20458750627236152 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  3 , 0 }},  0.05842269218108928 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  3 , 2 }},  0.2045875062723616 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  3 , 3 }},  -1.5031110423690183e-16 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  4 , 4 }},  -1.2097766581019774e-16 });
    pyScfEriMap.insert({{{ 2 ,  3 ,  5 , 5 }},  -1.2097766581019774e-16 });
    pyScfEriMap.insert({{{ 2 ,  4 ,  0 , 4 }},  0.03540434128645107 });
    pyScfEriMap.insert({{{ 2 ,  4 ,  2 , 4 }},  0.20458750627236152 });
    pyScfEriMap.insert({{{ 2 ,  4 ,  3 , 4 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 2 ,  4 ,  4 , 0 }},  0.03540434128645107 });
    pyScfEriMap.insert({{{ 2 ,  4 ,  4 , 2 }},  0.2045875062723616 });
    pyScfEriMap.insert({{{ 2 ,  4 ,  4 , 3 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 2 ,  5 ,  0 , 5 }},  0.03540434128645107 });
    pyScfEriMap.insert({{{ 2 ,  5 ,  2 , 5 }},  0.20458750627236152 });
    pyScfEriMap.insert({{{ 2 ,  5 ,  3 , 5 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 2 ,  5 ,  5 , 0 }},  0.03540434128645107 });
    pyScfEriMap.insert({{{ 2 ,  5 ,  5 , 2 }},  0.2045875062723616 });
    pyScfEriMap.insert({{{ 2 ,  5 ,  5 , 3 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  0 , 0 }},  -0.17972801468352317 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  0 , 2 }},  -0.0558286200205692 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  0 , 3 }},  0.061099830724552655 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  2 , 0 }},  -0.055828620020569215 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  2 , 2 }},  -0.15983432018807736 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  2 , 3 }},  0.05842269218108926 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  3 , 0 }},  0.061099830724552696 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  3 , 2 }},  0.05842269218108927 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  3 , 3 }},  -0.17293644755814705 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  4 , 4 }},  -0.15319127169355667 });
    pyScfEriMap.insert({{{ 3 ,  0 ,  5 , 5 }},  -0.15319127169355665 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  0 , 0 }},  -0.0928145966676756 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  0 , 2 }},  -0.03174049181613549 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  0 , 3 }},  0.05842269218108927 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  2 , 0 }},  -0.03174049181613549 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  2 , 2 }},  -1.6190681720401595e-16 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  2 , 3 }},  0.20458750627236152 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  3 , 0 }},  0.05842269218108928 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  3 , 2 }},  0.2045875062723616 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  3 , 3 }},  -1.5031110423690183e-16 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  4 , 4 }},  -1.2097766581019774e-16 });
    pyScfEriMap.insert({{{ 3 ,  2 ,  5 , 5 }},  -1.2097766581019774e-16 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  0 , 0 }},  0.45338219363319804 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  0 , 2 }},  0.21292012829372944 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  0 , 3 }},  -0.17293644755814708 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  2 , 0 }},  0.2129201282937294 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  2 , 2 }},  0.9259589345574372 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  2 , 3 }},  -1.5031110423690188e-16 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  3 , 0 }},  -0.17293644755814708 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  3 , 2 }},  -1.5031110423690183e-16 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  3 , 3 }},  0.9975136259296558 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  4 , 4 }},  0.8899728857331787 });
    pyScfEriMap.insert({{{ 3 ,  3 ,  5 , 5 }},  0.8899728857331787 });
    pyScfEriMap.insert({{{ 3 ,  4 ,  0 , 4 }},  -0.0065353535182581185 });
    pyScfEriMap.insert({{{ 3 ,  4 ,  2 , 4 }},  -1.4666719213352058e-17 });
    pyScfEriMap.insert({{{ 3 ,  4 ,  3 , 4 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 3 ,  4 ,  4 , 0 }},  -0.006535353518258118 });
    pyScfEriMap.insert({{{ 3 ,  4 ,  4 , 2 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 3 ,  4 ,  4 , 3 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 3 ,  5 ,  0 , 5 }},  -0.0065353535182581185 });
    pyScfEriMap.insert({{{ 3 ,  5 ,  2 , 5 }},  -1.4666719213352058e-17 });
    pyScfEriMap.insert({{{ 3 ,  5 ,  3 , 5 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 3 ,  5 ,  5 , 0 }},  -0.0065353535182581185 });
    pyScfEriMap.insert({{{ 3 ,  5 ,  5 , 2 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 3 ,  5 ,  5 , 3 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 4 ,  0 ,  0 , 4 }},  0.008391975738708628 });
    pyScfEriMap.insert({{{ 4 ,  0 ,  2 , 4 }},  0.03540434128645106 });
    pyScfEriMap.insert({{{ 4 ,  0 ,  3 , 4 }},  -0.006535353518258115 });
    pyScfEriMap.insert({{{ 4 ,  0 ,  4 , 0 }},  0.00839197573870863 });
    pyScfEriMap.insert({{{ 4 ,  0 ,  4 , 2 }},  0.03540434128645108 });
    pyScfEriMap.insert({{{ 4 ,  0 ,  4 , 3 }},  -0.006535353518258115 });
    pyScfEriMap.insert({{{ 4 ,  2 ,  0 , 4 }},  0.03540434128645108 });
    pyScfEriMap.insert({{{ 4 ,  2 ,  2 , 4 }},  0.20458750627236152 });
    pyScfEriMap.insert({{{ 4 ,  2 ,  3 , 4 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 4 ,  2 ,  4 , 0 }},  0.03540434128645107 });
    pyScfEriMap.insert({{{ 4 ,  2 ,  4 , 2 }},  0.2045875062723616 });
    pyScfEriMap.insert({{{ 4 ,  2 ,  4 , 3 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 4 ,  3 ,  0 , 4 }},  -0.0065353535182581185 });
    pyScfEriMap.insert({{{ 4 ,  3 ,  2 , 4 }},  -1.4666719213352058e-17 });
    pyScfEriMap.insert({{{ 4 ,  3 ,  3 , 4 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 4 ,  3 ,  4 , 0 }},  -0.006535353518258118 });
    pyScfEriMap.insert({{{ 4 ,  3 ,  4 , 2 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 4 ,  3 ,  4 , 3 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  0 , 0 }},  0.4138158434935223 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  0 , 2 }},  0.2037768541199685 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  0 , 3 }},  -0.15319127169355673 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  2 , 0 }},  0.2037768541199684 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  2 , 2 }},  0.9259589345574372 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  2 , 3 }},  -1.2097766581019777e-16 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  3 , 0 }},  -0.15319127169355665 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  3 , 2 }},  -1.2097766581019772e-16 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  3 , 3 }},  0.8899728857331787 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  4 , 4 }},  0.9975136259296558 });
    pyScfEriMap.insert({{{ 4 ,  4 ,  5 , 5 }},  0.8899728857331787 });
    pyScfEriMap.insert({{{ 4 ,  5 ,  4 , 5 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 4 ,  5 ,  5 , 4 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 5 ,  0 ,  0 , 5 }},  0.008391975738708628 });
    pyScfEriMap.insert({{{ 5 ,  0 ,  2 , 5 }},  0.03540434128645106 });
    pyScfEriMap.insert({{{ 5 ,  0 ,  3 , 5 }},  -0.006535353518258115 });
    pyScfEriMap.insert({{{ 5 ,  0 ,  5 , 0 }},  0.00839197573870863 });
    pyScfEriMap.insert({{{ 5 ,  0 ,  5 , 2 }},  0.03540434128645108 });
    pyScfEriMap.insert({{{ 5 ,  0 ,  5 , 3 }},  -0.006535353518258115 });
    pyScfEriMap.insert({{{ 5 ,  2 ,  0 , 5 }},  0.03540434128645108 });
    pyScfEriMap.insert({{{ 5 ,  2 ,  2 , 5 }},  0.20458750627236152 });
    pyScfEriMap.insert({{{ 5 ,  2 ,  3 , 5 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 5 ,  2 ,  5 , 0 }},  0.03540434128645107 });
    pyScfEriMap.insert({{{ 5 ,  2 ,  5 , 2 }},  0.2045875062723616 });
    pyScfEriMap.insert({{{ 5 ,  2 ,  5 , 3 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 5 ,  3 ,  0 , 5 }},  -0.0065353535182581185 });
    pyScfEriMap.insert({{{ 5 ,  3 ,  2 , 5 }},  -1.4666719213352058e-17 });
    pyScfEriMap.insert({{{ 5 ,  3 ,  3 , 5 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 5 ,  3 ,  5 , 0 }},  -0.0065353535182581185 });
    pyScfEriMap.insert({{{ 5 ,  3 ,  5 , 2 }},  -1.4666719213352055e-17 });
    pyScfEriMap.insert({{{ 5 ,  3 ,  5 , 3 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 5 ,  4 ,  4 , 5 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 5 ,  4 ,  5 , 4 }},  0.05377037009823802 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  0 , 0 }},  0.4138158434935223 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  0 , 2 }},  0.2037768541199685 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  0 , 3 }},  -0.15319127169355673 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  2 , 0 }},  0.2037768541199684 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  2 , 2 }},  0.9259589345574372 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  2 , 3 }},  -1.2097766581019777e-16 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  3 , 0 }},  -0.15319127169355665 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  3 , 2 }},  -1.2097766581019772e-16 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  3 , 3 }},  0.8899728857331787 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  4 , 4 }},  0.8899728857331787 });
    pyScfEriMap.insert({{{ 5 ,  5 ,  5 , 5 }},  0.9975136259296558 });
    // clang-format on
  }
};

TEST_F(AnIntegralsTest, LibintAtomVectorIsCorrectlyInitialized) {
  const auto atoms = Libint::toLibintAtoms(structure);
  EXPECT_EQ(atoms[0].atomic_number, static_cast<int>(Utils::ElementInfo::Z(structure.getElement(0))));
  EXPECT_EQ(atoms[1].atomic_number, static_cast<int>(Utils::ElementInfo::Z(structure.getElement(1))));
  EXPECT_EQ(atoms[0].x, structure.getPosition(0).x());
  EXPECT_EQ(atoms[0].y, structure.getPosition(0).y());
  EXPECT_EQ(atoms[0].z, structure.getPosition(0).z());
}

TEST_F(AnIntegralsTest, CanInitializeBasiSetController) {
  basisSet.setBasisSet("def2-svp");
  basisSet.setStructure(structure);

  EXPECT_EQ(basisSet.nBasisFunction(), 10);
  EXPECT_EQ(basisSet.getStructure().getPositions(), structure.getPositions());
  EXPECT_EQ(basisSet.getStructure().getElements(), structure.getElements());
}

TEST_F(AnIntegralsTest, PrintOverlap) {
  basisSet.setBasisSet("sto-3g");
  Utils::ElementTypeCollection els{Utils::ElementType::C, Utils::ElementType::C};
  Utils::PositionCollection pos(2, 3);
  pos << 0, 0, 0, 0, 0, 1.2 * Utils::Constants::bohr_per_angstrom;
  basisSet.setStructure({els, pos});

  auto& libint = Libint::getInstance();
  auto result = libint.compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet);
  std::cout << result.get0() << std::endl;
}

TEST_F(AnIntegralsTest, CanCalculateOverlapIntegrals) {
  basisSet.setBasisSet("def2-svp");
  basisSet.setStructure(structure);

  auto& libint = Libint::getInstance();
  auto result = libint.compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet);

  for (int row = 0; row < 10; ++row) {
    for (int col = 0; col < 10; ++col) {
      EXPECT_THAT(result.get<Utils::DerivativeOrder::Zero>()(row, col), DoubleNear(pyScfOverlapH2Def2SVP(row, col), 1e-8));
    }
  }
}

TEST_F(AnIntegralsTest, CanCalculateOverlapIntegralsFromCopiedBasisSet) {
  basisSet.setBasisSet("def2-svp");
  basisSet.setStructure(structure);
  auto basisSet2 = basisSet;

  auto& libint = Libint::getInstance();
  auto result = libint.compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet2);

  for (int row = 0; row < 10; ++row) {
    for (int col = 0; col < 10; ++col) {
      EXPECT_THAT(result.get<Utils::DerivativeOrder::Zero>()(row, col), DoubleNear(pyScfOverlapH2Def2SVP(row, col), 1e-8));
    }
  }
  BasisSetController basisSet3;
  basisSet3 = basisSet2;
  result = libint.compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet3);

  for (int row = 0; row < 10; ++row) {
    for (int col = 0; col < 10; ++col) {
      EXPECT_THAT(result.get<Utils::DerivativeOrder::Zero>()(row, col), DoubleNear(pyScfOverlapH2Def2SVP(row, col), 1e-8));
    }
  }
}

TEST_F(AnIntegralsTest, CanCalculateHCoreIntegrals) {
  basisSet.setBasisSet("def2-svp");
  basisSet.setStructure(structure);

  auto& libint = Libint::getInstance();
  auto kinetic = libint.compute1BodyIntegrals<Operator::kinetic, Utils::DerivativeOrder::Zero>(basisSet);
  auto nuclear = libint.compute1BodyIntegrals<Operator::nuclear, Utils::DerivativeOrder::Zero>(basisSet);

  auto hCore = kinetic + nuclear;
  auto hCore2 = libint.computeHCore(basisSet);

  for (int row = 0; row < 10; ++row) {
    for (int col = 0; col < 10; ++col) {
      EXPECT_THAT(hCore.get<Utils::DerivativeOrder::Zero>()(row, col), DoubleNear(pyScfHCoreH2Def2SVP(row, col), 1e-8));
      EXPECT_THAT(hCore2(row, col), DoubleNear(pyScfHCoreH2Def2SVP(row, col), 1e-8));
    }
  }
}

TEST_F(AnIntegralsTest, CanCalculateDipoleIntegrals) {
  basisSet.setBasisSet("def2-svp");
  basisSet.setStructure(structure);

  auto& libint = Libint::getInstance();
  auto dipole = libint.compute1BodyIntegrals<Operator::emultipole1, Utils::DerivativeOrder::Zero>(basisSet);

  for (int row = 0; row < 10; ++row) {
    for (int col = 0; col < 10; ++col) {
      EXPECT_THAT(dipole.x().get<Utils::DerivativeOrder::Zero>()(row, col), DoubleNear(pyScfDipoleH2Def2SVP(row, col), 1e-8));
    }
  }
}

TEST_F(AnIntegralsTest, CorrectlyCalculatesNddoERIs) {
  std::vector<Utils::AtomicGtos> gtos;
  Utils::AtomicGtos gtoH;
  gtoH.s = Utils::GtoExpansion{0, {{0, 3.42525091, 0.15432897}, {0, 0.62391373, 0.53532814}, {0, 0.16885540, 0.44463454}}};
  Utils::AtomicGtos gtoF;
  gtoF.s = Utils::GtoExpansion{0, {{0, 6.4648032, -0.09996723}, {0, 1.5022812, 0.39951283}, {0, 0.4885885, 0.70011547}}};
  gtoF.p = Utils::GtoExpansion{1, {{1, 6.4648032, 0.15591627}, {1, 1.5022812, 0.60768372}, {1, 0.4885885, 0.39195739}}};
  gtos.push_back(gtoH);
  gtos.push_back(gtoF);

  // Accuracy is limited by accuracy of the Angstrom->Bohr transformation,
  // so copy the coordinates from pySCF
  Utils::ElementTypeCollection elements{Utils::ElementType::H, Utils::ElementType::F};
  Utils::PositionCollection positions(2, 3);
  positions << 0, 0, 0, 2.267671349478074, 0, 0;
  basisSet = BasisSetController({elements, positions}, gtos, "sto-3g-NOCORE");

  NddoEriSaverDigester erisDigester(basisSet);

  auto oneCenterEvaluator = std::make_shared<MockHFOneCenterIntegralsEvaluator>();

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(oneCenterEvaluator);

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();

  // H-H block
  ASSERT_NO_THROW(eriCalculator.getResult().at({0, 0}).get0());
  Eigen::MatrixXd hhBlock = eriCalculator.getResult().at({0, 0}).get0();
  EXPECT_EQ(hhBlock.cols(), 1);
  EXPECT_EQ(hhBlock.rows(), 1);
  EXPECT_EQ(hhBlock, Eigen::MatrixXd::Identity(1, 1));

  // H-F block
  // Assert that only the lower diagonal is saved
  ASSERT_THROW(eriCalculator.getResult().at({0, 1}), std::exception);
  ASSERT_NO_THROW(eriCalculator.getResult().at({1, 0}));
  // get the block
  Eigen::MatrixXd hfBlock = eriCalculator.getResult().at({1, 0}).get0();

  // H has only ss, F has ss, spx spy, spz pxpx pxpy pxpz pypy pypz pzpz
  EXPECT_EQ(hfBlock.rows(), 10);
  EXPECT_EQ(hfBlock.cols(), 1);
  EXPECT_THAT(hfBlock(Orbitals::OrbitalPair::s_s, Orbitals::OrbitalPair::s_s), DoubleNear(pyScfEriMap[{{0, 0, 2, 2}}], 1e-13));
  EXPECT_THAT(hfBlock(Orbitals::OrbitalPair::s_x, Orbitals::OrbitalPair::s_s), DoubleNear(pyScfEriMap[{{0, 0, 2, 3}}], 1e-13));
  EXPECT_THAT(hfBlock(Orbitals::OrbitalPair::s_y, Orbitals::OrbitalPair::s_s), DoubleNear(pyScfEriMap[{{0, 0, 2, 4}}], 1e-13));
  EXPECT_THAT(hfBlock(Orbitals::OrbitalPair::s_z, Orbitals::OrbitalPair::s_s), DoubleNear(pyScfEriMap[{{0, 0, 2, 5}}], 1e-13));
  EXPECT_THAT(hfBlock(Orbitals::OrbitalPair::x_x, Orbitals::OrbitalPair::s_s), DoubleNear(pyScfEriMap[{{0, 0, 3, 3}}], 1e-13));
  EXPECT_THAT(hfBlock(Orbitals::OrbitalPair::x_y, Orbitals::OrbitalPair::s_s), DoubleNear(pyScfEriMap[{{0, 0, 3, 4}}], 1e-13));
  EXPECT_THAT(hfBlock(Orbitals::OrbitalPair::x_z, Orbitals::OrbitalPair::s_s), DoubleNear(pyScfEriMap[{{0, 0, 3, 5}}], 1e-13));
  EXPECT_THAT(hfBlock(Orbitals::OrbitalPair::y_y, Orbitals::OrbitalPair::s_s), DoubleNear(pyScfEriMap[{{0, 0, 4, 4}}], 1e-13));
  EXPECT_THAT(hfBlock(Orbitals::OrbitalPair::y_z, Orbitals::OrbitalPair::s_s), DoubleNear(pyScfEriMap[{{0, 0, 4, 5}}], 1e-13));
  EXPECT_THAT(hfBlock(Orbitals::OrbitalPair::z_z, Orbitals::OrbitalPair::s_s), DoubleNear(pyScfEriMap[{{0, 0, 5, 5}}], 1e-13));
  // F-F block
  ASSERT_NO_THROW(eriCalculator.getResult().at({1, 1}));
  Eigen::MatrixXd ffBlock = eriCalculator.getResult().at({1, 1}).get0();

  // H has only ss, F has ss, spx spy, spz pxpx pxpy pxpz pypy pypz pzpz
  EXPECT_EQ(ffBlock.rows(), 10);
  EXPECT_EQ(ffBlock.cols(), 10);
  EXPECT_EQ(ffBlock, Eigen::MatrixXd::Identity(10, 10));
}

// TEST_F(AnIntegralsTest, CanPerform4IndexTransformationOnefoldSymmetry) {
//  Eigen::MatrixXd moMat(2, 2);
//  moMat << 0.60040137, 0.90316332, 0.60040137, -0.90316332;
//  EriType eriAO;
//  Utils::MolecularOrbitals mos = Utils::MolecularOrbitals::createFromRestrictedCoefficients(moMat);
//  eriAO[{{0, 0, 0, 0}}] = 0.77460594;
//  eriAO[{{0, 0, 0, 1}}] = 0.22676029;
//  eriAO[{{0, 0, 1, 0}}] = 0.22676029;
//  eriAO[{{0, 0, 1, 1}}] = 0.41787409;
//  eriAO[{{0, 1, 0, 0}}] = 0.22676029;
//  eriAO[{{0, 1, 0, 1}}] = 0.09124653;
//  eriAO[{{0, 1, 1, 0}}] = 0.09124653;
//  eriAO[{{0, 1, 1, 1}}] = 0.22676029;
//  eriAO[{{1, 0, 0, 0}}] = 0.22676029;
//  eriAO[{{1, 0, 0, 1}}] = 0.09124653;
//  eriAO[{{1, 0, 1, 0}}] = 0.09124653;
//  eriAO[{{1, 0, 1, 1}}] = 0.22676029;
//  eriAO[{{1, 1, 0, 0}}] = 0.41787409;
//  eriAO[{{1, 1, 0, 1}}] = 0.22676029;
//  eriAO[{{1, 1, 1, 0}}] = 0.22676029;
//  eriAO[{{1, 1, 1, 1}}] = 0.77460594;
//  AO2MOTransformer<Utils::SecondQuantization::ERISymmetry::onefold> transformer(eriAO, mos);
//
//  auto moEri = transformer.transform();
//  for (auto mo : moEri) {
//    Utils::SecondQuantization::IndexType<4> index = mo.first;
//    EXPECT_THAT(mo.second, DoubleNear(pyScfMOEriMap[mo.first], 2e-8));
//  }
//}
//
//// General correctness is tested here, for numerical stability, see next test
// TEST_F(AnIntegralsTest, CanPerform4IndexTransformationEightfoldSymmetry) {
//  basisSet.setBasisSet("sto-3g");
//  basisSet.setStructure(structure);
//
//  auto& libint = Libint::getInstance();
//  Eigen::MatrixXd moMat(2, 2);
//  moMat << 0.60040137, 0.90316332, 0.60040137, -0.90316332;
//  Utils::MolecularOrbitals mos = Utils::MolecularOrbitals::createFromRestrictedCoefficients(moMat);
//
//  AO2MODigester<Utils::SecondQuantization::ERISymmetry::eightfold> digester(basisSet, mos);
//
//  TwoBodiesIntegralsEvaluator<AO2MODigester<Utils::SecondQuantization::ERISymmetry::eightfold>> evaluator(basisSet,
//  std::move(digester), {}); evaluator.evaluateTwoBodiesIntegrals<Operator::coulomb>();
//
//  auto moEri = evaluator.getResult();
//  for (auto mo : moEri) {
//    Utils::SecondQuantization::IndexType<4> index = mo.first;
//    EXPECT_THAT(mo.second, DoubleNear(pyScfMOEriMap[mo.first], 2e-8));
//  }
//}

// Checks numerical stability: coordinates and MOs are printed from pySCF with higher accuracy.
/*
 * from pyscf import gto,scf
 * HF = 'H 0 0 0; F 1.2 0 0'
 *
 * mol = gto.M(atom=HF, basis='sto-3g')
 * mf = scf.RHF(mol)
 * mf.init_guess='hcore'
 * mf.diis = False
 *
 * mf.kernel()
 *
 * print(mf.get_fock())
 * print(mf.energy_elec())
 * print(mf.make_rdm1())
 * mo_ints = ao2mo.kernel(mol, mf.mo_coeff, compact=False)
 * print(mo_ints)
 *
 *
 * for i in range(0,6):
 *     for j in range(0,6):
 *         for k in range(0,6):
 *             for l in range(0,6):
 *                 index1 = 6*i+j
 *                 index2 = 6*k+l
 *                 if (mo_ints[index1,index2] != 0):
 *                     print("pyScfMOEriMap2.insert({{{", i, ",", j, ",", k,
 *                         ",", l, "}}, ", mo_ints[index1,index2],
 *                         "});")
 *
 * import numpy as np
 * np.set_printoptions(precision=16)
 * print(mf.mo_coeff)
 * print(mol.atom_coords())
 */

TEST_F(AnIntegralsTest, CanSymmetrizeIndicesOnefold) {
  auto indices =
      Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::onefold>({{0, 0, 0, 0}});
  EXPECT_EQ(indices.size(), 1);
  EXPECT_EQ(indices[0], Utils::SecondQuantization::IndexType<4>({0, 0, 0, 0}));
  indices = Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::onefold>({{1, 0, 0, 0}});
  EXPECT_EQ(indices.size(), 1);
  EXPECT_EQ(indices[0], Utils::SecondQuantization::IndexType<4>({1, 0, 0, 0}));
  indices = Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::onefold>({{3, 2, 1, 0}});
  EXPECT_EQ(indices.size(), 1);
  EXPECT_EQ(indices[0], Utils::SecondQuantization::IndexType<4>({3, 2, 1, 0}));
}
TEST_F(AnIntegralsTest, CanSymmetrizeIndicesTwofold) {
  auto indices =
      Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::twofold>({{0, 0, 0, 0}});
  EXPECT_EQ(indices.size(), 1);
  EXPECT_EQ(indices[0], Utils::SecondQuantization::IndexType<4>({0, 0, 0, 0}));
  indices = Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::twofold>({{1, 0, 0, 0}});
  EXPECT_EQ(indices.size(), 2);
  EXPECT_EQ(indices[0], Utils::SecondQuantization::IndexType<4>({1, 0, 0, 0}));
  EXPECT_EQ(indices[1], Utils::SecondQuantization::IndexType<4>({0, 0, 1, 0}));
  indices = Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::twofold>({{3, 2, 1, 0}});
  EXPECT_EQ(indices.size(), 2);
  EXPECT_EQ(indices[0], Utils::SecondQuantization::IndexType<4>({3, 2, 1, 0}));
  EXPECT_EQ(indices[1], Utils::SecondQuantization::IndexType<4>({1, 0, 3, 2}));
}
TEST_F(AnIntegralsTest, CanSymmetrizeIndicesFourfold) {
  auto indices =
      Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::fourfold>({{0, 0, 0, 0}});
  EXPECT_EQ(indices.size(), 1);
  EXPECT_EQ(indices[0], Utils::SecondQuantization::IndexType<4>({0, 0, 0, 0}));
  indices = Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::fourfold>({{1, 0, 0, 0}});
  EXPECT_EQ(indices.size(), 2);
  EXPECT_EQ(indices[0], Utils::SecondQuantization::IndexType<4>({1, 0, 0, 0}));
  EXPECT_EQ(indices[1], Utils::SecondQuantization::IndexType<4>({0, 1, 0, 0}));
  indices = Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::fourfold>({{3, 2, 1, 0}});
  EXPECT_EQ(indices.size(), 4);
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({3, 2, 1, 0})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({2, 3, 1, 0})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({3, 2, 0, 1})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({2, 3, 0, 1})), indices.end());
}
TEST_F(AnIntegralsTest, CanSymmetrizeIndicesEightfold) {
  auto indices =
      Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::eightfold>({{0, 0, 0, 0}});
  EXPECT_EQ(indices.size(), 1);
  EXPECT_EQ(indices[0], Utils::SecondQuantization::IndexType<4>({0, 0, 0, 0}));
  indices =
      Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::eightfold>({{1, 1, 0, 0}});
  EXPECT_EQ(indices.size(), 2);
  EXPECT_EQ(indices[0], Utils::SecondQuantization::IndexType<4>({1, 1, 0, 0}));
  EXPECT_EQ(indices[1], Utils::SecondQuantization::IndexType<4>({0, 0, 1, 1}));
  indices =
      Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::eightfold>({{1, 0, 0, 0}});
  EXPECT_EQ(indices.size(), 4);
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({1, 0, 0, 0})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({0, 1, 0, 0})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({0, 0, 1, 0})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({0, 0, 0, 1})), indices.end());
  indices =
      Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::eightfold>({{1, 1, 1, 0}});
  EXPECT_EQ(indices.size(), 4);
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({1, 1, 0, 1})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({0, 1, 1, 1})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({1, 0, 1, 1})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({1, 1, 1, 0})), indices.end());
  indices =
      Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::eightfold>({{3, 2, 1, 0}});
  EXPECT_EQ(indices.size(), 8);
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({3, 2, 1, 0})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({2, 3, 1, 0})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({3, 2, 0, 1})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({2, 3, 0, 1})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({1, 0, 3, 2})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({1, 0, 2, 3})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({0, 1, 3, 2})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({0, 1, 2, 3})), indices.end());
  indices =
      Utils::SecondQuantization::getSymmetricIndices<Utils::SecondQuantization::ERISymmetry::eightfold>({{3, 2, 3, 2}});
  EXPECT_EQ(indices.size(), 4);
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({3, 2, 3, 2})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({3, 2, 2, 3})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({2, 3, 3, 2})), indices.end());
  EXPECT_NE(std::find(indices.begin(), indices.end(), Utils::SecondQuantization::IndexType<4>({2, 3, 2, 3})), indices.end());
}

TEST_F(AnIntegralsTest, CalculatesCoreOverlapAndDerivatives) {
  std::vector<Utils::AtomicGtos> gtos;
  Utils::AtomicGtos gtoH;
  gtoH.s = Utils::GtoExpansion{0, {{0, 3.42525091, 0.15432897}, {0, 0.62391373, 0.53532814}, {0, 0.16885540, 0.44463454}}};
  Utils::AtomicGtos gtoF;
  gtoF.s = Utils::GtoExpansion{0, {{0, 6.4648032, -0.09996723}, {0, 1.5022812, 0.39951283}, {0, 0.4885885, 0.70011547}}};
  gtoF.p = Utils::GtoExpansion{1, {{1, 6.4648032, 0.15591627}, {1, 1.5022812, 0.60768372}, {1, 0.4885885, 0.39195739}}};
  gtos.push_back(gtoH);
  gtos.push_back(gtoF);

  auto coreOrbital = Utils::SlaterToGaussian::getGTOExpansion(3, 1, 0, 4.5);
  // Accuracy is limited by accuracy of the Angstrom->Bohr transformation,
  // so copy the coordinates from pySCF
  Utils::ElementTypeCollection elements{Utils::ElementType::H, Utils::ElementType::F};
  Utils::PositionCollection positions(2, 3);
  positions << 0, 0, 0, 2.267671349478074, 0, 0;
  basisSet = BasisSetController({elements, positions}, gtos, "sto-3g-NOCORE");

  auto res = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);

  EXPECT_EQ(res.cols(), 1);
  EXPECT_EQ(res.rows(), 5);
  EXPECT_EQ(res(0, 0), 0);
  EXPECT_EQ(res(3, 0), 0);
  EXPECT_EQ(res(4, 0), 0);

  res = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::One>(0, basisSet, coreOrbital);

  EXPECT_EQ(res.cols(), 6);
  EXPECT_EQ(res.rows(), 5);
  EXPECT_EQ(res(0, 0), 0);
  EXPECT_EQ(res(3, 0), 0);
  EXPECT_EQ(res(4, 0), 0);

  positions(0, 0) += 1e-3;
  basisSet.modifyPositions(positions);
  Eigen::MatrixXd resPlus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(0, 0) -= 2e-3;
  basisSet.modifyPositions(positions);
  Eigen::MatrixXd resMinus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(0, 0) += 1e-3;
  Eigen::VectorXd resDev = -(resPlus - resMinus) / 2e-3;
  for (int i = 0; i < 5; ++i) {
    EXPECT_THAT(res(i, 0), DoubleNear(resDev(i), 1e-5));
  }

  positions(0, 1) += 1e-3;
  basisSet.modifyPositions(positions);
  resPlus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(0, 1) -= 2e-3;
  basisSet.modifyPositions(positions);
  resMinus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(0, 1) += 1e-3;

  resDev = -(resPlus - resMinus) / 2e-3;
  for (int i = 0; i < 5; ++i) {
    EXPECT_THAT(res(i, 1), DoubleNear(resDev(i), 1e-5));
  }

  positions(0, 2) += 1e-3;
  basisSet.modifyPositions(positions);
  resPlus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(0, 2) -= 2e-3;
  basisSet.modifyPositions(positions);
  resMinus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(0, 2) += 1e-3;

  resDev = -(resPlus - resMinus) / 2e-3;
  for (int i = 0; i < 5; ++i) {
    EXPECT_THAT(res(i, 2), DoubleNear(resDev(i), 1e-5));
  }

  positions(1, 0) += 1e-3;
  basisSet.modifyPositions(positions);
  resPlus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(1, 0) -= 2e-3;
  basisSet.modifyPositions(positions);
  resMinus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);

  positions(1, 0) += 1e-3;

  resDev = -(resPlus - resMinus) / 2e-3;
  for (int i = 0; i < 5; ++i) {
    EXPECT_THAT(res(i, 3), DoubleNear(resDev(i), 1e-5));
  }

  positions(1, 1) += 1e-3;
  basisSet.modifyPositions(positions);
  resPlus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(1, 1) -= 2e-3;
  basisSet.modifyPositions(positions);
  resMinus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(1, 1) += 1e-3;

  resDev = -(resPlus - resMinus) / 2e-3;
  for (int i = 0; i < 5; ++i) {
    EXPECT_THAT(res(i, 4), DoubleNear(resDev(i), 1e-5));
  }

  positions(1, 2) += 1e-3;
  basisSet.modifyPositions(positions);
  resPlus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(1, 2) -= 2e-3;
  basisSet.modifyPositions(positions);
  resMinus = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, basisSet, coreOrbital);
  positions(1, 2) += 1e-3;

  resDev = -(resPlus - resMinus) / 2e-3;
  for (int i = 0; i < 5; ++i) {
    EXPECT_THAT(res(i, 5), DoubleNear(resDev(i), 1e-5));
  }
}
} // namespace Integrals
} // namespace Sparrow
} // namespace Scine
