/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef SPARROW_MOLDENFILEGENERATOR_H
#define SPARROW_MOLDENFILEGENERATOR_H

#include "GenericMethodWrapper.h"
#include <Sparrow/Implementations/Nddo/Utils/IntegralsEvaluationUtils/GeneralTypes.h>
#include <Utils/DataStructures/AtomicGtos.h>
#include <Utils/Math/IterativeDiagonalizer/SpinAdaptedEigenContainer.h>
#include <Eigen/Core>
#include <map>
#include <memory>
#include <ostream>
namespace Scine {
namespace Sparrow {
class GenericMethodWrapper;
/**
 * @class MoldenFileGenerator @file MoldenFileGenerator.h
 * @brief Class to create the wavefunction information needed for outputting densities,...
 * Note that NDDO methods have their own STO-6G expansion, fine tuned according to their parameters,
 * while for DFTB methods the STO-6G expansion of the PM6 method was used.
 * Since STO-6G expansions are very similar one-another, the implementation ease of this was
 * deemed a satisfactory compromise.
 * If the underlying calculator has not been initialized, then care must be taken that outside of this
 * function a calculation is performed.
 * This class handles only s, p and d orbitals.
 */
class MoldenFileGenerator {
 public:
  /**
   * @brief Sets the calculator from which to generate a molden file.
   * @param calculator a reference to a GenericMethodWrapper instance.
   */
  explicit MoldenFileGenerator(const GenericMethodWrapper& calculator);
  ~MoldenFileGenerator() = default;

  /**
   * @brief Method to generate the molden input and print it somewhere.
   * @param out The stream the molden input is written to.
   */
  void generateWavefunctionInformation(std::ostream& out) const;

  /**
   * @brief Method to generate the molden input and print it somewhere.
   * @param out The stream the molden input is written to.
   */
  template<Utils::Reference restrictedness>
  void generateWavefunctionInformation(std::ostream& out, const Utils::MolecularOrbitals& mos,
                                       const Utils::SpinAdaptedContainer<restrictedness, Eigen::VectorXd>& occupations) const;

 private:
  void generateAtomBlock(std::ostream& out) const;
  void generateGTOBlock(std::ostream& out) const;
  void generateMolecularOrbitalsBlock(std::ostream& out) const;
  template<Utils::Reference restrictedness>
  void generateMolecularOrbitalsBlock(std::ostream& out, const Utils::MolecularOrbitals& mos,
                                      const Utils::SpinAdaptedContainer<restrictedness, Eigen::VectorXd>& occupations) const;
  template<Utils::Reference restrictedness>
  void write(std::ostream& out, const Utils::MolecularOrbitals& mos,
             const Utils::SpinAdaptedContainer<restrictedness, Eigen::VectorXd>& occupations,
             const Utils::SpinAdaptedContainer<restrictedness, Eigen::VectorXd>& moEnergies) const;
  void writeMOBlock(std::ostream& out, Eigen::MatrixXd moMatrix, const Eigen::VectorXd& occupation,
                    const std::vector<double>& moEnergies, const std::string& spin) const;
  const GenericMethodWrapper& calculator_;
  static std::map<int, nddo::GeneralTypes::orb_t> indexMap_;
};

template<Utils::Reference restrictedness>
inline void MoldenFileGenerator::generateWavefunctionInformation(
    std::ostream& out, const Utils::MolecularOrbitals& mos,
    const Utils::SpinAdaptedContainer<restrictedness, Eigen::VectorXd>& occupations) const {
  out << "[Molden Format]" << std::endl;
  out << "Written by Sparrow, the semiempirical library "
         "of the SCINE software package"
      << std::endl;
  generateAtomBlock(out);
  generateGTOBlock(out);
  generateMolecularOrbitalsBlock<restrictedness>(out, mos, occupations);
}

template<Utils::Reference restrictedness>
inline void MoldenFileGenerator::generateMolecularOrbitalsBlock(
    std::ostream& out, const Utils::MolecularOrbitals& mos,
    const Utils::SpinAdaptedContainer<restrictedness, Eigen::VectorXd>& occupations) const {
  out << "[5D]" << std::endl;
  out << "[7F]" << std::endl;
  out << "[9G]" << std::endl;
  out << "[MO]" << std::endl;

  // use occupation as mo energy to have orbitals in ascending occupation order!
  write(out, mos, occupations, occupations);
}
template<>
inline void MoldenFileGenerator::write<Utils::Reference::Restricted>(
    std::ostream& out, const Utils::MolecularOrbitals& mos,
    const Utils::SpinAdaptedContainer<Utils::Reference::Restricted, Eigen::VectorXd>& occupations,
    const Utils::SpinAdaptedContainer<Utils::Reference::Restricted, Eigen::VectorXd>& moEnergies) const {
  int size = moEnergies.restricted.size();
  writeMOBlock(out, mos.restrictedMatrix(), occupations.restricted,
               std::vector<double>(moEnergies.restricted.data(), moEnergies.restricted.data() + size), "Alpha");
}
template<>
inline void MoldenFileGenerator::write<Utils::Reference::Unrestricted>(
    std::ostream& out, const Utils::MolecularOrbitals& mos,
    const Utils::SpinAdaptedContainer<Utils::Reference::Unrestricted, Eigen::VectorXd>& occupations,
    const Utils::SpinAdaptedContainer<Utils::Reference::Unrestricted, Eigen::VectorXd>& moEnergies) const {
  int size = moEnergies.alpha.size();
  writeMOBlock(out, mos.alphaMatrix(), occupations.alpha,
               std::vector<double>(moEnergies.alpha.data(), moEnergies.alpha.data() + size), "Alpha");
  writeMOBlock(out, mos.betaMatrix(), occupations.beta,
               std::vector<double>(moEnergies.beta.data(), moEnergies.beta.data() + size), "Beta");
}
} // namespace Sparrow
} // namespace Scine

#endif // SPARROW_MOLDENFILEGENERATOR_H
