/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef SPARROW_HEATOFFORMATIONEVALUATOR_H
#define SPARROW_HEATOFFORMATIONEVALUATOR_H

#include "Omx/OmxElementParameters.h"
#include "Utils/Constants.h"
#include "Utils/ParameterUtils/ElementParameters.h"
#include "Utils/Typenames.h"
#include <iostream>
namespace Scine {
namespace Sparrow {
namespace HeatOfFormationEvaluation {

using Occupation = std::array<int, 4>;
enum OccupationIndex { s = 0, x = 1, y = 2, z = 3 };
/**
 * @brief Standard Heat of Formation for elements important for the OMx methods.
 * This are experimental data, but are not the most updated. But, since the
 * parameters of the OMx methods are also not the most updated, these are the
 * one that allow for the best fit to the original OMx method results.
 * Data from MOPAC
 *   Data are from CRC Handbook 1981-1982. (Indirectly through MOPAC).
 */
static const std::map<Utils::ElementType, int> standardHeatsOfFormation{
    {Utils::ElementType::H, Utils::Constants::hartree_per_kCalPerMol * 52.102},
    {Utils::ElementType::C, Utils::Constants::hartree_per_kCalPerMol * 170.890},
    {Utils::ElementType::N, Utils::Constants::hartree_per_kCalPerMol * 113.000},
    {Utils::ElementType::O, Utils::Constants::hartree_per_kCalPerMol * 59.559},
    {Utils::ElementType::F, Utils::Constants::hartree_per_kCalPerMol * 18.890}};
/**
 * @brief Calculates the Standard Heat of Formation, a typical semiempirical result.
 * This is calculated as
 *
 * \Delta H_f = E_{tot} - \sum_A (E_isol,A + \Delta H_f,A)
 *
 * \Delta H_f: The standard heat of formation of the calculated system.
 * E_{tot}: The total (electronic + nuclear repulsion) energy of the system.
 *          This quantity is what comes out of the SCF.
 * E_isol,A: This is the energy of a single atom A. Can be calculated with an
 *           SCF of the single isolated atom, or just analytically derived from
 *           the parameters of the model.
 * H_f,A: This is the standard heat of formation of the atom A. This is a parameter
 *        that can be found in the CODATA.
 */
inline auto calculateHeatOfFormation(const nddo::OmxElementParameters& parameters, double scfEnergy,
                                     const Utils::ElementTypeCollection& elements) -> double {
  return 0.0;
}
inline auto calculateIsolatedAtomEnergy(const nddo::ParametersOmx::Values::OneCenterTwoElectronIntegrals& oneCInt,
                                        const nddo::ParametersOmx::Values::Sp& oneCenterEnergies, const Occupation& o) -> double {
  if (o[x] + o[y] + o[z] == 0) {
    return o[s] * oneCenterEnergies.s;
  }

  int ns = o[s];
  int np = o[x] + o[y] + o[z];
  double hpp = 0.5 * (oneCInt.gpp - oneCInt.gp2);
  int cp[6] = {0, -1, -3, -1, 0, 0};

  double sEnergy, spEnergy, pEnergy;
  sEnergy = (ns - 1) * oneCInt.gss;
  spEnergy = ns * np * oneCInt.gsp;
  if (np >= 1) {
    pEnergy += 0.5 * np * (np -1) * oneCInt.gp2 + cp[np-1] * hpp;
    if (ns == 1) {
      spEnergy -= std::min(np, 3) * oneCInt.hsp;
    }
    else {
      spEnergy -= np * oneCInt.hsp;
    }
  }
  double twoElectronMatrixEnergy = sEnergy + spEnergy + pEnergy;

  double oneElectronMatrixEnergy = ns * oneCenterEnergies.s + np * oneCenterEnergies.p;
  return oneElectronMatrixEnergy + twoElectronMatrixEnergy;
}

} // namespace HeatOfFormationEvaluation
} // namespace Sparrow
} // namespace Scine

#endif // SPARROW_HEATOFFORMATIONEVALUATOR_H
