/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef SPARROW_NDDOSETTINGS_H
#define SPARROW_NDDOSETTINGS_H

#include <Utils/Settings.h>
#include <Utils/UniversalSettings/SettingPopulator.h>
#include <Utils/UniversalSettings/SettingsNames.h>

namespace Scine {
namespace Sparrow {
namespace nddo {

/**
 * @class OmxSettings OmxSettings.h
 * @brief The Settings specific to the Omx method.
 */
class NddoSettings : public Scine::Utils::Settings {
 public:
  NddoSettings() : Settings("OmxMethod") {
    Utils::UniversalSettings::SettingPopulator::populateLcaoSettings(_fields);
    Utils::UniversalSettings::SettingPopulator::populateScfSettings(_fields);
    Utils::UniversalSettings::SettingPopulator::populateSemiEmpiricalSettings(_fields, "");

    Utils::UniversalSettings::BoolDescriptor useNDDODipoleApprox(
        "Sets use of NDDO dipole approximation. Does nothing for OMx methods.");
    useNDDODipoleApprox.setDefaultValue(true);
    _fields.push_back(Utils::SettingsNames::NDDODipoleApproximation, std::move(useNDDODipoleApprox));
    /// CAS options
    Utils::UniversalSettings::BoolDescriptor autoCas("Do automatic CAS determination with UNO-CAS.");
    autoCas.setDefaultValue(false);
    _fields.push_back(Utils::SettingsNames::casWithUnoCI, std::move(autoCas));

    Utils::UniversalSettings::DoubleDescriptor minimalOccupationUnoCi("Occupation threshold for UNO-CI Cas detection."
                                                                      "The higher boundary is 2 - low_boundary.");
    minimalOccupationUnoCi.setMinimum(0);
    minimalOccupationUnoCi.setMaximum(1);
    minimalOccupationUnoCi.setDefaultValue(0.005);
    _fields.push_back(Utils::SettingsNames::unoCiLowThresholdOption, std::move(minimalOccupationUnoCi));

    Utils::UniversalSettings::IntDescriptor casOrbitalsAroundFermi("Number of orbitals around the fermi level "
                                                                   "to include in the cas");
    casOrbitalsAroundFermi.setDefaultValue(0);
    casOrbitalsAroundFermi.setMinimum(0);
    _fields.push_back(Utils::SettingsNames::casOrbitalsAroundFermiOption, std::move(casOrbitalsAroundFermi));
    Utils::UniversalSettings::StringDescriptor casOrbitalIndices("Comma-separated list of orbital indices to include"
                                                                 "in cas.");
    casOrbitalIndices.setDefaultValue("");
    _fields.push_back(Utils::SettingsNames::casOrbitalIndicesOption, std::move(casOrbitalIndices));
    /// CAS options end

    // FCI Dump options
    Utils::UniversalSettings::BoolDescriptor dumpIntegrals("Flags whether to perform AO2MO and dump the integrals.");
    dumpIntegrals.setDefaultValue(false);
    _fields.push_back(Utils::SettingsNames::dumpIntegralsOption, std::move(dumpIntegrals));

    Utils::UniversalSettings::FileDescriptor fciDumpFile("Name of the fciDumpFile.");
    fciDumpFile.setDefaultValue("fcidump.fci");
    _fields.push_back(Utils::SettingsNames::fcidumpFilenameOption, std::move(fciDumpFile));
    // FCI Dump options end

    resetToDefaults();
  };
};

} // namespace nddo
} // namespace Sparrow
} // namespace Scine

#endif // SPARROW_NDDOSETTINGS_H
