/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "ObjectWithIntegrals.h"
#include <Sparrow/Implementations/Nddo/Utils/FockMatrix.h>
#include <Sparrow/Implementations/Nddo/Utils/IntegralsEvaluationUtils/Global2c2eMatrix.h>
#include <Sparrow/Implementations/Nddo/Utils/IntegralsEvaluationUtils/TwoCenterIntegralContainer.h>
#include <Sparrow/Implementations/Nddo/Utils/IntegralsEvaluationUtils/oneCenterIntegralContainer.h>
#include <Sparrow/Implementations/Nddo/Utils/IntegralsEvaluationUtils/oneCenterTwoElectronIntegrals.h>
#include <Sparrow/Implementations/Nddo/Utils/TwoElectronMatrix.h>

namespace Scine {
namespace Sparrow {
namespace nddo {

auto ObjectWithIntegrals::transformIntegrals(const Utils::ElementTypeCollection& elements,
                                             const Utils::AtomsOrbitalsIndexes& aoInfos) const -> Integrals::EriType {
  const auto& oceri = getTwoElectronMatrix().getOneCenterIntegrals();
  const auto& tceri = getTwoElectronMatrix().getTwoCenterIntegrals();
  assert(int(elements.size()) == aoInfos.getNAtoms());
  Integrals::EriType eri;
  for (int a = 0; a < aoInfos.getNAtoms(); ++a) {
    int nAOsA = aoInfos.getNOrbitals(a);
    int dimA = nAOsA == 1 ? 1 : (nAOsA == 4 ? 10 : 40);

    // One center two electron integrals block
    const auto& ocInts = oceri.get(elements[a]);
    Eigen::MatrixXd eriBlock = Eigen::MatrixXd::Zero(dimA, dimA);
    for (int p = 0; p < nAOsA; ++p) {
      for (int q = 0; q < nAOsA; ++q) {
        auto bra = Integrals::Orbitals::pair(Integrals::Orbitals::Orbital(p), Integrals::Orbitals::Orbital(q));
        for (int r = 0; r < nAOsA; ++r) {
          for (int s = 0; s < nAOsA; ++s) {
            auto ket = Integrals::Orbitals::pair(Integrals::Orbitals::Orbital(r), Integrals::Orbitals::Orbital(s));
            eriBlock(bra, ket) = ocInts.get(p, q, r, s);
          }
        }
      }
    }

    Utils::MatrixWithDerivatives m;
    m.setBaseMatrix(eriBlock);
    eri[{a, a}] = m;

    // Two center two electron integrals block
    // Indices: in 2c2e first index must be smaller than second, in
    // eritype first index must be greater than second.
    // for a pair between two atoms A:1 and B:2, one has to get
    // tcInts.get(A,B) and in EriType eri.at(B,A).
    // The matrix must therefore also be transposed.
    for (int b = a + 1; b < aoInfos.getNAtoms(); ++b) {
      int nAOsB = aoInfos.getNOrbitals(b);
      int dimB = nAOsB == 1 ? 1 : (nAOsB == 4 ? 10 : 40);
      // a < b
      const auto& tcInts = *tceri.get(a, b);
      Eigen::MatrixXd eriBlockAB = Eigen::MatrixXd::Zero(dimB, dimA);
      for (int p = 0; p < nAOsB; ++p) {
        for (int q = 0; q < nAOsB; ++q) {
          auto bra = Integrals::Orbitals::pair(Integrals::Orbitals::Orbital(p), Integrals::Orbitals::Orbital(q));
          for (int r = 0; r < nAOsA; ++r) {
            for (int s = 0; s < nAOsA; ++s) {
              auto ket = Integrals::Orbitals::pair(Integrals::Orbitals::Orbital(r), Integrals::Orbitals::Orbital(s));
              eriBlock(bra, ket) = tcInts.get(r, s, p, q);
            }
          }
        }
      }
      // Canonical order with i >= j
      Utils::MatrixWithDerivatives m;
      m.setBaseMatrix(eriBlock);
      eri[{b, a}] = m;
    }
  }
  return eri;
}
} // namespace nddo
} // namespace Sparrow
} // namespace Scine
