/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef SPARROW_BASISSET_HANDLER_H
#define SPARROW_BASISSET_HANDLER_H

#include "Utils/DataStructures/AtomicGtos.h"
#include <Sparrow/Implementations/Nddo/Omx/OmxElementParameters.h>
#include <Utils/DataStructures/SlaterToGaussian.h>

namespace Scine {
namespace Sparrow {
namespace nddo {
namespace BasisSet {

namespace {
// Lambda function
inline void scaleGtfs(Utils::GtoExpansion& expansion, double exponentScaling) {
  for (auto& gtf : expansion.gtfs) {
    gtf.exponent *= exponentScaling * exponentScaling;
  }
}
} // namespace

/**
 * @brief This function returns the basis set for all elements for some custom parameters.
 */
inline auto omx(const ParametersOmx& params) {
  std::unordered_map<int, Utils::AtomicGtos> omxBasis;

  auto basis = ecp3g();
  for (auto& atomicGtos : basis) {
    int z = atomicGtos.first;
    auto gtos = atomicGtos.second;
    if (params.atomic.find(z) != params.atomic.end()) {
      double scaling = params.atomic.at(z).exponent;
      if (gtos.s) {
        scaleGtfs(*gtos.s, scaling);
      }
      if (gtos.p) {
        scaleGtfs(*gtos.p, scaling);
      }
      omxBasis.insert({z, gtos});
    }
  }
  return omxBasis;
}

/**
 * @brief This function returns the basis set for an elements for some arbitrary scaling factor.
 */
inline auto om2(Utils::ElementType element, const ParametersOmx::Values& parameters) -> Utils::AtomicGtos {
  unsigned int elementIndex = Utils::ElementInfo::Z(element);
  auto basis = ecp3g();
  double scaling = parameters.exponent;
  auto gto = basis[elementIndex];
  if (gto.s) {
    scaleGtfs(*gto.s, scaling);
  }
  if (gto.p) {
    scaleGtfs(*gto.p, scaling);
  }
  // no d as there is only up to p orbitals in OMx methods

  return gto;
}

} // namespace BasisSet
} // namespace nddo
} // namespace Sparrow
} // namespace Scine

#endif // SPARROW_BASISSET_HANDLER_H
