/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "EffectiveCorePotentialEvaluator.h"
#include "BasisHandler.h"
#include <Integrals/BasisSetController.h>
#include <Integrals/Libint.h>
#include <Integrals/TwoBodiesIntegrals/Orbitals.h>
#include <Utils/Constants.h>
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Utils/DataStructures/SlaterToGaussian.h>
#include <numeric>

namespace Scine {
namespace Sparrow {
namespace nddo {

namespace {
inline auto generateAoInfo(const OmxElementParameters& pars, const Utils::ElementTypeCollection& els) {
  auto aoInfos = Utils::AtomsOrbitalsIndexes(els.size());
  for (auto el : els) {
    aoInfos.addAtom(pars.get(el).nAOs);
  }
  return aoInfos;
}
} // namespace

EffectiveCorePotentialEvaluator::EffectiveCorePotentialEvaluator(const Utils::ElementTypeCollection& elements,
                                                                 const Utils::PositionCollection& positions,
                                                                 const OmxElementParameters& pars)
  : elements_(elements), positions_(positions), pars_(pars) {
  assert(Eigen::Index(elements.size()) == positions.rows());
  assembleECPMatrix();
}

auto EffectiveCorePotentialEvaluator::getECPMatrix() const -> const std::map<std::pair<int, int>, Eigen::VectorXd>& {
  return ecp_;
}

void EffectiveCorePotentialEvaluator::assembleECPMatrix() {
  for (unsigned atom_a = 0; atom_a < elements_.size(); ++atom_a) {
    Utils::Position posA = positions_.row(atom_a);
    const OmxElementParameters::AtomicParameters& parA = pars_.get(elements_[atom_a]);

    for (unsigned atom_b = 0; atom_b < elements_.size(); ++atom_b) {
      if (atom_a == atom_b || nCoreOrbitalOnElement(elements_[atom_b]) == 0) {
        continue;
      }
      Utils::Position posB = positions_.row(atom_b);
      const OmxElementParameters::AtomicParameters& parB = pars_.get(elements_[atom_b]);

      assert(nCoreOrbitalOnElement(elements_[atom_b]) == 1);
      ecp_.insert({{atom_a, atom_b}, getBlock(posA, posB, parA, parB)});
    }
  }
}

auto EffectiveCorePotentialEvaluator::nCoreOrbitalOnElement(Utils::ElementType el) -> int {
  return Utils::ElementInfo::Z(el) == 1 ? 0 : 1;
}

auto EffectiveCorePotentialEvaluator::getBlock(const Utils::Position& posA, const Utils::Position& posB,
                                               const OmxElementParameters::AtomicParameters& parA,
                                               const OmxElementParameters::AtomicParameters& parB) -> Eigen::VectorXd {
  Eigen::VectorXd blockAB = Eigen::VectorXd::Zero(parA.nAOs);

  auto integralValue = [](double distance, double bA, double bB, double aA, double aB) -> double {
    return 0.5 * (bA + bB) * std::sqrt(distance) * std::exp(-(aA + aB) * std::pow(distance, 2));
  };
  Utils::Position rab = posB - posA;
  double rabNorm = (rab).norm();
  const ParametersOmx::Values::ResonanceIntegrals& resA = parA.parameters.resonance;
  const ParametersOmx::Values::Pseudopotential& ecpB = parB.parameters.pseudopotential;

  blockAB(Integrals::Orbitals::Orbital::s) = integralValue(rabNorm, resA.beta.s, ecpB.beta, resA.alpha.s, ecpB.alpha);
  if (parA.nAOs == 4) {
    double localIntegral = integralValue(rabNorm, resA.beta.pSigma, ecpB.beta, resA.alpha.pSigma, ecpB.alpha);
    blockAB(Integrals::Orbitals::Orbital::x) = rabNorm < 1e-6 ? 0 : rab.x() / rabNorm * localIntegral;
    blockAB(Integrals::Orbitals::Orbital::y) = rabNorm < 1e-6 ? 0 : rab.y() / rabNorm * localIntegral;
    blockAB(Integrals::Orbitals::Orbital::z) = rabNorm < 1e-6 ? 0 : rab.z() / rabNorm * localIntegral;
  }
  if (parA.nAOs != 4 && parA.nAOs != 1) {
    throw std::runtime_error("Wrong number of atomic orbitals (atom A: " + std::to_string(parA.nAOs) +
                             ", atom B: 1 core orbital) in ECP Integral calculator.");
  }
  return blockAB;
}

auto EffectiveCorePotentialEvaluator::evaluate(int atom_a, int atom_b, const Integrals::BasisSetController& basisSet)
    -> Eigen::MatrixXd {
  auto aoInfos = generateAoInfo(pars_, elements_);

  int nAOsOnA = aoInfos.getNOrbitals(atom_a);
  int firstAoOnA = aoInfos.getFirstOrbitalIndex(atom_a);
  Eigen::MatrixXd result = Eigen::MatrixXd::Zero(nAOsOnA, nAOsOnA);

  if (aoInfos.getNOrbitals(atom_b) == 1) {
    return result;
  }

  Eigen::MatrixXd coreOverlap = Integrals::Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(
      atom_b, basisSet, generateCoreOrbital(atom_b));
  const Eigen::VectorXd& gVector = getECPMatrix().at({atom_a, atom_b});
  double faa = pars_.get(elements_[atom_b]).parameters.pseudopotential.faa;

  for (int mu = 0; mu < nAOsOnA; ++mu) {
    int m = mu + firstAoOnA;
    for (int nu = 0; nu < nAOsOnA; ++nu) {
      int n = nu + firstAoOnA;
      result(mu, nu) = -(coreOverlap(m, 0) * gVector(nu) + gVector(mu) * coreOverlap(n, 0)) -
                       coreOverlap(m, 0) * faa * coreOverlap(n, 0);
    }
  }
  return result;
}

auto EffectiveCorePotentialEvaluator::generateCoreOrbital(int atomIdx) -> Utils::GtoExpansion {
  assert(atomIdx < int(elements_.size()));
  double coreExponent = pars_.get(elements_[atomIdx]).parameters.pseudopotential.exponent;
  return Utils::SlaterToGaussian::getGTOExpansion(3, 1, 0, coreExponent);
}
} // namespace nddo
} // namespace Sparrow
} // namespace Scine
