/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef SPARROW_EFFECTIVECOREPOTENTIALEVALUTATOR_H
#define SPARROW_EFFECTIVECOREPOTENTIALEVALUTATOR_H

#include "OmxElementParameters.h"
#include <Utils/Typenames.h>

namespace Scine {
namespace Utils {
class AtomsOrbitalsIndexes;
}
namespace Sparrow {
namespace Integrals {
class BasisSetController;
}
namespace nddo {

/**
 * @brief This class evaluates the Effective Core Potential matrix G
 * @class EffectiveCorePotentialEvaluator @file EffectiveCorePotentialEvaluator.h
 * The matrix G enters the Effective Core Potential term
 *
 * V_{mu, nu},B(ECP) = -(S_{mu, alpha}G_{alpha, nu} + G_{mu, alpha}S_{alpha, nu})
 *                     - S_{mu, alpha}S_{alpha, nu}F_{alpha, alpha}
 *  with mu, nu on atom A, alpha the core orbital on atom B.
 */
class EffectiveCorePotentialEvaluator {
 public:
  /**
   * @enum Containing the indices in the local coordinate system.
   */
  enum Types : unsigned { s = 0, pSigma = 1, pPiPlus = 2, pPiMinus = 3 };

  /**
   * @brief Constructor taking the structure and the necessary parameters.
   */
  EffectiveCorePotentialEvaluator(const Utils::ElementTypeCollection& elements,
                                  const Utils::PositionCollection& positions, const OmxElementParameters& pars);
  /**
   * @brief Destructor.
   */
  ~EffectiveCorePotentialEvaluator() = default;

  /**
   * @brief Returns the calculated ecp matrix in the molecular coordinate system.
   * The rows of the matrix are the ECP-3G basis functions, the columns the core orbital.
   * This is therefore a rectangular matrix with as many rows as basis functions
   * and as many columns as core orbitals!
   * The matrix is in the form of a map of vectors of diatomic contributions.
   * So, for example, the element {0,1} of the map will have the matrix containing
   * the contribution of the G matrix of of atom 0 with the core orbital of atom 1.
   * if no core orbital is present, no element is calculated. The map is directional:
   * the element {0,1} is not the same as the element {1,0}!
   */
  auto getECPMatrix() const -> const std::map<std::pair<int, int>, Eigen::VectorXd>&;

  /**
   * @brief Assembles the resonance integral matrix in the molecular coordinate system.
   */
  void assembleECPMatrix();

  /**
   * @brief Generates the rotation matrix from the local to the molecular coordinate system.
   */
  auto generateRotationMatrix(const Utils::Position& posA, const Utils::Position& posB) -> Eigen::MatrixXd;

  /**
   * @brief Evaluate the effective core potential term
   *
   * This function implements the formula:
   * V_{mu, nu},B(ECP) = -(S_{mu, alpha}G_{alpha, nu} + G_{mu, alpha}S_{alpha, nu})
   *                     - S_{mu, alpha}S_{alpha, nu}F_{alpha, alpha}
   *  with mu, nu on atom A, alpha the core orbital on atom B.
   *
   * @returns A nAOsOnA x nAOsOnA matrix with the one center terms
   *          of the one-electron part of the Fock matrix for the ECPs for atom A
   *          due to the core on atom B.
   */
  auto evaluate(int atom_a, int atom_b, const Integrals::BasisSetController& basisSet) -> Eigen::MatrixXd;

  /**
   * @brief Generates a core s-type orbital for an element by rescaling the basis function's exponent.
   */
  auto generateCoreOrbital(int atomIdx) -> Utils::GtoExpansion;

 private:
  /**
   * @brief Calculated a (local) diatomic block of the resonance matrix and rotates it in the molecular system.
   */
  static auto getBlock(const Utils::Position& posA, const Utils::Position& posB,
                       const OmxElementParameters::AtomicParameters& parA,
                       const OmxElementParameters::AtomicParameters& parB) -> Eigen::VectorXd;

  /**
   * @brief Get the number of core atoms for each element. Returns 0 for H and 1 otherwise.
   */
  static auto nCoreOrbitalOnElement(Utils::ElementType el) -> int;

  Utils::ElementTypeCollection elements_;
  Utils::PositionCollection positions_;
  OmxElementParameters pars_;
  std::map<std::pair<int, int>, Eigen::VectorXd> ecp_;
};
} // namespace nddo
} // namespace Sparrow
} // namespace Scine
#endif // SPARROW_EFFECTIVECOREPOTENTIALEVALUTATOR_H
