/**
 * @file Calculator.h
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher
 * Group.\n See LICENSE.txt for details.
 */

#include "FockMatrix.h"
#include "KlopmanFinalizer.h"
#include "KlopmanOhnoObserver.h"
#include "OmxOneCenterTwoElectronEvaluator.h"
#include "OmxOneElectronMatrix.h"
#include "ParametersProcessor.h"
#include "Sparrow/Implementations/Nddo/Omx/AbInitioOverlapCalculator.h"
#include "TwoElectronMatrixEvaluator.h"
#include "Utils/Constants.h"
#include "Utils/Math/DerivOrderEnum.h"
#include <Integrals/BasisSetController.h>
#include <Integrals/Libint.h>
#include <Integrals/TwoBodiesIntegrals/NddoEriSaverDigester.h>
#include <Integrals/TwoBodiesIntegrals/NddoTwoBodiesIntegralsEvaluator.h>
#include <Utils/DataStructures/DensityMatrix.h>
#include <Utils/Scf/MethodInterfaces/OverlapCalculator.h>
#include <Eigen/Eigenvalues>
#include <iostream>

namespace Scine {
namespace Sparrow {
namespace nddo {

OMxFockMatrix::OMxFockMatrix(const Integrals::BasisSetController& basisSet, const Utils::DensityMatrix& densityMatrix,
                             const OmxElementParameters& elParams, const Utils::OverlapCalculator& overlapCalculator,
                             Core::Log& log)
  : libintInstance_(Integrals::Libint::getInstance()),
    basisSet_(basisSet),
    densityMatrix_(densityMatrix),
    elementParameters_(elParams),
    overlapCalculator_(overlapCalculator),
    log_(log) {
}

void OMxFockMatrix::initialize() {
  oneElectronMatrix_ = Eigen::MatrixXd::Zero(basisSet_.nBasisFunction(), basisSet_.nBasisFunction());

  if (densityMatrix_.restricted()) {
    twoElectronMatrix_.setRestrictedMatrix(oneElectronMatrix_);
  }
  else {
    twoElectronMatrix_.setRestrictedMatrix(oneElectronMatrix_);
    twoElectronMatrix_.setAlphaMatrix(oneElectronMatrix_);
    twoElectronMatrix_.setBetaMatrix(oneElectronMatrix_);
  }
}

void OMxFockMatrix::calculateDensityIndependentPart(Utils::DerivativeOrder order) {
  if (order != Utils::DerivativeOrder::Zero) {
    throw std::runtime_error("Derivatives not implemented!");
  }
  auto elements = basisSet_.getStructure().getElements();
  auto positions = basisSet_.getStructure().getPositions();
  // Two electron integrals
  koFactors_ = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(elementParameters_),
                                                  elements, positions);

  Integrals::NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet_);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(elements, elementParameters_));

  if (order == Utils::DerivativeOrder::Zero) {
    eriCalculator.evaluate<Integrals::Operator::coulomb, Utils::DerivativeOrder::Zero>();
    eris_ = eriCalculator.getResult();
    koFactors_->finalize<Utils::DerivativeOrder::Zero>(eris_);
  }
  else if (order == Utils::DerivativeOrder::One) {
    eriCalculator.evaluate<Integrals::Operator::coulomb, Utils::DerivativeOrder::One>();
    eris_ = eriCalculator.getResult();
    koFactors_->finalize<Utils::DerivativeOrder::One>(eris_);
  }
  else {
    throw std::runtime_error("Derivative higher than 1 not implemented");
  }
  notify(koFactors_);

  OmxOneElectronMatrix oem(elements, positions, elementParameters_, *koFactors_, eris_, basisSet_);
  oem.calculate(overlapCalculator_.getOverlap().getMatrixXd(), log_);
  oneElectronMatrix_ = oem();
}

void OMxFockMatrix::calculateDensityDependentPart(Utils::DerivativeOrder /*order*/) {
  TwoElectronMatrixEvaluator evaluator(eris_, densityMatrix_, basisSet_.toAtomsOrbitalsIndexes());
  twoElectronMatrix_ = evaluator.contract();
}

void OMxFockMatrix::finalize(Utils::DerivativeOrder order) {
  // Recalculate the Fock matrix: make it consistent with the obtained density matrix
  calculateDensityDependentPart(order);
}

void OMxFockMatrix::addDensityDependentElectronicContribution(std::shared_ptr<Utils::AdditiveElectronicContribution> /*contribution*/) {
}
void OMxFockMatrix::addDensityIndependentElectronicContribution(std::shared_ptr<Utils::AdditiveElectronicContribution> /*contribution*/) {
}
Utils::SpinAdaptedMatrix OMxFockMatrix::getMatrix() const {
  Utils::SpinAdaptedMatrix fockMatrix;
  if (densityMatrix_.restricted()) {
    return Utils::SpinAdaptedMatrix::createRestricted(oneElectronMatrix_ + twoElectronMatrix_.restrictedMatrix());
  }
  return Utils::SpinAdaptedMatrix::createUnrestricted(oneElectronMatrix_ + twoElectronMatrix_.alphaMatrix(),
                                                      oneElectronMatrix_ + twoElectronMatrix_.betaMatrix());
}

double OMxFockMatrix::calculateElectronicEnergy() const {
  double energy = densityMatrix_.restrictedMatrix().cwiseProduct(oneElectronMatrix_).sum();
  log_.debug << "One-electron energy contribution: " << energy << Core::Log::endl;
  if (densityMatrix_.restricted()) {
    double twoElectronPartContribution =
        0.5 * densityMatrix_.restrictedMatrix().cwiseProduct(twoElectronMatrix_.restrictedMatrix()).sum();
    log_.debug << "Two-electron energy contribution: " << twoElectronPartContribution << Core::Log::endl;
    energy += twoElectronPartContribution;
  }
  else {
    energy += 0.5 * densityMatrix_.alphaMatrix().cwiseProduct(twoElectronMatrix_.alphaMatrix()).sum();
    energy += 0.5 * densityMatrix_.betaMatrix().cwiseProduct(twoElectronMatrix_.betaMatrix()).sum();
  }
  return energy;
}

void OMxFockMatrix::addDerivatives(Utils::AutomaticDifferentiation::DerivativeContainerType<Utils::Derivative::First>& /*derivatives*/) const {
}
void OMxFockMatrix::addDerivatives(
    Utils::AutomaticDifferentiation::DerivativeContainerType<Utils::Derivative::SecondAtomic>& /*derivatives*/) const {
}
void OMxFockMatrix::addDerivatives(
    Utils::AutomaticDifferentiation::DerivativeContainerType<Utils::Derivative::SecondFull>& /*derivatives*/) const {
}

Utils::SpinAdaptedMatrix OMxFockMatrix::evaluatePotentialForDensity(const Utils::DensityMatrix& density) const {
  TwoElectronMatrixEvaluator evaluator(eris_, density, basisSet_.toAtomsOrbitalsIndexes());
  return evaluator.contract();
}

auto OMxFockMatrix::getEris() const -> const Integrals::EriType& {
  return eris_;
}
auto OMxFockMatrix::getOneElectronMatrix() const -> const Eigen::MatrixXd& {
  return oneElectronMatrix_;
}

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
