/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef SPARROW_OMXFOCKMATRIX_H
#define SPARROW_OMXFOCKMATRIX_H

#include "KlopmanOhnoObserver.h"
#include <Integrals/TwoBodiesIntegrals/Orbitals.h>
#include <Utils/DataStructures/MatrixWithDerivatives.h>
#include <Utils/DataStructures/SpinAdaptedMatrix.h>
#include <Utils/Scf/MethodInterfaces/ElectronicContributionCalculator.h>

namespace Scine {
namespace Core {
struct Log;
}

namespace Utils {
class DensityMatrix;
class OverlapCalculator;
} // namespace Utils

namespace Sparrow {

namespace Integrals {
class BasisSetController;
class Libint;
class AbInitioOverlapCalculator;
} // namespace Integrals
namespace nddo {

class OmxElementParameters;

/**
 * @class FockMatrix @file FockMatrix.h
 * @brief Class representing the Fock matrix, with methods for its evaluation.
 */
class OMxFockMatrix : public Utils::ElectronicContributionCalculator, public KlopmanOhnoShower {
 public:
  OMxFockMatrix(const Integrals::BasisSetController& basisSet, const Utils::DensityMatrix& densityMatrix,
                const OmxElementParameters& elParams, const Utils::OverlapCalculator& overlapCalculator, Core::Log& log);
  ~OMxFockMatrix() override = default;

  /*! Reinitialize after a change in the elements. */
  void initialize() final;

  /** This function will be called only once per single-point calculation. */
  void calculateDensityIndependentPart(Utils::DerivativeOrder order) override;
  /** This function will be called once per SCF iteration. */
  void calculateDensityDependentPart(Utils::DerivativeOrder order) override;
  /** This function will be called after the last iteration has run. */
  void finalize(Utils::DerivativeOrder order) final;

  /**
   * This function adds an additive electronic contribution to the
   * Hamiltonian that will be evaluated each SCF iteration.
   */
  void addDensityDependentElectronicContribution(std::shared_ptr<Utils::AdditiveElectronicContribution> contribution) final;
  /**
   * This function adds an additive electronic contribution to the Hamiltonian
   * that will be evaluated once per single-point calculation.
   */
  void addDensityIndependentElectronicContribution(std::shared_ptr<Utils::AdditiveElectronicContribution> contribution) final;

  Utils::SpinAdaptedMatrix getMatrix() const final;

  double calculateElectronicEnergy() const final;
  void addDerivatives(Utils::AutomaticDifferentiation::DerivativeContainerType<Utils::Derivative::First>& derivatives) const final;
  void addDerivatives(Utils::AutomaticDifferentiation::DerivativeContainerType<Utils::Derivative::SecondAtomic>& derivatives) const final;
  void addDerivatives(Utils::AutomaticDifferentiation::DerivativeContainerType<Utils::Derivative::SecondFull>& derivatives) const final;

  Utils::SpinAdaptedMatrix evaluatePotentialForDensity(const Utils::DensityMatrix& density) const;
  auto getEris() const -> const Integrals::EriType&;
  auto getOneElectronMatrix() const -> const Eigen::MatrixXd&;

 protected:
  /* Class members */
  Integrals::Libint& libintInstance_;
  const Integrals::BasisSetController& basisSet_;
  const Utils::DensityMatrix& densityMatrix_;
  Utils::SpinAdaptedMatrix twoElectronMatrix_;
  Eigen::MatrixXd oneElectronMatrix_;
  Integrals::EriType eris_;
  std::shared_ptr<KlopmanFinalizer> koFactors_;
  const OmxElementParameters& elementParameters_;
  const Utils::OverlapCalculator& overlapCalculator_;
  Core::Log& log_;
};

} // namespace nddo
} // namespace Sparrow
} // namespace Scine

#endif // SPARROW_OMXFOCKMATRIX_H
