/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef SPARROW_KLOPMANFINALIZER_H
#define SPARROW_KLOPMANFINALIZER_H

#include "ParametersProcessor.h"
#include "Utils/Math/AutomaticDifferentiation/AutomaticDifferentiationTypesHelper.h"
#include <Integrals/TwoBodiesIntegrals/NddoEriSaverDigester.h>
#include <Utils/Math/AutomaticDifferentiation/AutomaticDifferentiationHelpers.h>
#include <Utils/Math/AutomaticDifferentiation/MethodsTypesHelper.h>
#include <Utils/Typenames.h>

namespace Scine {
namespace Sparrow {
namespace nddo {

struct KlopmanScalingFactors;
template<Utils::DerivativeOrder order>
using KlopmanOhnoContainer = std::map<Integrals::DiatomicKey, Utils::AutomaticDifferentiation::Value3DType<order>>;

/**
 * @file KlopmanFinalizer.h @class KlopmanFinalizer
 * @brief Finalizer for the evaluation and storage of NDDO 2-electron integrals.
 * This class takes the precomputed klopman scalings (1/(2*gamma_ss^I) + 1/(2*gamma_ss^J))^2
 * and the positions (that change automatically with the given object) and apply
 * the scaling <s^Is^I|s^Js^J>^semiemp / <s^Is^I|s^Js^J>^anal to the eri.
 * <s^Is^I|s^Js^J>^anal is taken from the eris object on-the-fly.
 *
 * The full klopman-ohno coefficients are saved after the first call to "finalize"
 * in the klopmanOhnoCoefficients_ map, and can be recovered with a call to "get".
 */
struct KlopmanFinalizer {
  using iterator = KlopmanOhnoContainer<Utils::DerivativeOrder::Zero>::iterator;
  using const_iterator = KlopmanOhnoContainer<Utils::DerivativeOrder::Zero>::const_iterator;
  auto begin() const -> const_iterator {
    return klopmanOhnoCoefficients_.begin();
  }
  auto begin() -> iterator {
    return klopmanOhnoCoefficients_.begin();
  }
  auto end() const -> const_iterator {
    return klopmanOhnoCoefficients_.end();
  }
  auto end() -> iterator {
    return klopmanOhnoCoefficients_.end();
  }
  KlopmanFinalizer() = default;
  KlopmanFinalizer(const KlopmanScalingFactors& klopmanScalings, Utils::ElementTypeCollection elements,
                   const Utils::PositionCollection& positions)
    : klopman_(klopmanScalings), elements_(std::move(elements)), positions_(positions) {
  }
  KlopmanFinalizer(KlopmanFinalizer&& rhs) noexcept = default;
  KlopmanFinalizer(const KlopmanFinalizer& rhs) = default;
  KlopmanFinalizer& operator=(KlopmanFinalizer&& rhs) noexcept = default;
  KlopmanFinalizer& operator=(const KlopmanFinalizer& rhs) = default;
  ~KlopmanFinalizer() = default;
  template<Utils::DerivativeOrder order>
  void finalize(Integrals::EriType& eris);

  template<Utils::DerivativeOrder order>
  auto get(Integrals::DiatomicKey /*index*/) const -> const Utils::AutomaticDifferentiation::Value3DType<order>& {
    throw std::runtime_error("Derivative not implemented");
  }
  template<Utils::DerivativeOrder order>
  void set(Integrals::DiatomicKey index, Utils::AutomaticDifferentiation::Value3DType<order> value);
  template<Utils::DerivativeOrder order>
  auto has(Integrals::DiatomicKey index) const -> bool;

 private:
  template<Utils::DerivativeOrder order>
  auto evaluateSemiempiricalIntegral(int atom_i, int atom_j) const -> Utils::AutomaticDifferentiation::Value3DType<order>;
  KlopmanScalingFactors klopman_;
  KlopmanOhnoContainer<Utils::DerivativeOrder::Zero> klopmanOhnoCoefficients_;
  KlopmanOhnoContainer<Utils::DerivativeOrder::One> klopmanOhnoCoefficientsFirst_;
  // Not implemented
  // KlopmanOhnoContainer<Utils::DerivativeOrder::Two> klopmanOhnoCoefficientsSecond_;
  Utils::ElementTypeCollection elements_;
  Utils::PositionCollection positions_;
};

template<Utils::DerivativeOrder order>
inline void KlopmanFinalizer::finalize(Integrals::EriType& eris) {
  klopmanOhnoCoefficients_ = {};
  klopmanOhnoCoefficientsFirst_ = {};
  constexpr const Integrals::Orbitals::OrbitalPair ss = Integrals::Orbitals::s_s;
  for (auto& eri : eris) {
    auto index = eri.first;
    int atom_i = index.first;
    int atom_j = index.second;
    if (atom_i != atom_j) { // Only scale two-center eri
      auto klopmanScaling = evaluateSemiempiricalIntegral<order>(atom_i, atom_j) / eri.second.get<order>()(ss, ss);
      set<order>(index, klopmanScaling);
      eri.second.get<order>() *= klopmanScaling;
    }
  }
}

template<Utils::DerivativeOrder order>
auto KlopmanFinalizer::evaluateSemiempiricalIntegral(int atom_i, int atom_j) const
    -> Utils::AutomaticDifferentiation::Value3DType<order> {
  using std::sqrt;
  assert(atom_j < positions_.rows() && atom_i < positions_.rows());
  Utils::Position rab = positions_.row(atom_j) - positions_.row(atom_i);
  auto RSquared = Utils::AutomaticDifferentiation::toRSquared<order>(rab.x(), rab.y(), rab.z());
  RSquared += klopman_.get({Utils::ElementInfo::Z(elements_[atom_i]), Utils::ElementInfo::Z(elements_[atom_j])});
  auto invR = 1. / RSquared;

  return sqrt(invR);
}

template<>
inline auto KlopmanFinalizer::get<Utils::DerivativeOrder::Zero>(Integrals::DiatomicKey index) const -> const double& {
  return klopmanOhnoCoefficients_.at(index);
}

template<>
inline auto KlopmanFinalizer::get<Utils::DerivativeOrder::One>(Integrals::DiatomicKey index) const
    -> const Utils::AutomaticDifferentiation::First3D& {
  return klopmanOhnoCoefficientsFirst_.at(index);
}

template<>
inline void KlopmanFinalizer::set<Utils::DerivativeOrder::Zero>(Integrals::DiatomicKey index, double value) {
  klopmanOhnoCoefficients_.insert({index, value});
}

template<>
inline void KlopmanFinalizer::set<Utils::DerivativeOrder::One>(Integrals::DiatomicKey index,
                                                               Utils::AutomaticDifferentiation::First3D value) {
  klopmanOhnoCoefficients_.insert({index, value.value()});
  klopmanOhnoCoefficientsFirst_.insert({index, std::move(value)});
}

template<>
inline auto KlopmanFinalizer::has<Utils::DerivativeOrder::Zero>(Integrals::DiatomicKey index) const -> bool {
  return klopmanOhnoCoefficients_.find(index) != klopmanOhnoCoefficients_.end();
}

template<>
inline auto KlopmanFinalizer::has<Utils::DerivativeOrder::One>(Integrals::DiatomicKey index) const -> bool {
  return klopmanOhnoCoefficientsFirst_.find(index) != klopmanOhnoCoefficientsFirst_.end();
}

} // namespace nddo
} // namespace Sparrow
} // namespace Scine

#endif // SPARROW_KLOPMANFINALIZER_H
