/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "MatrixTransformer.h"
#include <Integrals/TwoBodiesIntegrals/Orbitals.h>
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Utils/DataStructures/DensityMatrix.h>

namespace Scine {
namespace Sparrow {

MatrixTransformer::MatrixTransformer(const Integrals::EriType& eris, const Utils::DensityMatrix& dm,
                                     const Utils::AtomsOrbitalsIndexes& aoInfos) {
  formErisInfos(aoInfos, dm.restricted());
  formCoulombAndExchange(eris, dm, aoInfos);
}

void MatrixTransformer::formErisInfos(const Utils::AtomsOrbitalsIndexes& aoInfos, bool isRestricted) {
  int totDimCoulomb = 0;
  eriInfos_.resize(aoInfos.getNAtoms());

  for (int atom_a = 0; atom_a < aoInfos.getNAtoms(); ++atom_a) {
    eriInfos_[atom_a] = {totDimCoulomb, std::pow(aoInfos.getNOrbitals(atom_a), 2)};
    totDimCoulomb += std::pow(aoInfos.getNOrbitals(atom_a), 2);
  }
  densityCoulomb_.restricted = Eigen::VectorXd::Zero(totDimCoulomb);
  if (!isRestricted) {
    densityCoulomb_.alpha = Eigen::VectorXd::Zero(totDimCoulomb);
    densityCoulomb_.beta = Eigen::VectorXd::Zero(totDimCoulomb);
  }
  coulomb_ = Eigen::MatrixXd::Zero(totDimCoulomb, totDimCoulomb);
}

const Eigen::MatrixXd& MatrixTransformer::coulombEriSupermatrix() const {
  return coulomb_;
}
const DensityBlock& MatrixTransformer::coulombDensitySupervector() const {
  return densityCoulomb_;
}
const Eigen::MatrixXd& MatrixTransformer::exchangeEri(int atom_a, int atom_b) const {
  return exchange_.at({atom_a, atom_b}).get0();
}
const DensityBlock& MatrixTransformer::exchangeDensity(int atom_a, int atom_b) const {
  return densityExchange_.at({atom_a, atom_b});
}

Eigen::MatrixXd MatrixTransformer::extractBlockFromVector(const Eigen::VectorXd& vector, int atom, int nAOsA) const {
  return Eigen::Map<const Eigen::MatrixXd>(vector.segment(eriInfos_[atom].first, eriInfos_[atom].second).eval().data(),
                                           nAOsA, nAOsA);
}

void MatrixTransformer::fillDensityBlock(const Utils::DensityMatrix& dm, Integrals::DiatomicKey index, int firstAA,
                                         int firstOnA, int firstOnB, int nAOsA, int nAOsB) {
  DensityBlock dBlock;
  if (dm.restricted()) {
    dBlock = {DensityBlock::getDensityMatrixBlock(dm.restrictedMatrix(), firstOnB, firstOnA, nAOsB, nAOsA), {}, {}};
  }
  else {
    dBlock = {DensityBlock::getDensityMatrixBlock(dm.restrictedMatrix(), firstOnB, firstOnA, nAOsB, nAOsA),
              DensityBlock::getDensityMatrixBlock(dm.alphaMatrix(), firstOnB, firstOnA, nAOsB, nAOsA),
              DensityBlock::getDensityMatrixBlock(dm.betaMatrix(), firstOnB, firstOnA, nAOsB, nAOsA)};
  }
#pragma omp critical(dBlock)
  { densityExchange_[index] = dBlock; }

  if (index.first == index.second) {
    densityCoulomb_.restricted->segment(firstAA, nAOsA * nAOsA) = *dBlock.restricted;
    if (dm.unrestricted()) {
      densityCoulomb_.alpha->segment(firstAA, nAOsA * nAOsA) = *dBlock.alpha;
      densityCoulomb_.beta->segment(firstAA, nAOsA * nAOsA) = *dBlock.beta;
    }
  }
}

void MatrixTransformer::formCoulombAndExchange(const Integrals::EriType& eris_, const Utils::DensityMatrix& dm,
                                               const Utils::AtomsOrbitalsIndexes& aoInfos) {
  using namespace Integrals::Orbitals;

#pragma omp parallel for schedule(dynamic)
  for (int atom_a = 0; atom_a < aoInfos.getNAtoms(); ++atom_a) {
    for (int atom_b = 0; atom_b <= atom_a; ++atom_b) {
      Integrals::DiatomicKey index{atom_a, atom_b};
      Integrals::canonicalize(index);
      int firstAA = eriInfos_[index.first].first;
      int nTotA = eriInfos_[index.first].second;
      int firstBB = eriInfos_[index.second].first;
      int nTotB = eriInfos_[index.second].second;
      int nAOsA = aoInfos.getNOrbitals(index.first);
      int nAOsB = aoInfos.getNOrbitals(index.second);
      fillDensityBlock(dm, index, firstAA, aoInfos.getFirstOrbitalIndex(index.first),
                       aoInfos.getFirstOrbitalIndex(index.second), nAOsA, nAOsB);
      const Eigen::MatrixXd& eri = eris_.at(index).get0();
      Eigen::MatrixXd coulomb = Eigen::MatrixXd::Zero(nTotA, nTotB);
      Eigen::MatrixXd exchange = Eigen::MatrixXd::Zero(nAOsA * nAOsB, nAOsA * nAOsB);
      for (int i = 0; i < nAOsA; ++i) {
        for (int j = 0; j < nAOsA; ++j) {
          for (int k = 0; k < nAOsB; ++k) {
            for (int l = 0; l < nAOsB; ++l) {
              coulomb(i * nAOsA + j, k * nAOsB + l) = eri(pair(Orbital(i), Orbital(j)), pair(Orbital(k), Orbital(l)));
              exchange(i * nAOsB + k, j * nAOsB + l) = eri(pair(Orbital(i), Orbital(j)), pair(Orbital(k), Orbital(l)));
            }
          }
        }
      }

      // Insert in supermatrix
      coulomb_.block(firstAA, firstBB, nTotA, nTotB) = coulomb;
#pragma omp critical(fillMap)
      { exchange_[index].get0() = exchange; }
    }
  }
}

} // namespace Sparrow
} // namespace Scine
