/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef SPARROW_MATRIXTRANSFORMER_H
#define SPARROW_MATRIXTRANSFORMER_H

#include <Utils/DataStructures/MatrixWithDerivatives.h>
#include <Eigen/Core>
#include <boost/functional/hash.hpp>
#include <boost/optional.hpp>
#include <unordered_map>

namespace Scine {

namespace Utils {
class DensityMatrix;
class AtomsOrbitalsIndexes;
} // namespace Utils

namespace Sparrow {

namespace Integrals {
using DiatomicKey = std::pair<int, int>;
using EriType = std::unordered_map<DiatomicKey, Utils::MatrixWithDerivatives, boost::hash<DiatomicKey>>;
} // namespace Integrals

/**
 * @brief Struct representing a block of the density matrix as a vector
 * @struct DensityBlock @file MatrixTransformer.h
 *
 * This is important for carrying out the contractions for the two-electon matrix
 * in matrix form and not element-wise.
 */
struct DensityBlock {
  boost::optional<Eigen::VectorXd> restricted;
  boost::optional<Eigen::VectorXd> alpha;
  boost::optional<Eigen::VectorXd> beta;
  static Eigen::VectorXd getDensityMatrixBlock(const Eigen::MatrixXd& densityMatrix, int firstAOOnA, int firstAOOnB,
                                               int nAOsA, int nAOsB) {
    return Eigen::Map<const Eigen::VectorXd>(densityMatrix.block(firstAOOnA, firstAOOnB, nAOsA, nAOsB).eval().data(),
                                             nAOsA * nAOsB);
  }
};

/**
 * @brief Class responsible for the formation of the coulomb and exchange density and integrals supermatrices.
 * @class MatrixTransformer @file MatrixTransformer.h
 */
class MatrixTransformer {
 public:
  /**
   * @brief Constructor.
   *
   * In the constructor the eris supermatrix/maps + the density counterparts are formed.
   */
  MatrixTransformer(const Integrals::EriType& eris, const Utils::DensityMatrix& dm, const Utils::AtomsOrbitalsIndexes& aoInfos);

  /**
   * @brief Contructs the AtomsOrbitalsIndexes counterpart of the supermatrix.
   */
  void formErisInfos(const Utils::AtomsOrbitalsIndexes& aoInfos, bool isRestricted);

  /**
   * @brief Getter for the coulomb ERI supermatrix.
   * It is a matrix with
   *
   * < \mu^I \nu^I |
   *
   * as row-index (\mu *nAOsOnI + \nu), and
   *
   * | \lambda^J \sigma^J >
   *
   * as col-index (\lambda *nAOsOnJ + \sigma).
   */
  auto coulombEriSupermatrix() const -> const Eigen::MatrixXd&;
  /**
   * @brief Getter for the coulomb Density vector.
   * It is a vector with
   *
   * | \lambda^J \sigma^J >
   *
   * as index (\lambda *nAOsOnJ + \sigma).
   * This will be contracted with the coulomb supermatrix to obtain the
   *
   * G_{\mu^I \nu^I} elements. G is the two-electron part of the Fock-matrix.
   */
  auto coulombDensitySupervector() const -> const DensityBlock&;
  /**
   * @brief Getter for the exchange ERI I-J block.
   * It is a matrix with
   *
   * < \mu^I \lambda^J|
   *
   * as row-index (\mu *nAOsOnJ + \lambda), and
   *
   * | \nu^I \sigma^J >
   *
   * as col-index (\nu *nAOsOnJ + \sigma).
   * the element < \mu^I \lambda^J|\nu^I \sigma^J > corresponds to the ERI
   * <\mu^I \nu^I|\lambda^J \sigma^J >. One index is transposed to allow for the
   * contraction with matrix-vector multiplications.
   *
   * The returned diatomic-block is a nAOsOnI*nAOsOnJ x nAOsOnI*nAOsOnJ dimensional matrix.
   *
   * It is not a supermatrix because no supermatrix exists after the transposition.
   * The NDDO forces a biatomic block to be formed.
   */
  auto exchangeEri(int atom_a, int atom_b) const -> const Eigen::MatrixXd&;
  /**
   * @brief Getter for the exchange density matrix I-J block as a vector.
   * It is a vector with
   *
   * | \nu^I \sigma^J >
   *
   * as index (\nu *nAOsOnJ + \sigma).
   * One index is transposed to allow for the contraction with matrix-vector multiplications.
   *
   * This will be contracted with the exchange I-J block to obtain the
   * G_{\mu^I \lambda^J} elements. G is the two-electron part of the Fock-matrix.
   */
  auto exchangeDensity(int atom_a, int atom_b) const -> const DensityBlock&;
  /**
   * @brief Forms a diagomal matrix block from a supervector.
   * in the linearized two-electron matrix with elements
   *
   * < \mu^I \nu^I |,
   *
   * transform a vector segment corresponding to atom I to a matrix diagonal block
   * with elements G_{\mu^I\nu^I}
   */
  auto extractBlockFromVector(const Eigen::VectorXd& vector, int atom, int nAOsA) const -> Eigen::MatrixXd;
  /**
   * @brief Forms the Coulomb supermatrix/density vector and Exchange blocks and density vectors.
   *
   * To form the coulomb and exchange matrices from a I-J block, one does
   * Matrix coulomb dimension (nAOsOnI^2, nAOsOnJ^2)
   * Matrix exchange dimension (nAOsOnI*nAOsOnJ, nAOsOnI*nAOsOnJ)
   *
   * for mu \on I,
   *   for nu \on I,
   *     for lambda \on J,
   *       for sigma \on J,
   *         coulomb(mu * nAOsOnI + nu, lambda * nAOsOnJ + sigma) = block(mu, nu, lambda, sigma)
   *         exchange(mu * nAOsOnK + lambda, nu * nAOsOnJ + sigma) = block(mu, nu, lambda, sigma)
   *
   * coulombSupermatrix.block(AA,BB) = coulomb
   * exchangeBlocksCollection.block(AB,AB) = exchange
   */
  void formCoulombAndExchange(const Integrals::EriType& eris_, const Utils::DensityMatrix& dm,
                              const Utils::AtomsOrbitalsIndexes& aoInfos);

 private:
  void fillDensityBlock(const Utils::DensityMatrix& dm, Integrals::DiatomicKey index, int firstAA, int firstOnA,
                        int firstOnB, int nAOsA, int nAOsB);
  std::vector<std::pair<int, int>> eriInfos_;
  Eigen::MatrixXd coulomb_;
  Integrals::EriType exchange_;
  DensityBlock densityCoulomb_;
  std::unordered_map<Integrals::DiatomicKey, DensityBlock, boost::hash<Integrals::DiatomicKey>> densityExchange_;
};

} // namespace Sparrow
} // namespace Scine

#endif
