/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef SPARROW_OMXELEMENTPARAMETERS_H
#define SPARROW_OMXELEMENTPARAMETERS_H

#include <Sparrow/Implementations/Nddo/Omx/ParametersOmx.h>
#include <Utils/DataStructures/AtomicGtos.h>
#include <Utils/Geometry/ElementInfo.h>
#include <Utils/Geometry/ElementTypes.h>
#include <map>

namespace Scine {
namespace Sparrow {
namespace nddo {

/*!
 * This class holds the pointers to the element-specific
 * runtime parameters
 */

class OmxElementParameters {
 public:
  /**
   * @brief contains the atomic parameters of the OMx methods.
   */
  struct AtomicParameters {
    Utils::ElementType element;
    ParametersOmx::Values parameters;
    int nAOs{0};
    Utils::AtomicGtos gtos{};
  };

  OmxElementParameters() = default;
  ~OmxElementParameters() = default;
  OmxElementParameters(OmxElementParameters&& rhs) noexcept = default;
  OmxElementParameters(const OmxElementParameters& rhs) = default;
  OmxElementParameters& operator=(OmxElementParameters&& rhs) noexcept = default;
  OmxElementParameters& operator=(const OmxElementParameters& rhs) = default;

  void clear() {
    parameters_.clear();
  }

  bool has(Utils::ElementType e) const {
    return parameters_.find(Utils::ElementInfo::Z(e)) != parameters_.end();
  }

  void insert(Utils::ElementType e, AtomicParameters parameters) {
    parameters_[Utils::ElementInfo::Z(e)] = std::move(parameters);
  }

  const AtomicParameters& get(Utils::ElementType e) const {
    if (!has(e)) {
      throw std::runtime_error("Atomic parameter for element " + Utils::ElementInfo::symbol(e) + " not set.");
    }
    return parameters_.at(Utils::ElementInfo::Z(e));
  }

  /**
   * Iterators
   * @{
   */
  using ContainerType = std::map<int, AtomicParameters>;
  using iterator = ContainerType::iterator;
  using const_iterator = ContainerType::const_iterator;
  iterator begin() {
    return parameters_.begin();
  }
  const_iterator begin() const {
    return parameters_.begin();
  }
  iterator end() {
    return parameters_.end();
  }
  const_iterator end() const {
    return parameters_.end();
  }
  /*
   * @}
   */

 private:
  std::map<int, AtomicParameters> parameters_;
};

namespace BasisSet {
std::unordered_map<int, Utils::AtomicGtos> ecp3g();
}

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
#endif // SPARROW_OMXELEMENTPARAMETERS_H
