/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "OmxInitializer.h"
#include "ParametersProcessor.h"
#include <Sparrow/Implementations/Nddo/Utils/IntegralsEvaluationUtils/oneCenterTwoElectronIntegrals.h>
#include <Sparrow/Implementations/Nddo/Utils/ParameterUtils/AtomicParameters.h>
#include <Utils/Geometry/ElementInfo.h>
#include <Utils/Scf/MethodExceptions.h>
#include <Utils/Typenames.h>
#include <cmath>
#include <numeric>

namespace Scine {
namespace Sparrow {
namespace nddo {

OmxInitializer::OmxInitializer(ModelType type) : modelType_(type) {
  //rawParameters_ = modelType_ == ModelType::Om1 ? om1() : (modelType_ == ModelType::Om2 ? om2() : om3());
  
  if (modelType_ == ModelType::Om1) {
    rawParameters_ = om1();
  }
  else if (modelType_ == ModelType::Om2) {
    rawParameters_ = om2();
  }
  else if (modelType_ == ModelType::Om3) {
    rawParameters_ = om3();
  }
  else if (modelType_ == ModelType::Odm2) {
    rawParameters_ = odm2();
  }
  else if (modelType_ == ModelType::Odm3) {
    rawParameters_ = odm3();
  }
  else if (modelType_ == ModelType::Aiqm1) {
    rawParameters_ = odm2();
  }
}

auto OmxInitializer::getBasis(const Utils::ElementTypeCollection& elements) const -> std::vector<Utils::AtomicGtos> {
  using BasisType = std::vector<Utils::AtomicGtos>;
  BasisType basis;
  std::transform(elements.begin(), elements.end(), std::back_inserter<BasisType>(basis),
                 [&](Utils::ElementType el) { return elementParameters_.get(el).gtos; });
  return basis;
}

void OmxInitializer::applyRawParameters(const Utils::ElementTypeCollection& elements) {
  elementParameters_.clear();

  // Fill atomic parameters
  for (auto e : elements) {
    if (!elementParameters_.has(e)) {
      if (rawParameters_.atomic.count(Utils::ElementInfo::Z(e)) == 0) {
        throw Utils::Methods::ParametersDoNotExistForElementException(e);
      }

      elementParameters_.insert(e, ParametersProcessor::processAtomicParameters(e, rawParameters_));
    }
  }
}

void OmxInitializer::initialize(const Utils::ElementTypeCollection& elements) {
  applyRawParameters(elements);
  nElectronsForUnchargedSpecies_ =
      std::accumulate(elements.begin(), elements.end(), 0,
                      [&](int current, Utils::ElementType el) { return current + Utils::ElementInfo::valElectrons(el); });
  coreCharges_.clear();
  aoIndexes_ = Utils::AtomsOrbitalsIndexes(elements.size());
  for (auto element : elements) {
    aoIndexes_.addAtom(elementParameters_.get(element).nAOs);
    coreCharges_.push_back(Utils::ElementInfo::valElectrons(element));
  }
}

ParametersOmx& OmxInitializer::getRawParameters() {
  return rawParameters_;
}

const ParametersOmx& OmxInitializer::getRawParameters() const {
  return rawParameters_;
}

Utils::AtomsOrbitalsIndexes OmxInitializer::getAtomsOrbitalsIndexes() const {
  return aoIndexes_;
}

unsigned OmxInitializer::getNumberElectronsForUnchargedSpecies() const {
  return nElectronsForUnchargedSpecies_;
}

std::vector<double> OmxInitializer::getCoreCharges() const {
  return coreCharges_;
}

const OmxElementParameters& OmxInitializer::getElementParameters() {
  return elementParameters_;
}

bool OmxInitializer::unrestrictedCalculationPossible() const {
  return true;
}

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
