/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef SPARROW_OMXINITIALIZER_H
#define SPARROW_OMXINITIALIZER_H

#include <Sparrow/Implementations/Nddo/Omx/OmxElementParameters.h>
#include <Sparrow/Implementations/Nddo/Omx/ParametersOmx.h>
#include <Sparrow/Implementations/Nddo/Utils/IntegralsEvaluationUtils/oneCenterIntegralContainer.h>
#include <Utils/DataStructures/AtomicGtos.h>
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Utils/Scf/MethodInterfaces/StructureDependentInitializer.h>
#include <string>
#include <vector>

namespace Scine {
namespace Sparrow {

namespace nddo {

enum class ModelType { Om1, Om2, Om3, Odm2, Odm3, Aiqm1 };
/**
 * @brief Settings for OMx methods.
 * Reads the parameters and applies them to the system of interest.
 */
class OmxInitializer : public Utils::StructureDependentInitializer {
 public:
  explicit OmxInitializer(ModelType type);

  /**
   * @brief returns which flavour of OMx is being initialized.
   * important for the parameters.
   */
  ModelType getModelType() const {
    return modelType_;
  }

  /**
   * @brief returns the gto basis for a structure.
   */
  auto getBasis(const Utils::ElementTypeCollection& elements) const -> std::vector<Utils::AtomicGtos>;

  /**
   * @brief Generates OmxElementParameters from OmxParameters.
   * i.e. converts units, precalculates stuff,...
   */
  void applyRawParameters(const Utils::ElementTypeCollection& elements);
  //! Implementation of base class methods
  // @{
  /*! Initialize the method <b>after</b> the parameters have been set or loaded. */
  void initialize(const Utils::ElementTypeCollection& elements) override;

  Utils::AtomsOrbitalsIndexes getAtomsOrbitalsIndexes() const override;
  unsigned getNumberElectronsForUnchargedSpecies() const override;
  std::vector<double> getCoreCharges() const override;
  bool unrestrictedCalculationPossible() const override;
  //! @}

  /*! Get reference to the class for raw NDDO parameters. */
  ParametersOmx& getRawParameters();
  /*! Get const reference to the class for raw NDDO parameters. */
  const ParametersOmx& getRawParameters() const;

  const OmxElementParameters& getElementParameters();

 private:
  ModelType modelType_;
  OmxElementParameters elementParameters_;
  ParametersOmx rawParameters_;

  unsigned int nElectronsForUnchargedSpecies_ = 0;
  std::vector<double> coreCharges_;
  Utils::AtomsOrbitalsIndexes aoIndexes_;
};

} // namespace nddo
} // namespace Sparrow
} // namespace Scine

#endif // SPARROW_OMXINITIALIZER_H
