/**
 * @file OmxMethod.cpp
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "OmxMethod.h"
#include "../ResultsAssembler.h"
#include "AbInitioOverlapCalculator.h"
#include "FockMatrix.h"
#include "OmxRepulsion.h"
#include "Dftd4Correction.h"
#include "NNCorrection.h"
#include <Integrals/AbInitioDensityMatrixGuessGenerator.h>
#include <Integrals/BasisSetController.h>
#include <Integrals/Libint.h>
#include <Integrals/TwoBodiesIntegrals/AO2MOTransformation.h>
#include <Sparrow/Implementations/MoldenFileGenerator.h>
#include <Sparrow/Implementations/Nddo/NddoSettings.h>
#include <Sparrow/Implementations/Nddo/TimeDependent/UnoCAS.h>
#include <Sparrow/Implementations/Sto6gParameters.h>
#include <Utils/DataStructures/SecondQuantization/CasSpecifier.h>
#include <Utils/DataStructures/SecondQuantization/FciDumping.h>
#include <Utils/Scf/LcaoUtils/SpinMode.h>
#include <Utils/UniversalSettings/SettingPopulator.h>
#include <Utils/UniversalSettings/SettingsNames.h>
#include <fstream>
#include <memory>
#include <iomanip>

namespace Scine {
namespace Sparrow {
namespace nddo {

template<typename OmxType>
OmxMethod<OmxType>::OmxMethod() : ScfMethod(true, Utils::DerivativeOrder::Zero, true) {
  this->settings_ = std::make_unique<NddoSettings>();
  auto omxInitializer = std::make_shared<OmxInitializer>(CrtpTraits<OmxType>::getModelType());
  initializer_ = omxInitializer;
  basisSet_ = std::make_unique<Integrals::BasisSetController>();
  overlapCalculator_ = std::make_shared<Integrals::AbInitioOverlapCalculator>(*basisSet_);
  densityMatrixGuess_ =
      std::make_shared<Integrals::AbInitioDensityMatrixGuessGenerator>(*basisSet_, nElectrons_, *overlapCalculator_);
  auto electronicPart = std::make_shared<OMxFockMatrix>(*basisSet_, densityMatrix_, omxInitializer->getElementParameters(),
                                                        *overlapCalculator_, this->getLog());
  auto rep = std::make_shared<OmxRepulsion>(elementTypes_, positions_);
  electronicPart->subscribe(rep);
  electronicPart_ = std::move(electronicPart);
  rep_ = (rep);
}

template<typename OmxType>
OmxMethod<OmxType>::OmxMethod(const OmxMethod& rhs) : OmxMethod<OmxType>() {
  this->results() = rhs.results();
  this->settings() = rhs.settings();
  this->setStructure(*rhs.getStructure());
  this->loadState(rhs.getState());
}

template<typename OmxType>
Utils::DensityMatrix OmxMethod<OmxType>::getDensityMatrixGuess() const {
  return Utils::ScfMethod::getDensityMatrixGuess();
}

template<typename OmxType>
void OmxMethod<OmxType>::constructBasis() {
  auto basis = this->getInitializer()->getBasis(elementTypes_);
  basisSet_->setStructure({elementTypes_, positions_}, basis);
}

template<typename OmxType>
void OmxMethod<OmxType>::initialize() {
  initializer_->initialize(elementTypes_);
  constructBasis();
  Utils::ScfMethod::initialize();
}

template<typename OmxType>
void OmxMethod<OmxType>::modifyPositions(Utils::PositionCollection newPosition) {
  GenericMethodWrapper::modifyPositions(std::move(newPosition));
  constructBasis();
}

template<typename OmxType>
auto OmxMethod<OmxType>::successfulCalculation() const -> bool {
  return hasConverged();
}
template<typename OmxType>
auto OmxMethod<OmxType>::getZPVEInclusion() const -> bool {
  return false;
}

template<typename OmxType>
Utils::SpinAdaptedMatrix OmxMethod<OmxType>::evaluatePotentialForDensity(const Utils::DensityMatrix& density) const {
  return std::dynamic_pointer_cast<OMxFockMatrix>(electronicPart_)->evaluatePotentialForDensity(density);
}

template<typename OmxType>
void OmxMethod<OmxType>::applySettings() {
  GenericMethodWrapper::applySettings();
  if (this->settings_->valid()) {
    auto spinMode =
        Utils::SpinModeInterpreter::getSpinModeFromString(this->settings_->getString(Utils::SettingsNames::spinMode));
    auto scfMixerName = this->settings_->getString(Utils::SettingsNames::mixer);
    setMolecularCharge(this->settings_->getInt(Utils::SettingsNames::molecularCharge));
    setSpinMultiplicity(this->settings_->getInt(Utils::SettingsNames::spinMultiplicity));
    setConvergenceCriteria({this->settings_->getDouble(Utils::SettingsNames::selfConsistenceCriterion),
                            this->settings_->getDouble(Utils::SettingsNames::densityRmsdCriterion)});
    setMaxIterations(this->settings_->getInt(Utils::SettingsNames::maxScfIterations));
    setScfMixer(Utils::UniversalSettings::SettingPopulator::stringToScfMixer(scfMixerName));

    if (spinMode == Utils::SpinMode::Any) {
      if (spinMultiplicity() == 1) {
        setUnrestrictedCalculation(false);
      }
      else {
        setUnrestrictedCalculation(true);
      }
    }
    if (spinMode == Utils::SpinMode::Restricted) {
      setUnrestrictedCalculation(false);
    }
    if (spinMode == Utils::SpinMode::Unrestricted) {
      setUnrestrictedCalculation(true);
    }
  }
  else {
    this->settings_->throwIncorrectSettings();
  }
}

template<typename OmxType>
void OmxMethod<OmxType>::calculateImpl(Utils::Derivative requiredDerivative) {
  Utils::ScfMethod::calculate(requiredDerivative, this->getLog());
}

template<typename OmxType>
void OmxMethod<OmxType>::assembleResults(const std::string& description) {
  GenericMethodWrapper::assembleResults(description);

  // TODO: Maybe D4 should be added here in order to be persistent
  //if (this->name() == "AIQM1") {
  //  Dftd4Correction dftd4Correction = Dftd4Correction(this->getStructure(), this->molecularCharge_);
  //  dftd4Correction.calculate();
  //  this->results().template set<Utils::Property::Energy>(getLcaoMethod().getEnergy() + dftd4Correction.getDispersionEnergy());
  //}

  if (this->requiredProperties_.containsSubSet(Utils::Property::DipoleMatrixAO)) {
    this->results().template set<Utils::Property::DipoleMatrixAO>(
        Integrals::Libint::getInstance().compute1BodyIntegrals<Integrals::Operator::emultipole1, Utils::DerivativeOrder::Zero>(
            *basisSet_));
  }
  if (this->settings().getBool(Utils::SettingsNames::dumpIntegralsOption) ||
      this->requiredProperties_.containsSubSet(Utils::Property::SQSpecifier)) {
    ResultsAssembler<OmxMethod<OmxType>> assembler(
        *this, *this, std::dynamic_pointer_cast<OMxFockMatrix>(electronicPart_)->getEris(),
        std::dynamic_pointer_cast<OMxFockMatrix>(electronicPart_)->getOneElectronMatrix(), this->settings(),
        this->results(), this->requiredProperties_.containsSubSet(Utils::Property::SQSpecifier),
        this->settings().getBool(Utils::SettingsNames::dumpIntegralsOption));
    assembler();
  }
}

template<typename OmxType>
Utils::PropertyList OmxMethod<OmxType>::possibleProperties() const {
  return Utils::Property::Energy | Utils::Property::BondOrderMatrix | Utils::Property::DensityMatrix |
         Utils::Property::AtomicCharges | Utils::Property::OverlapMatrix | Utils::Property::OrbitalEnergies |
         Utils::Property::CoefficientMatrix | Utils::Property::ElectronicOccupation | Utils::Property::Description |
         Utils::Property::Dipole | Utils::Property::SuccessfulCalculation | Utils::Property::SQSpecifier;
}

template<typename OmxType>
OmxMethod<OmxType>::~OmxMethod() = default;

auto Om2Method::getBasisImpl() -> std::unordered_map<int, Utils::AtomicGtos> {
  return Sto6g::om2();
}
Om2Method::Om2Method() = default;
Om2Method::Om2Method(const Om2Method& rhs) = default;
Om2Method::~Om2Method() = default;
auto Om3Method::getBasisImpl() -> std::unordered_map<int, Utils::AtomicGtos> {
  return Sto6g::om3();
}
Om3Method::Om3Method() = default;
Om3Method::Om3Method(const Om3Method& rhs) = default;
Om3Method::~Om3Method() = default;
auto Odm2Method::getBasisImpl() -> std::unordered_map<int, Utils::AtomicGtos> {
  return Sto6g::odm2();
}
Odm2Method::Odm2Method() = default;
Odm2Method::Odm2Method(const Odm2Method& rhs) = default;
Odm2Method::~Odm2Method() = default;
auto Odm3Method::getBasisImpl() -> std::unordered_map<int, Utils::AtomicGtos> {
  return Sto6g::odm3();
}
Odm3Method::Odm3Method() = default;
Odm3Method::Odm3Method(const Odm3Method& rhs) = default;
Odm3Method::~Odm3Method() = default;
auto Aiqm1Method::getBasisImpl() -> std::unordered_map<int, Utils::AtomicGtos> {
  return Sto6g::odm2();
}
Aiqm1Method::Aiqm1Method() = default;
Aiqm1Method::Aiqm1Method(const Aiqm1Method& rhs) = default;
Aiqm1Method::~Aiqm1Method() = default;

const Utils::Results& Aiqm1Method::calculate(std::string description) {
  const Utils::Results& results = GenericMethodWrapper::calculate(description);
  Utils::AtomCollection structure = *this->getStructure();
  int natom = structure.size();
  double electronicEnergy = results.get<Utils::Property::Energy>();

  if (natom == 1) {
    auto e = structure.getElements()[0];
    int Znumber = Utils::ElementInfo::Z(e);
    electronicEnergy = atomEnergyAIQM1[Znumber-1] + electronicEnergy - atomEnergyODM2[Znumber-1];
    this->results().set<Utils::Property::Energy>(electronicEnergy);
  }
  else {
    // TODO: Add D4 energy (it might be that this should be placed in assembleResults)
    Dftd4Correction dftd4Correction = Dftd4Correction(this->getStructure(), this->molecularCharge_);
    dftd4Correction.calculate();
    double d4Energy = dftd4Correction.getDispersionEnergy();

    // add NN energy
    NNCorrection nnCorrection = NNCorrection(this->getStructure());
    nnCorrection.calculate();
    double nnEnergy = nnCorrection.getNNenergy();

    std::cout << "ODM2* contribution                      :";
    std::cout << std::right << std::setw(15) << std::fixed << std::setprecision(8) << electronicEnergy;
    std::cout << " Hartree" << std::endl;
    std::cout << "D4 contribution                         :";
    std::cout << std::right << std::setw(15) << std::fixed << std::setprecision(8) << d4Energy;
    std::cout << " Hartree" << std::endl;

    this->results().set<Utils::Property::Energy>(electronicEnergy + d4Energy + nnEnergy);

  }

  return results;
}


} // namespace nddo
} // namespace Sparrow
} // namespace Scine
