/**
 * @file OmxMethod.h
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef SPARROW_OMXMETHOD_H
#define SPARROW_OMXMETHOD_H

#include "OmxInitializer.h"
#include <Sparrow/Implementations/GenericMethodWrapper.h>
#include <Utils/Scf/MethodInterfaces/ScfMethod.h>

namespace Scine {
namespace Utils {
namespace SecondQuantization {
template<Utils::Reference res>
class CasSpecifier;
} // namespace SecondQuantization
} // namespace Utils
namespace Sparrow {
namespace Integrals {
class BasisSetController;
} // namespace Integrals
namespace nddo {

class OmxRepulsion;
class Om2Method;
class Om3Method;
class Odm2Method;
class Odm3Method;
class Aiqm1Method;
enum class ModelType;

/**
 * @brief Crtp helper structs to define the model type as trait.
 * This avoids having to have a constructor with an argument which is problematic
 * with the clone interface.
 */
template<typename Crtp>
struct CrtpTraits {
  static_assert(true, "Crtp not implemented!");
};
/**
 * @brief Crtp helper struct to define the Om2 model type as trait.
 * This avoids having to have a constructor with an argument which is problematic
 * with the clone interface.
 */
template<>
struct CrtpTraits<Om2Method> {
  static constexpr ModelType getModelType() {
    return ModelType::Om2;
  }
};
/**
 * @brief Crtp helper struct to define the Om3 model type as trait.
 * This avoids having to have a constructor with an argument which is problematic
 * with the clone interface.
 */
template<>
struct CrtpTraits<Om3Method> {
  static constexpr ModelType getModelType() {
    return ModelType::Om3;
  }
};
/**
 * @brief Crtp helper struct to define the Odm2 model type as trait.
 * This avoids having to have a constructor with an argument which is problematic
 * with the clone interface.
 */
template<>
struct CrtpTraits<Odm2Method> {
  static constexpr ModelType getModelType() {
    return ModelType::Odm2;
  }
};
/**
 * @brief Crtp helper struct to define the Odm3 model type as trait.
 * This avoids having to have a constructor with an argument which is problematic
 * with the clone interface.
 */
template<>
struct CrtpTraits<Odm3Method> {
  static constexpr ModelType getModelType() {
    return ModelType::Odm3;
  }
};
/**
 * @brief Crtp helper struct to define the AIQM1 model type as trait.
 * This avoids having to have a constructor with an argument which is problematic
 * with the clone interface.
 */
template<>
struct CrtpTraits<Aiqm1Method> {
  static constexpr ModelType getModelType() {
    return ModelType::Aiqm1;
  }
};

template<typename OmxType>
class OmxMethod : public Utils::CloneInterface<Utils::Abstract<OmxMethod<OmxType>>, GenericMethodWrapper>,
                  public Utils::ScfMethod {
 public:
  using GenericMethodWrapper::calculate;
  using GenericMethodWrapper::getPositions;
  OmxMethod();
  OmxMethod(const OmxMethod& rhs);
  ~OmxMethod() override;

  Utils::DensityMatrix getDensityMatrixGuess() const final;
  void addElectronicContribution(std::shared_ptr<Utils::AdditiveElectronicContribution> /*contribution*/) final {
  }

  Utils::LcaoMethod& getLcaoMethod() final {
    return *this;
  }
  const Utils::LcaoMethod& getLcaoMethod() const final {
    return *this;
  }

  auto possibleProperties() const -> Utils::PropertyList final;
  void modifyPositions(Utils::PositionCollection newPosition) final;

  void applySettings() final;

  auto evaluatePotentialForDensity(const Utils::DensityMatrix& density) const -> Utils::SpinAdaptedMatrix;

  auto getInitializer() const -> std::shared_ptr<OmxInitializer> {
    return std::dynamic_pointer_cast<OmxInitializer>(initializer_);
  }

  /**
   * @brief Returns an object containing the basis coefficients and exponent.
   */
  auto getBasis() const -> std::unordered_map<int, Utils::AtomicGtos> final {
    return OmxType::getBasisImpl();
  }

 private:
  void assembleResults(const std::string& description) final;
  void initialize() final;
  auto successfulCalculation() const -> bool final;
  auto getZPVEInclusion() const -> bool final;
  void constructBasis();
  void calculateImpl(Utils::Derivative requiredDerivative) final;
  std::unique_ptr<Integrals::BasisSetController> basisSet_;
};

class Om2Method final : public Utils::CloneInterface<Om2Method, OmxMethod<Om2Method>> {
 public:
  static constexpr const char* model = "OM2";
  Om2Method();
  Om2Method(const Om2Method& rhs);
  ~Om2Method() final;
  /**
   * @brief Getter for the name of the method.
   * @returns Returns the name of the method.
   */
  std::string name() const final {
    return model;
  }
  /**
   * @brief Returns an object containing the basis coefficients and exponent.
   */
  static auto getBasisImpl() -> std::unordered_map<int, Utils::AtomicGtos>;
};

class Om3Method final : public Utils::CloneInterface<Om3Method, OmxMethod<Om3Method>> {
 public:
  static constexpr const char* model = "OM3";
  Om3Method();
  Om3Method(const Om3Method& rhs);
  ~Om3Method() final;
  /**
   * @brief Getter for the name of the method.
   * @returns Returns the name of the method.
   */
  std::string name() const final {
    return model;
  }
  /**
   * @brief Returns an object containing the basis coefficients and exponent.
   */
  static auto getBasisImpl() -> std::unordered_map<int, Utils::AtomicGtos>;
};

class Odm2Method : public Utils::CloneInterface<Odm2Method, OmxMethod<Odm2Method>> {
 public:
  static constexpr const char* model = "ODM2";
  Odm2Method();
  Odm2Method(const Odm2Method& rhs);
  ~Odm2Method();
  /**
   * @brief Getter for the name of the method.
   * @returns Returns the name of the method.
   */
  std::string name() const {
    return model;
  }
  /**
   * @brief Returns an object containing the basis coefficients and exponent.
   */
  static auto getBasisImpl() -> std::unordered_map<int, Utils::AtomicGtos>;
};

class Odm3Method final : public Utils::CloneInterface<Odm3Method, OmxMethod<Odm3Method>> {
 public:
  static constexpr const char* model = "ODM3";
  Odm3Method();
  Odm3Method(const Odm3Method& rhs);
  ~Odm3Method() final;
  /**
   * @brief Getter for the name of the method.
   * @returns Returns the name of the method.
   */
  std::string name() const final {
    return model;
  }
  /**
   * @brief Returns an object containing the basis coefficients and exponent.
   */
  static auto getBasisImpl() -> std::unordered_map<int, Utils::AtomicGtos>;
};

class Aiqm1Method final : public Odm2Method {
 public:
  static constexpr const char* model = "AIQM1";
  Aiqm1Method();
  Aiqm1Method(const Aiqm1Method& rhs);
  ~Aiqm1Method() final;
  /**
   * @brief Getter for the name of the method.
   * @returns Returns the name of the method.
   */
  std::string name() const final {
    return model;
  }
  /**
   * @brief Returns an object containing the basis coefficients and exponent.
   */
  static auto getBasisImpl() -> std::unordered_map<int, Utils::AtomicGtos>;
  /**
   * @brief Return the result of the calculation.
   */
  const Utils::Results& calculate(std::string description) final;
  std::vector<double> atomEnergyODM2{
    -0.4589994395, 0.0, 0.0, 0.0, 0.0, -4.3614347247, -7.6095532253, -12.04627788
  };
  std::vector<double> atomEnergyAIQM1{
    -0.50088038, 0.0, 0.0, 0.0, 0.0, -37.79221710, -54.53360298, -75.00986203
  };
};
} // namespace nddo
} // namespace Sparrow
} // namespace Scine

#endif // SPARROW_OMXMETHOD_H
