/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "OmxOneCenterTwoElectronEvaluator.h"
#include "OmxElementParameters.h"
#include <Integrals/TwoBodiesIntegrals/NddoEriSaverDigester.h>
#include <cassert>

namespace Scine {
namespace Sparrow {
namespace nddo {

OmxOneCenterTwoElectronEvaluator::OmxOneCenterTwoElectronEvaluator(Utils::ElementTypeCollection elements,
                                                                   const OmxElementParameters& parameters)
  : elements_(std::move(elements)) {
  fill(parameters);
}

OmxOneCenterTwoElectronEvaluator::~OmxOneCenterTwoElectronEvaluator() = default;

void OmxOneCenterTwoElectronEvaluator::evaluate(Integrals::NddoEriSaverDigester& digester, Integrals::Operator op, int atomIndex) {
  if (op != Integrals::Operator::coulomb) {
    throw std::runtime_error("Only Coulomb operator supported in OmxOneCenterTwoElectron evaluation.");
  }
  assert(atomIndex < int(elements_.size()));
  assert(oneCenterIntegrals_.find(Utils::ElementInfo::Z(elements_[atomIndex])) != oneCenterIntegrals_.end());
  digester.digest(oneCenterIntegrals_.at(Utils::ElementInfo::Z(elements_[atomIndex])), atomIndex);
}

inline auto addElementToMap(Utils::ElementType el, const OmxElementParameters::AtomicParameters& p) -> Eigen::MatrixXd {
  assert(p.nAOs == 1 || p.nAOs == 4);

  using namespace Integrals;

  Eigen::MatrixXd oneCenterBlock;
  if (p.nAOs == 1) {
    oneCenterBlock = Eigen::MatrixXd::Zero(1, 1);
    oneCenterBlock(Orbitals::pair(Orbitals::s, Orbitals::s), Orbitals::pair(Orbitals::s, Orbitals::s)) =
        p.parameters.oneCenterTwoElectrons.gss;
  }
  else if (p.nAOs == 4) {
    oneCenterBlock = Eigen::MatrixXd::Zero(10, 10);

    constexpr const Orbitals::OrbitalPair ss = Orbitals::pair(Orbitals::s, Orbitals::s);
    oneCenterBlock(ss, Orbitals::pair(Orbitals::s, Orbitals::s)) = p.parameters.oneCenterTwoElectrons.gss;
    oneCenterBlock(ss, Orbitals::pair(Orbitals::x, Orbitals::x)) = p.parameters.oneCenterTwoElectrons.gsp;
    oneCenterBlock(ss, Orbitals::pair(Orbitals::y, Orbitals::y)) = p.parameters.oneCenterTwoElectrons.gsp;
    oneCenterBlock(ss, Orbitals::pair(Orbitals::z, Orbitals::z)) = p.parameters.oneCenterTwoElectrons.gsp;

    constexpr const Orbitals::OrbitalPair xx = Orbitals::pair(Orbitals::x, Orbitals::x);
    oneCenterBlock(xx, xx) = p.parameters.oneCenterTwoElectrons.gpp;
    oneCenterBlock(xx, Orbitals::pair(Orbitals::y, Orbitals::y)) = p.parameters.oneCenterTwoElectrons.gp2;
    oneCenterBlock(xx, Orbitals::pair(Orbitals::z, Orbitals::z)) = p.parameters.oneCenterTwoElectrons.gp2;

    constexpr const Orbitals::OrbitalPair yy = Orbitals::pair(Orbitals::y, Orbitals::y);
    oneCenterBlock(yy, yy) = p.parameters.oneCenterTwoElectrons.gpp;
    oneCenterBlock(yy, Orbitals::pair(Orbitals::z, Orbitals::z)) = p.parameters.oneCenterTwoElectrons.gp2;

    constexpr const Orbitals::OrbitalPair zz = Orbitals::pair(Orbitals::z, Orbitals::z);
    oneCenterBlock(zz, zz) = p.parameters.oneCenterTwoElectrons.gpp;

    constexpr const Orbitals::OrbitalPair sx = Orbitals::pair(Orbitals::s, Orbitals::x);
    constexpr const Orbitals::OrbitalPair sy = Orbitals::pair(Orbitals::s, Orbitals::y);
    constexpr const Orbitals::OrbitalPair sz = Orbitals::pair(Orbitals::s, Orbitals::z);
    oneCenterBlock(sx, sx) = p.parameters.oneCenterTwoElectrons.hsp;
    oneCenterBlock(sy, sy) = p.parameters.oneCenterTwoElectrons.hsp;
    oneCenterBlock(sz, sz) = p.parameters.oneCenterTwoElectrons.hsp;

    const double pp2pp2 = 0.5 * (p.parameters.oneCenterTwoElectrons.gpp - p.parameters.oneCenterTwoElectrons.gp2);
    constexpr const Orbitals::OrbitalPair xy = Orbitals::pair(Orbitals::x, Orbitals::y);
    constexpr const Orbitals::OrbitalPair xz = Orbitals::pair(Orbitals::x, Orbitals::z);
    constexpr const Orbitals::OrbitalPair yz = Orbitals::pair(Orbitals::y, Orbitals::z);
    oneCenterBlock(xy, xy) = pp2pp2;
    oneCenterBlock(xz, xz) = pp2pp2;
    oneCenterBlock(yz, yz) = pp2pp2;

    oneCenterBlock = oneCenterBlock.selfadjointView<Eigen::Upper>();
  }
  else {
    throw std::runtime_error("Element " + Utils::ElementInfo::symbol(el) + " has " + std::to_string(p.nAOs) +
                             " AOs and is not supported in OMx.");
  }
  return oneCenterBlock;
}
void OmxOneCenterTwoElectronEvaluator::fill(const OmxElementParameters& parameters) {
  for (Utils::ElementType el : elements_) {
    auto it = oneCenterIntegrals_.find(Utils::ElementInfo::Z(el));
    if (it == oneCenterIntegrals_.end()) {
      oneCenterIntegrals_.insert({Utils::ElementInfo::Z(el), addElementToMap(el, parameters.get(el))});
    }
  }
}

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
