/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "OmxOneElectronMatrix.h"
#include "CoreElectronEvaluation.h"
#include "EffectiveCorePotentialEvaluator.h"
#include "OrthogonalizationCorrection.h"
#include "ResonanceIntegralEvaluator.h"
#include "Utils/IO/FormattedIOUtils.h"
#include <Core/Log.h>
#include <Utils/Constants.h>
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Utils/IO/FormattedIOUtils.h>
#include <functional>
#include <numeric>

namespace Scine {
namespace Sparrow {
namespace nddo {
OmxOneElectronMatrix::OmxOneElectronMatrix(const Utils::ElementTypeCollection& elements,
                                           const Utils::PositionCollection& positions,
                                           const OmxElementParameters& params, const KlopmanFinalizer& koFactors,
                                           const Integrals::EriType& eris, const Integrals::BasisSetController& basisSet)
  : elements_(elements), positions_(positions), params_(params), ko_(koFactors), eris_(eris), basisSet_(basisSet) {
  // if G2 and F2 are 0, then it is Om3.
  isOm3_ = std::accumulate(elements_.begin(), elements_.end(), true, [&](bool oldBool, Utils::ElementType el) {
    return (oldBool &&
            ((params_.get(el).parameters.orthogonalization.f2 + params_.get(el).parameters.orthogonalization.g2) < 1e-12));
  });
}
//! @brief Fills the one-electron matrix H by reading and calculating the required integrals.
void OmxOneElectronMatrix::calculate(const Eigen::MatrixXd& ovlp, Core::Log& log) {
  auto aoInfos = Utils::AtomsOrbitalsIndexes(elements_.size());
  for (auto el : elements_) {
    aoInfos.addAtom(params_.get(el).nAOs);
  }
  H_ = Eigen::MatrixXd::Zero(aoInfos.getNAtomicOrbitals(), aoInfos.getNAtomicOrbitals());

  ResonanceIntegralEvaluator resEval(elements_, positions_, params_);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix();

  calculateCoreElectron(aoInfos, log);
  calculateECP(aoInfos, log);
  calculateOrthogonalization(ovlp, resMat, aoInfos, log);
  calculateRemainingTerms(resMat, aoInfos);
}
/**
 * @brief Calculates the Core-Electron component of the one electron matrix
 */
void OmxOneElectronMatrix::calculateCoreElectron(const Utils::AtomsOrbitalsIndexes& aoInfos, Core::Log& log) {
  for (int atom_a = 0; atom_a < aoInfos.getNAtoms(); ++atom_a) {
    int firstAO = aoInfos.getFirstOrbitalIndex(atom_a);
    int nAOs = aoInfos.getNOrbitals(atom_a);
    Eigen::MatrixXd nuclearElectron = Eigen::MatrixXd::Zero(nAOs, nAOs);

    for (int atom_b = 0; atom_b < aoInfos.getNAtoms(); ++atom_b) {
      if (atom_b != atom_a) {
        double chargeOnB = Utils::ElementInfo::valElectrons(elements_[atom_b]);
        nuclearElectron += CoreElectronEvaluation::analyticalCoreElectron(atom_a, atom_b, chargeOnB, basisSet_, ko_);
      }
    }
    log.debug << "Core-Electron, diagonal block of atom " << atom_a << " [hartree]:" << Core::Log::nl;
    log.debug << [&nuclearElectron](std::ostream& out) { out << nuclearElectron; } << Core::Log::endl;
    H_.block(firstAO, firstAO, nAOs, nAOs) += nuclearElectron;
  }
}
/**
 * @brief Calculates the ECP component of the one electron matrix
 */
void OmxOneElectronMatrix::calculateECP(const Utils::AtomsOrbitalsIndexes& aoInfos, Core::Log& log) {
  EffectiveCorePotentialEvaluator ecpEval(elements_, positions_, params_);
  for (int atom_a = 0; atom_a < aoInfos.getNAtoms(); ++atom_a) {
    int firstAO = aoInfos.getFirstOrbitalIndex(atom_a);
    int nAOs = aoInfos.getNOrbitals(atom_a);
    Eigen::MatrixXd ecp = Eigen::MatrixXd::Zero(nAOs, nAOs);
    for (int atom_b = 0; atom_b < aoInfos.getNAtoms(); ++atom_b) {
      if (atom_b != atom_a) {
        ecp += ecpEval.evaluate(atom_a, atom_b, basisSet_);
      }
    }
    log.debug << "Effective Core Potential, diagonal block of atom " << atom_a << " [hartree]:" << Core::Log::nl;
    log.debug << [&ecp](std::ostream& out) { out << ecp; } << Core::Log::endl;
    H_.block(firstAO, firstAO, nAOs, nAOs) += ecp;
  }
}
void OmxOneElectronMatrix::calculateOrthogonalization(const Eigen::MatrixXd& ovlp, const Eigen::MatrixXd& resMat,
                                                      const Utils::AtomsOrbitalsIndexes& aoInfos, Core::Log& log) {
  OrthogonalizationCorrection orth(elements_, basisSet_, params_, ovlp, resMat);
  if (!isOm3_) {
    orth.formH(eris_);
    orth.calcLocalIntegral(ko_);
  }

  Eigen::MatrixXd orthComponent = Eigen::MatrixXd::Zero(H_.rows(), H_.cols());
  for (int atom_a = 0; atom_a < aoInfos.getNAtoms(); ++atom_a) {
    int firstAoOnA = aoInfos.getFirstOrbitalIndex(atom_a);
    int nAOsA = aoInfos.getNOrbitals(atom_a);

    orthComponent.block(firstAoOnA, firstAoOnA, nAOsA, nAOsA) += orth.getTwoCenterFirstOrderBlock(atom_a);
    if (!isOm3_) {
      orthComponent.block(firstAoOnA, firstAoOnA, nAOsA, nAOsA) += orth.getTwoCenterSecondOrderBlock(atom_a);
    }
    for (int atom_b = 0; atom_b < aoInfos.getNAtoms(); ++atom_b) {
      if (atom_b != atom_a) {
        int firstAoOnB = aoInfos.getFirstOrbitalIndex(atom_b);
        int nAOsB = aoInfos.getNOrbitals(atom_b);
        orthComponent.block(firstAoOnA, firstAoOnB, nAOsA, nAOsB) += orth.getThreeCenterFirstOrderBlock(atom_a, atom_b);
        if (!isOm3_) {
          orthComponent.block(firstAoOnA, firstAoOnB, nAOsA, nAOsB) += orth.getThreeCenterSecondOrderBlock(atom_a, atom_b);
        }
      }
    }
  }
  log.debug << "Orthogonalization correction [hartree]:" << Core::Log::nl;
  log.debug << [&orthComponent](std::ostream& out) { out << orthComponent; } << Core::Log::endl;
  H_ += orthComponent;
}

void OmxOneElectronMatrix::calculateRemainingTerms(const Eigen::MatrixXd& resMat, const Utils::AtomsOrbitalsIndexes& aoInfos) {
  H_ += resMat;
  for (int atom_a = 0; atom_a < aoInfos.getNAtoms(); ++atom_a) {
    int firstAO = aoInfos.getFirstOrbitalIndex(atom_a);

    H_.diagonal()(firstAO) += params_.get(elements_[atom_a]).parameters.oneCenterEnergy.s;
    if (aoInfos.getNOrbitals(atom_a) == 4) {
      H_.diagonal().segment(firstAO + 1, 3).array() += params_.get(elements_[atom_a]).parameters.oneCenterEnergy.p;
    }
  }
}
} // namespace nddo
} // namespace Sparrow
} // namespace Scine
