/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef SPARROW_OMXONEELECTRONMATRIX_H
#define SPARROW_OMXONEELECTRONMATRIX_H

#include <Core/Log.h>
#include <Integrals/TwoBodiesIntegrals/Orbitals.h>
#include <Sparrow/Implementations/Nddo/Omx/ParametersProcessor.h>
#include <Utils/Math/AutomaticDifferentiation/MethodsHelpers.h>
#include <Utils/Typenames.h>
#include <Eigen/Core>

namespace Scine {

namespace Core {
struct Log;
}

namespace Utils {
class MatrixWithDerivatives;
class AtomsOrbitalsIndexes;
} // namespace Utils

namespace Sparrow {

namespace Integrals {
class BasisSetController;
}

namespace nddo {
class AtomicParameters;
class ElementParameters;
class TwoCenterIntegralContainer;
class KlopmanFinalizer;

/**
 * @brief This class generates the one-electron matrix H for semi-empirical methods.
 */
class OmxOneElectronMatrix {
 public:
  //! @brief Constructor.
  OmxOneElectronMatrix(const Utils::ElementTypeCollection& elements, const Utils::PositionCollection& positions,
                       const OmxElementParameters& params, const KlopmanFinalizer& koFactors,
                       const Integrals::EriType& eris, const Integrals::BasisSetController& basisSet);
  //! @brief Fills the one-electron matrix H by reading and calculating the required integrals.
  void calculate(const Eigen::MatrixXd& ovlp, Core::Log& log);
  /**
   * @brief Calculates the Core-Electron component of the one electron matrix
   */
  void calculateCoreElectron(const Utils::AtomsOrbitalsIndexes& aoInfos, Core::Log& log);
  /**
   * @brief Calculates the ECP component of the one electron matrix
   */
  void calculateECP(const Utils::AtomsOrbitalsIndexes& aoInfos, Core::Log& log);
  /**
   * @brief Calculates the orthogonalization component of the one electron matrix.
   */
  void calculateOrthogonalization(const Eigen::MatrixXd& ovlp, const Eigen::MatrixXd& resMat,
                                  const Utils::AtomsOrbitalsIndexes& aoInfos, Core::Log& log);
  /**
   * @brief Calculates the remaining terms.
   * These are the diagonal orbital energies and the off-diagonal resonances.
   */
  void calculateRemainingTerms(const Eigen::MatrixXd& resMat, const Utils::AtomsOrbitalsIndexes& aoInfos);

  // TODO: implement derivatives
  template<Utils::Derivative O>
  void addDerivatives(Utils::AutomaticDifferentiation::DerivativeContainerType<O>& derivativeContainer,
                      const Utils::MatrixWithDerivatives& S) const;
  //! @brief Getter for the one-electron matrix H.
  const Eigen::MatrixXd& operator()() const {
    return H_;
  }

 private:
  auto calculateResonanceIntegrals() const -> Eigen::MatrixXd;
  template<Utils::Derivative O>
  void addDerivativesContribution1(Utils::AutomaticDifferentiation::DerivativeContainerType<O>& derivativeContainer,
                                   int a, int startIndex, int nAOs) const;
  template<Utils::Derivative O>
  void addDerivativesContribution2(Utils::AutomaticDifferentiation::DerivativeContainerType<O>& derivativeContainer,
                                   int a, int b, int indexA, int indexB, int nAOsA, int nAOsB,
                                   const Utils::MatrixWithDerivatives& S) const;

  const Utils::ElementTypeCollection& elements_;
  const Utils::PositionCollection& positions_;
  const OmxElementParameters& params_;
  const KlopmanFinalizer& ko_;
  const Integrals::EriType& eris_;
  const Integrals::BasisSetController& basisSet_;

  bool isOm3_{false};
  Eigen::MatrixXd H_;
};

} // namespace nddo

} // namespace Sparrow
} // namespace Scine
#endif // SPARROW_OMXONEELECTRONMATRIX_H
