/**
 * @file OmxMethod.cpp
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "OmxRepulsion.h"
#include "KlopmanFinalizer.h"
#include <Utils/Math/AutomaticDifferentiation/MethodsHelpers.h>

namespace Scine {
namespace Sparrow {
namespace nddo {

OmxRepulsion::OmxRepulsion(const Utils::ElementTypeCollection& elements, const Utils::PositionCollection& positions)
  : Utils::RepulsionCalculator(elements, positions) {
}

void OmxRepulsion::calculateRepulsion(Utils::DerivativeOrder order) {
  repulsionEnergy_ = 0;
  repulsionGradients_ = Utils::GradientCollection::Zero(elements_.size(), 3);

  // prepare charges
  std::vector<double> charges(elements_.size());
  for (unsigned i = 0; i < elements_.size(); ++i) {
    charges[i] = Utils::ElementInfo::valElectrons(elements_[i]);
  }

  // prepare hessian
  if (order == Utils::DerivativeOrder::Two) {
    Utils::AutomaticDifferentiation::Second3D dummy;
    const unsigned N = elements_.size();
    for (unsigned i = 0; i < N; ++i) {
      for (unsigned j = i + 1; j < N; ++j) {
        repulsionHessian_.insert({{i, j}, dummy});
      }
    }
  }

#if defined(_OPENMP)
  const int nThreads = omp_get_max_threads();
#else
  const int nThreads = 1;
#endif

  std::vector<double> repulsionBuffer(nThreads, 0.0);
  std::vector<Utils::GradientCollection> gradientBuffer(nThreads, repulsionGradients_);

  // parallel evaluation
  const int N = elements_.size();
  for (int i = 0; i < N; ++i) {
#pragma omp parallel for schedule(static)
    for (int j = i + 1; j < N; ++j) {
#if defined(_OPENMP)
      const int thread = omp_get_thread_num();
#else
      const int thread = 0;
#endif
      Eigen::RowVector3d distanceVector = positions_.row(j) - positions_.row(i);
      const double distance = distanceVector.norm();
      const double repulsionConstant = charges[i] * charges[j];
      Integrals::DiatomicKey index{i, j};
      Integrals::canonicalize(index);

      if (order == Utils::DerivativeOrder::Zero) {
        auto koCoef = koFactors_ ? koFactors_->get<Utils::DerivativeOrder::Zero>(index) : 1.0;
        repulsionBuffer[thread] +=
            calculatePairwiseCoreRepulsion<Utils::DerivativeOrder::Zero>(distance, koCoef * repulsionConstant);
      }
      else if (order == Utils::DerivativeOrder::One) {
        auto gradientData = Utils::AutomaticDifferentiation::get3Dfrom1D<Utils::DerivativeOrder::One>(
            calculatePairwiseCoreRepulsion<Utils::DerivativeOrder::One>(distance, repulsionConstant), distanceVector);
        if (koFactors_) {
          gradientData *= koFactors_->get<Utils::DerivativeOrder::One>(index);
        }
        repulsionBuffer[thread] += gradientData.value();
        Utils::Gradient pairwiseGradient = Utils::Gradient(gradientData.derivatives().transpose());
        Utils::AutomaticDifferentiation::addDerivativeToContainer<Utils::Derivative::First>(gradientBuffer[thread], i,
                                                                                            j, pairwiseGradient);
      }
      else if (order == Utils::DerivativeOrder::Two) {
        auto pairwiseHessian = Utils::AutomaticDifferentiation::get3Dfrom1D<Utils::DerivativeOrder::Two>(
            calculatePairwiseCoreRepulsion<Utils::DerivativeOrder::Two>(distance, repulsionConstant), distanceVector);
        if (koFactors_) {
          pairwiseHessian *= koFactors_->get<Utils::DerivativeOrder::Two>(index);
        }
        repulsionBuffer[thread] += pairwiseHessian.value();
        Utils::AutomaticDifferentiation::addDerivativeToContainer<Utils::Derivative::First>(
            gradientBuffer[thread], i, j, Utils::Gradient(pairwiseHessian.deriv()));
        repulsionHessian_.at({i, j}) = pairwiseHessian;
      }
    }
  }
  // reduction of parallel region buffers
  for (int i = 0; i < nThreads; ++i) {
    repulsionEnergy_ += repulsionBuffer[i];
    repulsionGradients_ += gradientBuffer[i];
  }
}

void OmxRepulsion::update(std::shared_ptr<KlopmanFinalizer> koFactors) {
  koFactors_ = std::move(koFactors);
}

OmxRepulsion::~OmxRepulsion() = default;
} // namespace nddo
} // namespace Sparrow
} // namespace Scine
