/**
 * @file OmxMethod.h
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef SPARROW_OMXREPULSION_H
#define SPARROW_OMXREPULSION_H

#include "KlopmanOhnoObserver.h"
#include <Utils/Math/AutomaticDifferentiation/AutomaticDifferentiationTypesHelper.h>
#include <Utils/Scf/MethodInterfaces/RepulsionCalculator.h>
#include <memory>
namespace Scine {
namespace Sparrow {
namespace nddo {

class KlopmanFinalizer;
class OMxFockMatrix;

/**
 * @class RepulsionCalculator @file RepulsionCalculator.h
 * @brief Class representing a repulsion calculator between (classical) nuclei.
 */
class OmxRepulsion final : public Utils::RepulsionCalculator, public KlopmanOhnoObserver {
 public:
  OmxRepulsion(const Utils::ElementTypeCollection& elements, const Utils::PositionCollection& positions);
  ~OmxRepulsion() final;

  /**
   * @brief Method to calculate the pairwise core-core repulsion.
   * This functions dispatches the call to the right PairwiseCoreRepulsion class
   * computing the interaction between two cores up to the derivative order specified.
   */
  void calculateRepulsion(Utils::DerivativeOrder order) final;

  /**
   * @brief Updates the Klopman Ohno factors in the observer pattern.
   */
  void update(std::shared_ptr<KlopmanFinalizer> koFactors) final;

 private:
  std::shared_ptr<KlopmanFinalizer> koFactors_;
  /**
   * @brief Use AutomaticDifferentiation to get the right values for the energy derivatives.
   */
  template<Utils::DerivativeOrder order>
  Utils::AutomaticDifferentiation::Value1DType<order> calculatePairwiseCoreRepulsion(double distance, double repulsionConstant) {
    auto R = Utils::AutomaticDifferentiation::variableWithUnitDerivative<order>(distance);
    // The result of this depends on the derivative order!
    auto inverseR = repulsionConstant / R;
    return inverseR;
  }
};

} // namespace nddo
} // namespace Sparrow
} // namespace Scine

#endif // SPARROW_OMXREPULSION_H
