/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "OrthogonalizationCorrection.h"
#include "CoreElectronEvaluation.h"
#include "OmxElementParameters.h"
#include "OmxOneCenterTwoElectronEvaluator.h"
#include "KlopmanFinalizer.h"
#include <Utils/Constants.h>
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Integrals/BasisSetController.h>
#include <Utils/Geometry/AtomCollection.h>
#include <Integrals/Libint.h>
#include <iostream>

namespace Scine {
namespace Sparrow {
namespace nddo {

OrthogonalizationCorrection::OrthogonalizationCorrection(Utils::ElementTypeCollection elements,
                                                         const Integrals::BasisSetController& basisSet,       
                                                         const OmxElementParameters& pars, const Eigen::MatrixXd& overlapMatrix,
                                                         const Eigen::MatrixXd& resonanceMatrix)
  : elements_(std::move(elements)),
    basisSet_(basisSet),
    pars_(pars),
    resonanceMatrix_(resonanceMatrix),
    overlapMatrix_(overlapMatrix),
    aoInfos_(std::make_unique<Utils::AtomsOrbitalsIndexes>(elements_.size())) {
  for (auto el : elements_) {
    int nAOs = pars_.get(el).nAOs;
    aoInfos_->addAtom(nAOs);
  }
  removeDiagonalBlock();
  threeCenterMatrix_ = formThreeCenterMatrix(resonanceMatrix_);
}

auto OrthogonalizationCorrection::getTwoCenterFirstOrderBlock(int atom_a) -> Eigen::MatrixXd {
  int nAOsOnA = aoInfos_->getNOrbitals(atom_a);
  int firstIndexOnA = aoInfos_->getFirstOrbitalIndex(atom_a);

  Eigen::MatrixXd result = Eigen::MatrixXd::Zero(nAOsOnA, nAOsOnA);
  result = overlapMatrix_.middleRows(firstIndexOnA, nAOsOnA) * resonanceMatrix_.middleCols(firstIndexOnA, nAOsOnA);
  return -pars_.get(elements_[atom_a]).parameters.orthogonalization.f1 * 0.5 * (result + result.transpose());
}

auto OrthogonalizationCorrection::getTwoCenterSecondOrderBlock(int atom_a) -> Eigen::MatrixXd {
  double f2 = pars_.get(elements_[atom_a]).parameters.orthogonalization.f2;

  int nAOsOnA = aoInfos_->getNOrbitals(atom_a);
  int firstIndexOnA = aoInfos_->getFirstOrbitalIndex(atom_a);
  Eigen::MatrixXd result = Eigen::MatrixXd::Zero(nAOsOnA, nAOsOnA);

  for (int atom_b = 0; atom_b < aoInfos_->getNAtoms(); ++atom_b) {
    if (atom_a != atom_b) {
      formHlocal(atom_a, atom_b);
      auto rotmat = generateRotationMatrix(atom_a, atom_b);
      int nAOsOnB = aoInfos_->getNOrbitals(atom_b);
      int firstIndexOnB = aoInfos_->getFirstOrbitalIndex(atom_b);
      Eigen::MatrixXd temp = Eigen::MatrixXd::Zero(nAOsOnA, nAOsOnA);
      for (int mu = 0; mu < nAOsOnA; ++mu) {
        double h_uu = HlocalB[mu];
        for (int nu = mu; nu < nAOsOnA; ++nu) {
          double h_vv = HlocalB[nu];
          for (int lambda = 0; lambda < nAOsOnB; ++lambda) {
            double h_ll = HlocalA[lambda];
            temp(mu, nu) += localOvlp(mu + firstIndexOnA, lambda + firstIndexOnB) *
                            localOvlp(nu + firstIndexOnA, lambda + firstIndexOnB) * (h_uu + h_vv - 2.0 * h_ll);
          }
          temp(nu, mu) = temp(mu, nu);
        }
      }
      if (nAOsOnA == 1) {
        result += temp;
      }
      else if (nAOsOnA == 4) {
        result += (rotmat.transpose() * temp * rotmat).eval();
      }
    }
  }
  return f2 / 8. * result;
}


void OrthogonalizationCorrection::removeDiagonalBlock() {
  int currentAO = 0;
  for (auto el : elements_) {
    int nAOs = pars_.get(el).nAOs;
    overlapMatrix_.block(currentAO, currentAO, nAOs, nAOs) = Eigen::MatrixXd::Zero(nAOs, nAOs);
    currentAO += nAOs;
  }
}

auto OrthogonalizationCorrection::formThreeCenterMatrix(const Eigen::MatrixXd& resonanceMatrix) const -> Eigen::MatrixXd {
  Eigen::MatrixXd sb = overlapMatrix_.selfadjointView<Eigen::Lower>() * resonanceMatrix;
  return sb + sb.transpose();
}

auto OrthogonalizationCorrection::getThreeCenterFirstOrderBlock(int atom_a, int atom_b) -> Eigen::MatrixXd {
  int nAOsOnA = aoInfos_->getNOrbitals(atom_a);
  int firstIndexOnA = aoInfos_->getFirstOrbitalIndex(atom_a);
  int nAOsOnB = aoInfos_->getNOrbitals(atom_b);
  int firstIndexOnB = aoInfos_->getFirstOrbitalIndex(atom_b);
  return -0.5 * getG1(atom_a, atom_b) * threeCenterMatrix_.block(firstIndexOnA, firstIndexOnB, nAOsOnA, nAOsOnB);
}

auto OrthogonalizationCorrection::getThreeCenterSecondOrderBlock(int atom_a, int atom_b) -> Eigen::MatrixXd {
  int nAOsOnA = aoInfos_->getNOrbitals(atom_a);
  int firstIndexOnA = aoInfos_->getFirstOrbitalIndex(atom_a);
  int nAOsOnB = aoInfos_->getNOrbitals(atom_b);
  int firstIndexOnB = aoInfos_->getFirstOrbitalIndex(atom_b);
  assert(atom_a != atom_b);

  Eigen::MatrixXd result = Eigen::MatrixXd::Zero(nAOsOnA, nAOsOnB);

  for (int mu = 0; mu < nAOsOnA; ++mu) {
    for (int lambda = 0; lambda < nAOsOnB; ++lambda) {
      for (int atom_c = 0; atom_c < aoInfos_->getNAtoms(); ++atom_c) {
        if (atom_c != atom_a && atom_c != atom_b) {
          int nAOsOnC = aoInfos_->getNOrbitals(atom_c);
          int firstIndexOnC = aoInfos_->getFirstOrbitalIndex(atom_c);
          auto h_ac = diagonalTerms_.at({atom_a, atom_c});
          auto h_ca = diagonalTerms_.at({atom_c, atom_a});
          auto h_bc = diagonalTerms_.at({atom_b, atom_c});
          auto h_cb = diagonalTerms_.at({atom_c, atom_b});
          double h_uu = mu == 0 ? *h_ac.first : *h_ac.second;
          double h_ll = lambda == 0 ? *h_bc.first : *h_bc.second;

          for (int rho = 0; rho < nAOsOnC; ++rho) {
            double h_rr_A = rho == 0 ? *h_ca.first : *h_ca.second;
            double h_rr_B = rho == 0 ? *h_cb.first : *h_cb.second;
            result(mu, lambda) += overlapMatrix_(mu + firstIndexOnA, rho + firstIndexOnC) *
                                  overlapMatrix_(rho + firstIndexOnC, lambda + firstIndexOnB) *
                                  (h_uu + h_ll - h_rr_A - h_rr_B);
          }
        }
      }
    }
  }
  return getG2(atom_a, atom_b) / 8. * result;
}

void OrthogonalizationCorrection::formH(const Integrals::EriType& eri) {
  using namespace Integrals::Orbitals;
  diagonalTerms_ = {};
  for (int atom_a = 0; atom_a < aoInfos_->getNAtoms(); ++atom_a) {
    const auto& par = pars_.get(elements_[atom_a]);
    int nAOs = par.nAOs;
    for (int atom_b = 0; atom_b < aoInfos_->getNAtoms(); ++atom_b) {
      if (atom_a != atom_b) {
        // Only core charge!
        double coreChargeOnB = Utils::ElementInfo::valElectrons(elements_[atom_b]);
        std::pair<boost::optional<double>, boost::optional<double>> diagElements;
        // This needs to be calculated anyhow
        diagElements.first = par.parameters.oneCenterEnergy.s +
                             CoreElectronEvaluation::getSemiempiricalCoreElectron(Orbital::s, Orbital::s,
                                                                                  {atom_a, atom_b}, eri, coreChargeOnB);
        if (nAOs == 4) {
          diagElements.second =
              par.parameters.oneCenterEnergy.p +
              1. / 3. *
                  (CoreElectronEvaluation::getSemiempiricalCoreElectron(Orbital::x, Orbital::x, {atom_a, atom_b}, eri,
                                                                        coreChargeOnB) +
                   CoreElectronEvaluation::getSemiempiricalCoreElectron(Orbital::y, Orbital::y, {atom_a, atom_b}, eri,
                                                                        coreChargeOnB) +
                   CoreElectronEvaluation::getSemiempiricalCoreElectron(Orbital::z, Orbital::z, {atom_a, atom_b}, eri,
                                                                        coreChargeOnB));
        }
        diagonalTerms_.insert({{atom_a, atom_b}, diagElements});
      }
    }
  }
}

void OrthogonalizationCorrection::formHlocal(int atom_a, int atom_b) {
  const auto& para = pars_.get(elements_[atom_a]);
  int nAOsOnA = para.nAOs;
  const auto& parb = pars_.get(elements_[atom_b]);
  int nAOsOnB = parb.nAOs;
  // Only core charge!
  double coreChargeOnB = Utils::ElementInfo::valElectrons(elements_[atom_b]);
  double coreChargeOnA = Utils::ElementInfo::valElectrons(elements_[atom_a]);

  int firstIndexOnA = aoInfos_->getFirstOrbitalIndex(atom_a);
  int firstIndexOnB = aoInfos_->getFirstOrbitalIndex(atom_b);
  
  HlocalB[0] = para.parameters.oneCenterEnergy.s - coreChargeOnB * localEri(atom_b, firstIndexOnA);
  HlocalA[0] = parb.parameters.oneCenterEnergy.s - coreChargeOnA * localEri(atom_a, firstIndexOnB);
  
  if (nAOsOnA == 4) {
    HlocalB[1] = para.parameters.oneCenterEnergy.p - coreChargeOnB * localEri(atom_b, firstIndexOnA + 1);
    HlocalB[2] = para.parameters.oneCenterEnergy.p - coreChargeOnB * localEri(atom_b, firstIndexOnA + 2);
    HlocalB[3] = para.parameters.oneCenterEnergy.p - coreChargeOnB * localEri(atom_b, firstIndexOnA + 3);
  }
  if (nAOsOnB == 4) {
    HlocalA[1] = parb.parameters.oneCenterEnergy.p - coreChargeOnA * localEri(atom_a, firstIndexOnB + 1);
    HlocalA[2] = parb.parameters.oneCenterEnergy.p - coreChargeOnA * localEri(atom_a, firstIndexOnB + 2);
    HlocalA[3] = parb.parameters.oneCenterEnergy.p - coreChargeOnA * localEri(atom_a, firstIndexOnB + 3);
  }
}

auto OrthogonalizationCorrection::getG1(int atom_a, int atom_b) const -> double {
  assert(atom_a < int(elements_.size()) && atom_b < int(elements_.size()));
  auto getG = [&](Utils::ElementType el) { return pars_.get(el).parameters.orthogonalization.g1; };
  return 0.5 * (getG(elements_[atom_a]) + getG(elements_[atom_b]));
}

auto OrthogonalizationCorrection::getG2(int atom_a, int atom_b) const -> double {
  assert(atom_a < int(elements_.size()) && atom_b < int(elements_.size()));
  auto getG = [&](Utils::ElementType el) { return pars_.get(el).parameters.orthogonalization.g2; };
  return 0.5 * (getG(elements_[atom_a]) + getG(elements_[atom_b]));
}

void OrthogonalizationCorrection::calcLocalIntegral(const KlopmanFinalizer& kofactors) {
  localEri = Eigen::MatrixXd::Zero(aoInfos_->getNAtoms(), basisSet_.nBasisFunction());
  localOvlp = Eigen::MatrixXd::Zero(basisSet_.nBasisFunction(), basisSet_.nBasisFunction());
  
  for (int atom_a = 0; atom_a < aoInfos_->getNAtoms(); ++atom_a) {
    Utils::Position posA = basisSet_.getStructure().getPosition(atom_a);
    int nAOsOnA = aoInfos_->getNOrbitals(atom_a);
    int firstIndexOnA = aoInfos_->getFirstOrbitalIndex(atom_a);
    for (int atom_b = 0; atom_b < aoInfos_->getNAtoms(); ++atom_b) {
      if (atom_a != atom_b) {
        int nAOsOnB = aoInfos_->getNOrbitals(atom_b);
        int firstIndexOnB = aoInfos_->getFirstOrbitalIndex(atom_b);
        auto elements = Utils::ElementTypeCollection{basisSet_.getStructure().getElement(atom_a),
                                                     basisSet_.getStructure().getElement(atom_b)};
        Utils::Position posB = basisSet_.getStructure().getPosition(atom_b);
        double rab = (posB - posA).norm();
        Utils::PositionCollection pos(2, 3);
        pos << 0, 0, 0, 0, 0, rab;
        auto structure = Utils::AtomCollection{elements, pos};
        std::vector<Utils::AtomicGtos> gtos;
        gtos.push_back(pars_.get(elements[0]).gtos);
        gtos.push_back(pars_.get(elements[1]).gtos);
        auto basis = Integrals::BasisSetController(structure, gtos);
        
        auto& instance = Integrals::Libint::getInstance();
        auto buffer1 = instance.compute1BodyIntegrals<Integrals::Operator::overlap, Utils::DerivativeOrder::Zero>(basis).getMatrixXd();
        localOvlp.block(firstIndexOnA, firstIndexOnB, nAOsOnA, nAOsOnB) = buffer1.block(0, nAOsOnA, nAOsOnA, nAOsOnB);

        auto localEngine = instance.getEngine({Integrals::Operator::coulomb, Utils::DerivativeOrder::Zero}, basis);
        auto const& buffer2 = localEngine.results();
        const auto& localbasis = basis.getBasisSet();

        auto atomsToShells = basis.atomsToShells();
        auto shells_a = atomsToShells[0];
        auto shells_b = atomsToShells[1];
        const auto shellIndex_a = int(shells_a[0]);
        const auto& indicesOfFirstBFinShell = localbasis.shell2bf();

        Integrals::DiatomicKey index{atom_a, atom_b};
        Integrals::canonicalize(index);
        using IntType = decltype(shells_b)::size_type;
        for (IntType i = 0; i < shells_b.size(); ++i) {
          const auto shellIndex_b = int(shells_b[i]);
          const auto iSize = localbasis[shellIndex_b].size();
          localEngine.compute2<toLibintOperator(Integrals::Operator::coulomb), libint2::BraKet::xx_xx, 0>(
              localbasis[shellIndex_b], localbasis[shellIndex_b], localbasis[shellIndex_a], localbasis[shellIndex_a]);
          int bf1 = indicesOfFirstBFinShell[i];
          Eigen::Map<const Eigen::Matrix<double, -1, -1, Eigen::RowMajor>> bufferMatrix(buffer2[0], iSize, iSize);
          Eigen::RowVectorXd bufferDiag = bufferMatrix.diagonal();
          bufferDiag *= kofactors.get<Utils::DerivativeOrder::Zero>(index);
          localEri.block(atom_a, firstIndexOnB + bf1, 1, iSize) = bufferDiag;
        }
      }
    }
  }
}

auto OrthogonalizationCorrection::generateRotationMatrix(int atomA, int atomB)
    -> Eigen::MatrixXd {
  Utils::Position posA = basisSet_.getStructure().getPosition(atomA);
  Utils::Position posB = basisSet_.getStructure().getPosition(atomB);
  Utils::Position Rab = posB - posA;
  double planeProjection = 0;
  double cosa = 1.0;
  double sina = 0.0;
  double cosb = 0;
  double sinb = 0;
  planeProjection = std::sqrt(Rab.x() * Rab.x() + Rab.y() * Rab.y());
  if (planeProjection != 0) {
    cosa = Rab.x() / planeProjection;
    sina = Rab.y() / planeProjection;
  }
  cosb = (Rab.z() / Rab.norm());
  sinb = planeProjection / Rab.norm();

  enum Cartesian : unsigned { s = 0, x = 1, y = 2, z = 3 };
  // Diatomic x Cartesian coordinate system
  Eigen::MatrixXd rotationMatrix = Eigen::MatrixXd::Zero(4, 4);
  rotationMatrix(Types::s, Cartesian::s) = 1;
  rotationMatrix(Types::pSigma, Cartesian::z) = cosb;
  rotationMatrix(Types::pSigma, Cartesian::x) = sinb * cosa;
  rotationMatrix(Types::pSigma, Cartesian::y) = sinb * sina;
  rotationMatrix(Types::pPiPlus, Cartesian::z) = -sinb;
  rotationMatrix(Types::pPiPlus, Cartesian::x) = cosb * cosa;
  rotationMatrix(Types::pPiPlus, Cartesian::y) = cosb * sina;
  rotationMatrix(Types::pPiMinus, Cartesian::z) = 0;
  rotationMatrix(Types::pPiMinus, Cartesian::x) = -sina;
  rotationMatrix(Types::pPiMinus, Cartesian::y) = cosa;

  return rotationMatrix;
}

OrthogonalizationCorrection::~OrthogonalizationCorrection() = default;

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
