/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "KlopmanOhnoObserver.h"
#include <Integrals/TwoBodiesIntegrals/Orbitals.h>
#include <Utils/Typenames.h>
#include <boost/optional.hpp>
namespace Scine {
namespace Utils {
class AtomsOrbitalsIndexes;
}
namespace Sparrow {
namespace Integrals {
class BasisSetController;
} // namespace Integrals
namespace nddo {
class OmxElementParameters;
/**
 * @brief Class responsible for the calculation of the orthogonalization corrections.
 * @class OthogonalizationCorrection @file OthogonalizationCorrection.h
 * The orthogonalization correction is split in 2 parts: a 2-center part and
 * a 3-center part.
 *
 * This is the two-center part.
 * That means, this class evaluates
 *
 * V_{mu, nu},B(ORT), with mu and nu on A.
 *
 * \f$ V_{\mu,\nu},B(ORT) = -1/2 * F_1^A \sum_{\lambda \in B}
 * (S_{\mu \lambda}\Beta_{\lambda \nu} +\Beta_{\mu \lambda}S_{\lambda \nu})
 * + 1/8 F_2^A \sum_{\lambda \in B} (S_{\mu \lambda}S_{\lambda \nu}
 * (H_{\mu \mu} + H_{\nu \nu} - 2H_{\lambda\lambda}) )
 * \f$
 * The first part is calculated by multiplying together blocks of the overlap and resonance matrix.
 * The second one is done by calculating the local diagonal elements H_{\mu \mu}
 *
 * H_{\mu \mu} = U_{\mu \mu} + V_{\mu \mu},B^s
 * H_{\nu \nu} = U_{\nu \nu} + V_{\nu \nu},B^s
 * H_{\lambda \lambda} = U_{\lambda \lambda} + V_{\lambda \lambda},A^s
 * (see CoreElectronEvaluation, getSemiempiricalCoreElectron())
 *
 * This is the three-center part.
 * That means, this class evaluates
 *
 * V_{mu, lambda},C(ORT), with mu on A and lambda on B.
 *
 * \f$ V_{\mu,\lambda},C(ORT) = -1/2 * G_1^{AB} \sum_{\rho \in C}
 * (S_{\mu \rho}\Beta_{\rho \lambda} +\Beta_{\mu \rho}S_{\rho \lambda})
 * + 1/8 G_2^{AB} \sum_{\rho \in C} (S_{\mu \rho}S_{\rho \lambda}
 * (H_{\mu \mu} + H_{\lambda \lambda} - 2H_{\rho\rho}) )
 * \f$
 * Where the coefficients are given by:
 * G_1^{AB} = 1/2 * (G_1^A + G_1^B)
 * G_2^{AB} = 1/2 * (G_2^A + G_2^B)
 *
 * The first part is calculated by multiplying together blocks of the overlap and resonance matrix.
 * The second one is done by calculating the local diagonal elements H_{\mu \mu}
 * H_{\mu \mu} + H_{\lambda \lambda} - 2H_{\rho\rho} =
 * H_{\mu \mu}^{AC} + H_{\lambda \lambda}^{BC} - H_{\rho\rho}^{CA} - H_{\rho\rho}^{C}
 *
 * H_{\mu \mu}^{AC} = U_{\mu \mu} + V_{\mu \mu},C^s
 * H_{\lambda \lambda}^{BC} = U_{\lambda \lambda} + V_{\lambda \lambda},C^s
 * H_{\rho \rho}^{CA} = U_{\rho \rho} + V_{\rho \rho},A^s
 * H_{\rho \rho}^{CB} = U_{\rho \rho} + V_{\rho \rho},B^s
 * (see CoreElectronEvaluation, getSemiempiricalCoreElectron())
 *
 * Note: the three-center correction is calculated only for C =/= A and C =/= B.
 *
 */
class OrthogonalizationCorrection {
 public:
  enum Types : unsigned { s = 0, pPiPlus = 1, pPiMinus = 2, pSigma = 3 };
  OrthogonalizationCorrection(Utils::ElementTypeCollection elements, const Integrals::BasisSetController& basisSet,
                              const OmxElementParameters& pars,
                              const Eigen::MatrixXd& overlapMatrix, const Eigen::MatrixXd& resonanceMatrix);
  ~OrthogonalizationCorrection();

  /**
   * @brief Gets the mu,nu (both on A) block with all the contributions of the B atoms
   */
  auto getTwoCenterFirstOrderBlock(int atom_a) -> Eigen::MatrixXd;

  /**
   * @brief Get A-A block of the second order two center orthogonalization correction.
   * So, calculate:
   *
   * 1/8 F_2^A \sum_{\lambda \in B} (S_{\mu \lambda}S_{\lambda \nu}
   * (H_{\mu \mu} + H_{\nu \nu} - 2H_{\lambda\lambda}) )
   * for all mu, nu (on A) in the A-A block.
   */
  auto getTwoCenterSecondOrderBlock(int atom_a) -> Eigen::MatrixXd;

  /**
   * @brief Gets the mu on A, lambda on B block with all the contributions of the C atoms
   */
  auto getThreeCenterFirstOrderBlock(int atom_a, int atom_b) -> Eigen::MatrixXd;

  /**
   * @brief Get A-B block of the second order three center orthogonalization correction.
   * So, calculate:
   *
   * 1/8 G_2^{AB} \sum_{\rho \in C} (S_{\mu \rho}S_{\rho \lambda}
   * (H_{\mu \mu} + H_{\lambda \lambda} - 2H_{\rho\rho}) )
   * for all mu (on A) and lambda (on B) in the A-B block.
   */
  auto getThreeCenterSecondOrderBlock(int atom_a, int atom_b) -> Eigen::MatrixXd;

  void formHlocal(int atom_a, int atom_b);
  void calcLocalIntegral(const KlopmanFinalizer& kofactors);
  void formH(const Integrals::EriType& eri);

  auto getG1(int atom_a, int atom_b) const -> double;
  auto getG2(int atom_a, int atom_b) const -> double;
  auto generateRotationMatrix(int atom_a, int atom_b) -> Eigen::MatrixXd;

 private:
  void removeDiagonalBlock();
  auto formThreeCenterMatrix(const Eigen::MatrixXd& resonanceMatrix) const -> Eigen::MatrixXd;
  Utils::ElementTypeCollection elements_;
  const OmxElementParameters& pars_;
  const Eigen::MatrixXd& resonanceMatrix_;
  Eigen::MatrixXd overlapMatrix_;
  Eigen::MatrixXd threeCenterMatrix_;
  std::unique_ptr<Utils::AtomsOrbitalsIndexes> aoInfos_;
  std::map<Integrals::DiatomicKey, std::pair<boost::optional<double>, boost::optional<double>>> diagonalTerms_;
  Eigen::MatrixXd localOvlp;
  Eigen::MatrixXd localEri;
  double HlocalA[4];
  double HlocalB[4];

 protected:
  const Integrals::BasisSetController& basisSet_;
  std::shared_ptr<KlopmanFinalizer> koFactors_;
};

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
