/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef INCLUDE_SPARROW_NDDO_PARAMETERSOMX_H
#define INCLUDE_SPARROW_NDDO_PARAMETERSOMX_H

#include "Utils/Geometry/ElementTypes.h"
#include "boost/functional/hash.hpp"
#include <unordered_map>

namespace Scine {
namespace Sparrow {
namespace nddo {

//! Nddo method parameters
struct ParametersOmx {
  struct Values {
    //!@name Member types
    //!@{
    struct Sp {
      double s = 0;
      double p = 0;

      template<class Archive>
      void serialize(Archive& archive);
    };
    struct Sppi {
      double s = 0;
      double pSigma = 0;
      double pPi = 0;

      template<class Archive>
      void serialize(Archive& archive);
    };
    //!@}

    /**
     * Parameters for the evaluation of resonance integrals
     * in a local coordinate system.
     * The integrals are evaluated in a local coordinate system, therefore
     * the paremters are divided in s, p(sigma), p(pi).
     * All paremters are in eV.
     */
    struct ResonanceIntegrals {
      //! Alpha parameter for each orbital type in a.u. (as, ap,sigma, ap,pi)
      //! for evaluation in a local coordinate system of resonance integrals.
      Sppi alpha;
      //! Beta parameter for each orbital type in eV bohr\f$^{-1/2}\f$ (bs, bp,sigma, bp,pi)
      //! for evaluation in a local coordinate system of resonance integrals.
      Sppi beta;
      //! Alpha parameter for each orbital type in a.u. (as, ap,sigma, ap,pi)
      //! for evaluation in a local coordinate system of resonance integrals
      //! between an element X and an hydrogen atom. ap,pi is then 0.
      Sppi alphaXH;
      //! Beta parameter for each orbital type in eV bohr\f$^{-1/2}\f$ (bs, bp,sigma, bp,pi)
      //! for evaluation in a local coordinate system of resonance integrals
      //! between an element X and an hydrogen atom. ap,pi is then 0.
      Sppi betaXH;

      template<class Archive>
      void serialize(Archive& archive);
    };

    /**
     * Parameters for the evaluation of resonance integrals
     * in a local coordinate system.
     * The integrals are evaluated in a local coordinate system, therefore
     * the paremters are divided in s, p(sigma), p(pi).
     * All paremters are in eV.
     */
    struct OrthogonalizationFactors {
      //! Coefficient for one center orthogonalization contribution
      double f1 = 0;
      //! Coefficient for two center orthogonalization contribution
      double g1 = 0;
      //! Coefficient for second order one center orthogonalization contribution
      double f2 = 0;
      //! Coefficient for second order two center orthogonalization contribution
      double g2 = 0;

      template<class Archive>
      void serialize(Archive& archive);
    };

    /**
     * Parameters for the evaluation of pseudopotential integrals
     * in a local coordinate system.
     * The integrals are evaluated in a local coordinate system.
     * All the core orbitals are s orbitals, so just 1 parameter is present.
     */
    struct Pseudopotential {
      //! Exponent to scale the core pseudo-orbital in a.u..
      double exponent = 0;
      //! Core energies in eV.
      double faa = 0;
      //! Alpha parameter for core orbitals in a.u. for evaluation in a
      //! local coordinate system of effective core potential integrals.
      double alpha = 0;
      //! Beta parameter for core orbitals in eV bohr\f$^{-1/2}\f$ for evaluation in a
      //! local coordinate system of effective core potential integrals.
      double beta = 0;

      template<class Archive>
      void serialize(Archive& archive);
    };

    /**
     * Parameters for the evaluation of one-cente-two-electon integrals.
     * In a sp basis only these 5 terms appear.
     */
    struct OneCenterTwoElectronIntegrals {
      //! One-center Coulomb integral s-s (eV)
      double gss = 0;
      //! One-center Coulomb integral p-p (eV)
      double gpp = 0;
      //! One-center Coulomb integral s-p (eV)
      double gsp = 0;
      //! One-center Coulomb integral p-p' (eV)
      double gp2 = 0;
      //! One-center exchange integral s-s (eV)
      double hsp = 0;

      template<class Archive>
      void serialize(Archive& archive);
    };

    //!@name Members
    //!@{
    //! Parameters to evaluate resonance integrals
    ResonanceIntegrals resonance;
    //! Parameters to e valuate the orthogonalization corrections.
    OrthogonalizationFactors orthogonalization;
    //! Parameters for pseudopotentials
    Pseudopotential pseudopotential;
    //! Parameters for evaluation of one center two electron integrals
    OneCenterTwoElectronIntegrals oneCenterTwoElectrons;
    //! One-center energy for each orbital in eV (Uss, Upp)
    Sp oneCenterEnergy;
    //! Orbital exponent to scale the basis exponent in bohr^(-1)
    double exponent = 0;
    //!@}
    template<class Archive>
    void serialize(Archive& archive);
  };

  //! Read parameters from a JSON file
  static ParametersOmx read(const std::string& filename);
  //!@}

  //!@name Observers
  //!@{
  //! Writes the parameters in JSON format to file
  void write(const std::string& filename) const;
  //!@}

  //!@name Members
  //!@{
  //! Map from atomic number to atomic parameters
  std::unordered_map<int, Values> atomic;
};
//! Default parameters for OM1
ParametersOmx om1();
//! Default parameters for OM2
ParametersOmx om2();
//! Default parameters for OM3
ParametersOmx om3();
//! Default parameters for ODM2
ParametersOmx odm2();
//! Default parameters for ODM3
ParametersOmx odm3();

template<typename LhsType>
inline ParametersOmx::Values::Sppi operator*(ParametersOmx::Values::Sppi lhs, LhsType rhs) {
  return {lhs.s * rhs, lhs.pSigma * rhs, lhs.pPi * rhs};
}

template<typename LhsType>
inline ParametersOmx::Values::Sp operator*(ParametersOmx::Values::Sp lhs, LhsType rhs) {
  return {lhs.s * rhs, lhs.p * rhs};
}

} // namespace nddo
} // namespace Sparrow
} // namespace Scine

#endif // INCLUDE_SPARROW_NDDO_PARAMETERSOMX_H
