/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "ParametersProcessor.h"
#include "ParametersOmx.h"
#include <Sparrow/Implementations/Nddo/Omx/BasisHandler.h>
#include <Utils/Constants.h>

namespace Scine {
namespace Sparrow {
namespace nddo {

double KlopmanScalingFactors::get(Integrals::DiatomicKey index) const {
  Integrals::canonicalize(index);
  auto it = factors.find(index);
  if (it == factors.end()) {
    throw std::runtime_error("No Klopman--Ohno precomputed factor for indices " + std::to_string(index.first) +
                             " and " + std::to_string(index.second) + ".");
  }
  return it->second;
}

void KlopmanScalingFactors::set(Integrals::DiatomicKey index, double value) {
  Integrals::canonicalize(index);
  factors[index] = value;
}

KlopmanScalingFactors ParametersProcessor::getKlopmanScalingFactors(const OmxElementParameters& elParameters) {
  // collect indices
  std::vector<int> indices;
  std::transform(elParameters.begin(), elParameters.end(), std::back_inserter<std::vector<int>>(indices),
                 [](auto el) { return el.first; });

  return getKlopmanScalingFactorsImpl(indices, [&elParameters](int z) {
    return elParameters.get(Utils::ElementInfo::element(z)).parameters.oneCenterTwoElectrons.gss;
  });
}
KlopmanScalingFactors ParametersProcessor::getKlopmanScalingFactors(const ParametersOmx& rawParameters) {
  // collect indices
  std::vector<int> indices;
  std::transform(rawParameters.atomic.begin(), rawParameters.atomic.end(),
                 std::back_inserter<std::vector<int>>(indices), [](auto el) { return el.first; });

  return getKlopmanScalingFactorsImpl(indices, [&rawParameters](int z) {
    return rawParameters.atomic.at(z).oneCenterTwoElectrons.gss * Utils::Constants::hartree_per_ev;
  });
}

OmxElementParameters::AtomicParameters ParametersProcessor::processAtomicParameters(Utils::ElementType e,
                                                                                    const ParametersOmx& rawParameters) {
  const ParametersOmx::Values& v = rawParameters.atomic.at(Utils::ElementInfo::Z(e));
  OmxElementParameters::AtomicParameters pars;
  pars.element = e;
  pars.nAOs = Utils::ElementInfo::Z(e) == 1 ? 1 : 4;

  auto& p = pars.parameters;
  p = v;
  p.resonance.beta = v.resonance.beta * Utils::Constants::hartree_per_ev;
  p.resonance.betaXH = v.resonance.betaXH * Utils::Constants::hartree_per_ev;
  p.pseudopotential.faa = v.pseudopotential.faa * Utils::Constants::hartree_per_ev;
  p.pseudopotential.beta = v.pseudopotential.beta * Utils::Constants::hartree_per_ev;
  p.oneCenterEnergy = v.oneCenterEnergy * Utils::Constants::hartree_per_ev;
  p.oneCenterTwoElectrons.gss = v.oneCenterTwoElectrons.gss * Utils::Constants::hartree_per_ev;
  p.oneCenterTwoElectrons.gpp = v.oneCenterTwoElectrons.gpp * Utils::Constants::hartree_per_ev;
  p.oneCenterTwoElectrons.gsp = v.oneCenterTwoElectrons.gsp * Utils::Constants::hartree_per_ev;
  p.oneCenterTwoElectrons.gp2 = v.oneCenterTwoElectrons.gp2 * Utils::Constants::hartree_per_ev;
  p.oneCenterTwoElectrons.hsp = v.oneCenterTwoElectrons.hsp * Utils::Constants::hartree_per_ev;

  pars.gtos = BasisSet::om2(e, p);
  return pars;
}
} // namespace nddo
} // namespace Sparrow
} // namespace Scine
