/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#ifndef SPARROW_PARAMETERPROCESSOR_H
#define SPARROW_PARAMETERPROCESSOR_H

#include "boost/functional/hash.hpp"
#include <Integrals/TwoBodiesIntegrals/Orbitals.h>
#include <Sparrow/Implementations/Nddo/Omx/OmxElementParameters.h>
#include <unordered_map>

namespace Scine {
namespace Sparrow {
namespace nddo {

struct ParametersOmx;

/**
 * @brief Struct containing the precomputed part of the Klopman--Ohno scaling factors.
 * This struct has an array containing (1/(2*gamma_ss^I) + 1/(2*gamma_ss^J))^2
 */
struct KlopmanScalingFactors {
 public:
  /**
   * @brief getter function. The index is canonicalized such that i>j.
   */
  double get(Integrals::DiatomicKey index) const;
  /**
   * @brief setter function. The index is canonicalized such that i>j.
   */
  void set(Integrals::DiatomicKey index, double value);

 private:
  std::map<Integrals::DiatomicKey, double> factors;
};
/*!
 * This class implements functions for the conversion between
 * raw parameters published for OMx and parameters useful at
 * runtime.
 */
class ParametersProcessor {
 public:
  using AtomicParameters = OmxElementParameters::AtomicParameters;
  static OmxElementParameters::AtomicParameters processAtomicParameters(Utils::ElementType e, const ParametersOmx& rawParameters);
  /**
   * @brief This function return the struct with an array containing (1/(2*gamma_ss^I) + 1/(2*gamma_ss^J))^2.
   * The index are element types (not index on the structure!)
   */
  static KlopmanScalingFactors getKlopmanScalingFactors(const ParametersOmx& rawParameters);

  /**
   * @brief This function return the struct with an array containing (1/(2*gamma_ss^I) + 1/(2*gamma_ss^J))^2.
   * Takes as arbument the already processed parameters. The index are element types (not index on the structure!)
   */
  static KlopmanScalingFactors getKlopmanScalingFactors(const OmxElementParameters& elParameters);

 private:
  template<typename FunctionType>
  static auto getKlopmanScalingFactorsImpl(const std::vector<int>& indices, FunctionType gammaSSGetter)
      -> KlopmanScalingFactors {
    KlopmanScalingFactors klopman;

    for (unsigned int i = 0; i < indices.size(); ++i) {
      for (unsigned int j = i; j < indices.size(); ++j) {
        Integrals::DiatomicKey index{indices[i], indices[j]};
        Integrals::canonicalize(index);
        double klopmanFactor = 1. / (2. * gammaSSGetter(index.first)) + 1. / (2. * gammaSSGetter(index.second));
        klopman.set(index, klopmanFactor * klopmanFactor);
      }
    }
    return klopman;
  }
  static void setGtoExpansion(Utils::ElementType e, AtomicParameters& par, const ParametersOmx::Values& p);
};

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
#endif // SPARROW_PARAMETERPROCESSOR_H
