/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "ResonanceIntegralEvaluator.h"
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>

namespace Scine {
namespace Sparrow {
namespace nddo {
ResonanceIntegralEvaluator::ResonanceIntegralEvaluator(const Utils::ElementTypeCollection& elements,
                                                       const Utils::PositionCollection& positions,
                                                       const OmxElementParameters& pars)
  : elements_(elements), positions_(positions), pars_(pars) {
  assert(Eigen::Index(elements.size()) == positions.rows());
  assembleResonanceIntegralMatrix();
}

auto ResonanceIntegralEvaluator::getResonanceIntegralMatrix() const -> Eigen::MatrixXd {
  return resonanceMatrix_;
}

void ResonanceIntegralEvaluator::assembleResonanceIntegralMatrix() {
  auto aoInfos = Utils::AtomsOrbitalsIndexes(elements_.size());
  for (auto el : elements_) {
    aoInfos.addAtom(pars_.get(el).nAOs);
  }
  resonanceMatrix_ = Eigen::MatrixXd::Zero(aoInfos.getNAtomicOrbitals(), aoInfos.getNAtomicOrbitals());

  for (unsigned atom_a = 0; atom_a < elements_.size(); ++atom_a) {
    Utils::Position posA = positions_.row(atom_a);
    const OmxElementParameters::AtomicParameters& parA = pars_.get(elements_[atom_a]);

    for (unsigned atom_b = 0; atom_b < atom_a; ++atom_b) {
      Utils::Position posB = positions_.row(atom_b);
      const OmxElementParameters::AtomicParameters& parB = pars_.get(elements_[atom_b]);

      resonanceMatrix_.block(aoInfos.getFirstOrbitalIndex(atom_a), aoInfos.getFirstOrbitalIndex(atom_b), parA.nAOs,
                             parB.nAOs) = getBlock(posA, posB, parA, parB);
    }
  }
  resonanceMatrix_ = resonanceMatrix_.selfadjointView<Eigen::Lower>();
}

auto ResonanceIntegralEvaluator::generateRotationMatrix(const Utils::Position& posA, const Utils::Position& posB)
    -> Eigen::MatrixXd {
  Utils::Position Rab = posB - posA;
  double planeProjection = 0;
  double cosa = 1.0;
  double sina = 0.0;
  double cosb = 0;
  double sinb = 0;
  planeProjection = std::sqrt(Rab.x() * Rab.x() + Rab.y() * Rab.y());
  if (planeProjection != 0) {
    cosa = Rab.x() / planeProjection;
    sina = Rab.y() / planeProjection;
  }
  cosb = (Rab.z() / Rab.norm());
  sinb = planeProjection / Rab.norm();

  enum Cartesian : unsigned { s = 0, x = 1, y = 2, z = 3 };
  // Diatomic x Cartesian coordinate system
  Eigen::MatrixXd rotationMatrix = Eigen::MatrixXd::Zero(4, 4);
  rotationMatrix(Types::s, Cartesian::s) = 1;
  rotationMatrix(Types::pSigma, Cartesian::z) = cosb;
  rotationMatrix(Types::pSigma, Cartesian::x) = sinb * cosa;
  rotationMatrix(Types::pSigma, Cartesian::y) = sinb * sina;
  rotationMatrix(Types::pPiPlus, Cartesian::z) = -sinb;
  rotationMatrix(Types::pPiPlus, Cartesian::x) = cosb * cosa;
  rotationMatrix(Types::pPiPlus, Cartesian::y) = cosb * sina;
  rotationMatrix(Types::pPiMinus, Cartesian::z) = 0;
  rotationMatrix(Types::pPiMinus, Cartesian::x) = -sina;
  rotationMatrix(Types::pPiMinus, Cartesian::y) = cosa;

  return rotationMatrix;
}

auto ResonanceIntegralEvaluator::getBlock(const Utils::Position& posA, const Utils::Position& posB,
                                          const OmxElementParameters::AtomicParameters& parA,
                                          const OmxElementParameters::AtomicParameters& parB) -> Eigen::MatrixXd {
  Eigen::MatrixXd blockAB = Eigen::MatrixXd::Zero(parA.nAOs, parB.nAOs);

  auto integralValue = [&](double distance, double bA, double bB, double aA, double aB) -> double {
    return 0.5 * (bA + bB) * std::sqrt(distance) * std::exp(-(aA + aB) * std::pow(distance, 2));
  };
  double rij = (posB - posA).norm();
  const ParametersOmx::Values::ResonanceIntegrals& resA = parA.parameters.resonance;
  const ParametersOmx::Values::ResonanceIntegrals& resB = parB.parameters.resonance;

  if (parA.nAOs == 1 && parB.nAOs == 1) {
    blockAB(Types::s, Types::s) = integralValue(rij, resA.beta.s, resB.beta.s, resA.alpha.s, resB.alpha.s);
  }
  else if (parA.nAOs == 1 && parB.nAOs == 4) {
    blockAB(Types::s, Types::s) = integralValue(rij, resA.beta.s, resB.betaXH.s, resA.alpha.s, resB.alphaXH.s);
    blockAB(Types::s, Types::pSigma) = -integralValue(rij, resA.beta.s, resB.betaXH.pSigma, resA.alpha.s, resB.alphaXH.pSigma);
    blockAB = blockAB * generateRotationMatrix(posA, posB);
  }
  else if (parA.nAOs == 4 && parB.nAOs == 1) {
    blockAB(Types::s, Types::s) = integralValue(rij, resA.betaXH.s, resB.beta.s, resA.alphaXH.s, resB.alpha.s);
    blockAB(Types::pSigma, Types::s) =
        integralValue(rij, resA.betaXH.pSigma, resB.beta.s, resA.alphaXH.pSigma, resB.alpha.s);
    blockAB = generateRotationMatrix(posA, posB).transpose() * blockAB;
  }
  else if (parA.nAOs == 4 && parB.nAOs == 4) {
    Eigen::MatrixXd rotMatrix = generateRotationMatrix(posA, posB);
    blockAB(Types::s, Types::s) = integralValue(rij, resA.beta.s, resB.beta.s, resA.alpha.s, resB.alpha.s);
    blockAB(Types::pSigma, Types::s) = integralValue(rij, resA.beta.pSigma, resB.beta.s, resA.alpha.pSigma, resB.alpha.s);
    blockAB(Types::s, Types::pSigma) = -integralValue(rij, resA.beta.s, resB.beta.pSigma, resA.alpha.s, resB.alpha.pSigma);
    blockAB(Types::pSigma, Types::pSigma) =
        -integralValue(rij, resA.beta.pSigma, resB.beta.pSigma, resA.alpha.pSigma, resB.alpha.pSigma);
    blockAB(Types::pPiPlus, Types::pPiPlus) = blockAB(Types::pPiMinus, Types::pPiMinus) =
        integralValue(rij, resA.beta.pPi, resB.beta.pPi, resA.alpha.pPi, resB.alpha.pPi);

    blockAB = (rotMatrix.transpose() * blockAB * rotMatrix).eval();
  }
  else {
    throw std::runtime_error("Wrong number of atomic orbitals (atom A: " + std::to_string(parA.nAOs) +
                             ", atom B: " + std::to_string(parB.nAOs) + ") in Resonance Integral calculator.");
  }
  return blockAB;
}

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
