/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "OmxElementParameters.h"
#include <Utils/Typenames.h>

#ifndef SPARROW_RESONANCEINTEGRALEVALUTATOR_H
#  define SPARROW_RESONANCEINTEGRALEVALUTATOR_H
namespace Scine {
namespace Utils {
class AtomsOrbitalsIndexes;
}
namespace Sparrow {
namespace nddo {

class ResonanceIntegralEvaluator {
 public:
  /**
   * @enum Containing the indices in the local coordinate system.
   */
  enum Types : unsigned { s = 0, pSigma = 1, pPiPlus = 2, pPiMinus = 3 };

  /**
   * @brief Constructor taking the structure and the necessary parameters.
   */
  ResonanceIntegralEvaluator(const Utils::ElementTypeCollection& elements, const Utils::PositionCollection& positions,
                             const OmxElementParameters& pars);
  /**
   * @brief Destructor.
   */
  ~ResonanceIntegralEvaluator() = default;

  /**
   * @brief Returns the calculated resonance matrix in the molecular coordinate system.
   */
  auto getResonanceIntegralMatrix() const -> Eigen::MatrixXd;

  /**
   * @brief Assembles the resonance integral matrix in the molecular coordinate system.
   */
  void assembleResonanceIntegralMatrix();

  /**
   * @brief Generates the rotation matrix from the local to the molecular coordinate system.
   */
  auto generateRotationMatrix(const Utils::Position& posA, const Utils::Position& posB) -> Eigen::MatrixXd;

 private:
  /**
   * @brief Calculated a (local) diatomic block of the resonance matrix and rotates it in the molecular system.
   */
  auto getBlock(const Utils::Position& posA, const Utils::Position& posB, const OmxElementParameters::AtomicParameters& parA,
                const OmxElementParameters::AtomicParameters& parB) -> Eigen::MatrixXd;

  Utils::ElementTypeCollection elements_;
  Utils::PositionCollection positions_;
  OmxElementParameters pars_;
  Eigen::MatrixXd resonanceMatrix_;
};
} // namespace nddo
} // namespace Sparrow
} // namespace Scine
#endif // SPARROW_RESONANCEINTEGRALEVALUTATOR_H
