/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "TwoElectronMatrixConstructor.h"
#include "MatrixTransformer.h"
#include <Integrals/TwoBodiesIntegrals/Orbitals.h>
#include <Utils/DataStructures/AtomsOrbitalsIndexes.h>
#include <Utils/DataStructures/DensityMatrix.h>

namespace Scine {
namespace Sparrow {

using namespace Utils::AutomaticDifferentiation;

TwoElectronMatrixConstructor::TwoElectronMatrixConstructor(const Utils::DensityMatrix& densityMatrix)
  : densityMatrix_(densityMatrix) {
  if (densityMatrix_.restricted()) {
    const auto& dm = densityMatrix_.restrictedMatrix();
    twoElectronMatrix_ = Utils::SpinAdaptedMatrix::createRestricted(Eigen::MatrixXd::Zero(dm.rows(), dm.cols()));
  }
  else {
    const auto& dm = densityMatrix_.alphaMatrix();
    twoElectronMatrix_ = Utils::SpinAdaptedMatrix::createUnrestricted(Eigen::MatrixXd::Zero(dm.rows(), dm.cols()),
                                                                      Eigen::MatrixXd::Zero(dm.rows(), dm.cols()));
  }
}

void TwoElectronMatrixConstructor::contractCoulomb(const MatrixTransformer& transformer,
                                                   const Utils::AtomsOrbitalsIndexes& aoInfos) {
  Eigen::VectorXd g = transformer.coulombEriSupermatrix().selfadjointView<Eigen::Lower>() *
                      *transformer.coulombDensitySupervector().restricted;
  for (int atom_a = 0; atom_a < aoInfos.getNAtoms(); ++atom_a) {
    int firstA = aoInfos.getFirstOrbitalIndex(atom_a);
    int nAOsA = aoInfos.getNOrbitals(atom_a);
    if (densityMatrix_.restricted()) {
      twoElectronMatrix_.restrictedMatrix().block(firstA, firstA, nAOsA, nAOsA) +=
          transformer.extractBlockFromVector(g, atom_a, nAOsA);
    }
    else {
      twoElectronMatrix_.alphaMatrix().block(firstA, firstA, nAOsA, nAOsA) +=
          transformer.extractBlockFromVector(g, atom_a, nAOsA);
      twoElectronMatrix_.betaMatrix().block(firstA, firstA, nAOsA, nAOsA) +=
          transformer.extractBlockFromVector(g, atom_a, nAOsA);
    }
  }
}
void TwoElectronMatrixConstructor::contractExchange(const MatrixTransformer& transformer, int atom_a, int firstOnA,
                                                    int nAOsA, int atom_b, int firstOnB, int nAOsB) {
  if (densityMatrix_.restricted()) {
    Eigen::VectorXd k =
        transformer.exchangeDensity(atom_a, atom_b).restricted->transpose() * transformer.exchangeEri(atom_a, atom_b);
    twoElectronMatrix_.restrictedMatrix().block(firstOnB, firstOnA, nAOsB, nAOsA) -=
        0.5 * Eigen::Map<const Eigen::MatrixXd>(k.data(), nAOsB, nAOsA);
  }
  else {
    Eigen::VectorXd kAlpha =
        transformer.exchangeDensity(atom_a, atom_b).alpha->transpose() * transformer.exchangeEri(atom_a, atom_b);
    Eigen::VectorXd kBeta =
        transformer.exchangeDensity(atom_a, atom_b).beta->transpose() * transformer.exchangeEri(atom_a, atom_b);
    twoElectronMatrix_.alphaMatrix().block(firstOnB, firstOnA, nAOsB, nAOsA) -=
        Eigen::Map<const Eigen::MatrixXd>(kAlpha.data(), nAOsB, nAOsA);
    twoElectronMatrix_.betaMatrix().block(firstOnB, firstOnA, nAOsB, nAOsA) -=
        Eigen::Map<const Eigen::MatrixXd>(kBeta.data(), nAOsB, nAOsA);
  }
}

void TwoElectronMatrixConstructor::finalizeEvaluation() {
  if (densityMatrix_.unrestricted()) {
    auto& gAlpha = twoElectronMatrix_.alphaMatrix();
    auto& gBeta = twoElectronMatrix_.betaMatrix();
    gAlpha = gAlpha.selfadjointView<Eigen::Upper>();
    gBeta = gBeta.selfadjointView<Eigen::Upper>();
  }
  else {
    auto& G = twoElectronMatrix_.restrictedMatrix();
    G = G.selfadjointView<Eigen::Upper>();
  }
}

const Utils::SpinAdaptedMatrix& TwoElectronMatrixConstructor::getResult() const {
  return twoElectronMatrix_;
}

template<Utils::Derivative O>
void TwoElectronMatrixConstructor::addDerivatives(DerivativeContainerType<O>& /*derivativeContainer*/) const {
  throw std::runtime_error("No 2-body integrals derivatives implemented");
  /*
  for (int i = 0; i < nAtoms_; ++i) {
    auto indexA = aoIndexes_.getFirstOrbitalIndex(i);
    auto nAOsA = aoIndexes_.getNOrbitals(i);
    for (int j = i + 1; j < nAtoms_; j++) {
      auto indexB = aoIndexes_.getFirstOrbitalIndex(j);
      auto nAOsB = aoIndexes_.getNOrbitals(j);

      addDerivativesForBlock<O>(derivativeContainer, i, j, indexA, indexB, nAOsA, nAOsB, *twoCenterIntegrals.get(i, j));
    }
  }
  */
}

template<Utils::Derivative O>
void TwoElectronMatrixConstructor::addDerivativesForBlock(DerivativeContainerType<O>& /*derivativeContainer*/, int /*a*/,
                                                          int /*b*/, int /*startA*/, int /*startB*/, int /*nAOsA*/,
                                                          int /*nAOsB*/, const Utils::MatrixWithDerivatives& /*m*/) const {
  throw std::runtime_error("No 2-body integrals derivatives implemented");
  /*
  int mu, nu, lambda, sigma;
  Utils::AutomaticDifferentiation::DerivativeType<O> integralD;
  Utils::AutomaticDifferentiation::DerivativeType<O> derivativeContribution;
  integralD.setZero();
  derivativeContribution.setZero();
  for (int i = 0; i < nAOsA; i++) {
    mu = startA + i;

    for (int j = 0; j <= i; j++) {
      nu = startA + j;

      for (int k = 0; k < nAOsB; k++) {
        lambda = startB + k;

        for (int l = 0; l <= k; l++) {
          sigma = startB + l;

          integralD = m.getDerivative<O>(i, j, k, l);

          // this multiple contains, in addition to what was above, a 1/2 factor for diagonal elements (NB: but not
          // visible because of two equal terms)
          double multiple = ((mu == nu) ? 1 : 2) * ((lambda == sigma) ? 1 : 2);
          double factor;

          if (!spinPolarized_) {
            factor = P(lambda, sigma) * P(mu, nu) * multiple - 0.5 * P(sigma, nu) * P(lambda, mu);
            if (mu > nu) {
              factor += (-0.5 * P(sigma, mu) * P(lambda, nu));
              if (lambda > sigma) {
                factor += (-0.5 * P(lambda, mu) * P(sigma, nu));
              }
            }
            if (lambda > sigma) {
              factor += (-0.5 * P(lambda, nu) * P(sigma, mu));
            }
          }
          else {
            factor = P(lambda, sigma) * P(mu, nu) * multiple - PAlpha_(sigma, nu) * PAlpha_(lambda, mu) -
                     PBeta_(sigma, nu) * PBeta_(lambda, mu);
            if (mu > nu) {
              factor -= PAlpha_(sigma, mu) * PAlpha_(lambda, nu) + PBeta_(sigma, mu) * PBeta_(lambda, nu);
              if (lambda > sigma) {
                factor -= PAlpha_(lambda, mu) * PAlpha_(sigma, nu) + PBeta_(lambda, mu) * PBeta_(sigma, nu);
              }
            }
            if (lambda > sigma) {
              factor -= PAlpha_(lambda, nu) * PAlpha_(sigma, mu) + PBeta_(lambda, nu) * PBeta_(sigma, mu);
            }
          }

          derivativeContribution += integralD * factor;
        }
      }
    }
  }
  Utils::AutomaticDifferentiation::addDerivativeToContainer<O>(derivativeContainer, a, b, derivativeContribution);
  */
}

template void TwoElectronMatrixConstructor::addDerivatives<Utils::Derivative::First>(
    Utils::AutomaticDifferentiation::DerivativeContainerType<Utils::Derivative::First>&) const;
template void TwoElectronMatrixConstructor::addDerivatives<Utils::Derivative::SecondAtomic>(
    Utils::AutomaticDifferentiation::DerivativeContainerType<Utils::Derivative::SecondAtomic>&) const;
template void TwoElectronMatrixConstructor::addDerivatives<Utils::Derivative::SecondFull>(
    Utils::AutomaticDifferentiation::DerivativeContainerType<Utils::Derivative::SecondFull>&) const;
} // namespace Sparrow
} // namespace Scine
