/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef ABINITIO_TWOELECTRONMATRIXCONSTRUCTOR_H
#define ABINITIO_TWOELECTRONMATRIXCONSTRUCTOR_H

#include <Utils/DataStructures/MatrixWithDerivatives.h>
#include <Utils/DataStructures/SpinAdaptedMatrix.h>
#include <Utils/Math/AutomaticDifferentiation/MethodsTypesHelper.h>
#include <boost/functional/hash.hpp>
#include <boost/optional.hpp>
#include <unordered_map>
namespace Scine {
namespace Utils {
class DensityMatrix;
class MatrixWithDerivatives;
class AtomsOrbitalsIndexes;
} // namespace Utils

namespace Integrals {
using DiatomicKey = std::pair<int, int>;
using EriType = std::unordered_map<DiatomicKey, Utils::MatrixWithDerivatives, boost::hash<DiatomicKey>>;
} // namespace Integrals
namespace Sparrow {

class MatrixTransformer;

/**
 * @class TwoElectronMatrixConstructor @file TwoElectronMatrixConstructor.h
 * @brief This class evaluates the two electron matrix elements.
 * This class evaluates the two electron matrix by evaluating integrals of
 * basis functions quartet and contracting them with the density.
 * It avoids having to duplicate the integral to matrix code every instance
 * it comes up.
 * The contraction of the ERIs and the density matrix is carried out through
 * matrix-vector operations offloaded to the Eigen library.
 * This enables more stability and implicit parallelism/vectorization.
 * The One-Center block with mu,nu,lambda,sigma on the same atom is
 * carried out coefficientwise.
 */
class TwoElectronMatrixConstructor {
 public:
  explicit TwoElectronMatrixConstructor(const Utils::DensityMatrix& densityMatrix);
  /**
   * @brief Evaluates the elements of the 2-electron matrix for two-center ERIs
   */
  void contractCoulomb(const MatrixTransformer& transformer, const Utils::AtomsOrbitalsIndexes& aoInfos);
  /**
   * @brief Evaluates the elements of the 2-electron matrix for two-center ERIs
   */
  void contractExchange(const MatrixTransformer& transformer, int atom_a, int firstOnA, int nAOsA, int atom_b,
                        int firstOnB, int nAOsB);
  /**
   * @brief Finalized the calculation, i.e. symmetrizes the result.
   */
  void finalizeEvaluation();
  /**
   * @brief Getter for the two electron matrix
   */
  const Utils::SpinAdaptedMatrix& getResult() const;

  template<Utils::Derivative O>
  void addDerivatives(Utils::AutomaticDifferentiation::DerivativeContainerType<O>& derivativeContainer) const;

 private:
  template<Utils::Derivative O>
  void addDerivativesForBlock(Utils::AutomaticDifferentiation::DerivativeContainerType<O>& derivativeContainer, int a, int b,
                              int startA, int startB, int nAOsA, int nAOsB, const Utils::MatrixWithDerivatives& m) const;

  Utils::SpinAdaptedMatrix twoElectronMatrix_;
  const Utils::DensityMatrix& densityMatrix_;
};

} // namespace Sparrow
} // namespace Scine

#endif // ABINITIO_TWOELECTRONMATRIXCONSTRUCTOR_H
