/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#include "TwoElectronMatrixEvaluator.h"
#include "MatrixTransformer.h"
#include "TwoElectronMatrixConstructor.h"
#include <Integrals/TwoBodiesIntegrals/Orbitals.h>
#include <Utils/DataStructures/DensityMatrix.h>
#include <omp.h>
#include <boost/optional.hpp>

namespace Scine {
namespace Sparrow {

TwoElectronMatrixEvaluator::TwoElectronMatrixEvaluator(const Integrals::EriType& eris, const Utils::DensityMatrix& densityMatrix,
                                                       Utils::AtomsOrbitalsIndexes aoInfos)
  : eris_(eris), densityMatrix_(densityMatrix), aoInfos_(std::move(aoInfos)) {
}

void TwoElectronMatrixEvaluator::initialize(Utils::SpinAdaptedMatrix& matrix) const {
  const auto& dm = densityMatrix_.restrictedMatrix();
  if (densityMatrix_.restricted()) {
    matrix = Utils::SpinAdaptedMatrix::createRestricted(Eigen::MatrixXd::Zero(dm.rows(), dm.cols()));
  }
  else {
    matrix = Utils::SpinAdaptedMatrix::createUnrestricted(Eigen::MatrixXd::Zero(dm.rows(), dm.cols()),
                                                          Eigen::MatrixXd::Zero(dm.rows(), dm.cols()));
  }
}

Utils::SpinAdaptedMatrix TwoElectronMatrixEvaluator::contract() const {
  MatrixTransformer transformer(eris_, densityMatrix_, aoInfos_);
  Utils::SpinAdaptedMatrix result;
  initialize(result);
  TwoElectronMatrixConstructor constructor(densityMatrix_);

  constructor.contractCoulomb(transformer, aoInfos_);

#pragma omp parallel for schedule(dynamic)
  for (int atom_a = 0; atom_a < aoInfos_.getNAtoms(); ++atom_a) {
    for (int atom_b = 0; atom_b <= atom_a; ++atom_b) {
      Integrals::DiatomicKey index{atom_a, atom_b};
      Integrals::canonicalize(index);
      constructor.contractExchange(transformer, index.first, aoInfos_.getFirstOrbitalIndex(index.first),
                                   aoInfos_.getNOrbitals(index.first), index.second,
                                   aoInfos_.getFirstOrbitalIndex(index.second), aoInfos_.getNOrbitals(index.second));
    }
  }

  constructor.finalizeEvaluation();

  return constructor.getResult();
}

Utils::SpinAdaptedMatrix TwoElectronMatrixEvaluator::evaluateDebug() const {
  const auto& dm = densityMatrix_.restrictedMatrix();
  Eigen::MatrixXd result = Eigen::MatrixXd::Zero(dm.rows(), dm.cols());
  using namespace Integrals::Orbitals;

  auto getBlock = [&](Integrals::DiatomicKey index) {
    return (index.first > index.second) ? eris_.at(index).get0() : eris_.at({index.second, index.first}).get0().transpose();
  };
  if (densityMatrix_.restricted()) {
    // diagonal elements
    for (int atom_a = 0; atom_a < aoInfos_.getNAtoms(); ++atom_a) {
      Eigen::MatrixXd eri1c = eris_.at({atom_a, atom_a}).get0();
      for (int mu = 0; mu < aoInfos_.getNOrbitals(atom_a); ++mu) {
        int m = mu + aoInfos_.getFirstOrbitalIndex(atom_a);
        const auto mumu = pair(Orbital(mu), Orbital(mu));
        for (int nu = 0; nu < aoInfos_.getNOrbitals(atom_a); ++nu) {
          int n = nu + aoInfos_.getFirstOrbitalIndex(atom_a);
          const auto nunu = pair(Orbital(nu), Orbital(nu));
          const auto munu = pair(Orbital(mu), Orbital(nu));
          result(m, m) += dm(n, n) * (eri1c(mumu, nunu) - 0.5 * eri1c(munu, munu));
        }
        for (int atom_b = 0; atom_b < aoInfos_.getNAtoms(); ++atom_b) {
          if (atom_b != atom_a) {
            Eigen::MatrixXd eri = getBlock({atom_a, atom_b});
            for (int lambda = 0; lambda < aoInfos_.getNOrbitals(atom_b); ++lambda) {
              int l = lambda + aoInfos_.getFirstOrbitalIndex(atom_b);
              for (int sigma = 0; sigma < aoInfos_.getNOrbitals(atom_b); ++sigma) {
                int s = sigma + aoInfos_.getFirstOrbitalIndex(atom_b);
                const auto lambdasigma = pair(Orbital(lambda), Orbital(sigma));
                result(m, m) += dm(l, s) * (eri(mumu, lambdasigma));
              }
            }
          }
        }
      }
    }
    // 1 center block non diagonal
    for (int atom_a = 0; atom_a < aoInfos_.getNAtoms(); ++atom_a) {
      Eigen::MatrixXd eri1c = eris_.at({atom_a, atom_a}).get0();
      for (int mu = 0; mu < aoInfos_.getNOrbitals(atom_a); ++mu) {
        int m = mu + aoInfos_.getFirstOrbitalIndex(atom_a);
        const auto mumu = pair(Orbital(mu), Orbital(mu));
        for (int nu = 0; nu < aoInfos_.getNOrbitals(atom_a); ++nu) {
          if (nu != mu) {
            int n = nu + aoInfos_.getFirstOrbitalIndex(atom_a);
            const auto munu = pair(Orbital(mu), Orbital(nu));
            const auto nunu = pair(Orbital(nu), Orbital(nu));
            result(m, n) += dm(m, n) * (1.5 * eri1c(munu, munu) - 0.5 * eri1c(mumu, nunu));
            for (int atom_b = 0; atom_b < aoInfos_.getNAtoms(); ++atom_b) {
              if (atom_b != atom_a) {
                Eigen::MatrixXd eri = getBlock({atom_a, atom_b});
                for (int lambda = 0; lambda < aoInfos_.getNOrbitals(atom_b); ++lambda) {
                  int l = lambda + aoInfos_.getFirstOrbitalIndex(atom_b);
                  for (int sigma = 0; sigma < aoInfos_.getNOrbitals(atom_b); ++sigma) {
                    int s = sigma + aoInfos_.getFirstOrbitalIndex(atom_b);
                    const auto lambdasigma = pair(Orbital(lambda), Orbital(sigma));
                    result(m, n) += dm(l, s) * (eri(munu, lambdasigma));
                  }
                }
              }
            }
          }
        }
      }
    }
    // 2center
    for (int atom_a = 0; atom_a < aoInfos_.getNAtoms(); ++atom_a) {
      for (int mu = 0; mu < aoInfos_.getNOrbitals(atom_a); ++mu) {
        int m = mu + aoInfos_.getFirstOrbitalIndex(atom_a);
        for (int nu = 0; nu < aoInfos_.getNOrbitals(atom_a); ++nu) {
          int n = nu + aoInfos_.getFirstOrbitalIndex(atom_a);
          const auto munu = pair(Orbital(mu), Orbital(nu));
          for (int atom_b = 0; atom_b < aoInfos_.getNAtoms(); ++atom_b) {
            if (atom_b != atom_a) {
              Eigen::MatrixXd eri = getBlock({atom_a, atom_b});
              for (int lambda = 0; lambda < aoInfos_.getNOrbitals(atom_b); ++lambda) {
                int l = lambda + aoInfos_.getFirstOrbitalIndex(atom_b);
                for (int sigma = 0; sigma < aoInfos_.getNOrbitals(atom_b); ++sigma) {
                  int s = sigma + aoInfos_.getFirstOrbitalIndex(atom_b);
                  const auto lambdasigma = pair(Orbital(lambda), Orbital(sigma));
                  result(m, l) -= 0.5 * dm(n, s) * (eri(munu, lambdasigma));
                }
              }
            }
          }
        }
      }
    }
    return Utils::SpinAdaptedMatrix::createRestricted(result);
  }
  Eigen::MatrixXd dmAlpha = densityMatrix_.alphaMatrix();
  Eigen::MatrixXd dmBeta = densityMatrix_.betaMatrix();
  Eigen::MatrixXd rAlpha = Eigen::MatrixXd::Zero(dm.rows(), dm.cols());
  Eigen::MatrixXd rBeta = Eigen::MatrixXd::Zero(dm.rows(), dm.cols());

  // diagonal elements
  for (int atom_a = 0; atom_a < aoInfos_.getNAtoms(); ++atom_a) {
    Eigen::MatrixXd eri1c = eris_.at({atom_a, atom_a}).get0();
    for (int mu = 0; mu < aoInfos_.getNOrbitals(atom_a); ++mu) {
      int m = mu + aoInfos_.getFirstOrbitalIndex(atom_a);
      const auto mumu = pair(Orbital(mu), Orbital(mu));
      for (int nu = 0; nu < aoInfos_.getNOrbitals(atom_a); ++nu) {
        int n = nu + aoInfos_.getFirstOrbitalIndex(atom_a);
        const auto nunu = pair(Orbital(nu), Orbital(nu));
        const auto munu = pair(Orbital(mu), Orbital(nu));
        rAlpha(m, m) += dm(n, n) * eri1c(mumu, nunu) - dmAlpha(n, n) * eri1c(munu, munu);
        rBeta(m, m) += dm(n, n) * eri1c(mumu, nunu) - dmBeta(n, n) * eri1c(munu, munu);
      }
      for (int atom_b = 0; atom_b < aoInfos_.getNAtoms(); ++atom_b) {
        if (atom_b != atom_a) {
          Eigen::MatrixXd eri = getBlock({atom_a, atom_b});
          for (int lambda = 0; lambda < aoInfos_.getNOrbitals(atom_b); ++lambda) {
            int l = lambda + aoInfos_.getFirstOrbitalIndex(atom_b);
            for (int sigma = 0; sigma < aoInfos_.getNOrbitals(atom_b); ++sigma) {
              int s = sigma + aoInfos_.getFirstOrbitalIndex(atom_b);
              const auto lambdasigma = pair(Orbital(lambda), Orbital(sigma));
              rAlpha(m, m) += dm(l, s) * (eri(mumu, lambdasigma));
              rBeta(m, m) += dm(l, s) * (eri(mumu, lambdasigma));
            }
          }
        }
      }
    }
  }
  // 1 center block non diagonal
  for (int atom_a = 0; atom_a < aoInfos_.getNAtoms(); ++atom_a) {
    Eigen::MatrixXd eri1c = eris_.at({atom_a, atom_a}).get0();
    for (int mu = 0; mu < aoInfos_.getNOrbitals(atom_a); ++mu) {
      int m = mu + aoInfos_.getFirstOrbitalIndex(atom_a);
      const auto mumu = pair(Orbital(mu), Orbital(mu));
      for (int nu = 0; nu < aoInfos_.getNOrbitals(atom_a); ++nu) {
        if (nu != mu) {
          int n = nu + aoInfos_.getFirstOrbitalIndex(atom_a);
          const auto munu = pair(Orbital(mu), Orbital(nu));
          const auto nunu = pair(Orbital(nu), Orbital(nu));
          rAlpha(m, n) += 2.0 * dm(m, n) * eri1c(munu, munu) - dmAlpha(m, n) * (eri1c(munu, munu) + eri1c(mumu, nunu));
          rBeta(m, n) += 2.0 * dm(m, n) * eri1c(munu, munu) - dmBeta(m, n) * (eri1c(munu, munu) + eri1c(mumu, nunu));
          for (int atom_b = 0; atom_b < aoInfos_.getNAtoms(); ++atom_b) {
            if (atom_b != atom_a) {
              Eigen::MatrixXd eri = getBlock({atom_a, atom_b});
              for (int lambda = 0; lambda < aoInfos_.getNOrbitals(atom_b); ++lambda) {
                int l = lambda + aoInfos_.getFirstOrbitalIndex(atom_b);
                for (int sigma = 0; sigma < aoInfos_.getNOrbitals(atom_b); ++sigma) {
                  int s = sigma + aoInfos_.getFirstOrbitalIndex(atom_b);
                  const auto lambdasigma = pair(Orbital(lambda), Orbital(sigma));
                  rAlpha(m, n) += dm(l, s) * (eri(munu, lambdasigma));
                  rBeta(m, n) += dm(l, s) * (eri(munu, lambdasigma));
                }
              }
            }
          }
        }
      }
    }
  }
  // 2center
  for (int atom_a = 0; atom_a < aoInfos_.getNAtoms(); ++atom_a) {
    for (int mu = 0; mu < aoInfos_.getNOrbitals(atom_a); ++mu) {
      int m = mu + aoInfos_.getFirstOrbitalIndex(atom_a);
      for (int nu = 0; nu < aoInfos_.getNOrbitals(atom_a); ++nu) {
        int n = nu + aoInfos_.getFirstOrbitalIndex(atom_a);
        const auto munu = pair(Orbital(mu), Orbital(nu));
        for (int atom_b = 0; atom_b < aoInfos_.getNAtoms(); ++atom_b) {
          if (atom_b != atom_a) {
            Eigen::MatrixXd eri = getBlock({atom_a, atom_b});
            for (int lambda = 0; lambda < aoInfos_.getNOrbitals(atom_b); ++lambda) {
              int l = lambda + aoInfos_.getFirstOrbitalIndex(atom_b);
              for (int sigma = 0; sigma < aoInfos_.getNOrbitals(atom_b); ++sigma) {
                int s = sigma + aoInfos_.getFirstOrbitalIndex(atom_b);
                const auto lambdasigma = pair(Orbital(lambda), Orbital(sigma));
                rAlpha(m, l) -= dmAlpha(n, s) * (eri(munu, lambdasigma));
                rBeta(m, l) -= dmBeta(n, s) * (eri(munu, lambdasigma));
              }
            }
          }
        }
      }
    }
  }
  return Utils::SpinAdaptedMatrix::createUnrestricted(rAlpha, rBeta);
}
} // namespace Sparrow
} // namespace Scine
