/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef SPARROW_RESULTSASSEMBLER_H
#define SPARROW_RESULTSASSEMBLER_H

#include <Integrals/TwoBodiesIntegrals/AO2MOTransformation.h>
#include <Integrals/TwoBodiesIntegrals/Orbitals.h>
#include <Sparrow/Implementations/Nddo/Utils/NddoExcitedStatesAssembler.h>
#include <Utils/CalculatorBasics.h>
#include <Utils/DataStructures/SecondQuantization/CasSpecifier.h>
#include <Utils/DataStructures/SecondQuantization/FciDumping.h>
#include <Utils/Scf/MethodInterfaces/ScfMethod.h>
#include <Utils/Settings.h>
namespace Scine {
namespace Sparrow {
namespace nddo {

template<typename MethodType>
class ResultsAssembler {
 public:
  enum class MatrixType { Coefficients, Density, Fock };
  ResultsAssembler(MethodType& method, const Utils::ScfMethod& scf, const Integrals::EriType& eris,
                   const Eigen::MatrixXd& hCore, const Utils::Settings& settings, Utils::Results& results,
                   bool requiredSQSpecifier, bool generateFcidump)
    : method_(method),
      scf_(scf),
      eris_(eris),
      hCore_(hCore),
      settings_(settings),
      results_(results),
      requiredSQSpecifier_(requiredSQSpecifier),
      generateFcidump_(generateFcidump),
      orbitalsAroundFermi_(settings_.getInt(Utils::SettingsNames::casOrbitalsAroundFermiOption)),
      casIndices_(Utils::SecondQuantization::CasGenerator::parseCasString(
          settings_.getString(Utils::SettingsNames::casOrbitalIndicesOption))) {
    if (orbitalsAroundFermi_ > scf_.getMolecularOrbitals().numberOrbitals()) {
      orbitalsAroundFermi_ = scf_.getMolecularOrbitals().numberOrbitals();
    }
  }

  void formWithUnos() {
    NddoExcitedStatesAssembler<MethodType> unosAssembler(method_);
    auto cas = unosAssembler.generateCasWithUNO(scf_.getNumberElectrons(), scf_.getNumberAtomicOrbitals());
    writeFciDumpFile(settings_.getString(Utils::SettingsNames::fcidumpFilenameOption), std::get<0>(cas), std::get<1>(cas));
  }

  void formWithCas() {
    const auto& mos = scf_.getMolecularOrbitals();
    Utils::SecondQuantization::CasGenerator generator(mos);
    auto filename = settings_.getString(Utils::SettingsNames::fcidumpFilenameOption);

    if (scf_.unrestrictedCalculationRunning()) {
      if (casIndices_.empty()) {
        assert(casOrbitalsAroundFermi_ != 0);
        // Assume indices alpha == indices beta
        writeFciDumpFile(filename, generator.generateUnrestricted(scf_.getNumberElectrons(), scf_.spinMultiplicity(),
                                                                  casIndices_, casIndices_));
      }
      else {
        writeFciDumpFile(filename, generator.generateUnrestricted(scf_.getNumberElectrons(), scf_.spinMultiplicity(),
                                                                  orbitalsAroundFermi_));
      }
    }
    // standard occupation
    Eigen::VectorXd occupations = Eigen::VectorXd::Zero(mos.numberOrbitals());
    occupations.head(scf_.getNumberElectrons() / 2).array() = 2;
    if (casIndices_.empty()) {
      assert(casOrbitalsAroundFermi_ != 0);
      writeFciDumpFile(filename, generator.generateRestricted(scf_.getNumberElectrons(), orbitalsAroundFermi_), occupations);
    }
    else {
      writeFciDumpFile(filename, generator.generateRestricted(scf_.getNumberElectrons(), casIndices_), occupations);
    }
  }

  void formWithoutCas() {
    auto filename = settings_.getString(Utils::SettingsNames::fcidumpFilenameOption);
    // Get coefficient matrix for whole space.
    const auto& mos = scf_.getMolecularOrbitals();
    Utils::SecondQuantization::FciDumpData fciDumpData;
    fciDumpData.coreEnergy = scf_.getRepulsionEnergy();
    fciDumpData.unrestrictedReference = scf_.unrestrictedCalculationRunning();
    fciDumpData.nElectrons = scf_.getNumberElectrons();
    fciDumpData.nOrbitals = mos.numberOrbitals();
    fciDumpData.orbitalSymmetry = std::vector<int>(fciDumpData.nOrbitals, 1);
    if (mos.isRestricted()) {
      Utils::SecondQuantization::MoIntegrals ints{
          Integrals::AO2MOTransformation::transform1eIntegrals<Utils::Reference::Restricted>(hCore_, mos),
          Integrals::AO2MOTransformation::ao2mo<Utils::Reference::Restricted>(eris_, mos, scf_.getAtomsOrbitalsIndexesHolder()),
          {}};
      Utils::SecondQuantization::FciDumping::write<Utils::Reference::Restricted>(filename, *ints.oneBodyIntegrals,
                                                                                 *ints.twoBodyIntegrals, fciDumpData);
      writeDipole(filename, mos, fciDumpData);
      std::vector<int> orbitalIndices(mos.numberOrbitals());
      std::iota(orbitalIndices.begin(), orbitalIndices.end(), 0);
      // standard occupation
      Eigen::VectorXd occupations = Eigen::VectorXd::Zero(mos.numberOrbitals());
      occupations.head(scf_.getNumberElectrons() / 2).array() = 2;

      results_.set<Utils::Property::SQSpecifier>(
          {std::move(ints),
           {Utils::SpinAdaptedContainer<Utils::Reference::Restricted, int>{fciDumpData.nElectrons},
            mos,
            Utils::SpinAdaptedContainer<Utils::Reference::Restricted, std::vector<int>>{std::move(orbitalIndices)},
            {}},
           occupations,
           fciDumpData.coreEnergy});
    }
    else {
      auto oneElectronInts =
          Integrals::AO2MOTransformation::transform1eIntegrals<Utils::Reference::Unrestricted>(hCore_, mos);
      auto twoElectronInts = Integrals::AO2MOTransformation::ao2mo<Utils::Reference::Unrestricted>(
          eris_, mos, scf_.getAtomsOrbitalsIndexesHolder());
      Utils::SecondQuantization::FciDumping::write<Utils::Reference::Unrestricted>(filename, oneElectronInts,
                                                                                   twoElectronInts, fciDumpData);
      if (requiredSQSpecifier_) {
        throw std::runtime_error("SecondQuantization specifier is not yet available for unrestricted calculations. "
                                 "Consider the UNO-CAS method.");
      }
    }
  }

  void operator()() {
    if (settings_.getBool(Utils::SettingsNames::casWithUnoCI)) { // with uno-ci
      formWithUnos();
    }
    else if (orbitalsAroundFermi_ == 0 && casIndices_.empty()) { // No Cas
      formWithoutCas();
    }
    else { // standard cas
      formWithCas();
    }
  }

  void writeDipole(const std::string& filename, const Utils::MolecularOrbitals& mos,
                   const Utils::SecondQuantization::FciDumpData& data) const {
    if (results_.has<Utils::Property::DipoleMatrixAO>()) {
      auto dip = results_.get<Utils::Property::DipoleMatrixAO>();
      Utils::SecondQuantization::FciDumping::write<Utils::Reference::Restricted>(
          "dipole_x_" + filename,
          Integrals::AO2MOTransformation::transform1eIntegrals<Utils::Reference::Restricted>(dip[0], mos), {}, data);
      Utils::SecondQuantization::FciDumping::write<Utils::Reference::Restricted>(
          "dipole_y_" + filename,
          Integrals::AO2MOTransformation::transform1eIntegrals<Utils::Reference::Restricted>(dip[1], mos), {}, data);
      Utils::SecondQuantization::FciDumping::write<Utils::Reference::Restricted>(
          "dipole_z_" + filename,
          Integrals::AO2MOTransformation::transform1eIntegrals<Utils::Reference::Restricted>(dip[2], mos), {}, data);
    }
  }

  void writeFciDumpFile(const std::string& filename,
                        const Utils::SecondQuantization::CasSpecifier<Utils::Reference::Restricted>& casSpecifier,
                        const Eigen::VectorXd& occupations) {
    auto activeBlock = casSpecifier.getActiveIndices().restricted;

    auto matrices = getCoreMatrices(casSpecifier);
    /*
     * E_core = E_nn + Tr(D_core*1elAO) + 0.5 * Tr(D_core*Fock)
     */
    double energyCorrection =
        matrices[MatrixType::Density].restricted.cwiseProduct(hCore_).sum() +
        0.5 * matrices[MatrixType::Fock].restricted.cwiseProduct(matrices[MatrixType::Density].restricted).sum();
    Utils::SecondQuantization::FciDumpData fciDumpData;
    fciDumpData.coreEnergy = scf_.getRepulsionEnergy() + energyCorrection;
    fciDumpData.unrestrictedReference = scf_.unrestrictedCalculationRunning();
    fciDumpData.nElectrons = casSpecifier.getCasElectrons().restricted;
    fciDumpData.nOrbitals = int(activeBlock.size());
    // No symmetry implemented.
    fciDumpData.orbitalSymmetry = std::vector<int>(fciDumpData.nOrbitals, 1);
    // Transform the 2el integrals (ERI) in molecular orbital basis.
    /**
     * h_{ab}^{CAS,MO} = h_{ab}^{MO} + F_core^{MO}
     * F_core = sum_{\lambda, \sigma} (2(\mu \nu | \lambda \sigma) - (\mu \sigma | \lambda \nu) ) * P_{\lambda, \sigma}
     * with P the core density matrix. MO means it is expressed in MO basis.
     */
    Utils::SecondQuantization::MoIntegrals ints = {
        {Integrals::AO2MOTransformation::transform1eIntegrals<Utils::Reference::Restricted>(
            hCore_ + matrices[MatrixType::Fock].restricted, casSpecifier)},
        {Integrals::AO2MOTransformation::ao2mo<Utils::Reference::Restricted>(
            eris_, casSpecifier.getActiveOrbitalsCoefficients(), scf_.getAtomsOrbitalsIndexesHolder())},
        {}};

    // Assemble
    if (generateFcidump_) {
      Utils::SecondQuantization::FciDumping::write<Utils::Reference::Restricted>(filename, *ints.oneBodyIntegrals,
                                                                                 *ints.twoBodyIntegrals, fciDumpData);
      writeDipole(filename, casSpecifier.getActiveOrbitalsCoefficients(), fciDumpData);
    }
    if (requiredSQSpecifier_) {
      results_.set<Utils::Property::SQSpecifier>({std::move(ints), casSpecifier, occupations, fciDumpData.coreEnergy});
    }
  }

  void writeFciDumpFile(const std::string& filename,
                        const Utils::SecondQuantization::CasSpecifier<Utils::Reference::Unrestricted>& casSpecifier) const {
    if (requiredSQSpecifier_) {
      throw std::runtime_error("SecondQuantization specifier is not yet available for unrestricted calculations. "
                               "Consider the UNO-CAS method.");
    }
    auto matrices = getCoreMatrices(casSpecifier);
    /*
     * E_core = E_nn + Tr(D_core*1elAO) + 0.5 * Tr(D_core*Fock)
     */
    double energyCorrection =
        (matrices[MatrixType::Density].alpha + matrices[MatrixType::Density].beta).cwiseProduct(hCore_).sum() +
        0.5 * (matrices[MatrixType::Fock].alpha.cwiseProduct(matrices[MatrixType::Density].alpha).sum() +
               matrices[MatrixType::Fock].beta.cwiseProduct(matrices[MatrixType::Density].beta).sum());
    Utils::SecondQuantization::FciDumpData fciDumpData;
    fciDumpData.coreEnergy = scf_.getRepulsionEnergy() + energyCorrection;
    fciDumpData.unrestrictedReference = scf_.unrestrictedCalculationRunning();
    fciDumpData.nElectrons = casSpecifier.getCasElectrons().alpha + casSpecifier.getCasElectrons().beta;
    // Assumes nAlpha == nBeta
    fciDumpData.nOrbitals = int(casSpecifier.getActiveIndices().alpha.size());
    // No symmetry implemented.
    fciDumpData.orbitalSymmetry = std::vector<int>(fciDumpData.nOrbitals, 1);
    // Transform the 2el integrals (ERI) in molecular orbital basis.

    /**
     * h_{ab}^{CAS,MO} = h_{ab}^{MO} + F_core^{MO}
     * F_core = sum_{\lambda, \sigma} (2(\mu \nu | \lambda \sigma) - (\mu \sigma | \lambda \nu) ) * P_{\lambda, \sigma}
     * with P the core density matrix. MO means it is expressed in MO basis.
     */
    auto hCoreMO = Integrals::AO2MOTransformation::transform1eIntegrals<Utils::Reference::Unrestricted>(hCore_, casSpecifier);
    hCoreMO.at(Utils::SecondQuantization::SpinComponent::Alpha) += matrices[MatrixType::Coefficients].alpha.transpose() *
                                                                   matrices[MatrixType::Fock].alpha *
                                                                   matrices[MatrixType::Coefficients].alpha;
    hCoreMO.at(Utils::SecondQuantization::SpinComponent::Beta) += matrices[MatrixType::Coefficients].beta.transpose() *
                                                                  matrices[MatrixType::Fock].beta *
                                                                  matrices[MatrixType::Coefficients].beta;

    Utils::SecondQuantization::FciDumping::write<Utils::Reference::Unrestricted>(
        filename, hCoreMO,
        Integrals::AO2MOTransformation::ao2mo<Utils::Reference::Unrestricted>(
            eris_, casSpecifier.getActiveOrbitalsCoefficients(), scf_.getAtomsOrbitalsIndexesHolder()),
        fciDumpData);
  }

  auto getCoreMatrices(const Utils::SecondQuantization::CasSpecifier<Utils::Reference::Restricted>& casSpecifier) const
      -> std::map<MatrixType, Utils::SpinAdaptedContainer<Utils::Reference::Restricted, Eigen::MatrixXd>> {
    std::map<MatrixType, Utils::SpinAdaptedContainer<Utils::Reference::Restricted, Eigen::MatrixXd>> res;
    // Get coefficient matrix for core space.
    res.insert({MatrixType::Coefficients, Utils::SpinAdaptedContainer<Utils::Reference::Restricted, Eigen::MatrixXd>{
                                              casSpecifier.getCoreOrbitalsCoefficients().restrictedMatrix()}});
    // Get density matrix for core space
    res.insert({MatrixType::Density, Utils::SpinAdaptedContainer<Utils::Reference::Restricted, Eigen::MatrixXd>{
                                         2.0 * res.at(MatrixType::Coefficients).restricted *
                                         res.at(MatrixType::Coefficients).restricted.transpose()}});
    Utils::DensityMatrix coreDM;
    coreDM.setDensity(Eigen::MatrixXd(res.at(MatrixType::Density).restricted),
                      scf_.getNumberElectrons() - casSpecifier.getCasElectrons().restricted);
    res.insert({MatrixType::Fock, Utils::SpinAdaptedContainer<Utils::Reference::Restricted, Eigen::MatrixXd>{
                                      method_.evaluatePotentialForDensity(coreDM).restrictedMatrix()}});
    return res;
  }

  auto getCoreMatrices(const Utils::SecondQuantization::CasSpecifier<Utils::Reference::Unrestricted>& casSpecifier) const
      -> std::map<MatrixType, Utils::SpinAdaptedContainer<Utils::Reference::Unrestricted, Eigen::MatrixXd>> {
    // Get coefficient matrix for core space.
    Utils::SpinAdaptedContainer<Utils::Reference::Unrestricted, Eigen::MatrixXd> coreCoeffs;
    coreCoeffs.alpha = casSpecifier.getCoreOrbitalsCoefficients().alphaMatrix();
    coreCoeffs.beta = casSpecifier.getCoreOrbitalsCoefficients().betaMatrix();
    // Get density matrix for core space
    Utils::SpinAdaptedContainer<Utils::Reference::Unrestricted, Eigen::MatrixXd> density;
    density.alpha = coreCoeffs.alpha * coreCoeffs.alpha.transpose();
    density.beta = coreCoeffs.beta * coreCoeffs.beta.transpose();
    Utils::DensityMatrix coreDM;
    coreDM.setDensity(Eigen::MatrixXd(density.alpha), Eigen::MatrixXd(density.beta), coreCoeffs.alpha.cols(),
                      coreCoeffs.beta.cols());
    Utils::SpinAdaptedContainer<Utils::Reference::Unrestricted, Eigen::MatrixXd> fock;
    fock.alpha = method_.evaluatePotentialForDensity(coreDM).alphaMatrix();
    fock.beta = method_.evaluatePotentialForDensity(coreDM).betaMatrix();
    return {{MatrixType::Coefficients, std::move(coreCoeffs)},
            {MatrixType::Density, std::move(density)},
            {MatrixType::Fock, std::move(fock)}};
  }

 private:
  MethodType& method_;
  const Utils::ScfMethod& scf_;
  const Integrals::EriType& eris_;
  const Eigen::MatrixXd& hCore_;
  const Utils::Settings& settings_;
  Utils::Results& results_;
  bool requiredSQSpecifier_;
  bool generateFcidump_;
  int orbitalsAroundFermi_{};
  std::vector<int> casIndices_;
}; // namespace nddo

} // namespace nddo
} // namespace Sparrow
} // namespace Scine

#endif
