/**
 * @file OmxMethod.h
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "UnoCAS.h"
#include <Core/Interfaces/Calculator.h>
#include <Sparrow/Implementations/Nddo/NddoSettings.h>
#include <Sparrow/Implementations/OrbitalSteeringCalculator.h>
#include <Utils/CalculatorBasics.h>
#include <Utils/DataStructures/SecondQuantization/CasSpecifier.h>
#include <Utils/DataStructures/SecondQuantization/UnoCI.h>
#include <Utils/Settings.h>
#include <Utils/UniversalSettings/SettingsNames.h>
namespace Scine {
namespace Sparrow {
namespace nddo {
namespace UnoCasGeneration {

auto hasOrthogonalBasis(const std::shared_ptr<Core::Calculator>& calc) -> bool {
  auto const& name = calc->name();
  std::vector<std::string> orthogonal{"MNDO", "AM1", "RM1", "PM3", "PM6", "OM2", "OM3", "ODM2", "ODM3", "AIQM1"};
  return std::find(orthogonal.begin(), orthogonal.end(), name) != orthogonal.end();
}

template<typename FilterType>
auto generate(const std::shared_ptr<Core::Calculator>& calc, FilterType filter, const Utils::Settings& steererSettings,
              const Core::Log& log)
    -> std::tuple<Utils::SecondQuantization::CasSpecifier<Utils::Reference::Restricted>, Eigen::VectorXd> {
  OrbitalSteeringCalculator steerer;
  steerer.settings() = steererSettings;
  steerer.setReferenceCalculator(calc->clone());
  steerer.getReferenceCalculator().settings().modifyString(Utils::SettingsNames::spinMode, "unrestricted");

  // Desactivate settings making the calculation recursive
  if (steerer.getReferenceCalculator().settings().valueExists(Utils::SettingsNames::dumpIntegralsOption)) {
    steerer.getReferenceCalculator().settings().modifyBool(Utils::SettingsNames::dumpIntegralsOption, false);
  }
  Utils::PropertyList properties = hasOrthogonalBasis(calc)
                                       ? Utils::Property::DensityMatrix
                                       : (Utils::Property::DensityMatrix | Utils::Property::OverlapMatrix);
  steerer.getReferenceCalculator().setRequiredProperties(properties);
  steerer.getReferenceCalculator().setLog(log);
  steerer.setLog(log);
  const auto& results = steerer.results();
  // The steering is undeterministic, so we need to do it a couple of times to be sure
  // that it happens. In the norm it is steered after the first iteration, but sometimes
  // also after the second one or so.
  for (int i = 0; i < 6; ++i) {
    steerer.calculate();
  }
  const auto& density = results.get<Utils::Property::DensityMatrix>();

  return Utils::SecondQuantization::UnoCI::getCas(
      density.alphaMatrix(), density.betaMatrix(),
      (hasOrthogonalBasis(calc) ? Eigen::MatrixXd(0, 0) : results.get<Utils::Property::OverlapMatrix>()), filter);
}

template auto generate<Utils::SecondQuantization::UnoCI::ThresholdFilter>(const std::shared_ptr<Core::Calculator>&,
                                                                          Utils::SecondQuantization::UnoCI::ThresholdFilter,
                                                                          const Utils::Settings&, const Core::Log&)
    -> std::tuple<Utils::SecondQuantization::CasSpecifier<Utils::Reference::Restricted>, Eigen::VectorXd>;
template auto generate<Utils::SecondQuantization::UnoCI::CasSizeFilter>(const std::shared_ptr<Core::Calculator>&,
                                                                        Utils::SecondQuantization::UnoCI::CasSizeFilter,
                                                                        const Utils::Settings&, const Core::Log&)
    -> std::tuple<Utils::SecondQuantization::CasSpecifier<Utils::Reference::Restricted>, Eigen::VectorXd>;
} // namespace UnoCasGeneration
} // namespace nddo
} // namespace Sparrow
} // namespace Scine
