/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef SPARROW_NDDOEXCITEDSTATESASSEMBLER_H
#define SPARROW_NDDOEXCITEDSTATESASSEMBLER_H

#include <Sparrow/Implementations/MoldenFileGenerator.h>
#include <Sparrow/Implementations/Nddo/TimeDependent/UnoCAS.h>
#include <Utils/DataStructures/MolecularOrbitals.h>
#include <Utils/DataStructures/SecondQuantization/CasSpecifier.h>

namespace Scine {
namespace Sparrow {
namespace nddo {
template<typename MethodType>
class NddoExcitedStatesAssembler {
 public:
  NddoExcitedStatesAssembler(MethodType& method);
  auto generateCasWithUNO(int nElectrons, int nAtomicOrbitals)
      -> std::tuple<Utils::SecondQuantization::CasSpecifier<Utils::Reference::Restricted>, Eigen::VectorXd>;
  void printUnoCiWavefunction(std::ostream& out, const Utils::MolecularOrbitals& mos, const Eigen::VectorXd& occupations) const;

  void printUnoCiWavefunction(const std::string& filename, const Utils::MolecularOrbitals& mos,
                              const Eigen::VectorXd& occupations) const;

 private:
  MethodType& method_;
};

template<typename MethodType>
NddoExcitedStatesAssembler<MethodType>::NddoExcitedStatesAssembler(MethodType& method) : method_(method) {
}

template<typename MethodType>
void NddoExcitedStatesAssembler<MethodType>::printUnoCiWavefunction(std::ostream& out, const Utils::MolecularOrbitals& mos,
                                                                    const Eigen::VectorXd& occupations) const {
  MoldenFileGenerator generator(method_);
  generator.generateWavefunctionInformation<Utils::Reference::Restricted>(
      out, mos, Utils::SpinAdaptedContainer<Utils::Reference::Restricted, Eigen::VectorXd>{occupations});
}

template<typename MethodType>
void NddoExcitedStatesAssembler<MethodType>::printUnoCiWavefunction(const std::string& filename,
                                                                    const Utils::MolecularOrbitals& mos,
                                                                    const Eigen::VectorXd& occupations) const {
  std::ofstream file(filename);
  if (!file.is_open()) {
    throw std::runtime_error("Impossible to open file " + filename + ".");
  }
  printUnoCiWavefunction(file, mos, occupations);
}
template<typename MethodType>
auto NddoExcitedStatesAssembler<MethodType>::generateCasWithUNO(int nElectrons, int nAtomicOrbitals)
    -> std::tuple<Utils::SecondQuantization::CasSpecifier<Utils::Reference::Restricted>, Eigen::VectorXd> {
  boost::optional<std::tuple<Utils::SecondQuantization::CasSpecifier<Utils::Reference::Restricted>, Eigen::VectorXd>> cas;
  int casOrbitalsAroundFermi = method_.settings().getInt(Utils::SettingsNames::casOrbitalsAroundFermiOption);
  OrbitalSteeringSettings settings;
  settings.modifyString(Utils::SettingsNames::mixer, Utils::SettingsNames::ScfMixers::ediisDiis);
  settings.modifyInt(Utils::SettingsNames::maxScfIterations, 500);

  if (casOrbitalsAroundFermi == 0) {
    double lowThreshold = method_.settings().getDouble(Utils::SettingsNames::unoCiLowThresholdOption);
    cas = UnoCasGeneration::generate(
        method_.clone(), Utils::SecondQuantization::UnoCI::ThresholdFilter{{lowThreshold, 2. - lowThreshold}}, settings);
  }
  else {
    cas = UnoCasGeneration::generate(method_.clone(),
                                     Utils::SecondQuantization::UnoCI::CasSizeFilter{casOrbitalsAroundFermi}, settings);
  }

  // If no orbitals were included in cas just do HOMO-LUMO UNO cas
  if (std::get<0>(*cas).empty()) {
    method_.getLog().warning << "No orbitals would be included in CAS. A minimal CAS(2,2) was created.";
    if (nElectrons < 2 || nAtomicOrbitals < 2) {
      throw std::runtime_error("Not enough electrons (" + std::to_string(nElectrons) + ") or orbitals(" +
                               std::to_string(nAtomicOrbitals) + ") to build a CAS(2,2). Are you sure you want to use UNO-CAS?");
    }
    Utils::SecondQuantization::CasGenerator generator(std::get<0>(*cas).getCasOrbitals());
    return {generator.generateRestricted(nElectrons, 2), (Eigen::VectorXd(2) << 2, 0).finished()};
  }

  std::string filename = "UNO_cas_" + std::to_string(std::get<0>(*cas).getCasElectrons().restricted) + "-" +
                         std::to_string(std::get<0>(*cas).getActiveIndices().restricted.size()) + ".molden.input";
  printUnoCiWavefunction(filename, std::get<0>(*cas).getCasOrbitals(), std::get<1>(*cas));

  return *cas;
}
} // namespace nddo
} // namespace Sparrow
} // namespace Scine
#endif // SPARROW_NDDOEXCITEDSTATESASSEMBLER_H
