/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include <Core/Interfaces/Calculator.h>
#include <Core/Log.h>
#include <Sparrow/Implementations/DipoleMatrixCalculator.h>
#include <Sparrow/Implementations/Nddo/Pm6/Wrapper/PM6MethodWrapper.h>
#include <Sparrow/Implementations/Nddo/Utils/DipoleUtils/AtomPairDipole.h>
#include <Sparrow/Implementations/Nddo/Utils/DipoleUtils/NDDODipoleMatrixCalculator.h>
#include <Sparrow/Implementations/Nddo/Utils/DipoleUtils/NDDODipoleMomentCalculator.h>
#include <Utils/CalculatorBasics.h>
#include <Utils/DataStructures/AtomicGtos.h>
#include <Utils/DataStructures/DipoleMatrix.h>
#include <Utils/DataStructures/SlaterToGaussian.h>
#include <Utils/Geometry/AtomCollection.h>
#include <Utils/IO/ChemicalFileFormats/XyzStreamHandler.h>
#include <Utils/Settings.h>
#include <Utils/Typenames.h>
#include <Utils/UniversalSettings/SettingsNames.h>
#include <gmock/gmock.h>
#include <Eigen/Core>
#include <array>
#include <chrono>

using namespace testing;

namespace Scine {
namespace Sparrow {

class SlowDipoleCalculation : public Test {
 public:
  std::shared_ptr<Core::Calculator> method;

 private:
  void SetUp() override {
    method = std::make_shared<PM6MethodWrapper>();
    method->setLog(Core::Log::silent());

    auto& settings = method->settings();
    settings.modifyDouble(Utils::SettingsNames::selfConsistenceCriterion, 1e-9);
    settings.modifyBool(Utils::SettingsNames::NDDODipoleApproximation, false);
    method->setRequiredProperties(Utils::Property::Energy);
  }
};

TEST_F(SlowDipoleCalculation, DipoleIsCorrectForBigOrganicMolecule) {
  method->setRequiredProperties(Utils::Property::Energy | Utils::Property::Dipole);
  std::stringstream BigMoleculeSS("113\n\n"
                                  "C    14.663000    29.225000     8.550000\n"
                                  "O    15.679000    23.530000     9.284000\n"
                                  "C    15.446000    24.383000     8.168000\n"
                                  "N    14.572000    30.082000    11.272000\n"
                                  "O    15.782000    29.646000     8.475000\n"
                                  "C    13.465000    29.993000     9.084000\n"
                                  "O    13.409000    23.131000     8.031000\n"
                                  "C    14.530000    23.550000     7.207000\n"
                                  "O    13.665000    31.396000     8.952000\n"
                                  "C    13.981000    24.239000     5.913000\n"
                                  "O    11.843000    25.089000     6.726000\n"
                                  "C    13.331000    29.725000    10.608000\n"
                                  "O    11.300000    22.870000     4.513000\n"
                                  "C    12.443000    24.165000     5.775000\n"
                                  "O    15.297000    27.944000    11.449000\n"
                                  "C    11.809000    24.239000     4.371000\n"
                                  "O    14.900000    24.579000     2.267000\n"
                                  "C    15.500000    29.146000    11.631000\n"
                                  "O    16.937000    23.198000     4.018000\n"
                                  "C    12.690000    24.440000     3.137000\n"
                                  "O    18.173000    25.452000     4.631000\n"
                                  "C    14.142000    24.761000     3.477000\n"
                                  "O    14.293000    27.968000     8.242000\n"
                                  "C    14.685000    23.777000     4.594000\n"
                                  "O    14.180000    21.013000     8.016000\n"
                                  "C    16.211000    24.019000     4.536000\n"
                                  "O    12.296000    26.819000     5.358000\n"
                                  "C    16.768000    25.373000     4.958000\n"
                                  "O    17.590000    25.630000     2.474000\n"
                                  "C    16.537000    25.716000     6.403000\n"
                                  "C    15.883000    26.853000     6.702000\n"
                                  "C    15.365000    26.987000     8.120000\n"
                                  "C    14.741000    25.692000     8.628000\n"
                                  "C    16.829000    24.694000     7.509000\n"
                                  "C    17.817000    25.283000     8.563000\n"
                                  "C    17.533000    23.387000     7.066000\n"
                                  "C    14.452000    22.283000     4.272000\n"
                                  "C    15.595000    27.978000     5.756000\n"
                                  "C    11.716000    22.817000     5.876000\n"
                                  "C    13.407000    21.834000     8.432000\n"
                                  "C    12.322000    21.565000     9.402000\n"
                                  "C    11.326000    22.468000     9.688000\n"
                                  "C    10.344000    22.126000    10.598000\n"
                                  "C    10.368000    20.934000    11.245000\n"
                                  "C    11.346000    20.041000    10.963000\n"
                                  "C    12.328000    20.345000    10.030000\n"
                                  "C    12.155000    30.467000    11.228000\n"
                                  "C    10.972000    30.594000    10.576000\n"
                                  "C     9.881000    31.266000    11.159000\n"
                                  "C    10.016000    31.826000    12.365000\n"
                                  "C    11.179000    31.735000    12.995000\n"
                                  "C    12.278000    31.046000    12.470000\n"
                                  "C    11.819000    26.403000     6.384000\n"
                                  "C    11.141000    27.185000     7.455000\n"
                                  "C    16.728000    29.689000    12.287000\n"
                                  "C    16.692000    30.723000    13.191000\n"
                                  "C    17.815000    31.137000    13.839000\n"
                                  "C    19.009000    30.546000    13.564000\n"
                                  "C    19.099000    29.524000    12.617000\n"
                                  "C    17.936000    29.083000    11.990000\n"
                                  "C    18.443000    25.606000     3.318000\n"
                                  "C    19.908000    25.754000     3.065000\n"
                                  "H    15.164000    23.911000    10.056000\n"
                                  "H    12.660000    29.703000     8.594000\n"
                                  "H    15.064000    22.834000     6.818000\n"
                                  "H    14.210000    25.180000     6.039000\n"
                                  "H    13.157000    28.794000    10.711000\n"
                                  "H    14.747000    31.020000    11.471000\n"
                                  "H    11.229000    24.988000     4.178000\n"
                                  "H    14.210000    25.658000     3.808000\n"
                                  "H    15.700000    25.275000     2.308000\n"
                                  "H    13.743000    31.792000     9.823000\n"
                                  "H    16.297000    26.065000     4.473000\n"
                                  "H    16.158000    27.238000     8.613000\n"
                                  "H    11.328000    23.337000     9.259000\n"
                                  "H     9.619000    22.738000    10.768000\n"
                                  "H     9.659000    20.728000    11.879000\n"
                                  "H    11.368000    19.196000    11.414000\n"
                                  "H    13.017000    19.699000     9.847000\n"
                                  "H    10.871000    30.182000     9.695000\n"
                                  "H     9.023000    31.355000    10.702000\n"
                                  "H     9.281000    32.288000    12.791000\n"
                                  "H    11.288000    32.169000    13.902000\n"
                                  "H    13.137000    31.044000    12.962000\n"
                                  "H    15.840000    31.211000    13.408000\n"
                                  "H    17.787000    31.834000    14.500000\n"
                                  "H    19.854000    30.900000    13.931000\n"
                                  "H    20.013000    29.129000    12.411000\n"
                                  "H    18.006000    28.339000    11.329000\n"
                                  "H    12.322000    25.156000     2.592000\n"
                                  "H    12.660000    23.648000     2.602000\n"
                                  "H    14.687000    25.706000     9.581000\n"
                                  "H    13.812000    25.658000     8.300000\n"
                                  "H    18.662000    25.467000     8.167000\n"
                                  "H    17.946000    24.653000     9.278000\n"
                                  "H    18.383000    23.576000     6.647000\n"
                                  "H    17.668000    22.810000     7.806000\n"
                                  "H    16.992000    22.906000     6.419000\n"
                                  "H    14.826000    21.733000     4.985000\n"
                                  "H    13.534000    22.092000     4.178000\n"
                                  "H    14.925000    22.044000     3.466000\n"
                                  "H    15.979000    27.765000     4.890000\n"
                                  "H    14.667000    28.076000     5.650000\n"
                                  "H    15.979000    28.770000     6.058000\n"
                                  "H    10.990000    22.810000     6.524000\n"
                                  "H    12.282000    22.068000     6.106000\n"
                                  "H    11.090000    28.124000     7.264000\n"
                                  "H    10.235000    26.879000     7.635000\n"
                                  "H    11.606000    27.094000     8.318000\n"
                                  "H    17.469000    26.089000     8.926000\n"
                                  "H    20.093000    25.850000     2.137000\n"
                                  "H    20.252000    26.520000     3.533000\n"
                                  "H    20.371000    24.964000     3.381000\n");

  auto BigMolecule = Utils::XyzStreamHandler::read(BigMoleculeSS);
  method->setStructure(BigMolecule);

  auto res = method->calculate("");
  const auto dipole = res.get<Utils::Property::Dipole>();
  auto dipoleFromPreciseCalculation = 12.8503; // Debye
  // Dipole must be similar to true one (in Debye-> *2.541746 D/a.u), allow for 10% error.
  // Result of precise calculation in TestScripts
  ASSERT_THAT(dipole.norm() * 2.541746 / dipoleFromPreciseCalculation, DoubleNear(1, 0.1));
}

TEST_F(SlowDipoleCalculation, NddoDipoleIsCorrectForBigOrganicMolecule) {
  std::stringstream BigMoleculeSS("113\n\n"
                                  " O         15.68894000     23.55459000      9.30853000\n"
                                  " O         15.81287000     29.65719000      8.45683000\n"
                                  " O         13.40468000     23.13785000      8.04487000\n"
                                  " O         13.66559000     31.43847000      8.92988000\n"
                                  " O         11.84316000     25.10056000      6.75053000\n"
                                  " O         11.27957000     22.92183000      4.48568000\n"
                                  " O         15.33904000     27.93590000     11.43104000\n"
                                  " O         14.89455000     24.61996000      2.26223000\n"
                                  " O         16.96517000     23.15749000      4.02856000\n"
                                  " O         18.19509000     25.47139000      4.66759000\n"
                                  " O         14.27744000     27.96112000      8.23079000\n"
                                  " O         14.19171000     20.98492000      8.01749000\n"
                                  " O         12.29558000     26.86046000      5.36613000\n"
                                  " O         17.56628000     25.65774000      2.46106000\n"
                                  " N         14.56699000     30.07902000     11.30982000\n"
                                  " C         14.68174000     29.22284000      8.56648000\n"
                                  " C         15.45143000     24.38602000      8.15052000\n"
                                  " C         13.50110000     30.02776000      9.09556000\n"
                                  " C         14.50265000     23.55220000      7.20651000\n"
                                  " C         13.97563000     24.22371000      5.91584000\n"
                                  " C         13.34539000     29.74221000     10.59931000\n"
                                  " C         12.43400000     24.18275000      5.79537000\n"
                                  " C         11.84834000     24.25174000      4.36145000\n"
                                  " C         15.49332000     29.13246000     11.64411000\n"
                                  " C         12.71374000     24.40802000      3.13631000\n"
                                  " C         14.15346000     24.71850000      3.47199000\n"
                                  " C         14.69562000     23.79842000      4.58623000\n"
                                  " C         16.22023000     24.00886000      4.53087000\n"
                                  " C         16.81442000     25.35470000      4.96225000\n"
                                  " C         16.55976000     25.70865000      6.39811000\n"
                                  " C         15.88346000     26.84571000      6.71155000\n"
                                  " C         15.33940000     27.00498000      8.10391000\n"
                                  " C         14.75248000     25.68359000      8.61465000\n"
                                  " C         16.82938000     24.68844000      7.50189000\n"
                                  " C         17.81199000     25.26213000      8.54280000\n"
                                  " C         17.53019000     23.39882000      7.06519000\n"
                                  " C         14.47601000     22.30914000      4.27596000\n"
                                  " C         15.59993000     27.96092000      5.75969000\n"
                                  " C         11.71549000     22.82509000      5.85431000\n"
                                  " C         13.43435000     21.82880000      8.41849000\n"
                                  " C         12.34736000     21.56724000      9.38923000\n"
                                  " C         11.33985000     22.48099000      9.67784000\n"
                                  " C         10.33828000     22.15278000     10.58677000\n"
                                  " C         10.35451000     20.93427000     11.24191000\n"
                                  " C         11.35932000     20.02687000     10.96377000\n"
                                  " C         12.34265000     20.33518000     10.03119000\n"
                                  " C         12.17664000     30.47797000     11.22155000\n"
                                  " C         10.95065000     30.57629000     10.56476000\n"
                                  " C          9.87791000     31.25309000     11.13399000\n"
                                  " C         10.00793000     31.83076000     12.37913000\n"
                                  " C         11.21090000     31.74058000     13.04578000\n"
                                  " C         12.27968000     31.06453000     12.48174000\n"
                                  " C         11.81973000     26.40896000      6.38046000\n"
                                  " C         11.13872000     27.18481000      7.46687000\n"
                                  " C         16.71749000     29.66167000     12.28018000\n"
                                  " C         16.66427000     30.72002000     13.18879000\n"
                                  " C         17.81868000     31.13753000     13.83690000\n"
                                  " C         19.03548000     30.55095000     13.53529000\n"
                                  " C         19.10009000     29.51933000     12.61263000\n"
                                  " C         17.94140000     29.06596000     11.99062000\n"
                                  " C         18.42320000     25.60374000      3.32230000\n"
                                  " C         19.89529000     25.74309000      3.07987000\n"
                                  " H         15.16627000     23.88279000     10.05888000\n"
                                  " H         12.61120000     29.71350000      8.54163000\n"
                                  " H         15.06941000     22.69588000      6.86985000\n"
                                  " H         14.19722000     25.27422000      6.01922000\n"
                                  " H         13.14478000     28.67651000     10.73243000\n"
                                  " H         14.76787000     31.06382000     11.43566000\n"
                                  " H         11.06762000     25.00235000      4.23170000\n"
                                  " H         14.21517000     25.75778000      3.80641000\n"
                                  " H         15.67245000     25.21900000      2.28643000\n"
                                  " H         16.27859000     26.06971000      4.35425000\n"
                                  " H         16.16746000     27.33282000      8.71910000\n"
                                  " H         11.31303000     23.45050000      9.20152000\n"
                                  " H          9.53106000     22.84162000     10.79023000\n"
                                  " H          9.57923000     20.69386000     11.95893000\n"
                                  " H         11.37607000     19.07270000     11.47077000\n"
                                  " H         13.11452000     19.60724000      9.81988000\n"
                                  " H         10.82137000     30.13843000      9.58713000\n"
                                  " H          8.93238000     31.34348000     10.61877000\n"
                                  " H          9.17891000     32.35272000     12.83531000\n"
                                  " H         11.31381000     32.21170000     14.00998000\n"
                                  " H         13.20619000     31.03173000     13.03773000\n"
                                  " H         15.74161000     31.22954000     13.42928000\n"
                                  " H         17.77621000     31.93087000     14.57134000\n"
                                  " H         19.94155000     30.92231000     13.99551000\n"
                                  " H         20.06383000     29.08383000     12.38096000\n"
                                  " H         18.00275000     28.26321000     11.26730000\n"
                                  " H         12.31093000     25.22577000      2.52791000\n"
                                  " H         12.65197000     23.52449000      2.50430000\n"
                                  " H         14.68841000     25.73408000      9.70223000\n"
                                  " H         13.71185000     25.65042000      8.29542000\n"
                                  " H         18.78044000     25.48765000      8.09474000\n"
                                  " H         17.96953000     24.56421000      9.36730000\n"
                                  " H         18.49572000     23.60736000      6.60120000\n"
                                  " H         17.69632000     22.73454000      7.91451000\n"
                                  " H         16.95213000     22.81777000      6.36140000\n"
                                  " H         14.87233000     21.66911000      5.06649000\n"
                                  " H         13.43326000     22.04626000      4.14831000\n"
                                  " H         14.97357000     22.01508000      3.35062000\n"
                                  " H         16.02277000     27.80580000      4.77422000\n"
                                  " H         14.52635000     28.09181000      5.63565000\n"
                                  " H         16.04339000     28.88920000      6.11321000\n"
                                  " H         10.87952000     22.80916000      6.55984000\n"
                                  " H         12.34095000     21.95832000      6.06644000\n"
                                  " H         11.10038000     28.24006000      7.20347000\n"
                                  " H         10.12125000     26.82392000      7.61594000\n"
                                  " H         11.68846000     27.07682000      8.40192000\n"
                                  " H         17.48064000     26.17473000      9.01138000\n"
                                  " H         20.08808000     25.86160000      2.01213000\n"
                                  " H         20.27627000     26.62537000      3.59684000\n"
                                  " H         20.41636000     24.85112000      3.42576000\n"
                                  " H         13.77125000     31.61639000      7.98284000\n");

  auto BigMolecule = Utils::XyzStreamHandler::read(BigMoleculeSS);
  method->setRequiredProperties(Utils::Property::Energy | Utils::Property::Dipole);
  method->settings().modifyBool(Utils::SettingsNames::NDDODipoleApproximation, true);
  method->setStructure(BigMolecule);

  auto res = method->calculate("");
  auto dipole = res.get<Utils::Property::Dipole>().norm() * 2.541746;

  // Reference value obtained with MOPAC (as implemented in ADF 2016.107)
  ASSERT_THAT(dipole, DoubleNear(11.693, 0.005));
}

} // namespace Sparrow
} // namespace Scine
