/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include <Integrals/Libint.h>
#include <Sparrow/Implementations/Nddo/Omx/EffectiveCorePotentialEvaluator.h>
#include <Sparrow/Implementations/Nddo/Omx/ParametersProcessor.h>
#include <Utils/Constants.h>
#include <Utils/DataStructures/SlaterToGaussian.h>
#include <Utils/Geometry/ElementTypes.h>
#include <Utils/IO/ChemicalFileFormats/XyzStreamHandler.h>
#include <Utils/Math/DerivOrderEnum.h>
#include <Utils/Typenames.h>
#include <gmock/gmock.h>

namespace Scine {
namespace Sparrow {
namespace nddo {

using namespace testing;
using namespace Integrals;

class AnECPTest : public Test {
 public:
  Utils::ElementTypeCollection elements;
  Utils::PositionCollection positions;

  OmxElementParameters elParams, elParamsOm3;

  std::vector<Utils::AtomicGtos> gtos, gtosOm3;
  Utils::AtomCollection structure;
  BasisSetController basisSet;

  void SetUp() final {
    elParams.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, om2()));
    elParams.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om2()));
    elParams.insert(Utils::ElementType::N, ParametersProcessor::processAtomicParameters(Utils::ElementType::N, om2()));
    elParams.insert(Utils::ElementType::O, ParametersProcessor::processAtomicParameters(Utils::ElementType::O, om2()));
    elParamsOm3.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, om3()));
    elParamsOm3.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om3()));
    elParamsOm3.insert(Utils::ElementType::N, ParametersProcessor::processAtomicParameters(Utils::ElementType::N, om3()));
    elParamsOm3.insert(Utils::ElementType::O, ParametersProcessor::processAtomicParameters(Utils::ElementType::O, om3()));
    Libint::getInstance();
  }
};

TEST_F(AnECPTest, ThrowsIfEmptyBasisConstructed) {
  gtos = {};
  ASSERT_THROW(BasisSetController({elements, positions}, gtos), std::runtime_error);
}

TEST_F(AnECPTest, ThrowsIfMismatchedGtosAndStructure) {
  gtos = {};
  elements = {Utils::ElementType::C, Utils::ElementType::C};
  positions = Utils::PositionCollection(2, 3);
  positions.row(0) = Utils::Position(0, 0, 0);
  positions.row(1) = Utils::Position(0, 0, 1.2 * Utils::Constants::bohr_per_angstrom);
  ASSERT_THROW(BasisSetController({elements, positions}, gtos), std::runtime_error);
}

TEST_F(AnECPTest, CalculatesECPPotentialMatrix) {
  elements = {Utils::ElementType::C, Utils::ElementType::C};
  positions = Utils::PositionCollection(2, 3);
  positions.row(0) = Utils::Position(0, 0, 0);
  positions.row(1) = Utils::Position(0, 0, 1.2 * Utils::Constants::bohr_per_angstrom);
  gtos = {};
  gtos.push_back(elParams.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::C).gtos);
  basisSet = BasisSetController({elements, positions}, gtos, "om2-3g");

  auto ecpEval = EffectiveCorePotentialEvaluator(elements, positions, elParams);
  Eigen::MatrixXd sum = ecpEval.evaluate(0, 1, basisSet);
  // Data from https://apps.automeris.io/wpd/
  // From W. Weber, "Ein neues emiempirisches NDDO-Verfahren mit Orthogonalisierungskorrekturen", 1996, Seite 80,
  // Figures 6.6, 6.8, 6.10
  EXPECT_THAT(sum(0, 0) * Utils::Constants::ev_per_hartree, DoubleNear(0.90, 2e-2));
  EXPECT_THAT(sum(0, 3) * Utils::Constants::ev_per_hartree, DoubleNear(1.70, 2e-2));
  EXPECT_THAT(sum(3, 3) * Utils::Constants::ev_per_hartree, DoubleNear(3.15, 2e-2));
}

TEST_F(AnECPTest, CanCalculateECPForCCOm2) {
  elements = Utils::ElementTypeCollection{Utils::ElementType::C, Utils::ElementType::C};
  positions = Utils::PositionCollection(2, 3);
  positions.row(0) = Utils::Position(0, 0, 0);
  positions.row(1) = Utils::Position(0, 0, 0.28803 * Utils::Constants::bohr_per_angstrom);
  auto pars = elParams.get(Utils::ElementType::C);
  gtos = {};
  gtos.push_back(pars.gtos);
  gtos.push_back(pars.gtos);
  basisSet = BasisSetController({elements, positions}, gtos, "om2-3g");
  auto ecpEval = EffectiveCorePotentialEvaluator(elements, positions, elParams);
  Eigen::MatrixXd sum = ecpEval.evaluate(0, 1, basisSet);
  // Data from https://apps.automeris.io/wpd/
  // From W. Weber, "Ein neues emiempirisches NDDO-Verfahren mit Orthogonalisierungskorrekturen", 1996, Seite 80,
  // Figure 6.7 uncertainty from reading the data from plot, jump of 0.2 per tick
  EXPECT_THAT(sum(0, 3) * Utils::Constants::ev_per_hartree, DoubleNear(51.856, 1e-1));

  positions.row(1) = Utils::Position(0, 0, 1.0 * Utils::Constants::bohr_per_angstrom);
  basisSet.modifyPositions(positions);
  ecpEval = EffectiveCorePotentialEvaluator(elements, positions, elParams);
  sum = ecpEval.evaluate(0, 1, basisSet);
  // Data from https://apps.automeris.io/wpd/
  // From W. Weber, "Ein neues emiempirisches NDDO-Verfahren mit Orthogonalisierungskorrekturen", 1996, Seite 80,
  // Figure 6.7 uncertainty from reading the data from plot, jump of 0.2 per tick
  EXPECT_THAT(sum(0, 3) * Utils::Constants::ev_per_hartree, DoubleNear(4.5296, 1e-1));
}

TEST_F(AnECPTest, CanCalculateECPForCCOm3) {
  elements = Utils::ElementTypeCollection{Utils::ElementType::C, Utils::ElementType::C};
  positions = Utils::PositionCollection(2, 3);
  positions.row(0) = Utils::Position(0, 0, 0);
  positions.row(1) = Utils::Position(0, 0, 0.29315 * Utils::Constants::bohr_per_angstrom);
  auto pars = elParamsOm3.get(Utils::ElementType::C);
  gtosOm3.push_back(pars.gtos);
  gtosOm3.push_back(pars.gtos);
  basisSet = BasisSetController({elements, positions}, gtosOm3, "om2-3g");
  auto ecpEval = EffectiveCorePotentialEvaluator(elements, positions, elParamsOm3);
  Eigen::MatrixXd sum = ecpEval.evaluate(0, 1, basisSet);
  // Data from https://apps.automeris.io/wpd/
  // Mirjam Scholten, Semiempirische Verfahren mit Orthogonalisierungskorrekturen: Die OM3 Methode
  // 2003 (The OM3 PhD thesis), Abbildung 6.11-6.13
  EXPECT_THAT(sum(0, 0) * Utils::Constants::ev_per_hartree, DoubleNear(29.5, 5e-1));
  EXPECT_THAT(sum(0, 3) * Utils::Constants::ev_per_hartree, DoubleNear(36.0, 5e-1));
  EXPECT_THAT(sum(3, 3) * Utils::Constants::ev_per_hartree, DoubleNear(44.1, 5e-1));

  positions.row(1) = Utils::Position(0, 0, 1.0 * Utils::Constants::bohr_per_angstrom);
  basisSet.modifyPositions(positions);
  ecpEval = EffectiveCorePotentialEvaluator(elements, positions, elParamsOm3);
  sum = ecpEval.evaluate(0, 1, basisSet);
  // Data from https://apps.automeris.io/wpd/
  // Mirjam Scholten, Semiempirische Verfahren mit Orthogonalisierungskorrekturen: Die OM3 Methode
  // 2003 (The OM3 PhD thesis), Abbildung 6.11
  EXPECT_THAT(sum(0, 0) * Utils::Constants::ev_per_hartree, DoubleNear(3.00, 1e-1));
  EXPECT_THAT(sum(0, 3) * Utils::Constants::ev_per_hartree, DoubleNear(4.90, 1e-1));
  EXPECT_THAT(sum(3, 3) * Utils::Constants::ev_per_hartree, DoubleNear(8.00, 1e-1));
}

TEST_F(AnECPTest, CanCalculateECPForOOOm3) {
  elements = Utils::ElementTypeCollection{Utils::ElementType::O, Utils::ElementType::O};
  positions = Utils::PositionCollection(2, 3);
  positions.row(0) = Utils::Position(0, 0, 0);
  positions.row(1) = Utils::Position(0, 0, 0.225 * Utils::Constants::bohr_per_angstrom);
  auto pars = elParamsOm3.get(Utils::ElementType::O);
  gtosOm3.push_back(pars.gtos);
  gtosOm3.push_back(pars.gtos);
  basisSet = BasisSetController({elements, positions}, gtosOm3, "om2-3g");
  auto ecpEval = EffectiveCorePotentialEvaluator(elements, positions, elParamsOm3);
  Eigen::MatrixXd sum = ecpEval.evaluate(0, 1, basisSet);
  // Data from https://apps.automeris.io/wpd/
  // Mirjam Scholten, Semiempirische Verfahren mit Orthogonalisierungskorrekturen: Die OM3 Methode
  // 2003 (The OM3 PhD thesis), Abbildung 6.22
  EXPECT_THAT(sum(0, 0) * Utils::Constants::ev_per_hartree, DoubleNear(45.85, 1e-1));

  positions.row(1) = Utils::Position(0, 0, 0.5 * Utils::Constants::bohr_per_angstrom);
  basisSet.modifyPositions(positions);
  ecpEval = EffectiveCorePotentialEvaluator(elements, positions, elParamsOm3);
  sum = ecpEval.evaluate(0, 1, basisSet);
  // Data from https://apps.automeris.io/wpd/
  // Mirjam Scholten, Semiempirische Verfahren mit Orthogonalisierungskorrekturen: Die OM3 Methode
  // 2003 (The OM3 PhD thesis), Abbildung 6.22
  EXPECT_THAT(sum(0, 0) * Utils::Constants::ev_per_hartree, DoubleNear(21.00, 1e-1));
}

TEST_F(AnECPTest, CanCalculateECPForOHOm3) {
  elements = Utils::ElementTypeCollection{Utils::ElementType::O, Utils::ElementType::H};
  positions = Utils::PositionCollection(2, 3);
  positions.row(0) = Utils::Position(0, 0, 0);
  positions.row(1) = Utils::Position(0, 0, 1.2 * Utils::Constants::bohr_per_angstrom);
  gtosOm3.push_back(elParamsOm3.get(Utils::ElementType::O).gtos);
  gtosOm3.push_back(elParamsOm3.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({elements, positions}, gtosOm3, "om2-3g");
  auto ecpEval = EffectiveCorePotentialEvaluator(elements, positions, elParamsOm3);
  Eigen::MatrixXd sum = ecpEval.evaluate(1, 0, basisSet);
  // Data from https://apps.automeris.io/wpd/
  // Mirjam Scholten, Semiempirische Verfahren mit Orthogonalisierungskorrekturen: Die OM3 Methode
  // 2003 (The OM3 PhD thesis), Abbildung 6.22
  EXPECT_THAT(sum(0, 0) * Utils::Constants::ev_per_hartree, DoubleNear(0.795, 1e-2));

  positions.row(1) = Utils::Position(0, 0, 2.0 * Utils::Constants::bohr_per_angstrom);
  basisSet.modifyPositions(positions);
  ecpEval = EffectiveCorePotentialEvaluator(elements, positions, elParamsOm3);
  sum = ecpEval.evaluate(1, 0, basisSet);
  // Data from https://apps.automeris.io/wpd/
  // Mirjam Scholten, Semiempirische Verfahren mit Orthogonalisierungskorrekturen: Die OM3 Methode
  // 2003 (The OM3 PhD thesis), Abbildung 6.22
  EXPECT_THAT(sum(0, 0) * Utils::Constants::ev_per_hartree, DoubleNear(0.019, 1e-3));
}

TEST_F(AnECPTest, CoreOverlapCalculatedCorrectlyOm2PyScf) {
  std::stringstream structure{"4\n\n"
                              "O        0.00000       0.00000        0.00000\n"
                              "C        1.00000       0.00000        0.00000\n"
                              "N        2.00000       0.00000        0.00000\n"
                              "H        1.00000       1.00000        0.00000\n"};

  auto s = Utils::XyzStreamHandler::read(structure);
  gtos = {};
  gtos.push_back(elParams.get(Utils::ElementType::O).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::N).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  auto bSet = BasisSetController(s, gtos, "om2-3g");

  auto coreOrbitalO = Utils::SlaterToGaussian::getGTOExpansion(3, 1, 0, 7.58579774);
  auto coreOrbitalN = Utils::SlaterToGaussian::getGTOExpansion(3, 1, 0, 6.939806);
  auto coreOrbitalC = Utils::SlaterToGaussian::getGTOExpansion(3, 1, 0, 5.16802668);
  Eigen::MatrixXd coreOverlapO = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(0, bSet, coreOrbitalO);
  Eigen::MatrixXd coreOverlapC = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(1, bSet, coreOrbitalC);
  Eigen::MatrixXd coreOverlapN = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(2, bSet, coreOrbitalN);
  Eigen::VectorXd pyscfO(13), pyscfN(13), pyscfC(13);
  pyscfO << 0.000000, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 4.36265121e-02, -8.37548472e-02, 0.00000000e+00,
      0.00000000e+00, 7.81826355e-04, -2.72431560e-03, 0.00000000e+00, 0.00000000e+00, 1.36997330e-02;
  pyscfN << 5.02347696e-05, 2.33865775e-04, 0.00000000e+00, 0.00000000e+00, 5.01166861e-02, 9.56408823e-02,
      0.00000000e+00, 0.00000000e+00, 0.0, 0.0, 0.00000000e+00, 0.00000000e+00, 1.57298068e-02;
  pyscfC << 3.75563399e-02, 8.46304489e-02, 0.00000000e+00, 0.00000000e+00, 0.0, 0.0, 0.00000000e+00, 0.00000000e+00,
      6.82002260e-02, -1.27640474e-01, 0.00000000e+00, 0.00000000e+00, 7.48301070e-02;
  /*
   * # From PyScf,
   *# Core overlap
   *
   *   from pyscf import gto
   *from pyscf.scf import _vhf
   *from pyscf import ao2mo
   *
   *mol = gto.Mole()
   *
   *bas = {
   *    'O': gto.basis.parse('''
   *    O    S
   *    0.2227660584e+01       0.1543289673E+00
   *    0.4057711562e+00       0.5353281423E+00
   *    0.1098175104e+00       0.4446345422E+00
   *    O    SP
   *    0.4784490000E+01      -0.192480000E+00     0.241580000E+00
   *    0.9986000000E+00       0.669520000E+00     0.558900000E+00
   *    0.2568700000E+00       0.502700000E+00     0.431600000E+00
   *    '''),
   *    'C': gto.basis.parse('''
   *    C    S
   *    0.2227660584e+01       0.1543289673E+00
   *    0.4057711562e+00       0.5353281423E+00
   *    0.1098175104e+00       0.4446345422E+00
   *    C    SP
   *    0.2644860000E+01      -0.191880000E+00     0.202590000E+00
   *    0.5421500000E+00       0.616280000E+00     0.558300000E+00
   *    0.1446600000E+00       0.548960000E+00     0.455140000E+00
   *    '''),
   *    'N': gto.basis.parse('''
   *    N    S
   *    0.2227660584e+01       0.1543289673E+00
   *    0.4057711562e+00       0.5353281423E+00
   *    0.1098175104e+00       0.4446345422E+00
   *    N    SP
   *    0.3688490000E+01      -0.192690000E+00     0.222810000E+00
   *    0.7753400000E+00       0.618880000E+00     0.560320000E+00
   *    0.2049800000E+00       0.549260000E+00     0.438590000E+00
   *    '''),
   *    'H': gto.basis.parse('''
   *    H    S
   *    0.2227660584e+01       0.1543289673E+00
   *    0.4057711562e+00       0.5353281423E+00
   *    0.1098175104e+00       0.4446345422E+00
   *    ''')
   *}
   *
   *for i in range(3):
   *    for j in range(1,4):
   *        expO = 7.58579774 if i == 0 else 1.55214516
   *        bas['O'][i][j][0] = bas['O'][i][j][0]*(expO**2)
   *        expN = 6.939806 if i == 0 else 1.33175233
   *        bas['N'][i][j][0] = bas['N'][i][j][0]*(expN**2)
   *        expC = 5.16802668 if i == 0 else 1.42036892
   *        bas['C'][i][j][0] = bas['C'][i][j][0]*(expC**2)
   *for j in range(1,4):
   *    bas['H'][0][j][0] = bas['H'][0][j][0]*(1.47386481**2)
   *
   *mol.build(
   *atom = '''
   *          O        0.00000       0.00000        0.00000
   *          C        1.00000       0.00000        0.00000
   *          N        2.00000       0.00000        0.00000
   *          H        1.00000       1.00000        0.00000
   *       ''',
   *basis = bas
   *)
   *
   *ovlp = mol.intor("int1e_ovlp")
   *print("Core O:", ovlp[0,:])
   *
   *print("Core C:", ovlp[5,:])
   *
   *print("Core N:", ovlp[10,:])
   *
   *print("Core H:", ovlp[15,:])
   *
   */
  for (int i = 0; i < 13; ++i) {
    EXPECT_NEAR(coreOverlapO(i), pyscfO(i), 1e-8);
    EXPECT_NEAR(coreOverlapC(i), pyscfC(i), 1e-8);
    EXPECT_NEAR(coreOverlapN(i), pyscfN(i), 1e-8);
  }
}

TEST_F(AnECPTest, CoreOverlapCalculatedCorrectly) {
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::C};
  Utils::PositionCollection pos(2, 3);
  pos << 0, 0, 0, 0, 0, 1.2 * Utils::Constants::bohr_per_angstrom;
  OmxElementParameters elParams2;
  auto cParams = ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om2());
  cParams.gtos.s =
      Utils::GtoExpansion{0, {{0, 2.9412494, -0.09996723}, {0, 0.6834831, 0.39951283}, {0, 0.2222899, 0.70011547}}};
  cParams.gtos.p =
      Utils::GtoExpansion{0, {{1, 2.9412494, 0.15591627}, {1, 0.6834831, 0.60768372}, {1, 0.2222899, 0.39195739}}};
  elParams2.insert(Utils::ElementType::C, cParams);
  gtos = {};
  gtos.push_back(elParams2.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParams2.get(Utils::ElementType::C).gtos);
  auto bSet = BasisSetController({els, pos}, gtos, "om2-3g");

  auto coreOrbital = Utils::SlaterToGaussian::getGTOExpansion(3, 1, 0, 5.67);
  Eigen::MatrixXd coreOverlap = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(1, bSet, coreOrbital);
  // From W. Weber, "Ein neues emiempirisches NDDO-Verfahren mit Orthogonalisierungskorrekturen", 1996, Seite 80,
  // Tabelle 6.9
  EXPECT_THAT(coreOverlap(0), DoubleNear(0.0598, 1e-4));

  // With scaled exponent
  double scaling = 1.4 * 1.4;
  gtos = {};
  gtos.push_back(
      {Utils::GtoExpansion{
           0, {{0, 2.9412494 * scaling, -0.09996723}, {0, 0.6834831 * scaling, 0.39951283}, {0, 0.2222899 * scaling, 0.70011547}}},
       Utils::GtoExpansion{
           0, {{1, 2.9412494 * scaling, 0.15591627}, {1, 0.6834831 * scaling, 0.60768372}, {1, 0.2222899 * scaling, 0.39195739}}},
       {}});
  gtos.push_back(
      {Utils::GtoExpansion{
           0, {{0, 2.9412494 * scaling, -0.09996723}, {0, 0.6834831 * scaling, 0.39951283}, {0, 0.2222899 * scaling, 0.70011547}}},
       Utils::GtoExpansion{
           0, {{1, 2.9412494 * scaling, 0.15591627}, {1, 0.6834831 * scaling, 0.60768372}, {1, 0.2222899 * scaling, 0.39195739}}},
       {}});
  bSet = BasisSetController({els, pos}, gtos, "om2-3g");
  coreOrbital = Utils::SlaterToGaussian::getGTOExpansion(3, 1, 0, 1.4 * 5.67);
  coreOverlap = Libint::computeOverlapWithCoreOrbitalOmx<Utils::DerivativeOrder::Zero>(1, bSet, coreOrbital);
  // From W. Weber, "Ein neues emiempirisches NDDO-Verfahren mit Orthogonalisierungskorrekturen", 1996, Seite 80,
  // Tabelle 6.9
  EXPECT_THAT(coreOverlap(0), DoubleNear(0.0185, 1e-4));
}

TEST_F(AnECPTest, CoreOrbitalsGeneratedCorrectly) {
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::C};
  Utils::PositionCollection pos(2, 3);
  pos << 0, 0, 0, 0, 0, 1.2 * Utils::Constants::bohr_per_angstrom;
  OmxElementParameters elParams2;
  auto cParams = ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om2());
  cParams.parameters.pseudopotential.exponent = 5.168;

  cParams.gtos.s =
      Utils::GtoExpansion{0, {{0, 2.9412494, -0.09996723}, {0, 0.6834831, 0.39951283}, {0, 0.2222899, 0.70011547}}};
  cParams.gtos.p =
      Utils::GtoExpansion{0, {{1, 2.9412494, 0.15591627}, {1, 0.6834831, 0.60768372}, {1, 0.2222899, 0.39195739}}};
  elParams2.insert(Utils::ElementType::C, cParams);
  gtos = {};
  gtos.push_back(elParams2.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParams2.get(Utils::ElementType::C).gtos);

  auto ecpEval = EffectiveCorePotentialEvaluator(els, pos, elParams2);

  auto coreOrbital = Utils::SlaterToGaussian::getGTOExpansion(3, 1, 0, cParams.parameters.pseudopotential.exponent);
  auto generatedCoreOrbital = ecpEval.generateCoreOrbital(0);
  EXPECT_DOUBLE_EQ(coreOrbital.gtfs[0].exponent, generatedCoreOrbital.gtfs[0].exponent);
  EXPECT_DOUBLE_EQ(coreOrbital.gtfs[0].coefficient, generatedCoreOrbital.gtfs[0].coefficient);
  EXPECT_DOUBLE_EQ(coreOrbital.gtfs[1].exponent, generatedCoreOrbital.gtfs[1].exponent);
  EXPECT_DOUBLE_EQ(coreOrbital.gtfs[1].coefficient, generatedCoreOrbital.gtfs[1].coefficient);
  EXPECT_DOUBLE_EQ(coreOrbital.gtfs[2].exponent, generatedCoreOrbital.gtfs[2].exponent);
  EXPECT_DOUBLE_EQ(coreOrbital.gtfs[2].coefficient, generatedCoreOrbital.gtfs[2].coefficient);
}

TEST_F(AnECPTest, CanCalculateOverlapSS) {
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::C};
  auto paramsOm1 = om2();
  paramsOm1.atomic[6].oneCenterEnergy.s = -50.15945;
  paramsOm1.atomic[6].oneCenterEnergy.p = -38.76257;
  // paramsOm1.atomic[6].exponent = 1.13551;
  paramsOm1.atomic[6].exponent = 1.0;
  paramsOm1.atomic[6].resonance.beta.s = -7.58632;
  paramsOm1.atomic[6].resonance.beta.pSigma = -4.49894;
  paramsOm1.atomic[6].resonance.beta.pPi = -5.91210;
  paramsOm1.atomic[6].resonance.alpha.s = 0.09325;
  paramsOm1.atomic[6].resonance.alpha.pSigma = 0.05399;
  paramsOm1.atomic[6].resonance.alpha.pPi = 0.10477;
  paramsOm1.atomic[6].resonance.alphaXH = paramsOm1.atomic[6].resonance.alpha;
  paramsOm1.atomic[6].resonance.betaXH = paramsOm1.atomic[6].resonance.beta;
  paramsOm1.atomic[6].orthogonalization.f1 = 0.50384;
  paramsOm1.atomic[6].orthogonalization.f2 = 0.66944;
  OmxElementParameters elParamsOm1;
  elParamsOm1.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, paramsOm1));

  Utils::PositionCollection pos(2, 3);
  pos.row(0) = Utils::Position{0, 0, 0};
  pos.row(1) = Utils::Position{0, 0, 1.25};
  pos *= Utils::Constants::bohr_per_angstrom;
  gtos = {};

  gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");
  Eigen::MatrixXd ovlpMatrix =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).get0();
  // Needed to read out from a graph in a paper, I am not so precise...
  // s,s
  EXPECT_THAT(ovlpMatrix(0, 4), DoubleNear(0.5, 2e-1));
  // pi,pi
  EXPECT_THAT(ovlpMatrix(2, 6), DoubleNear(0.375, 2e-1));
  // sigma,sigma
  EXPECT_THAT(ovlpMatrix(3, 7), DoubleNear(-0.2, 2e-1));
  // s,sigma
  EXPECT_THAT(ovlpMatrix(0, 7), DoubleNear(-0.5, 2e-1));
  //  Print this to see if it overlaps with plot in paper
  //  M. Kolb, W. Thiel, Beyond the MNDO Model: Methodical Considerations and Numerical Results, j comp chem, 1993
  //  figure 2, bottom right (eV are wrong unit obviously...)
  //  auto printOverlap = [&](double distanceInAngstrom) {
  //    pos.row(1) = Utils::Position{0, 0, distanceInAngstrom * Utils::Constants::bohr_per_angstrom};
  //    basisSet.modifyPositions(pos);
  //    std::cout
  //        << distanceInAngstrom << "    "
  //        << -10 * Libint::getInstance().compute1BodyIntegrals<Operator::overlap,
  //        Utils::DerivativeOrder::Zero>(basisSet).get0()(
  //                     0, 4)
  //        << "    "
  //        << -10 * Libint::getInstance().compute1BodyIntegrals<Operator::overlap,
  //        Utils::DerivativeOrder::Zero>(basisSet).get0()(
  //                     0, 7)
  //        << "    "
  //        << -10 * Libint::getInstance().compute1BodyIntegrals<Operator::overlap,
  //        Utils::DerivativeOrder::Zero>(basisSet).get0()(
  //                     3, 7)
  //        << "    "
  //        << -10 * Libint::getInstance().compute1BodyIntegrals<Operator::overlap,
  //        Utils::DerivativeOrder::Zero>(basisSet).get0()(
  //                     2, 6)
  //        << std::endl;
  //  };
  //  for (int i = 1; i < 30; ++i) {
  //    printOverlap(0.1 * i);
  //  }
}

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
