/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "Sparrow/Implementations/Nddo/Mndo/Wrapper/MNDOMethodWrapper.h"
#include "Utils/DataStructures/AtomsOrbitalsIndexes.h"
#include <Integrals/BasisSetController.h>
#include <Integrals/TwoBodiesIntegrals/NddoEriSaverDigester.h>
#include <Integrals/TwoBodiesIntegrals/NddoTwoBodiesIntegralsEvaluator.h>
#include <Integrals/TwoBodiesIntegrals/Orbitals.h>
#include <Sparrow/Implementations/Nddo/Mndo/Wrapper/MNDOMethodWrapper.h>
#include <Sparrow/Implementations/Nddo/NddoSettings.h>
#include <Sparrow/Implementations/Nddo/Omx/KlopmanFinalizer.h>
#include <Sparrow/Implementations/Nddo/Omx/MatrixTransformer.h>
#include <Sparrow/Implementations/Nddo/Omx/OmxElementParameters.h>
#include <Sparrow/Implementations/Nddo/Omx/OmxInitializer.h>
#include <Sparrow/Implementations/Nddo/Omx/OmxMethod.h>
#include <Sparrow/Implementations/Nddo/Omx/OmxOneCenterTwoElectronEvaluator.h>
#include <Sparrow/Implementations/Nddo/Omx/OmxOneElectronMatrix.h>
#include <Sparrow/Implementations/Nddo/Omx/OmxRepulsion.h>
#include <Sparrow/Implementations/Nddo/Omx/ParametersProcessor.h>
#include <Sparrow/Implementations/Nddo/Omx/TwoElectronMatrixConstructor.h>
#include <Sparrow/Implementations/Nddo/TimeDependent/UnoCAS.h>
#include <Sparrow/Implementations/Nddo/Utils/NddoExcitedStatesAssembler.h>
#include <Utils/Constants.h>
#include <Utils/DataStructures/DensityMatrix.h>
#include <Utils/IO/ChemicalFileFormats/XyzStreamHandler.h>
#include <Utils/Math/DerivOrderEnum.h>
#include <Utils/UniversalSettings/SettingsNames.h>
#include <gmock/gmock.h>

namespace Scine {
namespace Sparrow {
namespace nddo {

using namespace testing;
using namespace Integrals;

class AnOMxCalculation : public Test {
 public:
  nddo::ParametersOmx om2Params;
  OmxElementParameters elParams;

  std::vector<Utils::AtomicGtos> gtos;
  Utils::AtomCollection structure;
  BasisSetController basisSet;

  std::unique_ptr<nddo::OmxOneElectronMatrix> oneElMatrix_;

 protected:
  void SetUp() final {
    om2Params = om2();
    om2Params.atomic[1].exponent = om2Params.atomic[1].exponent * 1.24;

    Utils::ElementTypeCollection elements{Utils::ElementType::H, Utils::ElementType::F};
    Utils::PositionCollection positions(2, 3);
    positions << 0, 0, 0, 1, 0, 0;

    structure = Utils::AtomCollection{elements, positions};
    elParams.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, om2Params));
    elParams.insert(Utils::ElementType::F, ParametersProcessor::processAtomicParameters(Utils::ElementType::F, om2Params));

    gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
    gtos.push_back(elParams.get(Utils::ElementType::F).gtos);
    basisSet = BasisSetController(structure, gtos, "om2-3g");
  }
};

TEST_F(AnOMxCalculation, KlopmanFinalizerWorks) {
  EriType arbitraryEris;
  Eigen::MatrixXd HHBlock = Eigen::MatrixXd::Random(1, 1);
  Eigen::MatrixXd FFBlock = Eigen::MatrixXd::Random(10, 10);
  Eigen::MatrixXd HFBlock = Eigen::MatrixXd::Random(10, 1);
  arbitraryEris[{0, 0}].get0() = HHBlock;
  arbitraryEris[{1, 1}].get0() = FFBlock;
  arbitraryEris[{1, 0}].get0() = HFBlock;
  KlopmanScalingFactors klopman;
  // (1/2gamma^H + 1/2gamma^F)^2 = 3 (arbitrary)
  klopman.set({1, 9}, 3);

  KlopmanFinalizer finalizer(klopman, structure.getElements(), structure.getPositions());
  finalizer.finalize<Utils::DerivativeOrder::Zero>(arbitraryEris);

  EXPECT_EQ(arbitraryEris.at({0, 0}).get0(), HHBlock);
  EXPECT_EQ(arbitraryEris.at({1, 1}).get0(), FFBlock);
  EXPECT_EQ(arbitraryEris.at({1, 0}).get0(), 0.5 / HFBlock(0, 0) * HFBlock);
}

TEST_F(AnOMxCalculation, CanEvaluateNddoEri) {
  KlopmanFinalizer finalizer(ParametersProcessor::getKlopmanScalingFactors(om2Params), structure.getElements(),
                             structure.getPositions());

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(structure.getElements(), elParams));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer.finalize<Utils::DerivativeOrder::Zero>(eris);

  Eigen::MatrixXd MNDOHHBlock(1, 1);
  Eigen::MatrixXd MNDOFFBlock(10, 10);
  MNDOHHBlock(0, 0) = 0.472155;
  MNDOFFBlock << 0.621799, 0, 0, 0, 0.633926, 0, 0, 0.633926, 0, 0.633926, 0, 0.177499, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0.177499, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.177499, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.614081, 0, 0, 0.547932, 0,
      0.547932, 0, 0, 0, 0, 0, 0.0330744, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0330744, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0.614081, 0, 0.547932, 0, 0, 0, 0, 0, 0, 0, 0, 0.0330744, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.614081;
  MNDOFFBlock = MNDOFFBlock.selfadjointView<Eigen::Upper>();
  // One center blocks are correct
  for (const auto& block : eris) {
    int atom_i = block.first.first;
    int atom_j = block.first.second;
    if (atom_i == 0 && atom_j == 0) {
      EXPECT_THAT(block.second.get0()(0, 0), DoubleNear(MNDOHHBlock(0, 0), 1e-6));
    }
    else if (atom_i == 1 && atom_j == 1) {
      for (int row = 0; row < 10; ++row) {
        for (int col = 0; col < 10; ++col) {
          EXPECT_THAT(block.second.get0()(row, col), DoubleNear(MNDOFFBlock(row, col), 1e-6));
        }
      }
    }
  }
}

TEST_F(AnOMxCalculation, OneCenterTwoElectronMatrixCorrectlyConstruced) {
  Eigen::MatrixXd refMNDODensity(5, 5);
  Eigen::MatrixXd ref1c2eTwoElectronMatrix(5, 5);
  Eigen::MatrixXd ref2c2eTwoElectronMatrix(5, 5);

  refMNDODensity << 0.7086523907, 0.3595551905, -0.8864742724, 0, 0, 0.3595551905, 1.899887579, 0.2468246532, 0, 0,
      -0.8864742724, 0.2468246532, 1.39146003, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 2;
  ref1c2eTwoElectronMatrix << 0.1672969894, 0, 0, 0, 0, 0, 3.529969344, 0, 0, 0, 0, -0.01251748151, 3.588589009, 0, 0,
      0, 0, 0, 3.452060259, 0, 0, 0, 0, 0, 3.452060259;
  ref1c2eTwoElectronMatrix = ref1c2eTwoElectronMatrix.selfadjointView<Eigen::Lower>();

  ref2c2eTwoElectronMatrix << 3.32284, 0, 0, 0, 0, -0.134933, 0.335139, 0, 0, 0, 0.240707, -0.0797999, 0.352478, 0, 0,
      0, 0, 0, 0.316736, 0, 0, 0, 0, 0, 0.316736;

  ref2c2eTwoElectronMatrix = ref2c2eTwoElectronMatrix.selfadjointView<Eigen::Lower>();
  Utils::DensityMatrix refDenMat;
  refDenMat.setDensity(std::move(refMNDODensity), 8);

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(structure.getElements(), elParams));
  auto finalizer = std::make_unique<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(om2Params),
                                                      structure.getElements(), structure.getPositions());

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);

  Utils::AtomsOrbitalsIndexes aoInfos(2);
  aoInfos.addAtom(1);
  aoInfos.addAtom(4);
  MatrixTransformer transformer(eris, refDenMat, aoInfos);
  TwoElectronMatrixConstructor constructor(refDenMat);
  constructor.contractCoulomb(transformer, aoInfos);
  // FF exchange
  constructor.contractExchange(transformer, 0, 0, 1, 0, 0, 1);
  // HH exchange
  constructor.contractExchange(transformer, 1, 1, 4, 1, 1, 4);
  // HF exchange
  constructor.contractExchange(transformer, 1, 1, 4, 0, 0, 1);

  Eigen::MatrixXd actualMatrix = constructor.getResult().restrictedMatrix().selfadjointView<Eigen::Upper>();

  for (int row = 0; row < ref1c2eTwoElectronMatrix.rows(); ++row) {
    for (int col = 0; col < ref1c2eTwoElectronMatrix.cols(); ++col) {
      EXPECT_THAT(actualMatrix(row, col),
                  DoubleNear(ref2c2eTwoElectronMatrix(row, col) + ref1c2eTwoElectronMatrix(row, col), 1e-5));
    }
  }
}

TEST_F(AnOMxCalculation, CanCalculateOneElectronMatrix) {
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::H, Utils::ElementType::H,
                                      Utils::ElementType::H, Utils::ElementType::H};
  Utils::PositionCollection pos(5, 3);
  pos.row(0) = Utils::Position{0, 0, 0};
  pos.row(1) = Utils::Position{0, 0, 1.087};
  pos.row(2) = Utils::Position{1.0246513, 0, -0.362848056};
  pos.row(3) = Utils::Position{-0.51232565, 0.887374056, -0.362848056};
  pos.row(4) = Utils::Position{-0.51232565, -0.887374056, -0.362848056};

  pos *= Utils::Constants::bohr_per_angstrom * 0.999;
  gtos = {};

  auto params = om1();
  OmxElementParameters elParamsOm1;
  elParamsOm1.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, params));
  elParamsOm1.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, params));
  gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");
  auto finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(om1()), els, pos);

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParamsOm1));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);

  nddo::OmxOneElectronMatrix oem(els, pos, elParamsOm1, *finalizer, eris, basisSet);

  Eigen::MatrixXd ovlp =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).get0();
  auto log = Core::Log::silent();
  oem.calculate(ovlp, log);
  // Large deviations as the structure is not known...
  EXPECT_NEAR(oem()(0, 0) * Utils::Constants::ev_per_hartree, -86.79, 5e-2);
  EXPECT_NEAR(oem()(1, 1) * Utils::Constants::ev_per_hartree, -75.83, 5e-2);
  // Need to subtract pseudopotential part: -68.61 - 1.27 = -69.88
  EXPECT_NEAR(oem()(4, 4) * Utils::Constants::ev_per_hartree, -69.88, 5e-2);
}

TEST_F(AnOMxCalculation, DISABLED_CanCalculateKlopmanScalingFactorsDerivatives) {
  // this can be tested once 2-body integral derivatives are available
}
TEST_F(AnOMxCalculation, CanCalculateRepulsionEthylene) {
  std::stringstream ssethylene{"6\n\n"
                               "C 0  0  0\n"
                               "C 1.327  0  0\n"
                               "H -0.57496240169  0.92013218433  0\n"
                               "H -0.57496240169  -0.92013218433  0\n"
                               "H 1.90196240169  0.92013218433  0\n"
                               "H 1.90196240169  -0.92013218433  0\n"};
  auto structure = Utils::XyzStreamHandler::read(ssethylene);
  Utils::PositionCollection pos = structure.getPositions();
  auto els = structure.getElements();

  gtos = {};
  OmxElementParameters elParams;
  elParams.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om2()));
  elParams.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, om2()));
  gtos.push_back(elParams.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({els, pos}, gtos);

  auto finalizer = KlopmanFinalizer(ParametersProcessor::getKlopmanScalingFactors(om2()), els, pos);

  KlopmanOhnoShower shower;
  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParams));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer.finalize<Utils::DerivativeOrder::Zero>(eris);
  auto rep = std::make_shared<OmxRepulsion>(els, pos);

  shower.subscribe(std::shared_ptr<OmxRepulsion>(rep));
  double energy = 0.0;
  enum KlopmanOhnoType { Yes, No };
  auto pairEnergy = [&](int a, int b, KlopmanOhnoType type) -> double {
    Integrals::DiatomicKey index{a, b};
    canonicalize(index);
    double distance = (pos.row(a) - pos.row(b)).norm();
    double koCoef = type == Yes ? finalizer.get<Utils::DerivativeOrder::Zero>(index) : 1.0;
    return koCoef * Utils::ElementInfo::valElectrons(els[a]) * Utils::ElementInfo::valElectrons(els[b]) / distance;
  };
  for (int el_a = 0; el_a < 6; ++el_a) {
    for (int el_b = 0; el_b < el_a; ++el_b) {
      energy += pairEnergy(el_a, el_b, KlopmanOhnoType::No);
    }
  }

  rep->calculateRepulsion(Utils::DerivativeOrder::Zero);
  double repEn = rep->getRepulsionEnergy();
  EXPECT_NEAR(repEn, energy, 1e-6);

  shower.notify(std::make_shared<KlopmanFinalizer>(finalizer));
  energy = 0.0;
  for (int el_a = 0; el_a < 6; ++el_a) {
    for (int el_b = 0; el_b < el_a; ++el_b) {
      energy += pairEnergy(el_a, el_b, KlopmanOhnoType::Yes);
    }
  }
  rep->calculateRepulsion(Utils::DerivativeOrder::Zero);
  EXPECT_NEAR(rep->getRepulsionEnergy(), energy, 1e-6);
}

TEST_F(AnOMxCalculation, CanCalculateRepulsionMethane) {
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::H, Utils::ElementType::H,
                                      Utils::ElementType::H, Utils::ElementType::H};
  Utils::PositionCollection pos(5, 3);
  pos.row(0) = Utils::Position{0, 0, 0};
  pos.row(1) = Utils::Position{0, 0, 1.087};
  pos.row(2) = Utils::Position{1.0246513, 0, -0.362848056};
  pos.row(3) = Utils::Position{-0.51232565, 0.887374056, -0.362848056};
  pos.row(4) = Utils::Position{-0.51232565, -0.887374056, -0.362848056};

  pos *= Utils::Constants::bohr_per_angstrom;
  gtos = {};
  OmxElementParameters elParams;
  elParams.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om2()));
  elParams.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, om2()));
  gtos.push_back(elParams.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");

  auto finalizer = KlopmanFinalizer(ParametersProcessor::getKlopmanScalingFactors(om2()), els, pos);

  KlopmanOhnoShower shower;
  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParams));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer.finalize<Utils::DerivativeOrder::Zero>(eris);
  auto rep = std::make_shared<OmxRepulsion>(els, pos);

  shower.subscribe(std::shared_ptr<OmxRepulsion>(rep));
  double energy = 0.0;
  enum KlopmanOhnoType { Yes, No };
  auto pairEnergy = [&](int a, int b, KlopmanOhnoType type) -> double {
    Integrals::DiatomicKey index{a, b};
    canonicalize(index);
    double distance = (pos.row(a) - pos.row(b)).norm();
    double koCoef = type == Yes ? finalizer.get<Utils::DerivativeOrder::Zero>(index) : 1.0;
    return koCoef * Utils::ElementInfo::valElectrons(els[a]) * Utils::ElementInfo::valElectrons(els[b]) / distance;
  };
  for (int el_a = 0; el_a < 5; ++el_a) {
    for (int el_b = 0; el_b < el_a; ++el_b) {
      energy += pairEnergy(el_a, el_b, KlopmanOhnoType::No);
    }
  }

  rep->calculateRepulsion(Utils::DerivativeOrder::Zero);
  double repEn = rep->getRepulsionEnergy();
  EXPECT_NEAR(repEn, energy, 1e-6);

  shower.notify(std::make_shared<KlopmanFinalizer>(finalizer));
  energy = 0.0;
  for (int el_a = 0; el_a < 5; ++el_a) {
    for (int el_b = 0; el_b < el_a; ++el_b) {
      energy += pairEnergy(el_a, el_b, KlopmanOhnoType::Yes);
    }
  }
  rep->calculateRepulsion(Utils::DerivativeOrder::Zero);
  EXPECT_NEAR(rep->getRepulsionEnergy(), energy, 1e-6);
}

TEST_F(AnOMxCalculation, CanDoSinglePointOm2Hydrogen) {
  Om2Method om2;
  std::stringstream sshydrogen{"2\n\n"
                               "H 0  0  0\n"
                               "H 0  0  0.672\n"};
  om2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om2.setStructure(Utils::XyzStreamHandler::read(sshydrogen));
  om2.calculate("");
  EXPECT_NEAR(-om2.getSingleParticleEnergies().getRestrictedLevelEnergy(0) * Utils::Constants::ev_per_hartree, 15.68, 1e-2);
}

TEST_F(AnOMxCalculation, CanDoSinglePointOm2Nitrogen) {
  Om2Method om2;
  std::stringstream ssnitrogen{"2\n\n"
                               "N 0  0  0\n"
                               "N 0  0  1.098\n"};
  om2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  // om2.settings().modifyDouble(Utils::SettingsNames::selfConsistenceCriterion, 1e-5);
  om2.setStructure(Utils::XyzStreamHandler::read(ssnitrogen));
  om2.calculate("");
  for (auto en : om2.getSingleParticleEnergies().getRestrictedEnergies()) {
    std::cout << en * Utils::Constants::ev_per_hartree << std::endl;
  }
  EXPECT_NEAR(-om2.getSingleParticleEnergies().getRestrictedLevelEnergy(4) * Utils::Constants::ev_per_hartree, 15.46, 1e-2);
  EXPECT_NEAR(-om2.getSingleParticleEnergies().getRestrictedLevelEnergy(3) * Utils::Constants::ev_per_hartree, 16.80, 1e-2);
  EXPECT_NEAR(-om2.getSingleParticleEnergies().getRestrictedLevelEnergy(2) * Utils::Constants::ev_per_hartree, 16.80, 1e-2);
  EXPECT_NEAR(-om2.getSingleParticleEnergies().getRestrictedLevelEnergy(1) * Utils::Constants::ev_per_hartree, 21.88, 1e-2);
}

TEST_F(AnOMxCalculation, CanDoSinglePointOm3Nitrogen) {
  Om3Method om3;
  std::stringstream ssnitrogen{"2\n\n"
                               "N 0  0  0\n"
                               "N 0  0  1.170 \n"};
  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om3.setStructure(Utils::XyzStreamHandler::read(ssnitrogen));
  om3.calculate("");
  for (auto en : om3.getSingleParticleEnergies().getRestrictedEnergies()) {
    std::cout << en * Utils::Constants::ev_per_hartree << std::endl;
  }
  EXPECT_NEAR(-om3.getSingleParticleEnergies().getRestrictedLevelEnergy(4) * Utils::Constants::ev_per_hartree, 14.51, 1e-2);
  EXPECT_NEAR(-om3.getSingleParticleEnergies().getRestrictedLevelEnergy(3) * Utils::Constants::ev_per_hartree, 16.40, 1e-2);
  EXPECT_NEAR(-om3.getSingleParticleEnergies().getRestrictedLevelEnergy(2) * Utils::Constants::ev_per_hartree, 16.40, 1e-2);
  EXPECT_NEAR(-om3.getSingleParticleEnergies().getRestrictedLevelEnergy(1) * Utils::Constants::ev_per_hartree, 22.82, 1e-2);
}

TEST_F(AnOMxCalculation, CanDoSinglePointOm2CO) {
  Om2Method om2;
  std::stringstream ssCO{"2\n\n"
                         "C 0  0  0\n"
                         "O 0  0  1.171 \n"};
  om2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om2.setStructure(Utils::XyzStreamHandler::read(ssCO));
  om2.calculate("");
  for (auto en : om2.getSingleParticleEnergies().getRestrictedEnergies()) {
    std::cout << en * Utils::Constants::ev_per_hartree << std::endl;
  }
  EXPECT_NEAR(-om2.getSingleParticleEnergies().getRestrictedLevelEnergy(4) * Utils::Constants::ev_per_hartree, 13.60, 1e-2);
}

TEST_F(AnOMxCalculation, CanDoSinglePointOm2CO2) {
  Om2Method om2;
  std::stringstream ssCO2{"3\n\n"
                          "C 0  0  0\n"
                          "O 0  0  -1.199\n"
                          "O 0  0  1.199 \n"};
  om2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om2.setStructure(Utils::XyzStreamHandler::read(ssCO2));
  om2.calculate("");
  for (auto en : om2.getSingleParticleEnergies().getRestrictedEnergies()) {
    std::cout << en * Utils::Constants::ev_per_hartree << std::endl;
  }
  std::cout << om2.getMolecularOrbitals().restrictedMatrix() << std::endl;
  EXPECT_NEAR(-om2.getSingleParticleEnergies().getRestrictedLevelEnergy(4) * Utils::Constants::ev_per_hartree, 18.00, 1e-2);
}
TEST_F(AnOMxCalculation, CanDoSinglePointOm2Methane) {
  Om2Method om2;
  std::stringstream ssmethane{"5\n\n"
                              "C 0  0  0\n"
                              "H 0  0  1.087\n"
                              "H 1.0246513  0  -0.362848056\n"
                              "H -0.51232565  0.887374056  -0.362848056\n"
                              "H -0.51232565  -0.887374056  -0.362848056\n"};
  om2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om2.setStructure(Utils::XyzStreamHandler::read(ssmethane));

  auto result = om2.calculate("");
  EXPECT_NEAR(-om2.getSingleParticleEnergies().getRestrictedLevelEnergy(3) * Utils::Constants::ev_per_hartree, 13.64, 1e-2);
}

TEST_F(AnOMxCalculation, CanDoSinglePointOm3) {
  Om3Method om3;

  std::stringstream ssmethane{"5\n\n"
                              "C 0  0  0\n"
                              "H 0  0  1.087\n"
                              "H 1.0246513  0  -0.362848056\n"
                              "H -0.51232565  0.887374056  -0.362848056\n"
                              "H -0.51232565  -0.887374056  -0.362848056\n"};
  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-12);
  om3.settings().modifyDouble(Utils::SettingsNames::selfConsistenceCriterion, 1e-12);
  om3.setStructure(Utils::XyzStreamHandler::read(ssmethane));

  om3.calculate("");
  EXPECT_NEAR(-om3.getSingleParticleEnergies().getRestrictedLevelEnergy(3) * Utils::Constants::ev_per_hartree, 14.18, 1e-2);
}


TEST_F(AnOMxCalculation, SameRestrictedAndUnrestrictedEnergyOm3) {
  Om3Method om3;

  std::stringstream ssmethane{"5\n\n"
                              "C 0  0  0\n"
                              "H 0  0  1.087\n"
                              "H 1.0246513  0  -0.362848056\n"
                              "H -0.51232565  0.887374056  -0.362848056\n"
                              "H -0.51232565  -0.887374056  -0.362848056\n"};
  om3.setStructure(Utils::XyzStreamHandler::read(ssmethane));

  double energyRestricted = om3.calculate("").get<Utils::Property::Energy>();

  // Check that with unrestricted you get the same
  om3.settings().modifyString(Utils::SettingsNames::spinMode, "unrestricted");
  double energyUnrestricted = om3.calculate("").get<Utils::Property::Energy>();
  EXPECT_NEAR(energyRestricted, energyUnrestricted, 1e-6);
}

TEST_F(AnOMxCalculation, CanDOSinglePointOm2Acetylene) {
  Om2Method om2;

  std::stringstream acetylene{"4\n\n"
                              "C    0.00000   0.61114   0.00000\n"
                              "C    0.00000  -0.61114  -0.00000\n"
                              "H    0.00000   1.65239   0.00000\n"
                              "H    0.00000  -1.65239  -0.00000\n"};
  om2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om2.setStructure(Utils::XyzStreamHandler::read(acetylene));

  EXPECT_NEAR(om2.calculate("").get<Utils::Property::Energy>() * Utils::Constants::ev_per_hartree, -282.94213, 1e-2);
}

TEST_F(AnOMxCalculation, CanDOSinglePointOm2Biphenylene) {
  Om2Method om2;

  std::stringstream biphenyle{"20\n\n"
                              "C    0.00000   0.00000   0.00000\n"
                              "C    1.50478   0.00000   0.00000\n"
                              "C    0.00000   1.44173   0.00000\n"
                              "C    1.50478   1.44173   0.00000\n"
                              "C   -1.15145  -0.73566   0.00000\n"
                              "C    2.65622  -0.73566  -0.00000\n"
                              "C   -1.15145   2.17739  -0.00000\n"
                              "C    2.65622   2.17739   0.00000\n"
                              "C   -2.34400   0.02828   0.00000\n"
                              "C    3.84878   0.02828  -0.00000\n"
                              "C   -2.34400   1.41345  -0.00000\n"
                              "C    3.84878   1.41345   0.00000\n"
                              "H   -1.17436  -1.82562   0.00000\n"
                              "H    2.67914  -1.82562  -0.00000\n"
                              "H   -1.17436   3.26735  -0.00000\n"
                              "H    2.67914   3.26735   0.00000\n"
                              "H   -3.30186  -0.50666   0.00000\n"
                              "H    4.80664  -0.50666  -0.00000\n"
                              "H   -3.30186   1.94840  -0.00000\n"
                              "H    4.80664   1.94840   0.00000\n"};
  om2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om2.setStructure(Utils::XyzStreamHandler::read(biphenyle));

  EXPECT_NEAR(om2.calculate("").get<Utils::Property::Energy>() * Utils::Constants::ev_per_hartree, -1647.63341, 1e-2);
}

TEST_F(AnOMxCalculation, CanDoSinglePointOm2BiggerMolecule) {
  Om2Method om2;

  std::stringstream ss16epoxy10annulene{"19\n\n"
                                        "O    0.00000   0.00000   0.00000\n"
                                        "C   -0.88309   1.04917   0.00000\n"
                                        "C   -0.88309  -1.04917  -0.00000\n"
                                        "C   -1.30806   1.50084  -1.25050\n"
                                        "C   -1.30806  -1.50084  -1.25050\n"
                                        "C   -1.26289   1.52294   1.29122\n"
                                        "C   -1.26289  -1.52294   1.29122\n"
                                        "C   -1.22977   0.69695  -2.39066\n"
                                        "C   -1.22977  -0.69695  -2.39066\n"
                                        "C   -1.27309   0.71174   2.39547\n"
                                        "C   -1.27309  -0.71174   2.39547\n"
                                        "H   -1.71714   2.51170  -1.33682\n"
                                        "H   -1.71714  -2.51170  -1.33682\n"
                                        "H   -1.53201   2.58051   1.38390\n"
                                        "H   -1.53201  -2.58051   1.38390\n"
                                        "H   -1.33824   1.19614  -3.36633\n"
                                        "H   -1.33824  -1.19614  -3.36633\n"
                                        "H   -1.44114   1.18475   3.37655\n"
                                        "H   -1.44114  -1.18475   3.37655\n"};
  om2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om2.setStructure(Utils::XyzStreamHandler::read(ss16epoxy10annulene));

  EXPECT_NEAR(om2.calculate("").get<Utils::Property::Energy>() * Utils::Constants::ev_per_hartree, -1723.54629, 1e-2);

  std::stringstream ss4dimethylamino4cyanostilbene{"35\n\n"
                                                   "C    0.00000   0.00000   0.00000\n"
                                                   "C    1.38844   0.00000   0.00000\n"
                                                   "C    2.09946   1.21012   0.00000\n"
                                                   "C    1.40938   2.43078  -0.01609\n"
                                                   "C    0.02019   2.45199  -0.02033\n"
                                                   "C   -0.66475   1.23193  -0.00902\n"
                                                   "C    3.56490   1.17016   0.00727\n"
                                                   "C    4.30491   2.12867   0.59541\n"
                                                   "C    5.76841   2.10185   0.61337\n"
                                                   "C    6.44937   2.78246   1.63294\n"
                                                   "C    7.83424   2.77883   1.68691\n"
                                                   "C    8.58053   2.09607   0.70192\n"
                                                   "C    7.89316   1.42108  -0.32860\n"
                                                   "C    6.50622   1.42511  -0.36562\n"
                                                   "N    9.98440   2.06376   0.76691\n"
                                                   "C   10.66355   2.98861   1.63456\n"
                                                   "C   -2.10249   1.24359  -0.01249\n"
                                                   "N   -3.27446   1.25314  -0.01533\n"
                                                   "C   10.72472   1.56363  -0.36029\n"
                                                   "H    8.34210   3.30178   2.50079\n"
                                                   "H    5.87795   3.31530   2.40485\n"
                                                   "H    5.98822   0.90196  -1.17996\n"
                                                   "H    8.44701   0.88486  -1.10304\n"
                                                   "H    3.82986   2.97121   1.12089\n"
                                                   "H    4.02866   0.30063  -0.47435\n"
                                                   "H    1.93390  -0.94932   0.00986\n"
                                                   "H    1.96615   3.37300  -0.03696\n"
                                                   "H   -0.52988   3.39587  -0.03521\n"
                                                   "H   -0.56568  -0.93460   0.00698\n"
                                                   "H   11.81760   1.58965  -0.15537\n"
                                                   "H   10.45438   0.50148  -0.57514\n"
                                                   "H   10.53743   2.16433  -1.28974\n"
                                                   "H   11.76545   2.85430   1.56795\n"
                                                   "H   10.43322   4.05600   1.37584\n"
                                                   "H   10.37447   2.82190   2.70071\n"};
  om2.setStructure(Utils::XyzStreamHandler::read(ss4dimethylamino4cyanostilbene));

  EXPECT_NEAR(om2.calculate("").get<Utils::Property::Energy>() * Utils::Constants::ev_per_hartree, -2833.37117, 1e-2);
  std::stringstream ssacetaldehyde{"7\n\n"
                                   "C    0.00000   0.00000   0.00000\n"
                                   "C   1.50022   0.00000   0.00000\n"
                                   "O   2.12440   1.05019   0.00000\n"
                                   "H  -0.36723   1.02976   0.00000\n"
                                   "H  -0.36123  -0.52184   0.89354\n"
                                   "H  -0.36123  -0.52184  -0.89354\n"
                                   "H   2.01123  -1.01660   0.00000\n"};
  om2.setStructure(Utils::XyzStreamHandler::read(ssacetaldehyde));

  EXPECT_NEAR(om2.calculate("").get<Utils::Property::Energy>() * Utils::Constants::ev_per_hartree, -646.36576, 1e-2);
  for (auto charge : om2.getAtomicCharges()) {
    std::cout << charge << std::endl;
  }
}

TEST_F(AnOMxCalculation, CanDoSinglePointOm2Ethylene) {
  Om2Method om2;

  std::stringstream ssethylene{"6\n\n"
                               "C 0  0  0\n"
                               "C 1.339  0  0\n"
                               "H -0.56471727461  0.92879674835  0\n"
                               "H -0.56471727461  -0.92879674835  0\n"
                               "H 1.90371727461  0.92879674835  0\n"
                               "H 1.90371727461  -0.92879674835  0\n"};
  om2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om2.setStructure(Utils::XyzStreamHandler::read(ssethylene));

  auto result = om2.calculate("");
  EXPECT_NEAR(-om2.getSingleParticleEnergies().getRestrictedLevelEnergy(5) * Utils::Constants::ev_per_hartree, 10.66, 1e-2);
}

TEST_F(AnOMxCalculation, CanDoSinglePointOm3Ethylene) {
  Om3Method om3;

  std::stringstream ssethylene{"6\n\n"
                               "C 0  0  0\n"
                               "C 1.339  0  0\n"
                               "H -0.56471727461  0.92879674835  0\n"
                               "H -0.56471727461  -0.92879674835  0\n"
                               "H 1.90371727461  0.92879674835  0\n"
                               "H 1.90371727461  -0.92879674835  0\n"};
  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om3.setStructure(Utils::XyzStreamHandler::read(ssethylene));

  auto result = om3.calculate("");
  EXPECT_NEAR(-om3.getSingleParticleEnergies().getRestrictedLevelEnergy(5) * Utils::Constants::ev_per_hartree, 11.00, 1e-2);
}

TEST_F(AnOMxCalculation, ChargesOnEthyleneOm2) {
  Om2Method om2;

  std::stringstream ssethylene{"6\n\n"
                               "C 0  0  0\n"
                               "C 1.327  0  0\n"
                               "H -0.57496240169  0.92013218433  0\n"
                               "H -0.57496240169  -0.92013218433  0\n"
                               "H 1.90196240169  0.92013218433  0\n"
                               "H 1.90196240169  -0.92013218433  0\n"};
  om2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om2.setStructure(Utils::XyzStreamHandler::read(ssethylene));
  // With this geometry everything comes out correctly? Exact geometry is not given...
  //om2.modifyPositions(om2.getPositions() * 0.99);

  auto result = om2.calculate("");
  Eigen::VectorXd actualCharges(6);
  actualCharges << -0.43, -0.43, 0.22, 0.22, 0.22, 0.22;
  for (int i = 0; i < 6; ++i) {
    EXPECT_NEAR(om2.getAtomicCharge(i), actualCharges(i), 1e-2);
  }
}
TEST_F(AnOMxCalculation, ChargesOnEthyleneOm3) {
  Om3Method om3;

  std::stringstream ssethylene{"6\n\n"
                               "C 0  0  0\n"
                               "C 1.331  0  0\n"
                               "H -0.58512286011  0.92200609464  0\n"
                               "H -0.58512286011  -0.92200609464  0\n"
                               "H 1.91612286012  0.92200609464  0\n"
                               "H 1.91612286012  -0.92200609464  0\n"};
  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om3.setStructure(Utils::XyzStreamHandler::read(ssethylene));

  auto result = om3.calculate("");
  Eigen::VectorXd actualCharges(6);
  actualCharges << -0.39, -0.39, 0.19, 0.19, 0.19, 0.19;
  for (int i = 0; i < 6; ++i) {
    EXPECT_NEAR(om3.getAtomicCharge(i), actualCharges(i), 1e-2);
  }
}

TEST_F(AnOMxCalculation, ChargesOnAcetyleneOm3) {
  Om3Method om3;

  std::stringstream ssacetylene{"4\n\n"
                                "C        0.00000        0.00000        0.00000\n"
                                "C        1.22700        0.00000        0.00000\n"
                                "H       -1.05300        0.00000        0.00000\n"
                                "H        2.28000        0.00000        0.00000\n"};
  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om3.setStructure(Utils::XyzStreamHandler::read(ssacetylene));

  auto result = om3.calculate("");
  Eigen::VectorXd actualCharges(4);
  actualCharges << -0.35, -0.35, 0.35, 0.35;
  for (int i = 0; i < 4; ++i) {
    EXPECT_NEAR(om3.getAtomicCharge(i), actualCharges(i), 1e-2);
  }
}

TEST_F(AnOMxCalculation, CanDoSinglePointOdm2Ethylene) {
  Odm2Method odm2;

  std::stringstream ssethylene{"6\n\n"
                               "C 0  0  0\n"
                               "C 1.339  0  0\n"
                               "H -0.56471727461  0.92879674835  0\n"
                               "H -0.56471727461  -0.92879674835  0\n"
                               "H 1.90371727461  0.92879674835  0\n"
                               "H 1.90371727461  -0.92879674835  0\n"};
  odm2.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  odm2.setStructure(Utils::XyzStreamHandler::read(ssethylene));

  auto result = odm2.calculate("");
  EXPECT_NEAR(-odm2.getSingleParticleEnergies().getRestrictedLevelEnergy(5) * Utils::Constants::ev_per_hartree, 10.44, 1e-2);

  Eigen::VectorXd actualCharges(6);
  actualCharges << -0.42, -0.42, 0.21, 0.21, 0.21, 0.21;
  for (int i = 0; i < 6; ++i) {
    EXPECT_NEAR(odm2.getAtomicCharge(i), actualCharges(i), 1e-2);
  }
}

TEST_F(AnOMxCalculation, CanDoSinglePointOdm3Ethylene) {
  Odm3Method odm3;

  std::stringstream ssethylene{"6\n\n"
                               "C 0  0  0\n"
                               "C 1.339  0  0\n"
                               "H -0.56471727461  0.92879674835  0\n"
                               "H -0.56471727461  -0.92879674835  0\n"
                               "H 1.90371727461  0.92879674835  0\n"
                               "H 1.90371727461  -0.92879674835  0\n"};
  odm3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  odm3.setStructure(Utils::XyzStreamHandler::read(ssethylene));

  auto result = odm3.calculate("");
  EXPECT_NEAR(-odm3.getSingleParticleEnergies().getRestrictedLevelEnergy(5) * Utils::Constants::ev_per_hartree, 10.805, 1e-2);

  Eigen::VectorXd actualCharges(6);
  actualCharges << -0.386, -0.386, 0.193, 0.193, 0.193, 0.193;
  for (int i = 0; i < 6; ++i) {
    EXPECT_NEAR(odm3.getAtomicCharge(i), actualCharges(i), 1e-2);
  }
}

TEST_F(AnOMxCalculation, CanDOSinglePointAiqm1Acetylene) {
  Aiqm1Method aiqm1;

  std::stringstream acetylene{"4\n\n"
                              "C    0.00000   0.61114   0.00000\n"
                              "C    0.00000  -0.61114  -0.00000\n"
                              "H    0.00000   1.65239   0.00000\n"
                              "H    0.00000  -1.65239  -0.00000\n"};
  aiqm1.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  aiqm1.setStructure(Utils::XyzStreamHandler::read(acetylene));

  EXPECT_NEAR(aiqm1.calculate("").get<Utils::Property::Energy>(), -77.22987467, 1e-6);
}

TEST_F(AnOMxCalculation, CheckParallel) {
  Om3Method om3;
  Om3Method om32;

  std::stringstream ssacetylene{"4\n\n"
                                "C        0.00000        0.00000        0.00000\n"
                                "C        1.22700        0.00000        0.00000\n"
                                "H       -1.05300        0.00000        0.00000\n"
                                "H        2.28000        0.00000        0.00000\n"};
  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-16);
  om3.settings().modifyBool(Utils::SettingsNames::dumpIntegralsOption, true);
  om3.settings().modifyInt(Utils::SettingsNames::casOrbitalsAroundFermiOption, 2);
  om3.setStructure(Utils::XyzStreamHandler::read({ssacetylene}));
  om32.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-16);
  om32.settings().modifyBool(Utils::SettingsNames::dumpIntegralsOption, true);
  om32.settings().modifyInt(Utils::SettingsNames::casOrbitalsAroundFermiOption, 2);
  ssacetylene = std::stringstream{"4\n\n"
                                  "C        0.00000        0.00000        0.00000\n"
                                  "C        1.22700        0.00000        0.00000\n"
                                  "H       -1.05300        0.00000        0.00000\n"
                                  "H        2.28000        0.00000        0.00000\n"};
  om32.setStructure(Utils::XyzStreamHandler::read({ssacetylene}));

  // omp_set_num_threads(1);
  auto result = om3.calculate("");
  Eigen::MatrixXd denmat = om3.getDensityMatrix().restrictedMatrix();
  //  omp_set_num_threads(6);
  result = om32.calculate("");
  Eigen::MatrixXd denmat2 = om32.getDensityMatrix().restrictedMatrix();

  EXPECT_LE((denmat - denmat2).norm(), 1e-15);
}

TEST_F(AnOMxCalculation, CanDumpFciDumpFile) {
  Om3Method om3;

  std::stringstream ssacetylene{"4\n\n"
                                "C        0.00000        0.00000        0.00000\n"
                                "C        1.22700        0.00000        0.00000\n"
                                "H       -1.05300        0.00000        0.00000\n"
                                "H        2.28000        0.00000        0.00000\n"};
  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-16);
  om3.settings().modifyBool(Utils::SettingsNames::dumpIntegralsOption, true);
  om3.settings().modifyInt(Utils::SettingsNames::casOrbitalsAroundFermiOption, 2);
  om3.setStructure(Utils::XyzStreamHandler::read(ssacetylene));

  // 1 and 2 el integrals from direct transformation, core energy from 1 time calculation
  // NOTE: integrals smaller than 1e-16 are not written in the fci dump
  std::vector<std::tuple<double, int, int, int, int>> actual{
      {3.449612609662457e-01, 1, 1, 1, 1}, {1.209465225261747e-02, 2, 1, 2, 1},  {2.939926099290693e-01, 2, 2, 1, 1},
      {3.037057440399525e-01, 2, 2, 2, 2}, {-7.783498128417645e-01, 1, 1, 0, 0}, {-4.807205063750427e-01, 2, 2, 0, 0},
      {-8.989988208351228e+00, 0, 0, 0, 0}};
  auto result = om3.calculate("");
  std::ifstream fin("fcidump.fci");
  std::string buffer;
  std::getline(fin, buffer);
  std::getline(fin, buffer);
  std::getline(fin, buffer);
  std::getline(fin, buffer);
  int index = 0;
  while (std::getline(fin, buffer)) {
    std::stringstream line(buffer);
    double value = 0;
    int i, j, k, l;
    line >> value >> i >> j >> k >> l;
    EXPECT_NEAR(value, std::get<0>(actual[index]), 1e-11);
    EXPECT_EQ(i, std::get<1>(actual[index]));
    EXPECT_EQ(j, std::get<2>(actual[index]));
    EXPECT_EQ(k, std::get<3>(actual[index]));
    EXPECT_EQ(l, std::get<4>(actual[index]));
    ++index;
  }
  std::remove("fcidump.fci");
}

TEST_F(AnOMxCalculation, CanDoUnrestricted) {
  Om3Method om3;

  std::stringstream ssoctatetraene{"18\n\n"
                                   "C   1.58365383890105     -1.01977366210975      0.92247436419201\n"
                                   "C   2.73505565211047     -0.66433760211071      0.31609409662862\n"
                                   "C   2.92731453636819      0.55757081432472     -0.41912446207901\n"
                                   "C   4.09487910688951      0.90959532661804     -1.02658696645045\n"
                                   "C   4.29454579636180      2.12402891602624     -1.75622082690101\n"
                                   "C   5.46366087423553      2.46304575653813     -2.36483440286090\n"
                                   "C   5.73351533872446      3.68393787286481     -3.10266299681885\n"
                                   "C   5.00877216729167      4.81981286806325     -3.08649102860311\n"
                                   "H   1.49018519770001     -1.95889820258656      1.46723658060291\n"
                                   "H   0.70278428187383     -0.37473057075025      0.88705928752082\n"
                                   "H   3.59733787640963     -1.33824509181151      0.37446949668829\n"
                                   "H   2.07211748491966      1.24038464707200     -0.48506095831919\n"
                                   "H   4.94791876087749      0.22384327395107     -0.95684768561072\n"
                                   "H   3.43707710332457      2.80278445775672     -1.82741548983747\n"
                                   "H   6.28603135602525      1.74136832238316     -2.31335740163400\n"
                                   "H   6.63764759936980      3.66644514017172     -3.71987000388946\n"
                                   "H   4.11943257971621      4.93401458366712     -2.46350821160383\n"
                                   "H   5.29678044890082      5.67818314993179     -3.69298339102465\n"};

  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om3.setStructure(Utils::XyzStreamHandler::read(ssoctatetraene));
  om3.settings().modifyString(Utils::SettingsNames::mixer, "diis");
  om3.settings().modifyString(Utils::SettingsNames::spinMode, "restricted");
  double restrictedEnergy = om3.calculate("").get<Utils::Property::Energy>();
  om3.settings().modifyString(Utils::SettingsNames::spinMode, "unrestricted");
  double unrestrictedEnergy = om3.calculate("").get<Utils::Property::Energy>();

  EXPECT_NEAR(unrestrictedEnergy, restrictedEnergy, 1e-12);
}

TEST_F(AnOMxCalculation, CanDoUnoCI) {
  Om3Method om3;
  std::stringstream ssoctatetraene{"18\n\n"
                                   "C   1.58365383890105     -1.01977366210975      0.92247436419201\n"
                                   "C   2.73505565211047     -0.66433760211071      0.31609409662862\n"
                                   "C   2.92731453636819      0.55757081432472     -0.41912446207901\n"
                                   "C   4.09487910688951      0.90959532661804     -1.02658696645045\n"
                                   "C   4.29454579636180      2.12402891602624     -1.75622082690101\n"
                                   "C   5.46366087423553      2.46304575653813     -2.36483440286090\n"
                                   "C   5.73351533872446      3.68393787286481     -3.10266299681885\n"
                                   "C   5.00877216729167      4.81981286806325     -3.08649102860311\n"
                                   "H   1.49018519770001     -1.95889820258656      1.46723658060291\n"
                                   "H   0.70278428187383     -0.37473057075025      0.88705928752082\n"
                                   "H   3.59733787640963     -1.33824509181151      0.37446949668829\n"
                                   "H   2.07211748491966      1.24038464707200     -0.48506095831919\n"
                                   "H   4.94791876087749      0.22384327395107     -0.95684768561072\n"
                                   "H   3.43707710332457      2.80278445775672     -1.82741548983747\n"
                                   "H   6.28603135602525      1.74136832238316     -2.31335740163400\n"
                                   "H   6.63764759936980      3.66644514017172     -3.71987000388946\n"
                                   "H   4.11943257971621      4.93401458366712     -2.46350821160383\n"
                                   "H   5.29678044890082      5.67818314993179     -3.69298339102465\n"};

  OrbitalSteeringSettings settings;
  settings.modifyString(Utils::SettingsNames::mixer, Utils::SettingsNames::ScfMixers::ediisDiis);
  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om3.setStructure(Utils::XyzStreamHandler::read(ssoctatetraene));
  auto cas = std::get<0>(
      UnoCasGeneration::generate(om3.clone(), Utils::SecondQuantization::UnoCI::ThresholdFilter{{0.002, 1.998}}, settings));
  EXPECT_EQ(cas.getActiveIndices().restricted.size(), 8);
  EXPECT_EQ(cas.getCasElectrons().restricted, 8);
  EXPECT_EQ(cas.getInactiveIndices().restricted.size(), 17);
}

TEST_F(AnOMxCalculation, CanDoUnoCIFromCalculator) {
  Om3Method om3;
  std::stringstream ssoctatetraene{"18\n\n"
                                   "C   1.58365383890105     -1.01977366210975      0.92247436419201\n"
                                   "C   2.73505565211047     -0.66433760211071      0.31609409662862\n"
                                   "C   2.92731453636819      0.55757081432472     -0.41912446207901\n"
                                   "C   4.09487910688951      0.90959532661804     -1.02658696645045\n"
                                   "C   4.29454579636180      2.12402891602624     -1.75622082690101\n"
                                   "C   5.46366087423553      2.46304575653813     -2.36483440286090\n"
                                   "C   5.73351533872446      3.68393787286481     -3.10266299681885\n"
                                   "C   5.00877216729167      4.81981286806325     -3.08649102860311\n"
                                   "H   1.49018519770001     -1.95889820258656      1.46723658060291\n"
                                   "H   0.70278428187383     -0.37473057075025      0.88705928752082\n"
                                   "H   3.59733787640963     -1.33824509181151      0.37446949668829\n"
                                   "H   2.07211748491966      1.24038464707200     -0.48506095831919\n"
                                   "H   4.94791876087749      0.22384327395107     -0.95684768561072\n"
                                   "H   3.43707710332457      2.80278445775672     -1.82741548983747\n"
                                   "H   6.28603135602525      1.74136832238316     -2.31335740163400\n"
                                   "H   6.63764759936980      3.66644514017172     -3.71987000388946\n"
                                   "H   4.11943257971621      4.93401458366712     -2.46350821160383\n"
                                   "H   5.29678044890082      5.67818314993179     -3.69298339102465\n"};

  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om3.settings().modifyBool(Utils::SettingsNames::casWithUnoCI, true);
  om3.settings().modifyDouble(Utils::SettingsNames::unoCiLowThresholdOption, 0.002);
  om3.setStructure(Utils::XyzStreamHandler::read(ssoctatetraene));
  NddoExcitedStatesAssembler<Om3Method> unosAssembler(om3);
  auto res = unosAssembler.generateCasWithUNO(om3.getNumberElectrons(), om3.getNumberAtomicOrbitals());
  auto cas = std::get<0>(res);
  //  Eigen::VectorXd occs = std::get<1>(res);
  //  Eigen::VectorXd occsTest = Eigen::VectorXd::Constant(occs.size(), 1);
  //
  //  Eigen::VectorXd entropies = Eigen::VectorXd::Zero(occs.size());
  //  Eigen::VectorXd entropiesTest = Eigen::VectorXd::Zero(occs.size());
  //
  //  for (int i = 0; i < occs.size(); ++i) {
  //    entropies(i) = 2.0 * ((1 - occs(i) * 0.5) * std::log(1 - occs(i) * 0.5) + 0.5 * occs(i) * std::log(0.5 *
  //    occs(i))); entropiesTest(i) =
  //        2.0 * ((1 - occsTest(i) * 0.5) * std::log(1 - occsTest(i) * 0.5) + 0.5 * occsTest(i) * std::log(0.5 *
  //        occsTest(i)));
  //  }
  //  std::cout << "occs.transpose()" << std::endl;
  //  std::cout << occs.transpose() << std::endl;
  //  std::cout << "entropies.transpose()" << std::endl;
  //  std::cout << entropies.transpose() << std::endl;
  //  std::cout << "normalized.transpose()" << std::endl;
  //  std::cout << entropies.array().transpose() / std::log(4) << std::endl;
  //  std::cout << "normalizedTest.transpose()" << std::endl;
  //  std::cout << entropiesTest.array().transpose() / std::log(4) << std::endl;

  EXPECT_EQ(cas.getActiveIndices().restricted.size(), 8);
  EXPECT_EQ(cas.getCasElectrons().restricted, 8);
  EXPECT_EQ(cas.getInactiveIndices().restricted.size(), 17);
}

TEST_F(AnOMxCalculation, CanDoUnoCIFromCalculatorAndCasSize) {
  Om3Method om3;
  std::stringstream ssoctatetraene{"18\n\n"
                                   "C   1.58365383890105     -1.01977366210975      0.92247436419201\n"
                                   "C   2.73505565211047     -0.66433760211071      0.31609409662862\n"
                                   "C   2.92731453636819      0.55757081432472     -0.41912446207901\n"
                                   "C   4.09487910688951      0.90959532661804     -1.02658696645045\n"
                                   "C   4.29454579636180      2.12402891602624     -1.75622082690101\n"
                                   "C   5.46366087423553      2.46304575653813     -2.36483440286090\n"
                                   "C   5.73351533872446      3.68393787286481     -3.10266299681885\n"
                                   "C   5.00877216729167      4.81981286806325     -3.08649102860311\n"
                                   "H   1.49018519770001     -1.95889820258656      1.46723658060291\n"
                                   "H   0.70278428187383     -0.37473057075025      0.88705928752082\n"
                                   "H   3.59733787640963     -1.33824509181151      0.37446949668829\n"
                                   "H   2.07211748491966      1.24038464707200     -0.48506095831919\n"
                                   "H   4.94791876087749      0.22384327395107     -0.95684768561072\n"
                                   "H   3.43707710332457      2.80278445775672     -1.82741548983747\n"
                                   "H   6.28603135602525      1.74136832238316     -2.31335740163400\n"
                                   "H   6.63764759936980      3.66644514017172     -3.71987000388946\n"
                                   "H   4.11943257971621      4.93401458366712     -2.46350821160383\n"
                                   "H   5.29678044890082      5.67818314993179     -3.69298339102465\n"};

  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om3.settings().modifyBool(Utils::SettingsNames::casWithUnoCI, true);
  om3.settings().modifyInt(Utils::SettingsNames::casOrbitalsAroundFermiOption, 8);
  om3.setStructure(Utils::XyzStreamHandler::read(ssoctatetraene));
  NddoExcitedStatesAssembler<Om3Method> unosAssembler(om3);
  auto res = unosAssembler.generateCasWithUNO(om3.getNumberElectrons(), om3.getNumberAtomicOrbitals());
  auto cas = std::get<0>(res);
  EXPECT_EQ(cas.getActiveIndices().restricted.size(), 8);
  EXPECT_EQ(cas.getCasElectrons().restricted, 8);
  EXPECT_EQ(cas.getInactiveIndices().restricted.size(), 17);
  om3.calculate("");

  om3.settings().modifyInt(Utils::SettingsNames::casOrbitalsAroundFermiOption, 6);
  cas = std::get<0>(unosAssembler.generateCasWithUNO(om3.getNumberElectrons(), om3.getNumberAtomicOrbitals()));
  EXPECT_EQ(cas.getActiveIndices().restricted.size(), 6);
  EXPECT_EQ(cas.getCasElectrons().restricted, 6);
  EXPECT_EQ(cas.getInactiveIndices().restricted.size(), 18);
}

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
