/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include <Integrals/Libint.h>
#include <Integrals/TwoBodiesIntegrals/NddoTwoBodiesIntegralsEvaluator.h>
#include <Sparrow/Implementations/Nddo/Omx/BasisHandler.h>
#include <Sparrow/Implementations/Nddo/Omx/CoreElectronEvaluation.h>
#include <Sparrow/Implementations/Nddo/Omx/EffectiveCorePotentialEvaluator.h>
#include <Sparrow/Implementations/Nddo/Omx/KlopmanFinalizer.h>
#include <Sparrow/Implementations/Nddo/Omx/OmxOneCenterTwoElectronEvaluator.h>
#include <Sparrow/Implementations/Nddo/Omx/OrthogonalizationCorrection.h>
#include <Sparrow/Implementations/Nddo/Omx/ParametersProcessor.h>
#include <Sparrow/Implementations/Nddo/Omx/ResonanceIntegralEvaluator.h>
#include <Utils/Constants.h>
#include <Utils/DataStructures/SlaterToGaussian.h>
#include <Utils/Geometry/ElementTypes.h>
#include <Utils/IO/ChemicalFileFormats/XyzStreamHandler.h>
#include <Utils/Math/DerivOrderEnum.h>
#include <Utils/Typenames.h>
#include <gmock/gmock.h>
#include <Eigen/Eigenvalues>

namespace Scine {
namespace Sparrow {
namespace nddo {

using namespace testing;
using namespace Integrals;

class AResonanceTest : public Test {
 public:
  Utils::ElementTypeCollection elements;
  Utils::PositionCollection positions;

  OmxElementParameters elParamsOm1, elParams, elParamsOm3;

  std::vector<Utils::AtomicGtos> gtos, gtosOm3;
  Utils::AtomCollection structure;
  BasisSetController basisSet;

  Eigen::MatrixXd ovlpMatrix;

  void SetUp() final {
    elements = Utils::ElementTypeCollection{Utils::ElementType::C, Utils::ElementType::N};
    positions = Utils::PositionCollection(2, 3);
    positions << 0, 0, 0, 1, 0, 0;

    elParamsOm1.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, om1()));
    elParamsOm1.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om1()));

    structure = Utils::AtomCollection{elements, positions};
    elParams.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, om2()));
    elParams.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om2()));
    elParams.insert(Utils::ElementType::N, ParametersProcessor::processAtomicParameters(Utils::ElementType::N, om2()));
    gtos.push_back(elParams.get(Utils::ElementType::C).gtos);
    gtos.push_back(elParams.get(Utils::ElementType::N).gtos);
    elParamsOm3.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, om3()));
    elParamsOm3.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om3()));
    elParamsOm3.insert(Utils::ElementType::N, ParametersProcessor::processAtomicParameters(Utils::ElementType::N, om3()));
    elParamsOm3.insert(Utils::ElementType::O, ParametersProcessor::processAtomicParameters(Utils::ElementType::O, om3()));
    basisSet = BasisSetController(structure, gtos, "om2-3g");

    auto& instance = Libint::getInstance();
    ovlpMatrix = instance.compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();
  }
};

TEST_F(AResonanceTest, CanConstructResonanceIntegralEvaluator) {
  auto resEval = ResonanceIntegralEvaluator(elements, positions, elParams);
}

TEST_F(AResonanceTest, NoRotationIfAlignedToZAxis) {
  using LocalCoordinates = ResonanceIntegralEvaluator::Types;
  using GlobalCoordinates = Integrals::Orbitals::Orbital;
  positions.row(1) = Utils::Position{0, 0, 1};
  auto resEval = ResonanceIntegralEvaluator(elements, positions, elParams);
  Eigen::MatrixXd resonanceMatrix = resEval.getResonanceIntegralMatrix().block(0, 4, 4, 4);
  // Atom b is the fast index
  Eigen::MatrixXd rotMat = resEval.generateRotationMatrix({0, 0, 1}, {0, 0, 0});

  Eigen::MatrixXd rotatedMatrix = rotMat * resonanceMatrix * rotMat.transpose();

  EXPECT_DOUBLE_EQ(resonanceMatrix(GlobalCoordinates::s, GlobalCoordinates::s),
                   rotatedMatrix(LocalCoordinates::s, LocalCoordinates::s));
  EXPECT_DOUBLE_EQ(resonanceMatrix(GlobalCoordinates::s, GlobalCoordinates::z),
                   -rotatedMatrix(LocalCoordinates::s, LocalCoordinates::pSigma));
  EXPECT_DOUBLE_EQ(resonanceMatrix(GlobalCoordinates::s, GlobalCoordinates::x),
                   rotatedMatrix(LocalCoordinates::s, LocalCoordinates::pPiPlus));
  EXPECT_DOUBLE_EQ(resonanceMatrix(GlobalCoordinates::s, GlobalCoordinates::y),
                   rotatedMatrix(LocalCoordinates::s, LocalCoordinates::pPiMinus));
  EXPECT_DOUBLE_EQ(resonanceMatrix(GlobalCoordinates::x, GlobalCoordinates::x),
                   rotatedMatrix(LocalCoordinates::pPiPlus, LocalCoordinates::pPiPlus));
  EXPECT_DOUBLE_EQ(resonanceMatrix(GlobalCoordinates::y, GlobalCoordinates::y),
                   rotatedMatrix(LocalCoordinates::pPiMinus, LocalCoordinates::pPiMinus));
  EXPECT_DOUBLE_EQ(resonanceMatrix(GlobalCoordinates::z, GlobalCoordinates::z),
                   rotatedMatrix(LocalCoordinates::pSigma, LocalCoordinates::pSigma));
}

TEST_F(AResonanceTest, OtherCoordinatesEntartet) {
  positions.row(1) = Utils::Position{1, 0, 0};
  auto resEvalOnX = ResonanceIntegralEvaluator(elements, positions, elParams);
  positions.row(1) = Utils::Position{0, 1, 0};
  auto resEvalOnY = ResonanceIntegralEvaluator(elements, positions, elParams);
  positions.row(1) = Utils::Position{0, 0, 1};
  auto resEvalOnZ = ResonanceIntegralEvaluator(elements, positions, elParams);
  Eigen::MatrixXd resOnX = resEvalOnX.getResonanceIntegralMatrix().block(0, 4, 4, 4);
  Eigen::MatrixXd resOnY = resEvalOnY.getResonanceIntegralMatrix().block(0, 4, 4, 4);
  Eigen::MatrixXd resOnZ = resEvalOnZ.getResonanceIntegralMatrix().block(0, 4, 4, 4);

  // Y and Z are entartet
  EXPECT_DOUBLE_EQ(resOnX(0, 2), resOnX(0, 3));
  EXPECT_DOUBLE_EQ(resOnX(1, 2), resOnX(1, 3));
  EXPECT_DOUBLE_EQ(resOnX(2, 2), resOnX(3, 3));
  // X and Z are entartet
  EXPECT_DOUBLE_EQ(resOnY(0, 1), resOnY(0, 3));
  EXPECT_DOUBLE_EQ(resOnY(2, 1), resOnY(2, 3));
  EXPECT_DOUBLE_EQ(resOnY(1, 1), resOnY(3, 3));
  // X and Y are entartet
  EXPECT_DOUBLE_EQ(resOnZ(0, 1), resOnZ(0, 2));
  EXPECT_DOUBLE_EQ(resOnZ(3, 1), resOnZ(3, 2));
  EXPECT_DOUBLE_EQ(resOnZ(1, 1), resOnZ(2, 2));
}

TEST_F(AResonanceTest, ResonancePartEnergyIsInvariantToRotations) {
  /*
   * elements = Utils::ElementTypeCollection{Utils::ElementType::C, Utils::ElementType::C};
   * positions = Utils::PositionCollection(2, 3);
   * positions << 0, 0, 0, 1, 0, 0;
   */
  auto resEval = ResonanceIntegralEvaluator(elements, positions, elParams);
  Eigen::MatrixXd resMat1 = resEval.getResonanceIntegralMatrix();
  // sum_{l on B} S_mu,l Beta_l,nu + Beta_mu,l S_l,nu
  Eigen::MatrixXd hTerm1 =
      ovlpMatrix.block(0, 4, 4, 4) * resMat1.block(4, 0, 4, 4) + resMat1.block(0, 4, 4, 4) * ovlpMatrix.block(4, 0, 4, 4);
  Eigen::SelfAdjointEigenSolver<Eigen::MatrixXd> es(hTerm1);

  // Same system, other geometry
  positions.row(1) = Utils::Position{0, 0, 1};
  auto& instance = Libint::getInstance();
  basisSet.modifyPositions(positions);
  ovlpMatrix = instance.compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();

  resEval = ResonanceIntegralEvaluator(elements, positions, elParams);
  Eigen::MatrixXd resMat2 = resEval.getResonanceIntegralMatrix();
  Eigen::MatrixXd hTerm2 =
      ovlpMatrix.block(0, 4, 4, 4) * resMat2.block(4, 0, 4, 4) + resMat2.block(0, 4, 4, 4) * ovlpMatrix.block(4, 0, 4, 4);
  Eigen::SelfAdjointEigenSolver<Eigen::MatrixXd> es2(hTerm2);

  Eigen::VectorXd eigenvalues1 = es.eigenvalues();
  Eigen::VectorXd eigenvalues2 = es2.eigenvalues();
  for (int i = 0; i < eigenvalues1.size(); ++i) {
    EXPECT_THAT(eigenvalues1(i), DoubleNear(eigenvalues2(i), 1e-6));
  }
}

// Test that if I have a single carbon atom and calculate the interaction to a N core somewhere
// I get the same as if I habe a carbon-nitrogen compound and calculate
// V_{mu,nu on C}, N
TEST_F(AResonanceTest, TestVuvB) {
  Utils::MatrixWithDerivatives locResult;
  Utils::ElementTypeCollection elLoc{Utils::ElementType::C};
  Utils::PositionCollection posLoc(1, 3);
  posLoc << 0, 0, 0;
  std::vector<Utils::AtomicGtos> locGtos;
  locGtos.push_back(elParams.get(Utils::ElementType::C).gtos);
  std::pair<double, std::array<double, 3>> nuclearCharge{5.0, {{10, 1, 1}}};
  auto b = BasisSetController({elLoc, posLoc}, locGtos, "om2-3g");
  const auto& localBasis = b.getBasisSet();
  const auto indicesOfFirstBFinShell = localBasis.shell2bf();

  locResult.reset(static_cast<int>(localBasis.nbf()));

  auto localEngine = libint2::Engine(toLibintOperator(Operator::nuclear), static_cast<size_t>(localBasis.max_nprim()),
                                     localBasis.max_l(), 0);
  std::vector<std::pair<double, std::array<double, 3>>> nuclearCharges;
  nuclearCharges.push_back(nuclearCharge);
  localEngine.set_params(nuclearCharges);
  const auto& buffer = localEngine.results();
  // number of operators * number of derivatives
  // const auto resultsDimension = localEngine.nshellsets();

  // If oparators are hermitian (here, they are) there is index permutational symmetry:
  // <a|O|b> = <b|O|a>, so loop only on half the matrix
  for (size_t s1 = 0; s1 < localBasis.size(); ++s1) {
    int firstIndexOfS1 = indicesOfFirstBFinShell[s1];
    int sizeOfShell1 = localBasis[s1].size();

    for (size_t s2 = 0; s2 <= s1; ++s2) {
      int firstIndexOfS2 = indicesOfFirstBFinShell[s2];
      int sizeOfShell2 = localBasis[s2].size();

      localEngine.compute(localBasis[s1], localBasis[s2]);

      // If everything is screened out break out
      if (buffer[0] == nullptr) {
        break;
      }
      // Digest
      Eigen::Map<const Eigen::Matrix<double, -1, -1, Eigen::RowMajor>> bufferMatrix(buffer[0], sizeOfShell1, sizeOfShell2);
      fill<Utils::DerivativeOrder::Zero>(locResult, firstIndexOfS1, firstIndexOfS2, sizeOfShell1, sizeOfShell2,
                                         {{bufferMatrix}});
    }
  }

  Utils::ElementTypeCollection el{Utils::ElementType::C, Utils::ElementType::N};
  Utils::PositionCollection pos(2, 3);
  pos << 0, 0, 0, 10, 1, 1;

  auto str = Utils::AtomCollection{el, pos};
  auto basis = BasisSetController(str, gtos, "om2-3g");
  Eigen::MatrixXd result = CoreElectronEvaluation::analyticalCoreElectron(0, 1, 5.0, basis);

  for (int row = 0; row < result.rows(); ++row) {
    for (int col = 0; col < result.cols(); ++col) {
      EXPECT_THAT(locResult.get0()(row, col), DoubleNear(result(row, col), 1e-6));
    }
  }
}
TEST_F(AResonanceTest, TestVuvBKlopmanScaling) {
  Integrals::DiatomicKey index{1, 0};
  KlopmanFinalizer klopman;
  Integrals::canonicalize(index);
  klopman.set<Utils::DerivativeOrder::Zero>(index, 0.5);

  Eigen::MatrixXd result = CoreElectronEvaluation::analyticalCoreElectron(0, 1, 5.0, basisSet);

  Eigen::MatrixXd result2 = CoreElectronEvaluation::analyticalCoreElectron(0, 1, 5.0, basisSet, klopman);

  for (int row = 0; row < result.rows(); ++row) {
    for (int col = 0; col < result.cols(); ++col) {
      EXPECT_THAT(result(row, col), DoubleNear(2 * result2(row, col), 1e-6));
    }
  }
}

// Tests that Vuv,B^s = Z_B*<mu^A nu^A, s^Bs^B> is correctly calculated
TEST_F(AResonanceTest, VuvBSemiempiricalCorrectlyCalculated) {
  NddoEriSaverDigester digester(basisSet);
  KlopmanFinalizer finalizer(ParametersProcessor::getKlopmanScalingFactors(om2()), structure.getElements(),
                             structure.getPositions());

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(structure.getElements(), elParams));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer.finalize<Utils::DerivativeOrder::Zero>(eris);
  EXPECT_THAT(CoreElectronEvaluation::getSemiempiricalCoreElectron(Integrals::Orbitals::Orbital::s,
                                                                   Integrals::Orbitals::Orbital::s, {0, 1}, eris, 7.0),
              DoubleNear(-7.0 * eris.at({1, 0}).get0()(Integrals::Orbitals::OrbitalPair::s_s,
                                                       Integrals::Orbitals::OrbitalPair::s_s),
                         1e-6));
  EXPECT_THAT(CoreElectronEvaluation::getSemiempiricalCoreElectron(Integrals::Orbitals::Orbital::s,
                                                                   Integrals::Orbitals::Orbital::s, {1, 0}, eris, 6.0),
              DoubleNear(-6.0 * eris.at({1, 0}).get0()(Integrals::Orbitals::OrbitalPair::s_s,
                                                       Integrals::Orbitals::OrbitalPair::s_s),
                         1e-6));
  EXPECT_THAT(CoreElectronEvaluation::getSemiempiricalCoreElectron(Integrals::Orbitals::Orbital::s,
                                                                   Integrals::Orbitals::Orbital::x, {1, 0}, eris, 6.0),
              DoubleNear(-6.0 * eris.at({1, 0}).get0()(Integrals::Orbitals::OrbitalPair::s_x,
                                                       Integrals::Orbitals::OrbitalPair::s_s),
                         1e-6));
  EXPECT_THAT(CoreElectronEvaluation::getSemiempiricalCoreElectron(Integrals::Orbitals::Orbital::s,
                                                                   Integrals::Orbitals::Orbital::x, {0, 1}, eris, 7.0),
              DoubleNear(-7.0 * eris.at({1, 0}).get0()(Integrals::Orbitals::OrbitalPair::s_s,
                                                       Integrals::Orbitals::OrbitalPair::s_x),
                         1e-6));
}

TEST_F(AResonanceTest, TwoBodyOrthogonalizationCorrectionCorrect) {
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::H, Utils::ElementType::H,
                                      Utils::ElementType::H, Utils::ElementType::H};

  Utils::PositionCollection pos(5, 3);
  pos.row(0) = Utils::Position{0, 0, 0};
  pos.row(1) = Utils::Position{0, 0, 1.087};
  pos.row(2) = Utils::Position{1.0246513, 0, -0.362848056};
  pos.row(3) = Utils::Position{-0.51232565, 0.887374056, -0.362848056};
  pos.row(4) = Utils::Position{-0.51232565, -0.887374056, -0.362848056};
  pos *= Utils::Constants::bohr_per_angstrom * 0.998;
  gtos = {};

  auto params = om1();
  elParamsOm1 = {};
  elParamsOm1.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, params));
  elParamsOm1.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, params));
  gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParamsOm1));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  auto finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(om1()), els, pos);
  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);

  auto resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm1);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix();

  ovlpMatrix =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();
  OrthogonalizationCorrection VOrthB(els, basisSet, elParamsOm1, ovlpMatrix, resMat);
  VOrthB.formH(eris);

  // Give some slack as structure is not known...
  // From M. Kolb, "Ein neues semiempirisches Verfahren auf grundlage der nddo-naeherung", 1991, Page 109,
  // Tabelle 6.2
  Eigen::MatrixXd twoCenterAB =
      Utils::Constants::ev_per_hartree * (VOrthB.getTwoCenterFirstOrderBlock(0) + VOrthB.getTwoCenterSecondOrderBlock(0));
  EXPECT_NEAR(twoCenterAB(0, 0), 2.36, 1e-1);
  EXPECT_NEAR(twoCenterAB(3, 3), 1.20, 1e-1);
  Eigen::MatrixXd twoCenterBA = VOrthB.getTwoCenterFirstOrderBlock(1) + VOrthB.getTwoCenterSecondOrderBlock(1);
  EXPECT_NEAR(Utils::Constants::ev_per_hartree * twoCenterBA(0, 0), 2.78, 1e-1);
}

TEST_F(AResonanceTest, TwoBodyOrthogonalizationCorrectionCorrectOH2) {
  std::stringstream ssH2O{"3\n\n"
                          "O        0.00000        0.00000        0.00000\n"
                          "H        0.593849      0.7479093      0.00000\n"
                          "H        0.593849      -0.7479093     0.00000\n"};
  auto structure = Utils::XyzStreamHandler::read(ssH2O);
  auto pos = structure.getPositions();
  auto els = structure.getElements();
  gtos = {};

  auto params = om1();
  elParamsOm1 = {};
  elParamsOm1.insert(Utils::ElementType::O, ParametersProcessor::processAtomicParameters(Utils::ElementType::O, params));
  elParamsOm1.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, params));
  gtos.push_back(elParamsOm1.get(Utils::ElementType::O).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);

  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");
  auto finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(elParamsOm1), els, pos);

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParamsOm1));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  std::cout << eris[{1, 0}].get0() << std::endl;

  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);
  std::cout << eris[{1, 0}].get0() << std::endl;

  double semiempirical = 0.0;
  double anal = 0.0;
  Eigen::MatrixXd O = Eigen::MatrixXd::Zero(4, 4);
  Eigen::MatrixXd OOnlySemiempiricalCore = Eigen::MatrixXd::Zero(4, 4);
  for (int mu = 0; mu < 4; ++mu) {   // V(PI)_ss = (V_ss,a + sum_B Z_B * <s_Cs_C | s_B s_B>) * f_ko
    for (int nu = 0; nu < 4; ++nu) { // V(PI)_ss = (V_ss,a + sum_B Z_B * <s_Cs_C | s_B s_B>) * f_ko
      OOnlySemiempiricalCore(mu, nu) -= CoreElectronEvaluation::getSemiempiricalCoreElectron(mu, nu, {0, 1}, eris, 1.0);
      O(mu, nu) -= CoreElectronEvaluation::getSemiempiricalCoreElectron(mu, nu, {0, 1}, eris, 1.0);
      O(mu, nu) -= CoreElectronEvaluation::getSemiempiricalCoreElectron(mu, nu, {0, 2}, eris, 1.0);
      O(mu, nu) += CoreElectronEvaluation::analyticalCoreElectron(0, 1, 1.0, basisSet, *finalizer)(mu, nu);
      O(mu, nu) += CoreElectronEvaluation::analyticalCoreElectron(0, 2, 1.0, basisSet, *finalizer)(mu, nu);
    }
  }
  Eigen::MatrixXd H = Eigen::MatrixXd::Zero(1, 1);
  H(0, 0) -= CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {1, 0}, eris, 6.0);
  H(0, 0) -= CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {1, 2}, eris, 1.0);
  H(0, 0) += CoreElectronEvaluation::analyticalCoreElectron(1, 0, 6.0, basisSet, *finalizer)(0, 0);
  H(0, 0) += CoreElectronEvaluation::analyticalCoreElectron(1, 2, 1.0, basisSet, *finalizer)(0, 0);

  std::cout << "Vcoreelec(O)\n"
            << Utils::Constants::ev_per_hartree * CoreElectronEvaluation::analyticalCoreElectron(0, 1, 1.0, basisSet, *finalizer)
            << std::endl;
  std::cout << "Vcoreelec(O)\n" << CoreElectronEvaluation::analyticalCoreElectron(0, 2, 1.0, basisSet) << std::endl;
  std::cout << "Vcoreelec(H)\n" << CoreElectronEvaluation::analyticalCoreElectron(1, 0, 6.0, basisSet) << std::endl;
  std::cout << "Vsemi(O)\n" << Utils::Constants::ev_per_hartree * OOnlySemiempiricalCore << std::endl;
  std::cout << "PI(O)\n" << Utils::Constants::ev_per_hartree * O << std::endl;
  std::cout << "PI(H)\n" << Utils::Constants::ev_per_hartree * H << std::endl;
  std::cout << "PI(O)\n" << O << std::endl;
  std::cout << "PI(H)\n" << H << std::endl;
  auto resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm1);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix();

  ovlpMatrix =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();
  OrthogonalizationCorrection VOrthB(els, basisSet, elParamsOm1, ovlpMatrix, resMat);
  VOrthB.formH(eris);

  // From M. Kolb, "Ein neues semiempirisches Verfahren auf grundlage der nddo-naeherung", 1991, Page 109,
  // Tabelle 6.2
  std::cout << "O" << std::endl;
  Eigen::MatrixXd twoCenterO =
      Utils::Constants::ev_per_hartree * (VOrthB.getTwoCenterFirstOrderBlock(0) + VOrthB.getTwoCenterSecondOrderBlock(0));

  std::cout << "FirstOrderBlock O\n"
            << Utils::Constants::ev_per_hartree * VOrthB.getTwoCenterFirstOrderBlock(0) << std::endl;
  std::cout << "SecondOrderBlock O\n"
            << Utils::Constants::ev_per_hartree * VOrthB.getTwoCenterSecondOrderBlock(0) << std::endl;
  EXPECT_NEAR(twoCenterO(0, 0), 1.23, 2e-1);
  EXPECT_NEAR(twoCenterO(0, 1), 1.21, 2e-1);
  EXPECT_NEAR(twoCenterO(1, 1), 0.96, 2e-1);
  EXPECT_NEAR(twoCenterO(2, 2), 1.59, 2e-1);
  EXPECT_NEAR(twoCenterO(3, 3), 0.00, 2e-1);
  std::cout << "H" << std::endl;
  Eigen::MatrixXd twoCenterH =
      Utils::Constants::ev_per_hartree * (VOrthB.getTwoCenterFirstOrderBlock(1) + VOrthB.getTwoCenterSecondOrderBlock(1));
  EXPECT_NEAR(twoCenterH(0, 0), 4.34, 2e-1);
  std::cout << "H" << std::endl;
  Eigen::MatrixXd twoCenterH2 =
      Utils::Constants::ev_per_hartree * (VOrthB.getTwoCenterFirstOrderBlock(2) + VOrthB.getTwoCenterSecondOrderBlock(2));
  EXPECT_NEAR(twoCenterH2(0, 0), 4.34, 2e-1);

  std::cout << "twoCenterO\n" << twoCenterO << std::endl;
  std::cout << "twoCenterH\n" << twoCenterH << std::endl;
  std::cout << "twoCenterH2\n" << twoCenterH2 << std::endl;
}

TEST_F(AResonanceTest, TwoBodyOrthogonalizationCorrectionCorrectCOH2) {
  std::stringstream ssOCH2{"4\n\n"
                           "O        0.00000        0.00000        0.00000\n"
                           "C        1.19500        0.00000        0.00000\n"
                           "H        1.67145        1.01250        0.00000\n"
                           "H        1.67145       -1.01250        0.00000\n"};
  // std::stringstream ssOCH2{"4\n\n"
  //                         "O        0.00000        0.00000        0.00000\n"
  //                         "C        1.20800        0.00000        0.00000\n"
  //                         "H        1.70596        0.99875        0.00000\n"
  //                         "H        1.70596       -0.99875        0.00000\n"};

  auto structure = Utils::XyzStreamHandler::read(ssOCH2);
  auto pos = structure.getPositions();
  auto els = structure.getElements();
  gtos = {};

  auto params = om1();
  elParamsOm1 = {};
  elParamsOm1.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, params));
  elParamsOm1.insert(Utils::ElementType::O, ParametersProcessor::processAtomicParameters(Utils::ElementType::O, params));
  elParamsOm1.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, params));
  gtos.push_back(elParamsOm1.get(Utils::ElementType::O).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");
  auto finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(elParamsOm1), els, pos);

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParamsOm1));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);

  double semiempirical = 0.0;
  double anal = 0.0;
  Eigen::MatrixXd O = Eigen::MatrixXd(4, 4);
  Eigen::MatrixXd C = Eigen::MatrixXd(4, 4);
  Eigen::MatrixXd H = Eigen::MatrixXd(1, 1);
  for (int i = 0; i < 4; ++i) { // V(PI)_ss = (V_ss,a + sum_B Z_B * <s_Cs_C | s_B s_B>) * f_ko
    semiempirical -= CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {0, i}, eris, 1.0);
    anal += CoreElectronEvaluation::analyticalCoreElectron(0, i, 1.0, basisSet, *finalizer)(0, 0);
  }

  auto resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm1);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix();

  ovlpMatrix =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();
  OrthogonalizationCorrection VOrthB(els, basisSet, elParamsOm1, ovlpMatrix, resMat);
  VOrthB.formH(eris);

  // Give some slack as structure is not known...
  // I am willing to be very loose with the thresholds as many degrees of freedom and no idea what the structure
  // was... From M. Kolb, "Ein neues semiempirisches Verfahren auf grundlage der nddo-naeherung", 1991, Page 109,
  // Tabelle 6.2
  std::cout << "O" << std::endl;
  Eigen::MatrixXd twoCenterO =
      Utils::Constants::ev_per_hartree * (VOrthB.getTwoCenterFirstOrderBlock(0) + VOrthB.getTwoCenterSecondOrderBlock(0));
  EXPECT_NEAR(twoCenterO(0, 0), 0.85, 2e-1);
  EXPECT_NEAR(twoCenterO(0, 1), 1.48, 2e-1);
  EXPECT_NEAR(twoCenterO(1, 1), 1.74, 2e-1);
  EXPECT_NEAR(twoCenterO(2, 2), 0.40, 2e-1);
  std::cout << "C" << std::endl;
  Eigen::MatrixXd twoCenterC =
      Utils::Constants::ev_per_hartree * (VOrthB.getTwoCenterFirstOrderBlock(1) + VOrthB.getTwoCenterSecondOrderBlock(1));
  EXPECT_NEAR(twoCenterC(0, 0), 3.28, 2e-1);
  EXPECT_NEAR(twoCenterC(0, 1), -1.64, 2e-1);
  EXPECT_NEAR(twoCenterC(1, 1), 3.09, 3.5e-1);
  EXPECT_NEAR(twoCenterC(2, 2), 1.88, 2e-1);
  EXPECT_NEAR(twoCenterC(3, 3), 0.65, 2e-1);
  std::cout << "H" << std::endl;
  Eigen::MatrixXd twoCenterH =
      Utils::Constants::ev_per_hartree * (VOrthB.getTwoCenterFirstOrderBlock(2) + VOrthB.getTwoCenterSecondOrderBlock(2));
  EXPECT_NEAR(twoCenterH(0, 0), 2.64, 2.64 * 0.05);

  std::cout << "twoCenterO\n" << twoCenterO << std::endl;
  std::cout << "twoCenterC\n" << twoCenterC << std::endl;
  std::cout << "twoCenterH\n" << twoCenterH << std::endl;
}

TEST_F(AResonanceTest, TwoBodyOrthogonalizationCorrectionCorrectNH3) {
  Utils::ElementTypeCollection els = {Utils::ElementType::N, Utils::ElementType::H, Utils::ElementType::H,
                                      Utils::ElementType::H};

  Utils::PositionCollection pos(4, 3);
  pos.row(0) = Utils::Position{0.00000, 0.00000, 0.00000};
  pos.row(1) = Utils::Position{1.01300, 0.00000, 0.00000};
  pos.row(2) = Utils::Position{-0.29786, 0.96822, 0.00000};
  pos.row(3) = Utils::Position{-0.29786, -0.28470, -0.92542};
  pos *= Utils::Constants::bohr_per_angstrom * 0.998;
  gtos = {};

  auto params = om1();
  elParamsOm1 = {};
  elParamsOm1.insert(Utils::ElementType::N, ParametersProcessor::processAtomicParameters(Utils::ElementType::N, params));
  elParamsOm1.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, params));
  gtos.push_back(elParamsOm1.get(Utils::ElementType::N).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");
  auto finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(om1()), els, pos);

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParamsOm1));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);

  auto resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm1);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix();

  ovlpMatrix =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();
  OrthogonalizationCorrection VOrthB(els, basisSet, elParamsOm1, ovlpMatrix, resMat);
  VOrthB.formH(eris);

  // Give some slack as structure is not known...
  // From M. Kolb, "Ein neues semiempirisches Verfahren auf grundlage der nddo-naeherung", 1991, Page 109,
  // Tabelle 6.2
  Eigen::MatrixXd twoCenterN =
      Utils::Constants::ev_per_hartree * (VOrthB.getTwoCenterFirstOrderBlock(0) + VOrthB.getTwoCenterSecondOrderBlock(0));
  EXPECT_NEAR(twoCenterN(0, 0), 3.23, 1e-1);
  Eigen::MatrixXd twoCenterH =
      Utils::Constants::ev_per_hartree * (VOrthB.getTwoCenterFirstOrderBlock(1) + VOrthB.getTwoCenterSecondOrderBlock(1));
  EXPECT_NEAR(twoCenterH(0, 0), 3.86, 1e-1);
}

TEST_F(AResonanceTest, CalculatesResonanceMatrixCorrectly) {
  Utils::ElementTypeCollection els = {Utils::ElementType::H, Utils::ElementType::H, Utils::ElementType::H};
  Utils::PositionCollection pos(3, 3);
  pos << 0, 0, 0, 0, 0, 1.1 * Utils::Constants::bohr_per_angstrom, 0, 1.1 * Utils::Constants::bohr_per_angstrom,
      1.1 * Utils::Constants::bohr_per_angstrom;

  auto resEval = ResonanceIntegralEvaluator(els, pos, elParams);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix();
  Eigen::MatrixXd actualResMat(3, 3);
  // From W. Weber, "Ein neues emiempirisches NDDO-Verfahren mit Orthogonalisierungskorrekturen", 1996, Seite 60,
  // Tabelle 6.3
  actualResMat << 0, -2.786, -1.871, -2.786, 0, -2.786, -1.871, -2.786, 0;
  for (int row = 0; row < 3; ++row) {
    for (int col = 0; col < 3; ++col) {
      EXPECT_THAT(resMat(row, col) * Utils::Constants::ev_per_hartree, DoubleNear(actualResMat(row, col), 1e-3));
    }
  }
}

TEST_F(AResonanceTest, CalculatesResonanceMatrixCorrectlyOm3HH) {
  gtosOm3 = {};
  gtosOm3.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtosOm3.push_back(elParams.get(Utils::ElementType::H).gtos);
  Utils::ElementTypeCollection els = {Utils::ElementType::H, Utils::ElementType::H};
  Utils::PositionCollection pos(2, 3);
  pos << 0, 0, 0, 0, 0, 0.71233 * Utils::Constants::bohr_per_angstrom;

  auto resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix().array().abs();
  // Data from https://apps.automeris.io/wpd/
  // Mirjam Scholten, Semiempirische Verfahren mit Orthogonalisierungskorrekturen: Die OM3 Methode
  // 2003 (The OM3 PhD thesis), Abbildung 6.1
  ASSERT_EQ(resMat.cols(), 2);
  ASSERT_EQ(resMat.rows(), 2);
  EXPECT_THAT(resMat(0, 1) * Utils::Constants::ev_per_hartree, DoubleNear(3.0675, 5e-2));
  pos.row(1) = Utils::Position(0, 0, 2.000 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  // Data from https://apps.automeris.io/wpd/
  // Mirjam Scholten, Semiempirische Verfahren mit Orthogonalisierungskorrekturen: Die OM3 Methode
  // 2003 (The OM3 PhD thesis), Abbildung 6.1
  ASSERT_EQ(resMat.cols(), 2);
  ASSERT_EQ(resMat.rows(), 2);
  EXPECT_THAT(resMat(0, 1) * Utils::Constants::ev_per_hartree, DoubleNear(0.90365, 3e-2));
}

TEST_F(AResonanceTest, CalculatesResonanceMatrixCorrectlyOm3CC) {
  gtosOm3 = {};
  gtosOm3.push_back(elParams.get(Utils::ElementType::C).gtos);
  gtosOm3.push_back(elParams.get(Utils::ElementType::C).gtos);
  /*
   * Data from https://apps.automeris.io/wpd/
   * Mirjam Scholten, Semiempirische Verfahren mit Orthogonalisierungskorrekturen: Die OM3 Methode
   * 2003 (The OM3 PhD thesis), Abbildung 6.3
   * Distance [angstrom]   Energy [eV]
   * s-s:         0.6151079136690645, 6.010033444816054
   * s-sigma:     0.7158273381294964, 4.976588628762542
   * pi-pi:       0.6007194244604315, 4.65551839464883
   * sigma-sigma: 0.8093525179856114, 3.872909698996656
   */
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::C};
  Utils::PositionCollection pos(2, 3);
  pos << 0, 0, 0, 0, 0, 0.615107913669064 * Utils::Constants::bohr_per_angstrom;
  auto resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 8);
  ASSERT_THAT(resMat.rows(), 8);
  EXPECT_THAT(resMat(0, 4) * Utils::Constants::ev_per_hartree, DoubleNear(6.010033444816054, 5e-2));

  pos.row(1) = Utils::Position(0, 0, 0.7158273381294964 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 8);
  ASSERT_THAT(resMat.rows(), 8);
  EXPECT_THAT(resMat(0, 7) * Utils::Constants::ev_per_hartree, DoubleNear(4.976588628762542, 5e-2));

  pos.row(1) = Utils::Position(0, 0, 0.6007194244604315 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 8);
  ASSERT_THAT(resMat.rows(), 8);
  EXPECT_THAT(resMat(1, 5) * Utils::Constants::ev_per_hartree, DoubleNear(4.65551839464883, 5e-2));

  /*
   * sigma sigma integrals
   * 0.22987, 2.59486
   * 0.57653, 3.69453
   * 1.16469, 3.56913
   * 1.86277, 2.03537
   * 2.61631, 0.67524
   *
   */
  pos.row(1) = Utils::Position(0, 0, 0.22987 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 8);
  ASSERT_THAT(resMat.rows(), 8);
  EXPECT_THAT(resMat(3, 7) * Utils::Constants::ev_per_hartree, DoubleNear(2.59486, 5e-2));

  pos.row(1) = Utils::Position(0, 0, 0.57653 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 8);
  ASSERT_THAT(resMat.rows(), 8);
  EXPECT_THAT(resMat(3, 7) * Utils::Constants::ev_per_hartree, DoubleNear(3.69453, 5e-2));

  pos.row(1) = Utils::Position(0, 0, 1.16469 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 8);
  ASSERT_THAT(resMat.rows(), 8);
  EXPECT_THAT(resMat(3, 7) * Utils::Constants::ev_per_hartree, DoubleNear(3.56913, 5e-2));

  pos.row(1) = Utils::Position(0, 0, 1.86277 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 8);
  ASSERT_THAT(resMat.rows(), 8);
  EXPECT_THAT(resMat(3, 7) * Utils::Constants::ev_per_hartree, DoubleNear(2.03537, 5e-2));

  pos.row(1) = Utils::Position(0, 0, 2.61631 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 8);
  ASSERT_THAT(resMat.rows(), 8);
  EXPECT_THAT(resMat(3, 7) * Utils::Constants::ev_per_hartree, DoubleNear(0.67524, 5e-2));
}

TEST_F(AResonanceTest, CalculatesResonanceMatrixCorrectlyOm3CH) {
  gtosOm3 = {};
  gtosOm3.push_back(elParams.get(Utils::ElementType::C).gtos);
  gtosOm3.push_back(elParams.get(Utils::ElementType::H).gtos);
  /*
   * Data from https://apps.automeris.io/wpd/
   * Mirjam Scholten, Semiempirische Verfahren mit Orthogonalisierungskorrekturen: Die OM3 Methode
   * 2003 (The OM3 PhD thesis), Abbildung 6.3
   * Distance [angstrom]   Energy [eV]
   * s-s:      0.6458084506067059, 4.116546364092709
   * s-s:      2.0770571163155678, 0.696225903423286
   * s-sigma:  0.757467702942076, 3.53948636282333
   * s-sigma:  2.1283169756452196, 1.0290803344238455
   */
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::H};
  Utils::PositionCollection pos(2, 3);
  pos << 0, 0, 0, 0, 0, 0.6458084506067059 * Utils::Constants::bohr_per_angstrom;

  auto resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 5);
  ASSERT_THAT(resMat.rows(), 5);
  EXPECT_THAT(resMat(0, 4) * Utils::Constants::ev_per_hartree, DoubleNear(4.116546364092709, 5e-2));

  pos.row(1) = Utils::Position(0, 0, 2.0770571163155678 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 5);
  ASSERT_THAT(resMat.rows(), 5);
  EXPECT_THAT(resMat(0, 4) * Utils::Constants::ev_per_hartree, DoubleNear(0.696225903423286, 5e-2));

  pos.row(1) = Utils::Position(0, 0, 0.757467702942076 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 5);
  ASSERT_THAT(resMat.rows(), 5);
  EXPECT_THAT(resMat(3, 4) * Utils::Constants::ev_per_hartree, DoubleNear(3.53948636282333, 5e-2));

  pos.row(1) = Utils::Position(0, 0, 2.1283169756452196 * Utils::Constants::bohr_per_angstrom);
  resEval = ResonanceIntegralEvaluator(els, pos, elParamsOm3);
  resMat = resEval.getResonanceIntegralMatrix().array().abs();
  ASSERT_THAT(resMat.cols(), 5);
  ASSERT_THAT(resMat.rows(), 5);
  EXPECT_THAT(resMat(3, 4) * Utils::Constants::ev_per_hartree, DoubleNear(1.0290803344238455, 5e-2));
}

TEST_F(AResonanceTest, CalculatesResonancePotentialCorrectly) {
  Utils::ElementTypeCollection els = {Utils::ElementType::H, Utils::ElementType::H, Utils::ElementType::H};
  Utils::PositionCollection pos(3, 3);
  pos << 0, 0, 0, 0, 0, 2 * 1.1 * Utils::Constants::bohr_per_angstrom, 0, 0.0, 1.1 * Utils::Constants::bohr_per_angstrom;

  gtos = {};
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({els, pos}, gtos, "om2-3g");

  auto calcEris = [&](const Integrals::BasisSetController& bSet) {
    KlopmanFinalizer finalizer(ParametersProcessor::getKlopmanScalingFactors(om2()), els, pos);

    NddoTwoBodiesIntegralsEvaluator eriCalculator(bSet);

    eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
    auto eris = eriCalculator.getResult();
    finalizer.finalize<Utils::DerivativeOrder::Zero>(eris);
    return eris;
  };

  auto eris = calcEris(basisSet);
  ovlpMatrix =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();
  auto resEval = ResonanceIntegralEvaluator(els, pos, elParams);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix();
  OrthogonalizationCorrection ortEval(els, basisSet, elParams, ovlpMatrix, resMat);
  ortEval.formH(eris);
  Eigen::MatrixXd threeCenterAB = Utils::Constants::ev_per_hartree * (ortEval.getThreeCenterFirstOrderBlock(0, 1) +
                                                                      ortEval.getThreeCenterSecondOrderBlock(0, 1));
  Eigen::MatrixXd threeCenterAC = Utils::Constants::ev_per_hartree * (ortEval.getThreeCenterFirstOrderBlock(0, 2) +
                                                                      ortEval.getThreeCenterSecondOrderBlock(0, 2));
  // Resonance Beta_AB
  EXPECT_NEAR(resMat(0, 1) * Utils::Constants::ev_per_hartree, -0.710, 1e-3);
  // Three-body ortogonalization correction V_AB,C(ORT)
  EXPECT_NEAR(threeCenterAB(0, 0), 0.610, 1e-3);
  // Resonance Beta_AC
  EXPECT_NEAR(resMat(0, 2) * Utils::Constants::ev_per_hartree, -2.786, 1e-3);
  // Three-body ortogonalization correction V_AC,B(ORT)
  EXPECT_NEAR(threeCenterAC(0, 0), 0.114, 1e-3);
  pos << 0, 0, 0, 0, 1.1 * Utils::Constants::bohr_per_angstrom, 1.1 * Utils::Constants::bohr_per_angstrom, 0,
      1.1 * Utils::Constants::bohr_per_angstrom, 0.0;
  basisSet = {{els, pos}, gtos};
  eris = calcEris(basisSet);
  ovlpMatrix =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();
  resEval = ResonanceIntegralEvaluator(els, pos, elParams);
  resMat = resEval.getResonanceIntegralMatrix();
  OrthogonalizationCorrection ortEval90Degrees(els, basisSet, elParams, ovlpMatrix, resMat);
  ortEval90Degrees.formH(eris);
  threeCenterAB = Utils::Constants::ev_per_hartree * (ortEval90Degrees.getThreeCenterFirstOrderBlock(0, 1) +
                                                      ortEval90Degrees.getThreeCenterSecondOrderBlock(0, 1));
  threeCenterAC = Utils::Constants::ev_per_hartree * (ortEval90Degrees.getThreeCenterFirstOrderBlock(0, 2) +
                                                      ortEval90Degrees.getThreeCenterSecondOrderBlock(0, 2));
  // Resonance Beta_AB
  EXPECT_NEAR(resMat(0, 1) * Utils::Constants::ev_per_hartree, -1.871, 1e-3);
  // Three-body ortogonalization correction V_AB,C(ORT)
  EXPECT_NEAR(threeCenterAB(0, 0), 0.610, 1e-3);
  // Resonance Beta_AC
  EXPECT_NEAR(resMat(0, 2) * Utils::Constants::ev_per_hartree, -2.786, 1e-3);
  // Three-body ortogonalization correction V_AC,B(ORT)
  EXPECT_NEAR(threeCenterAC(0, 0), 0.344, 1e-3);
}

TEST_F(AResonanceTest, CanCalculateThreeCenterOrthogonalizationWithH4) {
  elements = {Utils::ElementType::H, Utils::ElementType::H, Utils::ElementType::H, Utils::ElementType::H};
  positions = {4, 3};
  positions.row(0) = Utils::Position{-0.26704548738, 0.75411319287, 0};
  positions.row(1) = Utils::Position{0, 0, 0};
  positions.row(2) = Utils::Position{1.0, 0, 0};
  positions.row(3) = Utils::Position{1.26704548738, 0.75411319287, 0};
  positions *= Utils::Constants::bohr_per_angstrom;

  gtos = {};
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({elements, positions}, gtos, "om2-3g");

  auto calcEris = [&](const Integrals::BasisSetController& bSet) {
    KlopmanFinalizer finalizer(ParametersProcessor::getKlopmanScalingFactors(om2()), elements, positions);

    NddoTwoBodiesIntegralsEvaluator eriCalculator(bSet);

    eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
    auto eris = eriCalculator.getResult();
    finalizer.finalize<Utils::DerivativeOrder::Zero>(eris);
    return eris;
  };

  auto eris = calcEris(basisSet);
  ovlpMatrix =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();
  auto resEval = ResonanceIntegralEvaluator(elements, positions, elParams);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix();
  OrthogonalizationCorrection ortEval(elements, basisSet, elParams, ovlpMatrix, resMat);
  ortEval.formH(eris);
  Eigen::MatrixXd threeCenterAB = Utils::Constants::ev_per_hartree * (ortEval.getThreeCenterFirstOrderBlock(0, 1) +
                                                                      ortEval.getThreeCenterSecondOrderBlock(0, 1));
  Eigen::MatrixXd threeCenterAC = Utils::Constants::ev_per_hartree * (ortEval.getThreeCenterFirstOrderBlock(0, 2) +
                                                                      ortEval.getThreeCenterSecondOrderBlock(0, 2));

  // Resonance Beta_AB
  EXPECT_NEAR(resMat(0, 1) * Utils::Constants::ev_per_hartree, -3.109, 1e-3);
  // Three-body ortogonalization correction V_AB,C(ORT)
  EXPECT_NEAR(threeCenterAB(0, 0), 0.648, 1e-3);
  // Resonance Beta_AC
  EXPECT_NEAR(resMat(0, 2) * Utils::Constants::ev_per_hartree, -2.046, 1e-3);
  // Three-body ortogonalization correction V_AC,B(ORT)
  EXPECT_NEAR(threeCenterAC(0, 0), 1.391, 1e-3);

  positions.row(0)(1) *= -1.0;
  basisSet.modifyPositions(positions);
  eris = calcEris(basisSet);
  ovlpMatrix =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();
  resEval = ResonanceIntegralEvaluator(elements, positions, elParams);
  resMat = resEval.getResonanceIntegralMatrix();
  OrthogonalizationCorrection ortEval2(elements, basisSet, elParams, ovlpMatrix, resMat);
  ortEval2.formH(eris);
  threeCenterAB = Utils::Constants::ev_per_hartree *
                  (ortEval2.getThreeCenterFirstOrderBlock(0, 1) + ortEval2.getThreeCenterSecondOrderBlock(0, 1));
  threeCenterAC = Utils::Constants::ev_per_hartree *
                  (ortEval2.getThreeCenterFirstOrderBlock(0, 2) + ortEval2.getThreeCenterSecondOrderBlock(0, 2));
  // Resonance Beta_AB
  EXPECT_NEAR(resMat(0, 1) * Utils::Constants::ev_per_hartree, -3.109, 1e-3);
  // Three-body ortogonalization correction V_AB,C(ORT)
  EXPECT_NEAR(threeCenterAB(0, 0), 0.506, 1e-3);
  // Resonance Beta_AC
  EXPECT_NEAR(resMat(0, 2) * Utils::Constants::ev_per_hartree, -2.046, 1e-3);
  // Three-body ortogonalization correction V_AC,B(ORT)
  EXPECT_NEAR(threeCenterAC(0, 0), 1.080, 1e-3);
}

TEST_F(AResonanceTest, CalculateKlopmanOhno) {
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::C, Utils::ElementType::H,
                                      Utils::ElementType::H, Utils::ElementType::H, Utils::ElementType::H};
  Utils::PositionCollection pos(6, 3);
  pos.row(0) = Utils::Position{0, 0, 0};
  pos.row(1) = Utils::Position{0, 0, 1.2};
  pos.row(2) = Utils::Position{0, 1, 1.86};
  pos.row(3) = Utils::Position{0, -1, 1.86};
  pos.row(4) = Utils::Position{0, 1, -0.66};
  pos.row(5) = Utils::Position{0, -1, 0.66};

  pos *= Utils::Constants::bohr_per_angstrom;
  auto rawParams = om2();
  rawParams.atomic[1].exponent = 1.24;
  rawParams.atomic[6].exponent = 1.0;
  OmxElementParameters elParams2;
  elParams2.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, rawParams));
  elParams2.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, rawParams));
  gtos = {};
  gtos.push_back(elParams2.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParams2.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParams2.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams2.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams2.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParams2.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({els, pos}, gtos, "om2-3g");

  auto finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(rawParams), els, pos);

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);
  //
  // From M. Kolb, "Ein neues semiempirisches Verfahren auf grundlage der nddo-naeherung", 1991, Seite 47,
  // Tabelle 3.1
  EXPECT_THAT(finalizer->get<Utils::DerivativeOrder::Zero>({1, 0}), DoubleNear(0.80, 1e-2));
  EXPECT_THAT(finalizer->get<Utils::DerivativeOrder::Zero>({2, 1}), DoubleNear(0.78, 1e-2));
}

TEST_F(AResonanceTest, CanCalculatePenetrationIntegrals) {
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::H, Utils::ElementType::H,
                                      Utils::ElementType::H, Utils::ElementType::H};

  Utils::PositionCollection pos(5, 3);
  pos.row(0) = Utils::Position{0, 0, 0};
  pos.row(1) = Utils::Position{0, 0, 1.087};
  pos.row(2) = Utils::Position{1.02465, 0, -0.36285};
  pos.row(3) = Utils::Position{-0.512325, 0.88737293, -0.36285};
  pos.row(4) = Utils::Position{-0.512325, -0.88737293, -0.36285};
  pos *= Utils::Constants::bohr_per_angstrom * 1.0065;
  gtos = {};

  auto params = om1();
  params.atomic[1].exponent = 1.24;
  params.atomic[6].exponent = 1.00;
  elParamsOm1 = {};
  elParamsOm1.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, params));
  elParamsOm1.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, params));
  gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");
  auto finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(om1()), els, pos);

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParamsOm1));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);
  double semiempirical = 0.0;
  double anal = 0.0;
  for (int i = 1; i < 5; ++i) { // V(PI)_ss = (V_ss,a + sum_B Z_B * <s_Cs_C | s_B s_B>) * f_ko
    semiempirical -= CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {0, i}, eris, 1.0);
    anal += CoreElectronEvaluation::analyticalCoreElectron(0, i, 1.0, basisSet, *finalizer)(0, 0);
  }
  // From M. Kolb, "Ein neues semiempirisches Verfahren auf Grundlage der NDDO-Naehrung: Entwicklung der Methode,
  // Parametrisierung und Anwendung", 1991, Seite 48, Tabelle 3.2
  EXPECT_THAT((anal + semiempirical) * Utils::Constants::ev_per_hartree, DoubleNear(-2.58, 1e-2));
  semiempirical = 0;
  anal = 0;
  for (int i = 1; i < 5; ++i) { // V(PI)_ss = (V_ss,a + sum_B Z_B * <s_Cs_C | s_B s_B>) * f_ko
    semiempirical -= CoreElectronEvaluation::getSemiempiricalCoreElectron(3, 3, {0, i}, eris, 1.0);
    anal += CoreElectronEvaluation::analyticalCoreElectron(0, i, 1.0, basisSet, *finalizer)(3, 3);
  }
  // From M. Kolb, "Ein neues semiempirisches Verfahren auf Grundlage der NDDO-Naehrung: Entwicklung der Methode,
  // Parametrisierung und Anwendung", 1991, Seite 48, Tabelle 3.2
  EXPECT_THAT((anal + semiempirical) * Utils::Constants::ev_per_hartree, DoubleNear(-2.48, 1e-2));
  semiempirical = 0;
  anal = 0;
  for (int i = 0; i < 5; ++i) { // V(PI)_ss = (V_ss,a + sum_B Z_B * <s_Cs_C | s_B s_B>) * f_ko
    if (i != 1) {
      semiempirical -= CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {1, i}, eris, i == 0 ? 4 : 1);
      anal += CoreElectronEvaluation::analyticalCoreElectron(1, i, i == 0 ? 4 : 1, basisSet, *finalizer)(0, 0);
    }
  }
  // From M. Kolb, "Ein neues semiempirisches Verfahren auf Grundlage der NDDO-Naehrung: Entwicklung der Methode,
  // Parametrisierung und Anwendung", 1991, Seite 48, Tabelle 3.2
  EXPECT_THAT((anal + semiempirical) * Utils::Constants::ev_per_hartree, DoubleNear(-3.62, 1e-2));

  // If far apart, penetration integral goes to 0.
  pos *= Utils::Constants::bohr_per_angstrom * 100;
  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");
  finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(om1()), els, pos);

  NddoTwoBodiesIntegralsEvaluator farEriCalculator{basisSet};
  farEriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParamsOm1));

  farEriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  eris = farEriCalculator.getResult();
  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);
  semiempirical = 0.0;
  anal = 0.0;
  for (int i = 0; i < 5; ++i) { // V(PI)_ss = (V_ss,a + sum_B Z_B * <s_Cs_C | s_B s_B>) * f_ko
    if (i != 1) {
      semiempirical -= CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {1, i}, eris, i == 0 ? 4 : 1);
      anal += CoreElectronEvaluation::analyticalCoreElectron(1, i, i == 0 ? 4 : 1, basisSet, *finalizer)(0, 0);
    }
  }
  EXPECT_THAT(anal + semiempirical, DoubleNear(0, 1e-5));
}

// This is to plot the distance dependence of the Penetration Integrals.
/*
 *TEST_F(AResonanceTest, DistanceDependence) {
 *  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::C};
 *
 *  Utils::PositionCollection pos(2, 3);
 *  pos.row(0) = Utils::Position{0, 0, 0};
 *  pos.row(1) = Utils::Position{1.202, 0, 0.0};
 *  pos *= Utils::Constants::bohr_per_angstrom;
 *  gtos = {};
 *
 *  auto params = om1();
 *  params.atomic[1].exponent = 1.24;
 *  params.atomic[6].exponent = 1.00;
 *  elParamsOm1 = {};
 *  elParamsOm1.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C,
 *params)); gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
 *  gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
 *
 *  auto printElements = [&](double distanceInAngstrom) {
 *    pos.row(1) = Utils::Position(distanceInAngstrom * Utils::Constants::bohr_per_angstrom, 0, 0);
 *    basisSet = BasisSetController({els, pos}, gtos);
 *    NddoEriSaverDigester digester(basisSet);
 *    auto finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(om1()), els,
 *pos); digester.setFinalizer(finalizer);
 *
 *    NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet, std::move(digester));
 *    eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParamsOm1));
 *
 *    eriCalculator.evaluate<Operator::coulomb>();
 *    const auto& eris = eriCalculator.getResult();
 *    auto coefs = finalizer->getCoefficients();
 *    double ss =
 *        (CoreElectronEvaluation::analyticalCoreElectron(0, 1, 4.0, basisSet,
 *coefs)(0, 0) - CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {0, 1}, eris, 4.0)) *
 *        Utils::Constants::ev_per_hartree;
 *    double ssigma =
 *        (CoreElectronEvaluation::analyticalCoreElectron(0, 1, 4.0, basisSet,
 *coefs)(0, 1) - CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 1, {0, 1}, eris, 4.0)) *
 *        Utils::Constants::ev_per_hartree;
 *    double sigmasigma =
 *        (CoreElectronEvaluation::analyticalCoreElectron(0, 1, 4.0, basisSet,
 *coefs)(1, 1) - CoreElectronEvaluation::getSemiempiricalCoreElectron(1, 1, {0, 1}, eris, 4.0)) *
 *        Utils::Constants::ev_per_hartree;
 *    double pipi =
 *        (CoreElectronEvaluation::analyticalCoreElectron(0, 1, 4.0, basisSet,
 *coefs)(3, 3) - CoreElectronEvaluation::getSemiempiricalCoreElectron(3, 3, {0, 1}, eris, 4.0)) *
 *        Utils::Constants::ev_per_hartree;
 *    std::cout << distanceInAngstrom << std::fixed << "\t" << ss << "\t" << ssigma << "\t" << sigmasigma << "\t" <<
 *pipi
 *              << std::endl;
 *  };
 *  for (int i = 0; i < 100; ++i) {
 *    printElements(1e-5 + i * 0.025);
 *  }
 *}
 */

TEST_F(AResonanceTest, CanCalculatePenetrationIntegralsAcetylene) {
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::C, Utils::ElementType::H,
                                      Utils::ElementType::H};

  Utils::PositionCollection pos(4, 3);
  pos.row(0) = Utils::Position{0, 0, 0};
  pos.row(1) = Utils::Position{1.202, 0, 0.0};
  pos.row(2) = Utils::Position{-1.063, 0.0, 0.0};
  pos.row(3) = Utils::Position{2.265, 0.0, 0.0};
  pos *= Utils::Constants::bohr_per_angstrom;
  gtos = {};

  auto params = om1();
  params.atomic[1].exponent = 1.24;
  params.atomic[6].exponent = 1.00;
  elParamsOm1 = {};
  elParamsOm1.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, params));
  elParamsOm1.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, params));
  gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);

  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");
  auto finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(elParamsOm1), els, pos);

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParamsOm1));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);
  double ss = 0.0;
  double sx = 0.0;
  double xx = 0.0;
  double yy = 0.0;
  for (int i = 1; i < 4; ++i) { // V(PI)_ss = (V_ss,a + sum_B Z_B * <s_Cs_C | s_B s_B>) * f_ko
    ss += -CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {0, i}, eris, i == 1 ? 4 : 1.0) +
          CoreElectronEvaluation::analyticalCoreElectron(0, i, i == 1 ? 4 : 1.0, basisSet, *finalizer)(0, 0);
    sx += -CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 1, {0, i}, eris, i == 1 ? 4 : 1.0) +
          CoreElectronEvaluation::analyticalCoreElectron(0, i, i == 1 ? 4 : 1.0, basisSet, *finalizer)(0, 1);
    xx += -CoreElectronEvaluation::getSemiempiricalCoreElectron(1, 1, {0, i}, eris, i == 1 ? 4 : 1.0) +
          CoreElectronEvaluation::analyticalCoreElectron(0, i, i == 1 ? 4 : 1.0, basisSet, *finalizer)(1, 1);
    yy += -CoreElectronEvaluation::getSemiempiricalCoreElectron(2, 2, {0, i}, eris, i == 1 ? 4 : 1.0) +
          CoreElectronEvaluation::analyticalCoreElectron(0, i, i == 1 ? 4 : 1.0, basisSet, *finalizer)(2, 2);
  }
  // Give some tolerance as exact structure is not given...
  // From M. Kolb, "Ein neues semiempirisches Verfahren auf Grundlage der NDDO-Naehrung: Entwicklung der Methode,
  // Parametrisierung und Anwendung", 1991, Seite 48, Tabelle 3.2
  EXPECT_THAT((ss)*Utils::Constants::ev_per_hartree, DoubleNear(-3.81, 2e-2));
  EXPECT_THAT((sx)*Utils::Constants::ev_per_hartree, DoubleNear(-3.32, 2e-2));
  EXPECT_THAT((xx)*Utils::Constants::ev_per_hartree, DoubleNear(-8.40, 2e-2));
  EXPECT_THAT((yy)*Utils::Constants::ev_per_hartree, DoubleNear(-1.41, 2e-2));
  ss = 0;
  for (int i = 0; i < 3; ++i) { // V(PI)_ss = (V_ss,a + sum_B Z_B * <s_Cs_C | s_B s_B>) * f_ko
    ss += -CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {3, i}, eris, (i == 0 || i == 1) ? 4 : 1) +
          CoreElectronEvaluation::analyticalCoreElectron(3, i, (i == 0 || i == 1) ? 4 : 1, basisSet, *finalizer)(0, 0);
  }
  // From M. Kolb, "Ein neues semiempirisches Verfahren auf Grundlage der NDDO-Naehrung: Entwicklung der Methode,
  // Parametrisierung und Anwendung", 1991, Seite 48, Tabelle 3.2
  // This is the furthest one, five some tolerance
  EXPECT_THAT((ss)*Utils::Constants::ev_per_hartree, DoubleNear(-3.94, 4e-2));
}

TEST_F(AResonanceTest, CanCalculatePenetrationIntegralsWithOm1Parameters) {
  Utils::ElementTypeCollection els = {Utils::ElementType::C, Utils::ElementType::H, Utils::ElementType::H,
                                      Utils::ElementType::H, Utils::ElementType::H};

  Utils::PositionCollection pos(5, 3);
  pos.row(0) = Utils::Position{0, 0, 0};
  pos.row(1) = Utils::Position{0, 0, 1.087};
  pos.row(2) = Utils::Position{1.02465, 0, -0.36285};
  pos.row(3) = Utils::Position{-0.512325, 0.88737293, -0.36285};
  pos.row(4) = Utils::Position{-0.512325, -0.88737293, -0.36285};
  pos *= Utils::Constants::bohr_per_angstrom;
  gtos = {};

  auto params = om1();
  elParamsOm1 = {};
  elParamsOm1.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, params));
  elParamsOm1.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, params));
  gtos.push_back(elParamsOm1.get(Utils::ElementType::C).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  gtos.push_back(elParamsOm1.get(Utils::ElementType::H).gtos);
  basisSet = BasisSetController({els, pos}, gtos, "om1-3g");
  auto finalizer = std::make_shared<KlopmanFinalizer>(ParametersProcessor::getKlopmanScalingFactors(om1()), els, pos);

  NddoTwoBodiesIntegralsEvaluator eriCalculator(basisSet);
  eriCalculator.set(std::make_shared<OmxOneCenterTwoElectronEvaluator>(els, elParamsOm1));

  eriCalculator.evaluate<Operator::coulomb, Utils::DerivativeOrder::Zero>();
  auto eris = eriCalculator.getResult();
  finalizer->finalize<Utils::DerivativeOrder::Zero>(eris);
  double ss = 0.0;
  double ssH = 0.0;
  double xx = 0.0;
  for (int i = 1; i < 5; ++i) { // V(PI)_ss = (V_ss,a + sum_B Z_B * <s_Cs_C | s_B s_B>) * f_ko
    ss += CoreElectronEvaluation::analyticalCoreElectron(0, i, 1.0, basisSet, *finalizer)(0, 0) -
          CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {0, i}, eris, 1.0);
    xx += CoreElectronEvaluation::analyticalCoreElectron(0, i, 1.0, basisSet, *finalizer)(3, 3) -
          CoreElectronEvaluation::getSemiempiricalCoreElectron(3, 3, {0, i}, eris, 1.0);
    int indexH = i - 1;
    double coreCharge = indexH == 0 ? 4.0 : 1.0;
    ssH += CoreElectronEvaluation::analyticalCoreElectron(4, indexH, coreCharge, basisSet, *finalizer)(0, 0) -
           CoreElectronEvaluation::getSemiempiricalCoreElectron(0, 0, {4, indexH}, eris, coreCharge);
  }
  // From M. Kolb, "Ein neues semiempirisches Verfahren auf Grundlage der NDDO-Naehrung: Entwicklung der Methode,
  // Parametrisierung und Anwendung", 1991, Seite 48, Tabelle 3.2
  EXPECT_THAT(ss * Utils::Constants::ev_per_hartree, DoubleNear(-2.56, 2e-2));
  EXPECT_THAT(xx * Utils::Constants::ev_per_hartree, DoubleNear(-2.55, 2e-2));
  EXPECT_THAT(ssH * Utils::Constants::ev_per_hartree, DoubleNear(-2.80, 2e-2));
}

TEST_F(AResonanceTest, G1G2CalculatedCorrectly) {
  ovlpMatrix =
      Libint::getInstance().compute1BodyIntegrals<Operator::overlap, Utils::DerivativeOrder::Zero>(basisSet).getMatrixXd();
  auto resEval = ResonanceIntegralEvaluator(elements, positions, elParams);
  Eigen::MatrixXd resMat = resEval.getResonanceIntegralMatrix();
  OrthogonalizationCorrection ortEval(elements, basisSet, elParams, ovlpMatrix, resMat);
  auto g1C = elParams.get(Utils::ElementType::C).parameters.orthogonalization.g1;
  auto g2C = elParams.get(Utils::ElementType::C).parameters.orthogonalization.g2;
  auto g1N = elParams.get(Utils::ElementType::N).parameters.orthogonalization.g1;
  auto g2N = elParams.get(Utils::ElementType::N).parameters.orthogonalization.g2;
  auto g1 = ortEval.getG1(0, 1);
  auto g2 = ortEval.getG2(0, 1);
  auto g1inv = ortEval.getG1(1, 0);
  auto g2inv = ortEval.getG2(1, 0);
  EXPECT_DOUBLE_EQ(g1, g1inv);
  EXPECT_DOUBLE_EQ(g2, g2inv);
  EXPECT_DOUBLE_EQ(g1, 0.5 * (g1C + g1N));
  EXPECT_DOUBLE_EQ(g2, 0.5 * (g2C + g2N));
}
} // namespace nddo
} // namespace Sparrow
} // namespace Scine
