/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include <Sparrow/Implementations/Nddo/HeatOfFormationEvaluator.h>
#include <Sparrow/Implementations/Nddo/Omx/OmxMethod.h>
#include <Sparrow/Implementations/Nddo/Omx/ParametersProcessor.h>
#include <Utils/Constants.h>
#include <Utils/IO/ChemicalFileFormats/XyzStreamHandler.h>
#include <Utils/Math/DerivOrderEnum.h>
#include <Utils/UniversalSettings/SettingsNames.h>
#include <gmock/gmock.h>

namespace Scine {
namespace Sparrow {
namespace nddo {

using namespace testing;
using namespace Integrals;

class AHeatOfFormationTest : public Test {
 public:
  Utils::AtomCollection structure;
  Om2Method om2Method;
  Om3Method om3Method;
  double isolHOm3;
  double isolOOm3;
  double isolCOm3;
  static double calculateHeatOfFormationByHandMethane(Core::Calculator& calculator) {
    std::stringstream ssmethane{"5\n\n"
                                "C 0  0  0\n"
                                "H 0  0  1.087\n"
                                "H 1.0246513  0  -0.362848056\n"
                                "H -0.51232565  0.887374056  -0.362848056\n"
                                "H -0.51232565  -0.887374056  -0.362848056\n"};
    calculator.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
    calculator.setStructure(Utils::XyzStreamHandler::read(ssmethane));
    double heatOfFormationH = 52.102;  // kCal/mol
    double heatOfFormationC = 170.890; // kCal/mol
    double energyRestricted = calculator.calculate("").get<Utils::Property::Energy>();

    // Isolated atoms
    Utils::PositionCollection pos(1, 3);
    pos = Utils::Position{0, 0, 0};
    calculator.setStructure({{Utils::ElementType::C}, pos});
    calculator.settings().modifyString(Utils::SettingsNames::spinMode, "unrestricted");
    calculator.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 3);
    double energyC = calculator.calculate("").get<Utils::Property::Energy>();
    calculator.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 2);
    calculator.setStructure({{Utils::ElementType::H}, pos});
    double energyH = calculator.calculate("").get<Utils::Property::Energy>();

    std::cout << "Energy Restricted: " << energyRestricted << std::endl;
    std::cout << "Energy C: " << energyC << std::endl;
    std::cout << "Energy H: " << energyH << std::endl;
    std::cout << "Energy hofH: " << heatOfFormationH << std::endl;
    std::cout << "Energy hofC: " << heatOfFormationC << std::endl;
    double heatOfFormation = (energyRestricted - energyC - 4.0 * energyH) * Utils::Constants::kCalPerMol_per_hartree +
                             heatOfFormationH * 4.0 + heatOfFormationC;
    return heatOfFormation;
  }

 protected:
  void SetUp() final {
    std::stringstream ssmethane{"5\n\n"
                                "C 0  0  0\n"
                                "H 0  0  1.087\n"
                                "H 1.0246513  0  -0.362848056\n"
                                "H -0.51232565  0.887374056  -0.362848056\n"
                                "H -0.51232565  -0.887374056  -0.362848056\n"};
    structure = Utils::XyzStreamHandler::read(ssmethane);
    om2Method.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
    om3Method.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
    om2Method.setStructure(structure);
    om3Method.setStructure(structure);

    Om3Method om3;
    std::stringstream ssH{"1\n\nH 0 0 0"};
    om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
    om3.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 2);
    om3.settings().modifyString(Utils::SettingsNames::spinMode, "unrestricted");
    om3.setStructure(Utils::XyzStreamHandler::read(ssH));
    isolHOm3 = om3.calculate("").get<Utils::Property::Energy>();

    std::stringstream ssC{"1\n\nC 0 0 0"};
    om3.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 3);
    om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
    om3.setStructure(Utils::XyzStreamHandler::read(ssC));
    isolCOm3 = om3.calculate("").get<Utils::Property::Energy>();

    std::stringstream ssO{"1\n\nO 0 0 0"};
    om3.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 3);
    om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
    om3.setStructure(Utils::XyzStreamHandler::read(ssO));
    isolOOm3 = om3.calculate("").get<Utils::Property::Energy>();
  }
};

TEST_F(AHeatOfFormationTest, CanCalculateHeatOfFormationByHandMethaneOm2) {
  EXPECT_NEAR(AHeatOfFormationTest::calculateHeatOfFormationByHandMethane(om2Method), -19.2, 1e-1);
}

TEST_F(AHeatOfFormationTest, CanCalculateHeatOfFormationByHandMethaneOm3) {
  EXPECT_NEAR(AHeatOfFormationTest::calculateHeatOfFormationByHandMethane(om3Method), -18.0, 1e-1);
}

TEST_F(AHeatOfFormationTest, CanCalculateIsolatedAtomEnergyOm2) {
  OmxElementParameters om2Params;
  om2Params.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, om2()));
  om2Params.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om2()));
  auto oneCIntC = om2Params.get(Utils::ElementType::C).parameters.oneCenterTwoElectrons;
  auto oneCIntH = om2Params.get(Utils::ElementType::H).parameters.oneCenterTwoElectrons;
  auto oneCEnC = om2Params.get(Utils::ElementType::C).parameters.oneCenterEnergy;
  auto oneCEnH = om2Params.get(Utils::ElementType::H).parameters.oneCenterEnergy;

  // Isolated atoms
  Utils::PositionCollection pos(1, 3);
  pos = Utils::Position{0, 0, 0};
  om2Method.setStructure({{Utils::ElementType::C}, pos});
  om2Method.settings().modifyString(Utils::SettingsNames::spinMode, "unrestricted");
  om2Method.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 3);
  double energyC = om2Method.calculate("").get<Utils::Property::Energy>();
  om2Method.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 2);
  om2Method.setStructure({{Utils::ElementType::H}, pos});
  double energyH = om2Method.calculate("").get<Utils::Property::Energy>();

  double isolEnergyH = HeatOfFormationEvaluation::calculateIsolatedAtomEnergy(oneCIntH, oneCEnH, {{1, 0, 0, 0}});
  double isolEnergyC = HeatOfFormationEvaluation::calculateIsolatedAtomEnergy(oneCIntC, oneCEnC, {{2, 1, 1, 0}});

  EXPECT_NEAR(isolEnergyH, energyH, 1e-5);
  EXPECT_NEAR(isolEnergyC, energyC, 1e-5);
}

TEST_F(AHeatOfFormationTest, CanCalculateIsolatedAtomEnergyOm3) {
  OmxElementParameters om3Params;
  om3Params.insert(Utils::ElementType::H, ParametersProcessor::processAtomicParameters(Utils::ElementType::H, om3()));
  om3Params.insert(Utils::ElementType::C, ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om3()));
  auto oneCIntC = om3Params.get(Utils::ElementType::C).parameters.oneCenterTwoElectrons;
  auto oneCIntH = om3Params.get(Utils::ElementType::H).parameters.oneCenterTwoElectrons;
  auto oneCEnC = om3Params.get(Utils::ElementType::C).parameters.oneCenterEnergy;
  auto oneCEnH = om3Params.get(Utils::ElementType::H).parameters.oneCenterEnergy;

  // Isolated atoms
  Utils::PositionCollection pos(1, 3);
  pos = Utils::Position{0, 0, 0};
  om3Method.setStructure({{Utils::ElementType::C}, pos});
  om3Method.settings().modifyString(Utils::SettingsNames::spinMode, "unrestricted");
  om3Method.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 3);
  double energyC = om3Method.calculate("").get<Utils::Property::Energy>();
  om3Method.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 2);
  om3Method.setStructure({{Utils::ElementType::H}, pos});
  double energyH = om3Method.calculate("").get<Utils::Property::Energy>();

  double isolEnergyH = HeatOfFormationEvaluation::calculateIsolatedAtomEnergy(oneCIntH, oneCEnH, {{1, 0, 0, 0}});
  double isolEnergyC = HeatOfFormationEvaluation::calculateIsolatedAtomEnergy(oneCIntC, oneCEnC, {{2, 1, 1, 0}});

  EXPECT_NEAR(isolEnergyH, energyH, 1e-5);
  EXPECT_NEAR(isolEnergyC, energyC, 1e-5);
}

TEST_F(AHeatOfFormationTest, CanCalculateHeatOfFormationByHandEthyleneOm2) {
  std::stringstream ssethylene{"6\n\n"
                               "C 0  0  0\n"
                               "C 1.339  0  0\n"
                               "H -0.56471727461  0.92879674835  0\n"
                               "H -0.56471727461  -0.92879674835  0\n"
                               "H 1.90371727461  0.92879674835  0\n"
                               "H 1.90371727461  -0.92879674835  0\n"};
  om2Method.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om2Method.setStructure(Utils::XyzStreamHandler::read(ssethylene));

  double heatOfFormationH = 52.102;  // kCal/mol
  double heatOfFormationC = 170.890; // kCal/mol

  double energyRestricted = om2Method.calculate("").get<Utils::Property::Energy>();

  // Isolated atoms
  Utils::PositionCollection pos(1, 3);
  pos = Utils::Position{0, 0, 0};
  om2Method.setStructure({{Utils::ElementType::C}, pos});
  om2Method.settings().modifyString(Utils::SettingsNames::spinMode, "unrestricted");
  om2Method.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 3);
  double energyC = om2Method.calculate("").get<Utils::Property::Energy>();
  om2Method.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 2);
  om2Method.setStructure({{Utils::ElementType::H}, pos});
  double energyH = om2Method.calculate("").get<Utils::Property::Energy>();

  double heatOfFormation = (energyRestricted - 2.0 * energyC - 4.0 * energyH) * Utils::Constants::kCalPerMol_per_hartree +
                           heatOfFormationH * 4.0 + heatOfFormationC * 2.0;
  EXPECT_NEAR(heatOfFormation, 13.1, 1e-1);
}

TEST_F(AHeatOfFormationTest, CanCalculateHeatOfFormationByHandEthyleneOm3) {
  std::stringstream ssethylene{"6\n\n"
                               "C        0.00000        0.00000        0.00000\n"
                               "C        1.33100        0.00000        0.00000\n"
                               "H       -0.58512        0.92201        0.00000\n"
                               "H       -0.58512       -0.92201       -0.00000\n"
                               "H        1.91612        0.92201        0.00000\n"
                               "H        1.91612       -0.92201        0.00000\n"};

  om3Method.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om3Method.setStructure(Utils::XyzStreamHandler::read(ssethylene));

  double heatOfFormationH = 52.102;  // kCal/mol
  double heatOfFormationC = 170.890; // kCal/mol

  double energyRestricted = om3Method.calculate("").get<Utils::Property::Energy>();

  double heatOfFormation = (energyRestricted - 2.0 * isolCOm3 - 4.0 * isolHOm3) * Utils::Constants::kCalPerMol_per_hartree +
                           heatOfFormationH * 4.0 + heatOfFormationC * 2.0;
  EXPECT_NEAR(heatOfFormation, 13.9, 1e-1);
}

TEST_F(AHeatOfFormationTest, HeatOfFormationKeteneOm3) {
  Om3Method om3;

  std::stringstream ssketene{"5\n\n"
                             "C        0.00000        0.00000        0.00000\n"
                             "H        1.07200        0.00000        0.00000\n"
                             "H       -0.50657        0.94476        0.00000\n"
                             "C       -0.67787       -1.13265       -0.00000\n"
                             "O       -1.29412       -2.16232       -0.00000\n"};
  om3.settings().modifyDouble(Utils::SettingsNames::densityRmsdCriterion, 1e-9);
  om3.settings().modifyInt(Utils::SettingsNames::spinMultiplicity, 1);
  om3.settings().modifyString(Utils::SettingsNames::spinMode, "restricted");
  om3.setStructure(Utils::XyzStreamHandler::read(ssketene));

  double enKetene = om3.calculate("").get<Utils::Property::Energy>();

  double heatOfFormationH = 52.102;  // kCal/mol
  double heatOfFormationC = 170.890; // kCal/mol
  double heatOfFormationO = 59.559;  // kCal/mol

  double stdHof = (enKetene - 2 * isolHOm3 - 2 * isolCOm3 - isolOOm3) * Utils::Constants::kCalPerMol_per_hartree +
                  2 * heatOfFormationH + 2 * heatOfFormationC + heatOfFormationO;
  EXPECT_NEAR(stdHof, -11.0, 1e-1);
}

} // namespace nddo
} // namespace Sparrow
} // namespace Scine
