/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include <Sparrow/Implementations/Nddo/Omx/ParametersProcessor.h>
#include <Sparrow/Implementations/Nddo/Utils/IntegralsEvaluationUtils/oneCenterTwoElectronIntegrals.h>
#include <Sparrow/Implementations/Nddo/Utils/ParameterUtils/AtomicParameters.h>
#include <Sparrow/Implementations/Nddo/Utils/ParameterUtils/PM6DiatomicParameters.h>
#include <Sparrow/Implementations/Nddo/Utils/ParameterUtils/RawParameterProcessor.h>
#include <Utils/Constants.h>
#include <Utils/Geometry/ElementInfo.h>
#include <gmock/gmock.h>
#include <algorithm>

namespace Scine {
namespace Sparrow {

#define SQRT2 1.41421356237309504880

using namespace testing;
using namespace nddo;

class ARawParameterProcessing : public Test {
 public:
  ARawParameterProcessing() : processor(parameters) {
  }

  ParametersOmx om2Params;
  ParametersOmx::Values carbon;

  Parameters parameters;
  Parameters::Atomic iodine;
  Parameters::Diatomic cadmiumChlorine;
  std::unique_ptr<AtomicParameters> atomicPar;
  std::shared_ptr<PM6DiatomicParameters> diatomicPar;
  std::shared_ptr<OneCenterTwoElectronIntegrals> integral;
  RawParameterProcessor processor;
  static auto getOm3Parameters() -> ParametersOmx {
    // OM3 Parameters
    auto om3 = om2();
    // H
    om3.atomic[1].exponent = 1.25906452;
    om3.atomic[1].oneCenterEnergy.s = -12.45828647;
    om3.atomic[1].resonance.beta.s = -3.40064659;
    om3.atomic[1].resonance.alpha.s = 0.06931667;
    om3.atomic[1].orthogonalization.f1 = 0.25393975;
    om3.atomic[1].orthogonalization.f2 = 0.;
    om3.atomic[1].orthogonalization.g1 = 0.35600772;
    om3.atomic[1].orthogonalization.g2 = 0.;
    // C
    om3.atomic[6].exponent = 1.27811536;
    om3.atomic[6].oneCenterEnergy.s = -50.55997310;
    om3.atomic[6].oneCenterEnergy.p = -39.60463506;
    om3.atomic[6].resonance.beta.s = -7.15007507;
    om3.atomic[6].resonance.beta.pSigma = -4.00965991;
    om3.atomic[6].resonance.beta.pPi = -5.63958651;
    om3.atomic[6].resonance.alpha.s = 0.09197146;
    om3.atomic[6].resonance.alpha.pSigma = 0.05274021;
    om3.atomic[6].resonance.alpha.pPi = 0.09864674;
    om3.atomic[6].resonance.betaXH.s = -6.19914817;
    om3.atomic[6].resonance.betaXH.pSigma = -4.23218526;
    om3.atomic[6].resonance.alphaXH.s = 0.10023679;
    om3.atomic[6].resonance.alphaXH.pSigma = 0.05492720;
    om3.atomic[6].orthogonalization.f1 = 0.41151269;
    om3.atomic[6].orthogonalization.f2 = 0.;
    om3.atomic[6].orthogonalization.g1 = 0.10398816;
    om3.atomic[6].orthogonalization.g2 = 0.;
    om3.atomic[6].pseudopotential.exponent = 5.700000;
    om3.atomic[6].pseudopotential.alpha = 0.15323932;
    om3.atomic[6].pseudopotential.beta = -22.48815939;
    om3.atomic[6].pseudopotential.faa = -283.81699000;
    // N
    om3.atomic[7].exponent = 1.30965521;
    om3.atomic[7].oneCenterEnergy.s = -75.98413465;
    om3.atomic[7].oneCenterEnergy.p = -57.38630489;
    om3.atomic[7].resonance.beta.s = -13.42485887;
    om3.atomic[7].resonance.beta.pSigma = -5.69143961;
    om3.atomic[7].resonance.beta.pPi = -8.25767437;
    om3.atomic[7].resonance.alpha.s = 0.09461210;
    om3.atomic[7].resonance.alpha.pSigma = 0.06941595;
    om3.atomic[7].resonance.alpha.pPi = 0.10511596;
    om3.atomic[7].resonance.betaXH.s = -11.40440049;
    om3.atomic[7].resonance.betaXH.pSigma = -7.87479008;
    om3.atomic[7].resonance.alphaXH.s = 0.11356707;
    om3.atomic[7].resonance.alphaXH.pSigma = 0.09244486;
    om3.atomic[7].orthogonalization.f1 = 0.58223498;
    om3.atomic[7].orthogonalization.f2 = 0.;
    om3.atomic[7].orthogonalization.g1 = 0.05928617;
    om3.atomic[7].orthogonalization.g2 = 0.;
    om3.atomic[7].pseudopotential.exponent = 6.73673665;
    om3.atomic[7].pseudopotential.alpha = 0.15915338;
    om3.atomic[7].pseudopotential.beta = -22.78185605;
    om3.atomic[7].pseudopotential.faa = -380.94261410;
    // O
    om3.atomic[8].exponent = 1.20838191;
    om3.atomic[8].oneCenterEnergy.s = -105.79319826;
    om3.atomic[8].oneCenterEnergy.p = -78.90502490;
    om3.atomic[8].resonance.beta.s = -14.42839639;
    om3.atomic[8].resonance.beta.pSigma = -8.77114206;
    om3.atomic[8].resonance.beta.pPi = -12.94995697;
    om3.atomic[8].resonance.alpha.s = 0.12962541;
    om3.atomic[8].resonance.alpha.pSigma = 0.09275135;
    om3.atomic[8].resonance.alpha.pPi = 0.16086067;
    om3.atomic[8].resonance.betaXH.s = -13.56403003;
    om3.atomic[8].resonance.betaXH.pSigma = -9.42200507;
    om3.atomic[8].resonance.alphaXH.s = 0.14516313;
    om3.atomic[8].resonance.alphaXH.pSigma = 0.10981869;
    om3.atomic[8].orthogonalization.f1 = 0.55266327;
    om3.atomic[8].orthogonalization.f2 = 0.;
    om3.atomic[8].orthogonalization.g1 = 0.06226814;
    om3.atomic[8].orthogonalization.g2 = 0.;
    om3.atomic[8].pseudopotential.exponent = 7.74239776;
    om3.atomic[8].pseudopotential.alpha = 0.18281098;
    om3.atomic[8].pseudopotential.beta = -26.13251784;
    om3.atomic[8].pseudopotential.faa = -512.5190000;
    om3.atomic[9].oneCenterEnergy.s = -120.65477058;
    om3.atomic[9].oneCenterEnergy.p = -107.50304443;
    om3.atomic[9].exponent = 1.20564838;
    om3.atomic[9].resonance.beta.s = -6.19918959;
    om3.atomic[9].resonance.beta.pSigma = -13.82075048;
    om3.atomic[9].resonance.beta.pPi = -18.96710976;
    om3.atomic[9].resonance.betaXH.s = -8.06286638;
    om3.atomic[9].resonance.betaXH.pSigma = -13.92380910;
    om3.atomic[9].resonance.alpha.s = 0.31128286;
    om3.atomic[9].resonance.alpha.pSigma = 0.12448702;
    om3.atomic[9].resonance.alpha.pPi = 0.21596283;
    om3.atomic[9].resonance.alphaXH.s = 0.32623663;
    om3.atomic[9].resonance.alphaXH.pSigma = 0.15497349;
    om3.atomic[9].pseudopotential.exponent = 8.70367110;
    om3.atomic[9].pseudopotential.alpha = 0.00010713;
    om3.atomic[9].pseudopotential.beta = -10.62649349;
    om3.atomic[9].pseudopotential.faa = -685.41999336;
    om3.atomic[9].orthogonalization.f1 = 1.03522357;
    om3.atomic[9].orthogonalization.g1 = 0.14035077;
    return om3;
  }

  static auto getOm1Parameters() -> ParametersOmx {
    // OM1 parameters
    auto paramsOm1 = om2();
    paramsOm1.atomic[1].oneCenterEnergy.s = -12.83851861;
    paramsOm1.atomic[1].exponent = 1.20948930;
    paramsOm1.atomic[1].resonance.beta.s = -4.89312435;
    paramsOm1.atomic[1].resonance.alpha.s = 0.09653898;
    paramsOm1.atomic[1].orthogonalization.f1 = 0.54128873;
    paramsOm1.atomic[1].orthogonalization.f2 = 0.84668969;
    paramsOm1.atomic[6].oneCenterEnergy.s = -50.15945;
    paramsOm1.atomic[6].oneCenterEnergy.p = -38.76257345;
    paramsOm1.atomic[6].exponent = 1.13551142;
    paramsOm1.atomic[6].resonance.beta.s = -7.58632270;
    paramsOm1.atomic[6].resonance.beta.pSigma = -4.49894163;
    paramsOm1.atomic[6].resonance.beta.pPi = -5.91210138;
    paramsOm1.atomic[6].resonance.alpha.s = 0.09325105;
    paramsOm1.atomic[6].resonance.alpha.pSigma = 0.05398748;
    paramsOm1.atomic[6].resonance.alpha.pPi = 0.10477244;
    paramsOm1.atomic[6].resonance.alphaXH = paramsOm1.atomic[6].resonance.alpha;
    paramsOm1.atomic[6].resonance.betaXH = paramsOm1.atomic[6].resonance.beta;
    paramsOm1.atomic[6].orthogonalization.f1 = 0.50383851;
    paramsOm1.atomic[6].orthogonalization.f2 = 0.66944409;
    paramsOm1.atomic[7].oneCenterEnergy.s = -71.33505463;
    paramsOm1.atomic[7].oneCenterEnergy.p = -56.58315267;
    paramsOm1.atomic[7].exponent = 1.16081665;
    paramsOm1.atomic[7].resonance.beta.s = -12.00586167;
    paramsOm1.atomic[7].resonance.beta.pSigma = -9.64950408;
    paramsOm1.atomic[7].resonance.beta.pPi = -10.16405908;
    paramsOm1.atomic[7].resonance.betaXH.s = -8.08332477;
    paramsOm1.atomic[7].resonance.betaXH.pSigma = -11.48923075;
    paramsOm1.atomic[7].resonance.alpha.s = 0.10185884;
    paramsOm1.atomic[7].resonance.alpha.pSigma = 0.08540515;
    paramsOm1.atomic[7].resonance.alpha.pPi = 0.14350678;
    paramsOm1.atomic[7].resonance.alphaXH.s = 0.07514671;
    paramsOm1.atomic[7].resonance.alphaXH.pSigma = 0.11556228;
    paramsOm1.atomic[7].orthogonalization.f1 = 0.63476395;
    paramsOm1.atomic[7].orthogonalization.f2 = 0.31135759;
    paramsOm1.atomic[8].oneCenterEnergy.s = -93.04158571;
    paramsOm1.atomic[8].oneCenterEnergy.p = -77.59792792;
    paramsOm1.atomic[8].exponent = 1.10190209;
    paramsOm1.atomic[8].resonance.beta.s = -6.22223757;
    paramsOm1.atomic[8].resonance.beta.pSigma = -9.94028730;
    paramsOm1.atomic[8].resonance.beta.pPi = -11.29342651;
    paramsOm1.atomic[8].resonance.betaXH.s = -6.45960578;
    paramsOm1.atomic[8].resonance.betaXH.pSigma = -12.47451386;
    paramsOm1.atomic[8].resonance.alpha.s = 0.10891616;
    paramsOm1.atomic[8].resonance.alpha.pSigma = 0.09666556;
    paramsOm1.atomic[8].resonance.alpha.pPi = 0.15255321;
    paramsOm1.atomic[8].resonance.alphaXH.s = 0.07953217;
    paramsOm1.atomic[8].resonance.alphaXH.pSigma = 0.12908958;
    paramsOm1.atomic[8].orthogonalization.f1 = 0.68193417;
    paramsOm1.atomic[8].orthogonalization.f2 = 0.47652748;
    paramsOm1.atomic[9].oneCenterEnergy.s = -121.69518463;
    paramsOm1.atomic[9].oneCenterEnergy.p = -106.37310634;
    paramsOm1.atomic[9].exponent = 1.16498140;
    paramsOm1.atomic[9].resonance.beta.s = -5.73558167;
    paramsOm1.atomic[9].resonance.beta.pSigma = -16.36168108;
    paramsOm1.atomic[9].resonance.beta.pPi = -17.22481680;
    paramsOm1.atomic[9].resonance.betaXH.s = -5.73558167;
    paramsOm1.atomic[9].resonance.betaXH.pSigma = -16.36168108;
    paramsOm1.atomic[9].resonance.alpha.s = 0.19374572;
    paramsOm1.atomic[9].resonance.alpha.pSigma = 0.13034754;
    paramsOm1.atomic[9].resonance.alpha.pPi = 0.22033729;
    paramsOm1.atomic[9].resonance.alphaXH.s = 0.19374572;
    paramsOm1.atomic[9].resonance.alphaXH.pSigma = 0.13034754;
    paramsOm1.atomic[9].orthogonalization.f1 = 1.19938976;
    paramsOm1.atomic[9].orthogonalization.f2 = 0.49484369;
    return paramsOm1;
  }

  void SetUp() override {
    // Those are the parameters for iodine
    iodine.pack.oneCenterEnergy = {-59.9732, -56.4598, -28.8226};
    iodine.pack.beta = {-30.5225, -5.94212, -7.67611};
    iodine.pack.gss = 7.23476;
    iodine.pack.gpp = 9.87747;
    iodine.pack.gsp = 9.15441;
    iodine.pack.gp2 = 8.03592;
    iodine.pack.hsp = 5.00422;
    iodine.pack.orbitalExponent = {4.49865, 1.91707, 1.87518};
    iodine.pack.internalExponent = {9.13524, 6.88819, 3.79152};
    iodine.pack.pcore = 1.8806;
    iodine.pack.f0sd = 20.495;
    iodine.pack.g2sd = 2.16014;
    // iodine.grepa = -0.035519;
    // iodine.grepb = 1.74439;
    // iodine.grepc = 1.22384;

    // Parameters Carbon OM2
    carbon.exponent = 1.42036892;
    carbon.oneCenterEnergy.s = -51.65550844;
    carbon.oneCenterEnergy.p = -39.74369825;

    carbon.resonance.beta.s = -7.21406021;
    carbon.resonance.beta.pSigma = -4.14394503;
    carbon.resonance.beta.pPi = -5.97107657;

    carbon.resonance.alpha.s = 0.09045297;
    carbon.resonance.alpha.pSigma = 0.05452192;
    carbon.resonance.alpha.pPi = 0.10204903;

    carbon.resonance.betaXH.s = -6.30164062;
    carbon.resonance.betaXH.pSigma = -4.04444703;

    carbon.resonance.alphaXH.s = 0.09668329;
    carbon.resonance.alphaXH.pSigma = 0.05283694;

    carbon.orthogonalization.f1 = 0.49949211;
    carbon.orthogonalization.f2 = 0.72261226;
    carbon.orthogonalization.g1 = 0.21284361;
    carbon.orthogonalization.g2 = 0.99250289;

    carbon.pseudopotential.exponent = 5.16802668;
    carbon.pseudopotential.faa = -305.68646337;
    carbon.pseudopotential.beta = -9.07185084;
    carbon.pseudopotential.alpha = 0.16985745;

    carbon.oneCenterTwoElectrons.gss = 12.23;
    carbon.oneCenterTwoElectrons.gpp = 11.08;
    carbon.oneCenterTwoElectrons.gsp = 11.47;
    carbon.oneCenterTwoElectrons.gp2 = 9.84;
    carbon.oneCenterTwoElectrons.hsp = 2.43;

    // Those are parameters for the Cd-Cl pair:
    cadmiumChlorine.exponent = 0.943547;
    cadmiumChlorine.factor = 0.140424;

    parameters.atomic.emplace(Utils::ElementInfo::Z(Utils::ElementType::I), iodine);
    parameters.diatomic.emplace(Parameters::key(Utils::ElementType::Cd, Utils::ElementType::Cl), cadmiumChlorine);
    auto par = processor.processAtomicParameters(Utils::ElementType::I);
    atomicPar = std::move(par.first);
    integral = std::move(par.second);
    diatomicPar = processor.runtimeDiatomicParameters(Utils::ElementType::Cd, Utils::ElementType::Cl);
  }
};

TEST_F(ARawParameterProcessing, SetsBetaValuesCorrectly) {
  ASSERT_THAT(atomicPar->betaS(), DoubleEq(iodine.pack.beta.s * Utils::Constants::hartree_per_ev));
  ASSERT_THAT(atomicPar->betaP(), DoubleEq(iodine.pack.beta.p * Utils::Constants::hartree_per_ev));
  ASSERT_THAT(atomicPar->betaD(), DoubleEq(iodine.pack.beta.d * Utils::Constants::hartree_per_ev));
}

TEST_F(ARawParameterProcessing, SetsUValuesCorrectly) {
  ASSERT_THAT(atomicPar->Uss(), DoubleEq(iodine.pack.oneCenterEnergy.s * Utils::Constants::hartree_per_ev));
  ASSERT_THAT(atomicPar->Upp(), DoubleEq(iodine.pack.oneCenterEnergy.p * Utils::Constants::hartree_per_ev));
  ASSERT_THAT(atomicPar->Udd(), DoubleEq(iodine.pack.oneCenterEnergy.d * Utils::Constants::hartree_per_ev));
}

TEST_F(ARawParameterProcessing, SetsChargeSeparationsCorrectly) {
  auto d = atomicPar->chargeSeparations();

  ASSERT_THAT(d.get(multipole::MultipolePair::sp1), DoubleNear(0.3746525, 1e-4));
  ASSERT_THAT(d.get(multipole::MultipolePair::pp2), DoubleNear(1.89517161 / SQRT2, 1e-5));
  ASSERT_THAT(d.get(multipole::MultipolePair::sd2), DoubleNear(0.77747018 / SQRT2, 1e-5));
  ASSERT_THAT(d.get(multipole::MultipolePair::pd1), DoubleNear(1.29634171, 1e-5));
  ASSERT_THAT(d.get(multipole::MultipolePair::dd2), DoubleNear(1.63749938 / SQRT2, 1e-5));
}

TEST_F(ARawParameterProcessing, GtoExpansionHasBeenPerformed) {
  ASSERT_TRUE(atomicPar->GTOs().s);
  ASSERT_TRUE(atomicPar->GTOs().p);
  ASSERT_TRUE(atomicPar->GTOs().d);
}

/*TEST_F(ARawParameterProcessing, SetsGaussianRepulsionParameters) {
  auto g = atomicPar->getGaussianRepulsionParameters();

  ASSERT_TRUE(atomicPar->hasGaussianRepulsionParameters());
  ASSERT_THAT(std::get<0>(g), DoubleEq(iodine.grepa * Utils::Constants::bohr_per_angstrom));
  ASSERT_THAT(std::get<1>(g), DoubleEq(iodine.grepb * Utils::Constants::angstrom_per_bohr *
Utils::Constants::angstrom_per_bohr)); ASSERT_THAT(std::get<2>(g), DoubleEq(iodine.grepc *
Utils::Constants::bohr_per_angstrom));
}*/

TEST_F(ARawParameterProcessing, SetsDiatomicExponentCorrectly) {
  ASSERT_THAT(diatomicPar->alpha(), Eq(cadmiumChlorine.exponent * Utils::Constants::angstrom_per_bohr));
}

TEST_F(ARawParameterProcessing, SetsDiatomicFactorCorrectly) {
  ASSERT_THAT(diatomicPar->x(), Eq(cadmiumChlorine.factor));
}

TEST_F(ARawParameterProcessing, SetsPCoreCorrectly) {
  ASSERT_THAT(atomicPar->pCore(), DoubleEq(iodine.pack.pcore));
}

TEST_F(ARawParameterProcessing, Returns1c2eIntegrals) {
  ASSERT_THAT(integral->getNumberIntegrals(), Eq(58));
  ASSERT_THAT(integral->get(0) * Utils::Constants::ev_per_hartree, DoubleNear(7.23480, 1e-4));
  ASSERT_THAT(integral->get(1) * Utils::Constants::ev_per_hartree, DoubleNear(9.15440, 1e-4));
  ASSERT_THAT(integral->get(2) * Utils::Constants::ev_per_hartree, DoubleNear(5.004200, 1e-4));
  ASSERT_THAT(integral->get(3), DoubleNear(9.87750 / Utils::Constants::ev_per_hartree, 1e-4));
  ASSERT_THAT(integral->get(4), DoubleNear(8.03590 / Utils::Constants::ev_per_hartree, 1e-4));
  ASSERT_THAT(integral->get(5), DoubleNear(0.920800 / Utils::Constants::ev_per_hartree, 1e-4));
  ASSERT_THAT(integral->get(11), DoubleNear(-0.815800 / Utils::Constants::ev_per_hartree, 1e-4));
  ASSERT_THAT(integral->get(24), DoubleNear(0.7065 / Utils::Constants::ev_per_hartree, 1e-4));
  ASSERT_THAT(integral->get(25), DoubleNear(-0.7065 / Utils::Constants::ev_per_hartree, 1e-4));
  ASSERT_THAT(integral->get(27), DoubleNear(0.815800 / Utils::Constants::ev_per_hartree, 1e-4));
  ASSERT_THAT(integral->get(30), DoubleNear(18.3654 / Utils::Constants::ev_per_hartree, 1e-4));
  ASSERT_THAT(integral->get(41), DoubleNear(17.2884 / Utils::Constants::ev_per_hartree, 1e-4));
  ASSERT_THAT(integral->get(52), DoubleNear(-0.962700 / Utils::Constants::ev_per_hartree, 1e-4));
  ASSERT_THAT(integral->get(56), DoubleNear(0.962700 / Utils::Constants::ev_per_hartree, 1e-4));
}

TEST_F(ARawParameterProcessing, AssignsInCodeOM2ParametersCorrectly) {
  om2Params = om2();
  ASSERT_EQ(om2Params.atomic.size(), 5);
  ASSERT_NE(om2Params.atomic.find(6), om2Params.atomic.end());

  auto carbtest = om2Params.atomic[6];
  const auto& rt = carbtest.resonance;
  const auto& r = carbon.resonance;
  EXPECT_DOUBLE_EQ(rt.alpha.s, r.alpha.s);
  EXPECT_DOUBLE_EQ(rt.alpha.pSigma, r.alpha.pSigma);
  EXPECT_DOUBLE_EQ(rt.alpha.pPi, r.alpha.pPi);
  EXPECT_DOUBLE_EQ(rt.beta.s, r.beta.s);
  EXPECT_DOUBLE_EQ(rt.beta.pSigma, r.beta.pSigma);
  EXPECT_DOUBLE_EQ(rt.beta.pPi, r.beta.pPi);
  EXPECT_DOUBLE_EQ(rt.alphaXH.s, r.alphaXH.s);
  EXPECT_DOUBLE_EQ(rt.alphaXH.pSigma, r.alphaXH.pSigma);
  EXPECT_DOUBLE_EQ(rt.alphaXH.pPi, r.alphaXH.pPi);
  EXPECT_DOUBLE_EQ(rt.betaXH.s, r.betaXH.s);
  EXPECT_DOUBLE_EQ(rt.betaXH.pSigma, r.betaXH.pSigma);
  EXPECT_DOUBLE_EQ(rt.betaXH.pPi, r.betaXH.pPi);

  const auto& ot = carbtest.orthogonalization;
  const auto& o = carbon.orthogonalization;
  EXPECT_DOUBLE_EQ(ot.f1, o.f1);
  EXPECT_DOUBLE_EQ(ot.f2, o.f2);
  EXPECT_DOUBLE_EQ(ot.g1, o.g1);
  EXPECT_DOUBLE_EQ(ot.g2, o.g2);

  const auto& pt = carbtest.pseudopotential;
  const auto& p = carbon.pseudopotential;
  EXPECT_DOUBLE_EQ(pt.exponent, p.exponent);
  EXPECT_DOUBLE_EQ(pt.faa, p.faa);
  EXPECT_DOUBLE_EQ(pt.alpha, p.alpha);
  EXPECT_DOUBLE_EQ(pt.beta, p.beta);

  EXPECT_DOUBLE_EQ(carbtest.oneCenterEnergy.s, carbon.oneCenterEnergy.s);
  EXPECT_DOUBLE_EQ(carbtest.oneCenterEnergy.p, carbon.oneCenterEnergy.p);
  EXPECT_DOUBLE_EQ(carbtest.exponent, carbon.exponent);
  EXPECT_DOUBLE_EQ(carbtest.oneCenterTwoElectrons.gss, 12.23);
  EXPECT_DOUBLE_EQ(carbtest.oneCenterTwoElectrons.gpp, 11.08);
  EXPECT_DOUBLE_EQ(carbtest.oneCenterTwoElectrons.gsp, 11.47);
  EXPECT_DOUBLE_EQ(carbtest.oneCenterTwoElectrons.gp2, 9.84);
  EXPECT_DOUBLE_EQ(carbtest.oneCenterTwoElectrons.hsp, 2.43);
}

TEST_F(ARawParameterProcessing, ProcessesInCodeOM2ParametersCorrectly) {
  auto carbonParams = ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om2());
  ASSERT_EQ(carbonParams.nAOs, 4);
  ASSERT_EQ(carbonParams.element, Utils::ElementType::C);

  const auto& rt = carbonParams.parameters.resonance;
  const auto& r = carbon.resonance;
  EXPECT_DOUBLE_EQ(rt.alpha.s, r.alpha.s);
  EXPECT_DOUBLE_EQ(rt.alpha.pSigma, r.alpha.pSigma);
  EXPECT_DOUBLE_EQ(rt.alpha.pPi, r.alpha.pPi);
  EXPECT_DOUBLE_EQ(rt.beta.s, r.beta.s * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(rt.beta.pSigma, r.beta.pSigma * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(rt.beta.pPi, r.beta.pPi * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(rt.alphaXH.s, r.alphaXH.s);
  EXPECT_DOUBLE_EQ(rt.alphaXH.pSigma, r.alphaXH.pSigma);
  EXPECT_DOUBLE_EQ(rt.alphaXH.pPi, r.alphaXH.pPi);
  EXPECT_DOUBLE_EQ(rt.betaXH.s, r.betaXH.s * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(rt.betaXH.pSigma, r.betaXH.pSigma * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(rt.betaXH.pPi, r.betaXH.pPi * Utils::Constants::hartree_per_ev);

  const auto& ot = carbonParams.parameters.orthogonalization;
  const auto& o = carbon.orthogonalization;
  EXPECT_DOUBLE_EQ(ot.f1, o.f1);
  EXPECT_DOUBLE_EQ(ot.f2, o.f2);
  EXPECT_DOUBLE_EQ(ot.g1, o.g1);
  EXPECT_DOUBLE_EQ(ot.g2, o.g2);

  const auto& pt = carbonParams.parameters.pseudopotential;
  const auto& p = carbon.pseudopotential;
  EXPECT_DOUBLE_EQ(pt.exponent, p.exponent);
  EXPECT_DOUBLE_EQ(pt.faa, p.faa * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(pt.alpha, p.alpha);
  EXPECT_DOUBLE_EQ(pt.beta, p.beta * Utils::Constants::hartree_per_ev);

  EXPECT_DOUBLE_EQ(carbonParams.parameters.oneCenterEnergy.s, carbon.oneCenterEnergy.s * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(carbonParams.parameters.oneCenterEnergy.p, carbon.oneCenterEnergy.p * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(carbonParams.parameters.exponent, carbon.exponent);
  EXPECT_DOUBLE_EQ(carbonParams.parameters.oneCenterTwoElectrons.gss, 12.23 * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(carbonParams.parameters.oneCenterTwoElectrons.gpp, 11.08 * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(carbonParams.parameters.oneCenterTwoElectrons.gsp, 11.47 * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(carbonParams.parameters.oneCenterTwoElectrons.gp2, 9.84 * Utils::Constants::hartree_per_ev);
  EXPECT_DOUBLE_EQ(carbonParams.parameters.oneCenterTwoElectrons.hsp, 2.43 * Utils::Constants::hartree_per_ev);
}

TEST_F(ARawParameterProcessing, CanReadInBasisFile) {
  auto basis = BasisSet::ecp3g();

  ASSERT_EQ(basis.size(), 9);
  auto carbonGtf = basis[6];
  EXPECT_TRUE(carbonGtf.s);
  EXPECT_TRUE(carbonGtf.p);
  // OMx methods do not have d orbitals
  EXPECT_FALSE(carbonGtf.d);

  const auto& s = *carbonGtf.s;
  ASSERT_EQ(s.gtfs.size(), 3);
  ASSERT_DOUBLE_EQ(s.gtfs[0].exponent, 0.2644860000E+01);
  ASSERT_DOUBLE_EQ(s.gtfs[1].exponent, 0.5421500000E+00);
  ASSERT_DOUBLE_EQ(s.gtfs[2].exponent, 0.1446600000E+00);
  ASSERT_DOUBLE_EQ(s.gtfs[0].coefficient, -0.191880000E+00);
  ASSERT_DOUBLE_EQ(s.gtfs[1].coefficient, 0.616280000E+00);
  ASSERT_DOUBLE_EQ(s.gtfs[2].coefficient, 0.548960000E+00);

  const auto& p = *carbonGtf.p;
  ASSERT_EQ(p.gtfs.size(), 3);
  ASSERT_DOUBLE_EQ(p.gtfs[0].exponent, 0.2644860000E+01);
  ASSERT_DOUBLE_EQ(p.gtfs[1].exponent, 0.5421500000E+00);
  ASSERT_DOUBLE_EQ(p.gtfs[2].exponent, 0.1446600000E+00);
  ASSERT_DOUBLE_EQ(p.gtfs[0].coefficient, 0.202590000E+00);
  ASSERT_DOUBLE_EQ(p.gtfs[1].coefficient, 0.558300000E+00);
  ASSERT_DOUBLE_EQ(p.gtfs[2].coefficient, 0.455140000E+00);
}

TEST_F(ARawParameterProcessing, CanProcessOM1Parameters) {
  auto paramsOm1 = om1();
  auto paramsOm2 = om2();
  auto paramsOm3 = om3();
  auto run_asserts_om1 = [&](auto el1, auto el2) {
    EXPECT_DOUBLE_EQ(el1.exponent, el2.exponent);
    EXPECT_DOUBLE_EQ(el1.oneCenterEnergy.s, el2.oneCenterEnergy.s);
    EXPECT_DOUBLE_EQ(el1.oneCenterEnergy.p, el2.oneCenterEnergy.p);
    EXPECT_DOUBLE_EQ(el1.oneCenterTwoElectrons.gp2, el2.oneCenterTwoElectrons.gp2);
    EXPECT_DOUBLE_EQ(el1.oneCenterTwoElectrons.gss, el2.oneCenterTwoElectrons.gss);
    EXPECT_DOUBLE_EQ(el1.oneCenterTwoElectrons.gpp, el2.oneCenterTwoElectrons.gpp);
    EXPECT_DOUBLE_EQ(el1.oneCenterTwoElectrons.gsp, el2.oneCenterTwoElectrons.gsp);
    EXPECT_DOUBLE_EQ(el1.orthogonalization.f1, el2.orthogonalization.f1);
    EXPECT_DOUBLE_EQ(el1.orthogonalization.f2, el2.orthogonalization.f2);
    EXPECT_DOUBLE_EQ(el1.orthogonalization.g1, 0);
    EXPECT_DOUBLE_EQ(el1.orthogonalization.g2, 0.);
    EXPECT_DOUBLE_EQ(el1.pseudopotential.alpha, 0.);
    EXPECT_DOUBLE_EQ(el1.pseudopotential.beta, 0.);
    EXPECT_DOUBLE_EQ(el1.pseudopotential.exponent, 0.);
    EXPECT_DOUBLE_EQ(el1.pseudopotential.faa, 0.);
    EXPECT_DOUBLE_EQ(el1.resonance.alpha.s, el2.resonance.alpha.s);
    EXPECT_DOUBLE_EQ(el1.resonance.alpha.pSigma, el2.resonance.alpha.pSigma);
    EXPECT_DOUBLE_EQ(el1.resonance.alpha.pPi, el2.resonance.alpha.pPi);
    EXPECT_DOUBLE_EQ(el1.resonance.beta.s, el2.resonance.beta.s);
    EXPECT_DOUBLE_EQ(el1.resonance.beta.pSigma, el2.resonance.beta.pSigma);
    EXPECT_DOUBLE_EQ(el1.resonance.beta.pPi, el2.resonance.beta.pPi);
    EXPECT_DOUBLE_EQ(el1.resonance.alphaXH.s, el2.resonance.alphaXH.s);
    EXPECT_DOUBLE_EQ(el1.resonance.alphaXH.pSigma, el2.resonance.alphaXH.pSigma);
    EXPECT_DOUBLE_EQ(el1.resonance.betaXH.s, el2.resonance.betaXH.s);
    EXPECT_DOUBLE_EQ(el1.resonance.betaXH.pSigma, el2.resonance.betaXH.pSigma);
  };
  auto run_asserts_om3 = [&](auto el1, auto el2) {
    EXPECT_DOUBLE_EQ(el1.exponent, el2.exponent);
    EXPECT_DOUBLE_EQ(el1.oneCenterEnergy.s, el2.oneCenterEnergy.s);
    EXPECT_DOUBLE_EQ(el1.oneCenterEnergy.p, el2.oneCenterEnergy.p);
    EXPECT_DOUBLE_EQ(el1.oneCenterTwoElectrons.gp2, el2.oneCenterTwoElectrons.gp2);
    EXPECT_DOUBLE_EQ(el1.oneCenterTwoElectrons.gss, el2.oneCenterTwoElectrons.gss);
    EXPECT_DOUBLE_EQ(el1.oneCenterTwoElectrons.gpp, el2.oneCenterTwoElectrons.gpp);
    EXPECT_DOUBLE_EQ(el1.oneCenterTwoElectrons.gsp, el2.oneCenterTwoElectrons.gsp);
    EXPECT_DOUBLE_EQ(el1.orthogonalization.f1, el2.orthogonalization.f1);
    EXPECT_DOUBLE_EQ(el1.orthogonalization.f2, 0);
    EXPECT_DOUBLE_EQ(el1.orthogonalization.g1, el2.orthogonalization.g1);
    EXPECT_DOUBLE_EQ(el1.orthogonalization.g2, 0);
    EXPECT_DOUBLE_EQ(el1.pseudopotential.alpha, el2.pseudopotential.alpha);
    EXPECT_DOUBLE_EQ(el1.pseudopotential.beta, el2.pseudopotential.beta);
    EXPECT_DOUBLE_EQ(el1.pseudopotential.exponent, el2.pseudopotential.exponent);
    EXPECT_DOUBLE_EQ(el1.pseudopotential.faa, el2.pseudopotential.faa);
    EXPECT_DOUBLE_EQ(el1.resonance.alpha.s, el2.resonance.alpha.s);
    EXPECT_DOUBLE_EQ(el1.resonance.alpha.pSigma, el2.resonance.alpha.pSigma);
    EXPECT_DOUBLE_EQ(el1.resonance.alpha.pPi, el2.resonance.alpha.pPi);
    EXPECT_DOUBLE_EQ(el1.resonance.beta.s, el2.resonance.beta.s);
    EXPECT_DOUBLE_EQ(el1.resonance.beta.pSigma, el2.resonance.beta.pSigma);
    EXPECT_DOUBLE_EQ(el1.resonance.beta.pPi, el2.resonance.beta.pPi);
    EXPECT_DOUBLE_EQ(el1.resonance.alphaXH.s, el2.resonance.alphaXH.s);
    EXPECT_DOUBLE_EQ(el1.resonance.alphaXH.pSigma, el2.resonance.alphaXH.pSigma);
    EXPECT_DOUBLE_EQ(el1.resonance.betaXH.s, el2.resonance.betaXH.s);
    EXPECT_DOUBLE_EQ(el1.resonance.betaXH.pSigma, el2.resonance.betaXH.pSigma);
  };
  std::vector<int> indices = {1, 6, 7, 8, 9};
  for (int index : indices) {
    run_asserts_om1(paramsOm1.atomic[index], ARawParameterProcessing::getOm1Parameters().atomic[index]);
    run_asserts_om3(paramsOm3.atomic[index], ARawParameterProcessing::getOm3Parameters().atomic[index]);
  }
}

TEST_F(ARawParameterProcessing, CanProcessBasisSetOM2) {
  auto carbonParams = ParametersProcessor::processAtomicParameters(Utils::ElementType::C, om2());

  auto carbonGtf = carbonParams.gtos;

  double scaling = carbonParams.parameters.exponent * carbonParams.parameters.exponent;

  EXPECT_TRUE(carbonGtf.s);
  EXPECT_TRUE(carbonGtf.p);
  // OMx methods do not have d orbitals
  EXPECT_FALSE(carbonGtf.d);

  const auto& s = *carbonGtf.s;
  ASSERT_EQ(s.gtfs.size(), 3);
  ASSERT_DOUBLE_EQ(s.gtfs[0].exponent, 0.2644860000E+01 * scaling);
  ASSERT_DOUBLE_EQ(s.gtfs[1].exponent, 0.5421500000E+00 * scaling);
  ASSERT_DOUBLE_EQ(s.gtfs[2].exponent, 0.1446600000E+00 * scaling);
  ASSERT_DOUBLE_EQ(s.gtfs[0].coefficient, -0.191880000E+00);
  ASSERT_DOUBLE_EQ(s.gtfs[1].coefficient, 0.616280000E+00);
  ASSERT_DOUBLE_EQ(s.gtfs[2].coefficient, 0.548960000E+00);

  const auto& p = *carbonGtf.p;
  ASSERT_EQ(p.gtfs.size(), 3);
  ASSERT_DOUBLE_EQ(p.gtfs[0].exponent, 0.2644860000E+01 * scaling);
  ASSERT_DOUBLE_EQ(p.gtfs[1].exponent, 0.5421500000E+00 * scaling);
  ASSERT_DOUBLE_EQ(p.gtfs[2].exponent, 0.1446600000E+00 * scaling);
  ASSERT_DOUBLE_EQ(p.gtfs[0].coefficient, 0.202590000E+00);
  ASSERT_DOUBLE_EQ(p.gtfs[1].coefficient, 0.558300000E+00);
  ASSERT_DOUBLE_EQ(p.gtfs[2].coefficient, 0.455140000E+00);
}
} // namespace Sparrow
} // namespace Scine
